
/*
  Mistral is a constraint satisfaction and optimisation library
  Copyright (C) 2003-2005  Emmanuel Hebrard
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  The author can be contacted electronically at 
  ehebrard@cse.unsw.edu.au.
*/

/** \file con.h
    \brief Header for the Constraint library.
*/

/**********************************************
 * Constraint and subclasses
 **********************************************/


#ifndef __CONSTRAINT_H
#define __CONSTRAINT_H

#include <csp.h>
#include <var.h>

#ifdef _COINLP

#include "ClpSimplex.hpp"
#include "ClpInterior.hpp"
#include "CoinHelperFunctions.hpp"
#include "CoinTime.hpp"
#include "CoinBuild.hpp"
#include "CoinModel.hpp"
#include "ClpCholeskyBase.hpp"

#endif


#ifdef _CPLEX

#include <ilcplex/cplex.h>

#endif


#include <assert.h>


/**********************************************
 * Top level Constraints
 **********************************************/


/**********************************************
 * Unary Constraint
 **********************************************/
/*! \class UnaryConstraint
  \brief  Unary Constraints.
*/
class UnaryConstraint {

 public:
  /**@name Parameters*/
  //@{ 
  Solver *solver;
  VariableInt *X;
  //@}

  /**@name Constructors*/
  //@{
  UnaryConstraint(Solver *s, VariableInt *v);
  virtual ~UnaryConstraint();
  //@}

  /**@name Solving*/
  //@{
  virtual bool propagate() = 0;
  void activate();
  //@}
};


/**********************************************
 * Unary Lower Bound Constraint
 **********************************************/
/*! \class UnaryConstraintMore
  \brief  Unary Constraints (lower bound).
*/
class UnaryConstraintMore : public UnaryConstraint {

 public:
  /**@name Parameters*/
  //@{ 
  int bound;
  //@}

  /**@name Constructors*/
  //@{
  UnaryConstraintMore(Solver*, VariableInt*, int);
  virtual ~UnaryConstraintMore() {};
  //@}

  /**@name Solving*/
  //@{
  virtual bool propagate();
  //@}
};


/**********************************************
 * Unary Lower Bound Constraint
 **********************************************/
/*! \class UnaryConstraintLess
  \brief  Unary Constraints (upper bound).
*/
class UnaryConstraintLess : public UnaryConstraint {

 public:
  /**@name Parameters*/
  //@{ 
  int bound;
  //@}

  /**@name Constructors*/
  //@{
  UnaryConstraintLess(Solver*, VariableInt*, int);
  virtual ~UnaryConstraintLess() {};
  //@}

  /**@name Solving*/
  //@{
  virtual bool propagate();
  //@}
};


/**********************************************
 * Clause Constraint
 **********************************************/ 
/*! \class ConstraintClause
 \brief  implementation of a disjunction of literals "x =/= v"
*/
class ConstraintClause : public Constraint {

 public:
  /**@name Parameters*/
  //@{ 
  int* conflict;
  //@}

  /**@name Constructors*/
  //@{
  ConstraintClause(Solver*, VariableInt** v, const int n, const int* c);
  virtual ~ConstraintClause();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;  
  inline bool propagate();
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{
  virtual void print(std::ostream& o) const;
  //@}
};


/**********************************************
 * Extensional Constraint
 **********************************************/ 
/*! \class ConstraintExtensional
 \brief  implementation of a n-ary extensional constraint

 The conflicts are encoded using
 a N-dimensional flatened matrix:
 an entry of the matrix equal to 0
 denotes an allowed combination (support), whilst
 1 denotes a forbidden combination (conflict).
*/
class ConstraintExtensional : public Constraint {

 private:
  /**@name Parameters*/
  //@{ 
  int *var_sizes; // NOTE: this has arity-1 elements
  MistralSet matrix;
  //@}

 public:
  /**@name Constructors*/
  //@{
  ConstraintExtensional(Solver*, VariableInt** v, 
			const int l, MistralSet& m, 
			int*);
  virtual ~ConstraintExtensional();
  //@}

  /**@name Accessors*/
  //@{
  void fillNogood () ;
  void fillSupport() ;

  void addNogood(const int* vals) ;
  void addSupport(const int* vals) ;

  int getpos(const int *vals) const;
  int getpos() const;
  //@}

  /**@name Solving*/
  //@{
  /*!
     Check the corresponding cell in the matrix
  */
  inline int check( const int* ) const ;
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const;
  //@}
};


/**********************************************
 * Large Extensionalbak Constraint
 **********************************************/ 
/*! \class ConstraintLargeExtensionalbak
 \brief  implementation of a n-ary extensionalbak constraint

 The conflicts are encoded using
 a list of supporting tuples for each
 value of each variable.
 This encoding is less time efficient
 but more space efficient than the matrix encoding
*/
class ConstraintLargeExtensionalbak : public Constraint {

 private:
  /**@name Parameters*/
  //@{ 
  MistralStack<int*>** supportList;
  VariableInt ***vars;
  VariableInt **X;
  MistralSet *D_X;
  int** supports;
  int** residues;
  int* themins;
  int index;

  MistralSet **isInDomain;
  

  inline void initSupport( const int var, const int val )
  {
    supports = supportList[var][val].stack_;
    index = supportList[var][val].size();
    X = vars[var];
    D_X = isInDomain[var];
  }
  //@}

 public:
  /**@name Constructors*/
  //@{
  ConstraintLargeExtensionalbak(Solver*, VariableInt** v, 
			     const int n, MistralStack<int*>** sl);
  virtual ~ConstraintLargeExtensionalbak();
  //@}

  /**@name Solving*/
  //@{
  /*!
     Check the corresponding cell in the matrix
  */
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const;
  virtual void printlong(std::ostream& o);
  //@}
};

/**********************************************
 * Large Extensional Constraint
 **********************************************/ 
/*! \class ConstraintLargeExtensional
 \brief  implementation of a n-ary extensional constraint

 The conflicts are encoded using
 a list of supporting tuples for each
 value of each variable.
 This encoding is less time efficient
 but more space efficient than the matrix encoding
*/
class ConstraintLargeExtensional : public Constraint {

 private:
  /**@name Parameters*/
  //@{ 
  MistralStack<int*>** supportList;
  int *order;
  MistralSet *D_X;
  int** supports_X;
  int*** residues;
  int* themins;
  //@}

 public:
  /**@name Constructors*/
  //@{
  ConstraintLargeExtensional(Solver*, VariableInt** v, 
			     const int n, MistralStack<int*>** sl);
  virtual ~ConstraintLargeExtensional();
  //@}

  /**@name Solving*/
  //@{
  /*!
     Check the corresponding cell in the matrix
  */
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const;
  virtual void printlong(std::ostream& o);
  //@}
};


/**********************************************
 * Extensional Binary Constraint
 **********************************************/ 
/*! \class ConstraintBinaryExtensional
 \brief  implementation of a binary extensional constraint

 Uses bit-wise operation to speed-up support checking.
 Lists of supports are stored in bitsets
*/
class ConstraintBinaryExtensional : public Constraint {

 private:

  /**@name Parameters*/
  //@{ 
  MistralSet *supportList[2];
  int size[2];
  int minSup[2];
  int *residualSupport[2];
  //@}

 public:
  /**@name Constructors*/
  //@{
  ConstraintBinaryExtensional(Solver*, VariableInt**, MistralSet**);
  virtual ~ConstraintBinaryExtensional();
  void init();  
  //@}

  /**@name Accessors*/
  //@{
  void fillNogood () ;
  void fillSupport() ;

  void addNogood(const int* vals) ;
  void addSupport(const int* vals) ;
  //@}

  /**@name Solving*/
  //@{
  /*!
     Check the corresponding cell in the matrix
  */
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const;
  //@}
};


/**********************************************
 * =/= Constraint
 **********************************************/ 
/*! \class ConstraintNotEqual
 \brief  Binary not-equal Constraint (x0 =/= x1).
*/
class ConstraintNotEqual : public Constraint {

 public:  
  /**@name Constructors*/
  //@{
  ConstraintNotEqual(Solver*, VariableInt** v);
  virtual ~ConstraintNotEqual();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate();
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const;
  //@}
};


/**********************************************
 * =/= Constraint with a wildcard
 **********************************************/ 
/*! \class ConstraintNotEqualWildcard
 \brief  Binary not-equal Constraint with a wildcard
(x0 =/= x1 or x0 == w).
*/
class ConstraintNotEqualWildcard : public Constraint {

 public:  
  /**@name Parameters*/
  //@{ 
  int wildcard;
  //@}

  /**@name Constructors*/
  //@{
  ConstraintNotEqualWildcard(Solver*, VariableInt**, const int);
  virtual ~ConstraintNotEqualWildcard();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate();
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const;
  //@}
};


/**********************************************
 * Inverse Constraint Decomposition
 **********************************************/ 
/*! \class ConstraintInverse
 \brief  Binary inverse Constraint (xi = j <-> xj = i)

 Use to encode primal/dual channeling
*/
class ConstraintInverse : public Constraint {

 public:  
  /**@name Parameters*/
  //@{ 
  int i_[2];
  //@}

  /**@name Constructors*/
  //@{
  ConstraintInverse(Solver*, VariableInt** v, int, int);
  virtual ~ConstraintInverse();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const;
  //@}
};


/**********************************************
 * Lex Constraint Decomposition
 **********************************************/ 
/*! \class ConstraintLex
 \brief  Quaternary constraint used to encode LexOrder
*/
class ConstraintLex : public Constraint {

 public:  
  /**@name Constructors*/
  //@{
  ConstraintLex(Solver*, VariableInt** v);
  virtual ~ConstraintLex();
  //@}

  /**@name Solving*/
  //@{
  virtual int check( const int* ) const ;
  virtual bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{    
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * <= Constraint
 **********************************************/ 
/*! \class ConstraintLess
 \brief  Binary Less Than Constraint (x0 + k <= x1).
*/
class ConstraintLess : public Constraint {

 public: 
  /**@name Parameters*/
  //@{  
  int offset;
  //@}

  /**@name Constructors*/
  //@{
  ConstraintLess(Solver*, VariableInt** v);
  ConstraintLess(Solver*, VariableInt** v, int ofs);
  virtual ~ConstraintLess();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const;
  //@}
};

/***********************************************
 * All Different Constraint (bounds consistency).
 ***********************************************/
typedef struct {
  int min, max;		// start, end of interval
  int minrank, maxrank; // rank of min & max in bounds[] of an adcsp
} Interval;

/*! \class ConstraintAllDiff
 \brief  AllDifferent Constraint.

 Constraint of difference on a set of variables.
 Ensures that a set of variables are assigned distinct 
 values (only Bounds Consistency is implemented)
 The code is from Lopez-Ortiz, Quimper, Tromp and van Beek
*/
class ConstraintAllDiff : public Constraint {
  
 private:
  /**@name Parameters*/
  //@{  
  int lastLevel;
  int *t;		// tree links
  int *d;		// diffs between critical capacities
  int *h;		// hall interval links
  Interval *iv;
  Interval **minsorted;
  Interval **maxsorted;
  int *bounds;  // bounds[1..nb] hold set of min & max in the niv intervals
                // while bounds[0] and bounds[nb+1] allow sentinels
  int nb;
  void sortit();
  int filterlower();
  int filterupper();
  void propagateValue();
  //@}

 public:
  /**@name Constructors*/
  //@{
  ConstraintAllDiff(Solver*, VariableInt** x, const int n);
  virtual ~ConstraintAllDiff();
  void init();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate();
  inline bool propagate(VariableInt *v, const int e); 
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/***********************************************
 * Global Cardinality Constraint (bounds consistency).
 ***********************************************/
typedef struct {
  int firstValue;
  int lastValue;
  int* sum;
  int* ds;
} partialSum;
   

/*! \class ConstraintGlobalCardinality
 \brief  Global Cardinality Constraint.

 User defined propagator for enforcing bounds consistency
 on the restricted gcc constraint when bounds on
 occurrences are [a_i,b_i].
 A value "v" must be assigned to at least
 minOccurrences[v - firstDomainValue] variables and at most
 maxOccurrences[v - firstDomainValue] variables
 The code is from Lopez-Ortiz, Quimper, Tromp and van Beek
*/   
class ConstraintGlobalCardinality : public Constraint
{

 private:
  /**@name Parameters*/
  //@{  
  int lastLevel;
  int *t;			// tree links
  int *d;			// diffs between critical capacities
  int *h;			// hall interval links
  int *stableInterval;	// stable sets
  int *potentialStableSets;	// links elements that potentialy belong to same stable set
  int *newMin;
  Interval *iv;
  Interval **minsorted;
  Interval **maxsorted;
  int *bounds;  // bounds[1..nb] hold set of min & max of the n intervals
  // while bounds[0] and bounds[nb+1] allow sentinels
  int nb;
  
  partialSum* l; 
  partialSum* u;
  partialSum* initializePartialSum(const int firstValue, 
				   int count, const int* elements);
  void destroyPartialSum(partialSum *p);
  int  sum(partialSum *p, int from, int to);
  int  searchValue(partialSum *p, int value);
  int  minValue(partialSum *p);
  int  maxValue(partialSum *p);
  int  skipNonNullElementsRight(partialSum *p, int value);
  int  skipNonNullElementsLeft(partialSum *p, int value);
  
  void sortit();
  int  filterLowerMax();
  int  filterUpperMax();
  int  filterLowerMin(int *tl, int *c,
		      int* stableAndUnstableSets,
		      int* stableInterval,
		      int* potentialStableSets,
		      int* newMin);
  int  filterUpperMin(int *tl, int *c,
		      int* stableAndUnstableSets,
		      int* stableInterval,
		      int* newMax);
  //@}  

 public:
  /**@name Constructors*/
  //@{
  ConstraintGlobalCardinality( Solver *s,
			       VariableInt **v,
			       const int n,
			       const int firstDomainValue,
			       const int lastDomainValue,
			       const int* minOccurrences,
			       const int* maxOccurrences);
  ~ConstraintGlobalCardinality();
  //@}

  /**@name Solving*/
  //@{  
  inline int check( const int* ) const ;
  inline bool propagate();
  inline bool propagate(VariableInt *v, const int e); 
  //@}

  /**@name Miscellaneous*/
  //@{    
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * EdgeFinder Constraint
 **********************************************/ 
/*! \class ConstraintEdgeFinder
 \brief  Constraint on a set of tasks sharing a unary resource.
*/
class ConstraintEdgeFinder : public Constraint {

 public:
  /**@name Parameters*/
  //@{ 
  /*!
   * The durations of the tasks
   */
  int *durations;
  /*!
   * The tasks' indices sorted by decreasing max due date
   */
  int *sorted;
  /*!
   * The residual durations at each release date computed 
   * during the JPS
   */
  int **Pk;
  //@}

  /**@name Constructors*/
  //@{
  ConstraintEdgeFinder(Solver*, VariableInt** v, const int n, const int *d);
  virtual ~ConstraintEdgeFinder();
  //@}

  /**@name Utils*/
  //@{
  /*!
   * Compute the Jackson Preemptive Schedule necessary 
   * to the propagation method
   */
  bool jacksonPSchedule();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate();
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const;
  //@}
};



/**********************************************
 * Or Constraint
 **********************************************/ 
/*! \class ConstraintOr
 \brief  Binary Or Constraint (x0 or x1).
*/
class ConstraintOr : public Constraint
{

 public:
  /**@name Constructors*/
  //@{
  ConstraintOr(Solver*, VariableInt**);
  virtual ~ConstraintOr();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/* /\********************************************** */
/*  * IfThenElse Constraint */
/*  **********************************************\/  */
/* /\*! \class ConstraintIfThenElse */
/*  \brief  if-then-else statement (x0 ? x1 : x2) */
/* *\/ */
/* class ConstraintIfThenElse : public Constraint */
/* { */

/*  public: */
/*   /\**@name Constructors*\/ */
/*   //@{ */
/*   ConstraintIfThenElse(Solver*, VariableInt**); */
/*   virtual ~ConstraintIfThenElse(); */
/*   //@} */

/*   /\**@name Solving*\/ */
/*   //@{ */
/*   inline int check( const int* ) const ; */
/*   inline bool propagate(VariableInt *v, const int e); */
/*   //@} */

/*   /\**@name Miscellaneous*\/ */
/*   //@{   */
/*   virtual void print(std::ostream& o) const ; */
/*   //@} */
/* }; */



/* /\********************************************** */
/*  * Nested Predicates */
/*  **********************************************\/ */

/**********************************************
 * BoolSum Equal Constraint
 **********************************************/
//  lb <= x1 + ... + xn <= ub
/// Constraint on the value of the sum of a set of variables.
class ConstraintBoolSumEqual : public Constraint {
 
 public:
  /**@name Parameters*/
  //@{
  int total;
  ReversibleInt min_;
  ReversibleInt max_;
  //@}

  /**@name Constructors*/
  //@{
  ConstraintBoolSumEqual(Solver*, VariableInt** v, 
			 const int n, const int t);
  virtual ~ConstraintBoolSumEqual();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate();
  inline bool propagate(VariableInt *v, const int e);
  //@}
  
  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * BoolSum Less Constraint
 **********************************************/
//  lb <= x1 + ... + xn <= ub
/// Constraint on the value of the sum of a set of variables.
class ConstraintBoolSumLess : public Constraint {
 
 public:
  /**@name Parameters*/
  //@{
  int total;
  ReversibleInt min_;
  //@}

  /**@name Constructors*/
  //@{
  ConstraintBoolSumLess(Solver*, VariableInt** v, 
			 const int n, const int t);
  virtual ~ConstraintBoolSumLess();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate();
  inline bool propagate(VariableInt *v, const int e);
  //@}
  
  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * BoolSum More Constraint
 **********************************************/
//  lb <= x1 + ... + xn <= ub
/// Constraint on the value of the sum of a set of variables.
class ConstraintBoolSumMore : public Constraint {
 
 public:
  /**@name Parameters*/
  //@{
  int total;
  ReversibleInt max_;
  //@}

  /**@name Constructors*/
  //@{
  ConstraintBoolSumMore(Solver*, VariableInt** v, 
			 const int n, const int t);
  virtual ~ConstraintBoolSumMore();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate();
  inline bool propagate(VariableInt *v, const int e);
  //@}
  
  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * SimpleSum Predicate
 **********************************************/
/*! \class PredicateSimpleSum
 \brief  Sum of variables (x1 + ... + xn = y)
*/
class PredicateSimpleSum : public Constraint {

 public:
  /**@name Constructors*/
  //@{
  PredicateSimpleSum(Solver*, VariableInt** v, const int n);
  virtual ~PredicateSimpleSum();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate();
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * SumEqual Constraint
 **********************************************/
/*! \class ConstraintSumEqual
 \brief  Constraint on a sum of variables (x1 + ... + xn = total)
*/
class ConstraintSumEqual : public Constraint {

 public:
  /**@name Parameters*/
  //@{ 
  int total;
  //@}

  /**@name Constructors*/
  //@{
  ConstraintSumEqual(Solver*, 
		     VariableInt** v,
		     const int n,
		     const int s);
  virtual ~ConstraintSumEqual();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate();
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};

/**********************************************
 * WeightedSumEqual Constraint
 **********************************************/
/*! \class ConstraintWeightedSumEqual
 \brief  Constraint on a sum of variables (a1 * x1 + ... + an * xn = total)
*/
class ConstraintWeightedSumEqual : public Constraint {

 public:
  /**@name Parameters*/
  //@{ 
  int *weights;
  //@}

  /**@name Constructors*/
  //@{
  ConstraintWeightedSumEqual(Solver*, 
			     VariableInt** v,
			     const int n,
			     const int *s);
  virtual ~ConstraintWeightedSumEqual();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate();
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * SumLess Constraint
 **********************************************/
/*! \class ConstraintSumLess
 \brief  Constraint on a sum of variables (x1 + ... + xn = total)
*/
class ConstraintSumLess : public Constraint {

 public:
  /**@name Parameters*/
  //@{ 
  int total;
  //@}

  /**@name Constructors*/
  //@{
  ConstraintSumLess(Solver*, 
		     VariableInt** v,
		     const int n,
		     const int s);
  virtual ~ConstraintSumLess();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate();
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * WeightedSum Predicate
 **********************************************/
/*! \class PredicateWeightedSum
 \brief  WeightedSum of variables with coefficients (a1 * x1 + ... + an * xn == y)
*/
class PredicateWeightedSum : public Constraint 
{

 public:
  /**@name Parameters*/
  //@{ 
  int *weights;
  //@}

  /**@name Constructors*/
  //@{
  PredicateWeightedSum( Solver*, 
			VariableInt** nvars,
			const int n,
			const int* os );
  virtual ~PredicateWeightedSum();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate();
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * Equality with offset Predicate
 **********************************************/
/*! \class PredicateOffset
 \brief  Equality with an offset (x + k = y)
*/
class PredicateOffset : public Constraint//, public BitVar
{

 private:
  MistralSet xdom;

 public:
  /**@name Parameters*/
  //@{ 
  int offset;
  //@}

  /**@name Constructors*/
  //@{
  PredicateOffset(Solver*, VariableInt**, const int);
  virtual ~PredicateOffset();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  inline bool awakeOnRange ( VariableInt * );
  inline bool awakeOnDomain( VariableInt * );  
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * Equality by a factor Predicate
 **********************************************/
/*! \class PredicateFactor
 \brief  Equality by a factor (x * k = y)
*/
class PredicateFactor : public Constraint//, public BitVar
{

 public:
  /**@name Parameters*/
  //@{ 
  int factor;
  //@}

  /**@name Constructors*/
  //@{
  PredicateFactor(Solver*, VariableInt**, const int);
  virtual ~PredicateFactor();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * Negation Predicate
 **********************************************/
/*! \class PredicateNegation
 \brief  Arithmetic negation (-x = y)
*/
class PredicateNegation : public Constraint
{

 private:
  /**@name Parameters*/
  //@{ 
  MistralSet dom[2];
  //@}

 public:
  /**@name Constructors*/
  //@{
  PredicateNegation(Solver*, VariableInt**);
  virtual ~PredicateNegation();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * Absolute value Predicate
 **********************************************/
/*! \class PredicateAbs
 \brief  Absolute value (|x| = y)
*/
class PredicateAbs : public Constraint
{

 public:
  /**@name Constructors*/
  //@{
  PredicateAbs(Solver*, VariableInt**);
  virtual ~PredicateAbs();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * Addition Predicate
 **********************************************/
/*! \class PredicateAdd
 \brief  Binary addition predicate (x0 + x1 = y)
*/
class PredicateAdd : public Constraint
{

 private:
  /**@name Parameters*/
  //@{ 
  MistralSet xdom;
  MistralSet tdom;
  //@}

 public:
  /**@name Constructors*/
  //@{
  PredicateAdd(Solver*, VariableInt**);
  virtual ~PredicateAdd();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * Substraction Predicate
 **********************************************/
/*! \class PredicateSub
 \brief  Binary Substraction predicate (x0 - x1 = y)
*/
class PredicateSub : public Constraint
{

 private:
  /**@name Parameters*/
  //@{ 
  MistralSet xdom;
  MistralSet tdom;
  //@}

 public:
  /**@name Constructors*/
  //@{
  PredicateSub(Solver*, VariableInt**);
  virtual ~PredicateSub();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * Multiplication Predicate
 **********************************************/
/*! \class PredicateMul
 \brief  Binary Multiplication predicate (x0 * x1 = y)
*/
class PredicateMul : public Constraint
{

 private:
  /**@name Parameters*/
  //@{ 
  MistralSet xdom;
  //@}

 public:
  /**@name Constructors*/
  //@{
  PredicateMul(Solver*, VariableInt**);
  virtual ~PredicateMul();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * Division Predicate
 **********************************************/
/*! \class PredicateDiv
 \brief  Binary Division predicate (x0 / x1 = y)
*/
class PredicateDiv : public Constraint
{

 private:
  /**@name Parameters*/
  //@{ 
  MistralSet xdom;
  //@}

 public:
  /**@name Constructors*/
  //@{
  PredicateDiv(Solver*, VariableInt**);
  virtual ~PredicateDiv();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * Modulo Predicate
 **********************************************/
/*! \class PredicateMod
 \brief  Binary Modulo predicate (x0 % x1 = y)
*/
class PredicateMod : public Constraint
{

 public:
  /**@name Constructors*/
  //@{
  PredicateMod(Solver*, VariableInt**);
  virtual ~PredicateMod();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * Modulo Predicate
 **********************************************/
/*! \class PredicateMod
 \brief  Binary Modulo predicate (x0 % x1 = y)
*/
class PredicateModConstant : public Constraint
{

 public:

  int modulo;

  /**@name Constructors*/
  //@{
  PredicateModConstant(Solver*, VariableInt**, const int mod);
  virtual ~PredicateModConstant();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * Power Predicate
 **********************************************/
/*! \class PredicatePow
 \brief  Binary Exponant predicate (x0 ^ x1 = y)
*/
class PredicatePow : public Constraint
{

 public:
  /**@name Constructors*/
  //@{
  PredicatePow(Solver*, VariableInt**);
  virtual ~PredicatePow();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * Minimum Predicate
 **********************************************/
/*! \class PredicateMin
 \brief  Minimum of a set of variables (min(x0,...x1) = y)
*/
class PredicateMin : public Constraint
{

 public:
  /**@name Constructors*/
  //@{
  PredicateMin(Solver*, VariableInt**, const int);
  virtual ~PredicateMin();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * Maximum Predicate
 **********************************************/
/*! \class PredicateMax
 \brief  Maximum of a set of variables (max(x0,...x1) = y)
*/
class PredicateMax : public Constraint
{

 public:
  /**@name Constructors*/
  //@{
  PredicateMax(Solver*, VariableInt**, int);
  virtual ~PredicateMax();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * Equality Predicate
 **********************************************/
/*! \class PredicateEqual
 \brief  Truth value of a binary equality ((x0 = x1) <-> y)
*/
class PredicateEqual : public Constraint
{

 public:
  /**@name Parameters*/
  //@{ 
  int spin;
  //@}

  /**@name Constructors*/
  //@{
  PredicateEqual(Solver*, VariableInt**, int);
  virtual ~PredicateEqual();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * Equality with a constant Predicate
 **********************************************/
/*! \class PredicateEqualConstant
 \brief  Truth value of a unary equality ((x0 = k) <-> y)
*/
class PredicateEqualConstant : public Constraint
{

 public:
  /**@name Parameters*/
  //@{ 
  int K;
  int spin;
  //@}

  /**@name Constructors*/
  //@{
  PredicateEqualConstant(Solver*, 
			 VariableInt**, 
			 const int,
			 const int);
  virtual ~PredicateEqualConstant();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * Precedence Predicate
 **********************************************/
/*! \class PredicatePrecedence
 \brief  Truth value of a binary precedence ((x0 + k <= x1) <-> y)
*/
class PredicateLess : public Constraint
{

 public:
  /**@name Parameters*/
  //@{ 
  int offset;
  //@}

  /**@name Constructors*/
  //@{
  PredicateLess(Solver*, VariableInt**, const int);
  virtual ~PredicateLess();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * Lower bound Predicate
 **********************************************/
/*! \class PredicatelowerBound
 \brief  Truth value of a binary precedence ((x0 >= k) <-> y)
*/
class PredicateLowerBound : public Constraint
{

 public:
  /**@name Parameters*/
  //@{ 
  int bound;
  //@}

  /**@name Constructors*/
  //@{
  PredicateLowerBound(Solver*, VariableInt**, const int);
  virtual ~PredicateLowerBound();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * Upper bound Predicate
 **********************************************/
/*! \class PredicateUpperBound
 \brief  Truth value of a binary precedence ((x0 <= k) <-> y)
*/
class PredicateUpperBound : public Constraint
{

 public:
  /**@name Parameters*/
  //@{ 
  int bound;
  //@}

  /**@name Constructors*/
  //@{
  PredicateUpperBound(Solver*, VariableInt**, const int);
  virtual ~PredicateUpperBound();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * Disjunctive Predicate
 **********************************************/
/*! \class PredicateDisjunctive
 \brief  Ensures that two intervals do not overlap

 The last variable takes 0 if the first interval precedes
 the second, and 1 otherwise : (y -> (x0 + k0 <= x1) and !y -> (x1 + k1 <= x0))
*/
class PredicateDisjunctive : public Constraint//, public BoolVar
{

 public:
  /**@name Parameters*/
  //@{ 
  int duration[2];
  //@}

  /**@name Constructors*/
  //@{
  PredicateDisjunctive(Solver*, VariableInt**, const int*);
  virtual ~PredicateDisjunctive();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * Not Predicate
 **********************************************/
/*! \class PredicateNot
 \brief  Logic negation (!x0 = x1)
*/
class PredicateNot : public Constraint
{

 public:
  /**@name Constructors*/
  //@{
  PredicateNot(Solver*, VariableInt**);
  virtual ~PredicateNot();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate();
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * And Predicate
 **********************************************/
/*! \class PredicateAnd
 \brief  Truth value of a binary conjunction (x0 & x1 <-> y)
*/
class PredicateAnd : public Constraint
{

 public:
  /**@name Constructors*/
  //@{
  PredicateAnd(Solver*, VariableInt**);
  virtual ~PredicateAnd();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * Or Predicate
 **********************************************/
/*! \class PredicateOr
 \brief  Truth value of a binary conjunction (x0 or x1 <-> y)
*/
class PredicateOr : public Constraint
{

 public:
  /**@name Constructors*/
  //@{
  PredicateOr(Solver*, VariableInt**);
  virtual ~PredicateOr();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * IfThenElse Predicate
 **********************************************/
/*! \class PredicateIfThenElse
 \brief  Truth value of a if then else statement (x0 ? x1 : x2)
*/
class PredicateIfThenElse : public Constraint
{

 public:
  /**@name Constructors*/
  //@{
  PredicateIfThenElse(Solver*, VariableInt**);
  virtual ~PredicateIfThenElse();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * Element Predicate
 **********************************************/
/*! \class PredicateElement
 \brief  Element predicate

 Predicate on a vector x of variables and two variables y and z, ensuring that x[y] = z.
*/
class PredicateElement : public Constraint
{

 private:
  /**@name Parameters*/
  //@{ 
  MistralSet aux_dom;
  //@}

 public:
  /**@name Constructors*/
  //@{
  PredicateElement(Solver*, VariableInt** v, const int n);
  virtual ~PredicateElement();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const;
  //@}
};


/**********************************************
 * LDS Constraint
 **********************************************/
/*! \class ConstraintLDS
 \brief  Constraint on a sum of variables (x1 + ... + xn = total)
*/
class ConstraintLDS : public Constraint {

 public:
  /**@name Parameters*/
  //@{
  int                *ideal;
  int             threshold;
  ReversibleInt     mindiff;
  ReversibleInt     maxdiff;
  ReversibleSet  maybeequal;
  ReversibleSet cannotequal;
  //@}

  /**@name Constructors*/
  //@{
  ConstraintLDS(Solver*);
  virtual ~ConstraintLDS();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * LDDS Constraint
 **********************************************/
/*! \class ConstraintLDDS
 \brief  Constraint on a sum of variables (x1 + ... + xn = total)
*/
class ConstraintLDDS : public Constraint {

 public:
  /**@name Parameters*/
  //@{
  int    **distance_to_val;
  int    **val_to_distance;
  int            threshold;
  ReversibleInt   *mindiff;
  ReversibleInt   *maxdiff;
  //ReversibleSet equalities;
  //@}

  /**@name Constructors*/
  //@{
  ConstraintLDDS(Solver*);
  virtual ~ConstraintLDDS();
  //@}

  /**@name Solving*/
  //@{
  int getMaxThreshold() { return maxdiff[arity]; }
  inline int check( const int* ) const ;
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{
  virtual void print(std::ostream& o) const ;
  //@}
};



#ifdef _CPLEX

/**********************************************
 * CPlex Constraint
 **********************************************/
class CPlexConstraint : public Constraint {

 protected:
  /**@name Parameters*/
  //@{ 
  int nvals;
  int nRows;
  int nColumns;
  int lb;
  int ub;
  int type;

  int *values2index;
  int *index2values;

  int **asgn2index;
  int *index2valasgn;
  int *index2varasgn;

  int *lbs;

  double epsilon;

  void updateModel();


/*************************
 * CPlex structs
 *************************/
  int status;

  char     *probname;  
  int      objsen;
  double   *obj;
  double   *rhs;
  char     *sense;
  int      *matbeg;
  int      *matcnt;
  int      *matind;
  double   *matval;
  double   *zlb;
  double   *zub;
  int      *qmatbeg;
  int      *qmatcnt;
  int      *qmatind;
  double   *qmatval;
  

  int      solstat;
  double   objval;
  double   *x;//[NUMCOLS];
  double   *pi;//[NUMROWS];
  double   *slack;//[NUMROWS];
  double   *dj;//[NUMCOLS];
  
  CPXENVptr     env;
  CPXLPptr      lp;
  /*************************
   *                       *
   *************************/
  //@}

 public:
  /**@name Constructors*/
  //@{
  CPlexConstraint(Solver*, VariableInt** v, const int n);
  virtual ~CPlexConstraint();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  virtual bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * CPlexMinDiff Predicate
 **********************************************/
/// Counts the maximum number of differences in a set of vars
class PredicateCPlexMinDiff : public CPlexConstraint
{  
 public:

  /**@name Parameters*/
  //@{ 
  int K;
  int *occurrence;
  int *orderedvals;
  VariableInt **vars;
  int  filtering;
  //@}

  /**@name Constructors*/
  //@{
  PredicateCPlexMinDiff(Solver*, const int f, VariableInt** x, const int n);
  virtual ~PredicateCPlexMinDiff();
  //@}

  /**@name Solving*/
  //@{
  int approximation(int&);

  inline bool propagate(VariableInt *v, const int e);
  //@}
  
  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};

#endif


#ifdef _COINLP

/**********************************************
 * LP Constraint
 **********************************************/
class LpConstraint : public Constraint {

 public:

  /**@name Parameters*/
  //@{ 
  static const int SIMPLEX_PRIMAL     = 0;
  static const int SIMPLEX_DUAL       = 1;
  static const int SIMPLEX_BARRIER    = 2;
  static const int SIMPLEX_PRIMAL_IT  = 3;
  static const int SIMPLEX_DUAL_IT    = 4;
  static const int SIMPLEX_BARRIER_IT = 5;
  static const int INTERIOR           = 6;

  static const int LPO = 0;
  static const int LPA = 1;
  static const int LPS = 2;
  static const int APX = 3;
  static const int UPB = 4;


 protected:
  
  ClpSimplex  smodel;
  ClpInterior imodel;

  int nvals;
  int nRows;
  int nColumns;
  int lb;
  int ub;
  int type;

  int *values2index;
  int *index2values;

  int **asgn2index;
  int *index2valasgn;
  int *index2varasgn;

  int *lbs;

  double epsilon;

  void updateModel();
  //@}

 public:
  /**@name Constructors*/
  //@{
  LpConstraint(Solver*);
  LpConstraint(Solver*, VariableInt** v, const int n, const int t=6);
  void init(const int t);
  virtual ~LpConstraint();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;
  virtual bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{    
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * MaxDiff Predicate
 **********************************************/
/// Counts the maximum number of differences in a set of vars
class PredicateMaxDiff : public LpConstraint
{  
  
 public:

  /**@name Parameters*/
  //@{ 
  MistralSet freevals;
  int *groundvals;
  int *orderedvals;
  int  filtering;
  //@}

  /**@name Constructors*/
  //@{
  PredicateMaxDiff(Solver*, VariableInt** v, const int n, const int t=6);
  PredicateMaxDiff(Solver*, const int f, VariableInt** v, const int n, const int t=6);
  void initStruct(const int t);
  virtual ~PredicateMaxDiff();
  //@}

  /**@name Solving*/
  //@{
  int upperbound();
  int approximation(int&);

  inline int check(int *sol) const;
  inline bool propagate(VariableInt *v, const int e);
  //@}
  
  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};


/**********************************************
 * MinDiff Predicate
 **********************************************/
/// Counts the maximum number of differences in a set of vars
class PredicateMinDiff : public LpConstraint
{  
 public:

  /**@name Parameters*/
  //@{ 
  int K;
  int *occurrence;
  int *orderedvals;
  VariableInt **vars;
  int  filtering;

  ClpCholeskyBase * cholesky;
  //@}

  /**@name Constructors*/
  //@{
  PredicateMinDiff(Solver*, VariableInt** x, const int n, const int t=6);
  PredicateMinDiff(Solver*, const int f, VariableInt** x, const int n, const int t=6);
  void initStruct(const int t);
  virtual ~PredicateMinDiff();
  //@}

  /**@name Solving*/
  //@{
  int approximation(int&);

  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{  
  virtual void print(std::ostream& o) const ;
  //@}
};

#endif // _COINLP


/**********************************************
 * Decomposition of Slide Constraint
 **********************************************/ 
/*! \class ConstraintSlideDecomp
 \brief  implementation of slide
*/
class ConstraintSlideDecomp : public Constraint {

 public:
  /**@name Parameters*/
  //@{ 
  int nbVals; 
  int sequenceWidth;
  int* conflicts;
  int* int2ind;
  int* ind2int;
  int* exposant;
  //@}

  /**@name Utilities*/
  //@{ 
  inline void tuple2int(int& x, int* tuple, int n) {
    x = 0;
    while( n-- )
      x += (tuple[n] * exposant[n]);  
  }  
  inline int ktuple( const int left, int const right ) const {
    return ind2int[left]%nbVals + ind2int[right]*nbVals;
  }
  //@{ 

  /**@name Constructors*/
  //@{
  ConstraintSlideDecomp(Solver*, VariableInt**, 
			const int, const int, const int, 
			const int, const int,  
			int*, int*, int*, int*);
  ~ConstraintSlideDecomp();
  //@{ 

  /**@name Solving*/
  //@{
  inline int check( const int* ) const {return 0;}  
  inline bool propagate(VariableInt *v, const int e) {return true;}
  //@}
};


/**********************************************
 * Slide Constraint
 **********************************************/ 
/*! \class ConstraintSlide
 \brief  implementation of tuple sliding
*/
class ConstraintSlide : public Constraint {

 protected:

  /**@name Parameters*/
  //@{ 
  int nbVals; 
  int sequenceWidth;
  int* conflicts;
  int* int2ind;
  int* ind2int;
  int* exposant;
  //@}

  /**@name Utilities*/
  //@{ 
  inline int appendLeft(const int t, const int e) const {
    return ind2int[t]*nbVals+e;
  }  
  inline int appendRight(const int t, const int e) const {
    return ind2int[t]+e*exposant[sequenceWidth-1];
  }
  //@}

 public:

  /**@name Constructors*/
  //@{
  ConstraintSlide(Solver*, VariableInt**, 
		  ConstraintSlideDecomp*);
  ~ConstraintSlide();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;  
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{
  virtual void print(std::ostream& o) const;
  //@}
};


/**********************************************
 * Tuple Constraint
 **********************************************/ 
/*! \class PredicateTuple
 \brief  implementation of tuple
*/
class PredicateTuple : public Constraint {

 protected:

  /**@name Parameters*/
  //@{ 
  int           nbVals;
  MistralSet     *vals;
  int         *int2ind;
  int         *ind2int;
  int        *exposant;
  //@}

 public:

  /**@name Constructors*/
  //@{
  PredicateTuple(Solver*, VariableInt**,
		 const int, const int,
		 int*, int*, int*);
    ~PredicateTuple();
  //@}

  /**@name Solving*/
  //@{
  inline int check( const int* ) const ;  
  inline bool propagate(VariableInt *v, const int e);
  //@}

  /**@name Miscellaneous*/
  //@{
  virtual void print(std::ostream& o) const;
  //@}
};



#endif // _CONSTRAINT_H
