
/*
  Mistral is a constraint satisfaction and optimisation library
  Copyright (C) 2003-2005  Emmanuel Hebrard
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  The author can be contacted electronically at 
  ehebrard@cse.unsw.edu.au.
*/



/*! \file csp.h
    \brief Header for the data structures representing a CSP.
*/

/**********************************************
 * The core: CSP, Constraint, Variable,
 **********************************************/

#ifndef _CSP_H
#define _CSP_H

#include <iostream>
#include <iomanip>
#include <vector>
#include <set.h>

#include <map>

class ValSelector;
class VariableInt;
class Constraint;
class Solver;
class CSP;

class UnaryConstraint;
class ObjectiveFunction;


class ValIterator {

 public:
  virtual bool next()=0;
  virtual operator const int() const=0;
};


/**********************************************
 * VariableInt
 **********************************************/
/*! \class VariableInt
  \brief Abstract Representation of Variables.

  Various informations, such as the list of Constraints
  constraining the Variable are stored. The important
  parts are the methods makeDecision() and
  makeComplementary(), which define the left and
  right branches of the search tree
*/
class VariableInt {

//class ConstrainedObject {
  /*!
    ConstrainedObject

   */

  // remove range and domain from VariableInt and domain from RangeVar.
 protected:

  MistralList<Constraint*> valueTrigger;
  MistralList<Constraint*> rangeTrigger;
  MistralList<Constraint*> domainTrigger;

 public: 

  /// The list of constraints triggered on value assignment
  virtual MistralList<Constraint*>* triggerOnValue() {return &valueTrigger;}
  virtual MistralNode<Constraint*>* constraintsOnValue() 
  {return &(valueTrigger.head);}
  /// The list of constraints triggered on bound(s) reduction
  virtual MistralList<Constraint*>* triggerOnRange() {return &rangeTrigger;}
  virtual MistralNode<Constraint*>* constraintsOnRange() 
  {return &(rangeTrigger.head);}
  /// The list of constraints triggered on value(s) removal
  virtual MistralList<Constraint*>* triggerOnDomain() {return &domainTrigger;}
  virtual MistralNode<Constraint*>* constraintsOnDomain() 
  {return &(domainTrigger.head);}
  /// The list of watched constraints
  MistralList<Constraint*> watchedConstraints;

  //};

  /*!
    DecisionVariable

   */

//class DecisionVariable : public ConstrainedObject {

 public:
  static const int CONST = 0;
  static const int BIT   = 1;
  static const int BOOL  = 2;
  static const int RANGE = 3;
  static const int LIST  = 4;
  static const int NOVAL = INT_MAX;

  /*!@name Parameters*/
  //@{

  // does the domain fit on a single memory word
  bool isWord;

  /// Direct pointer to the solver
  Solver *solver;

  /// Value (branching) selection;
  ValSelector *branch;

  /// Pointer to its cell in the array 'sequence' of the solver
  /// Used when linking/unlinking the variable.
  VariableInt** seqIdx;
  //DecisionVariable** seqIdx;
  /// Pointer to either the array 'sequence', or 'auxilliary'
  /// Used when linking/unlinking the variable.
  VariableInt*** seqBeg;
  //DecisionVariable*** seqBeg;

  /// The agreggated constraint weight
  long unsigned int weight;
  /// The index of this VariableInt.
  int id;
  /// Whether or not the variable is in the current constraint network
  bool isLinked() const { return (seqIdx >= *seqBeg); } 
  /// Number active Constraints on this Variable
  int degree;

  /*!@name Accessors and Iterators*/
  //@{
  /// Returns the assigned value if it exists
  virtual int value() const = 0;
  //@}

  // unLink when it become assigned and link when backtracking
  /// Acts as if the Variable was removed from the Constraint network
  void unLink();
  /// Restore the Variable in the Constraint network
  void link();
  /// Print the domain

  //};


  /*!
    VariableInt

   */

  //class VariableInt : public DecisionVariable {
  
 public:

  /// DomainIterator
  virtual ValIterator *getDomainIterator()=0;

  /*!@name Accessors and Iterators*/
  //@{
  /// Returns the first (not necessarily min) value in the domain
  virtual int first() const = 0;

  /// Returns the last (not necessarily max) value in the domain
  virtual int  last() const = 0;

  /// Returns the minimum value in the domain
  virtual int min() const = 0;

  /// Returns the maximum value in the domain
  virtual int max() const = 0;

  /// Returns the minimum absolute value in [0..max] \inter domain
  virtual int minPosAbs() const = 0;

  /// Returns the minimum absolute value in [min..0] \inter domain
  virtual int minNegAbs() const = 0;

  /// Returns the size of the domain
  virtual int domsize() const = 0;

  /// Returns the minimum value that could belong to the domain
  virtual int minCapacity() const = 0;

  /// Returns 1 + the maximum value that could belong to the domain
  virtual int maxCapacity() const = 0;

  /// Value increment
  /*!
    Used when iterating on values in the domain.
    Returns the value succeeding "v".
    Example:

    \code
  
    vali = vi->first();
    do {
    [something];
    vi = vi->getNext(vali) ;
    } while ( vi->good(vali) );

    \endcode

  */
  virtual int getNext(const int v) const = 0;

  /// Termination condition
  /*!
    Used when iterating on values in the domain.
    Returns true if "v" is valid value.
    Example:
  
    \code
  
    vali = vi->first();
    do {
    [something];
    vi = vi->getNext(vali) ;
    } while ( vi->good(vali) );
  
    \endcode

  */
  virtual bool good(const int v) const = 0;

  /// Value increment
  /*!
    Used when iterating on values in the domain.
    Set "v" to the value succeeding "v".
    Example:

    \code
  
    vali = vi->first();
    do {
    [something];
    } while ( vi->setNext(vali) );
  
    \endcode

  */
  virtual bool setNext(int& v) const = 0;
  //@}

  /*!@name Decision methods*/
  //@{
  virtual bool revise( MistralSet*, int*, VariableInt* ) = 0;
  virtual bool revise( Constraint*, const int ) = 0;
  //@}

  /*!@name Query methods*/
  //@{
  /// Whether or not the Variable is currently an interval
  virtual bool isRange() const = 0;

  /// Whether or not the Variable is bound to a ground value
  virtual bool isGround() const = 0;

  /*!
    Whether "v" is currently contained in the domain
  */
  virtual bool contain(const int v) const = 0;
  virtual bool fastContain(const int v) const = 0;

  // set operation functions
  /*!
    Whether the domain has a nonempty intersection
    with the interval [l..u]
  */
  virtual bool intersect(const int lo, const int up) const = 0;

  /*!
    Whether the domain is included in
    the interval [l..u]
  */
  virtual bool included(const int lo, const int up) const = 0;

  /*!
    Whether the interval [l..u] is included in
    the domain
  */
  //virtual bool include const int lo, const int up) const = 0;

  /*!
    Whether the domain has a nonempty intersection
    with the Variable x
  */
  virtual bool intersect(const VariableInt* x) const = 0;
  /*!
    Whether the domain is included
    in the Variable x
  */
  virtual bool included(const VariableInt* x) const = 0;
  /*!
    Whether the Variable x is included
    in the domain
  */
  virtual bool include(const VariableInt* x) const = 0;
  /*!
    Whether the domain has a nonempty intersection
    with the set s
  */
  virtual bool intersect(const MistralSet* s) const = 0;
  /*!
    Whether the domain is included
    in the set s
  */
  virtual bool included(const MistralSet* s) const = 0;
  /*!
    Whether the set s is included
    in the domain
  */
  virtual bool include(const MistralSet* s) const = 0;
  /*!
    Whether the domain has a nonempty intersection
    with the set s
  */
  virtual bool intersect(const MistralSet& s) const = 0;
  /*!
    Whether the domain has a nonempty intersection
    with the set s
  */
  virtual bool intersect(const MistralSet& s, int& r) const = 0;
  virtual bool wordIntersect(const MistralSet& s) const = 0;
  /*!
    Whether the domain is included
    in the set s
  */
  virtual bool included(const MistralSet& s) const = 0;
  /*!
    Whether the set s is included
    in the domain
  */
  virtual bool include(const MistralSet& s) const = 0;
  /*!
    Intersect its domain with a set s
  */
  virtual void intersectTo( MistralSet* s ) const = 0;
  /*!
    Do the union of its domain with a set s
  */
  virtual void unionTo( MistralSet* s ) const = 0;
  /*!
    Intersect its domain with a set s
  */
  virtual void intersectTo( MistralSet& s ) const = 0;
  /*!
    Do the union of its domain with a set s
  */
  virtual void unionTo( MistralSet& s ) const = 0;
  //@}


  /*!@name Domain handling methods*/
  //@{
  /// Remove the value "v"
  virtual bool remove(const int v) = 0;
  /// Remove all values but "v"
  virtual bool setDomain(const int v) = 0;
  /// Remove all values strictly lower than l
  virtual bool setMin(const int l) = 0;
  /// Remove all values strictly greater than u
  virtual bool setMax(const int u) = 0;
  /// Remove all values that do not appear in the array "a" of length "l"
  virtual bool setDomain(const int* a, const int l) = 0;
  /// Remove all values that do not appear in the set "s"
  virtual bool setDomain(const MistralSet& s) = 0;
  /// Remove all values that do not appear in the current domain of the Variable "x"
  virtual bool setDomain( VariableInt* x) const = 0;
  /// Remove all values that belong to the set "s"
  virtual bool removeSet(const MistralSet& s) = 0;
  /// Remove all values in the interval [l..u]
  virtual bool removeRange(const int l, const int u) = 0;
  //@}




  /*!@name Constructors*/
  //@{
  VariableInt() {}
  VariableInt(Solver*);
  virtual ~VariableInt();
  //@}




  /*!@name Utility methods*/
  virtual void print(std::ostream&) const = 0;
  virtual void printshort(std::ostream&) const = 0;
  void printConsList(std::ostream&) ;
  virtual int getType() const { return -1; }
  //@}
};


/**********************************************
 * Constraint
 **********************************************/
/*! \class Constraint
  \brief Representation of Constraints.

  The constrained variables are stored in the
  array scope[]. The method propagate() is
  called when computing the GAC closure. A
  generic propagate() method, using AC3
  with residual supports is implemented in
  the class Constraint.
*/
class Constraint {

 public:

  /*!@name Static members*/
  //@{
  static const int WEIGHTONE      = 1;
  static const int WEIGHTTWO      = 1;
  static const int WEIGHTTHREE    = 1;
  static const int DOMAINTRIGGER  = 1;
  static const int RANGETRIGGER   = 2;
  static const int VALUETRIGGER   = 4;
  static const int DELAYTRIGGER   = 8;
  //@}


  /*!@name Parameters*/
  //@{
  Solver *solver;

  /// delayed?
  bool delayed;
  /// An unique id
  int id;
  /// The number of constrained variables
  int arity;
  /// The number of constrained variables that are not yet assigned
  int unasgnVars;
  /// Weight, used in some heuristics
  long unsigned int weight;
  long unsigned int oldweight;
  /// Hack for the linkage
  int lastlinked;

  /// The constrained variables.
  VariableInt** scope;
  /// The indices of the two variables that are watching us
  VariableInt* watch[2];

  /// The type of trigger for each constrained variable
  /*!
    The possible triggers, for the ith variable in the scope are:
    - NULL
    - &(scope[i]->valueTrigger)
    - &(scope[i]->rangeTrigger)
    - &(scope[i]->domainTrigger)
  */
  MistralList<Constraint*> **triggers;
  /// The element of the list
  /*!
    This node points to this constraint and
    can be added or removed from the corresponding
    list in constant time.
  */
  MistralNode<Constraint*> *elements;
  
  /// used to manipulate supports in the generic AC algorithm
  int *supp; 
  /// structure to keep residual supports
  int ***supports; 
  //@}


  /*!@name Constructors*/
  //@{
  /// The scope is build by copying an array "scp" containing "l" variables
  void initScope(VariableInt **scp, const int l, const int t, const int w=1);
  Constraint() {}
  Constraint(Solver *s);
  Constraint(Solver *s, VariableInt **scp, const int l, const int t=DOMAINTRIGGER, const int w=1);
  virtual ~Constraint();
  //@}

  /*!@name Propagators*/
  //@{
  /*!
   * This methods returns 0 if the constraint is satisfied
   * by the combination and any positive value otherwise.  
   */
  virtual int check(const int*) const = 0;
  /*! 
   *  This method is called when the domain of a constrained 
   *  variable has been modified. Some domains in 
   *  scope may be reduced, and true 
   *  is returned on success (there are still at least one 
   *  consistent assignment) or false otherwise. 
   */
  virtual bool propagate();
  virtual bool propagate(VariableInt *, const int e);
  /*!
   *  Check if the cached support is valid.
   *  Otherwise initialize an iterator on supports
   */
  bool firstSupport(const int, const int);
  /*!
   *  Find the first valid support. 
   *  Return true if a support has been found, 
   *  or false if no such support exists.
   */
  bool findSupport(const int, const int);
  //@}


  /*!@name Manipulation methods*/
  //@{
  bool isLinked();
  void stopWatching( VariableInt* );
  void resumeWatching( VariableInt* );
  /// Whether the residual support should be used
  void useResidualSupports() ;
  /// Print the constraint
  virtual void print(std::ostream & os) const ;
  bool constrains( VariableInt *x ) const ;
  //@}
};


/**********************************************
 * MistralGraph
 **********************************************/

/*! \class MistralGraph
  \brief  A directed graph implementation.
*/
class MistralGraph
{
 public:
  int n;

  int **neighbor;
  int ***node;
  int **first;
  int **last;

  MistralGraph( VariableInt **vars, const int l );
  MistralGraph( CSP& model ); //TODO
  MistralGraph( const int l );
  MistralGraph( const int l, vector<int> E );
  MistralGraph( ) { n = 0; }
  void init(const int);
  void cleanUp();

  ~MistralGraph();

 private:
  int getAndRemoveMaxCand( MistralSet& cand );
  void removeNeighbours  ( const int x, MistralSet& neighbours );
  void updateGammaK      ( const int x, MistralSet& res );
  void storeClique       ( MistralSet& K, vector< vector< int > >& cliques );
  void BronKerbosch      ( MistralSet& cand,
			   MistralSet& K,
			   MistralSet& gammaK,
			   vector< vector< int > >& cliques,
			   const int limit );

 public:
  void getAllCliques( vector< vector< int > >& cliques, const int limit=INT_MAX );
  void getSomeCliques( vector< vector< int > >& cliques );

  inline bool isIn(const int x, const int y) const
  {
    return (bool)(node[x][y]);
  }

  int next(const int i, const int j)
  {
    return *(node[i][j]+1);
  }
  int degree(const int) const;
  void add(const int, const int);
  void remove(const int, const int);

  void dfs(const int) const;
  int AltBlumMehlorn( int nv );
  int greedyMaxClique() const;
  
  void print(std::ostream& o) const;
};



#endif // __CSP_H




/*! \mainpage Mistral
 *
 * \section intro_sec Introduction
 *
 * Welcome to <i><b>Mistral</b></i>'s web page.
 * <i><b>Mistral</b></i> is a C++ library for solving
 * Constraint satisfaction and optmisation problems.
 * If you are unsure about Constraint satisfaction,
 * you may want to check the <a href="http://kti.ms.mff.cuni.cz/~bartak/Constraints/index.html">online guide to Constraint Programming</a> by Roman Bart&aacute;k first.
 * This documentation has been
 * generated using <a href="http://www.doxygen.org/index.html">doxygen</a>.
 *
 * 
 *
 *
 * \section install_sec Installation
 * 
 * In principle, the makefile should
 * work ok as is.
 * <br>
 * To compile the library and all examples:
 * <br>
 * $ <tt>make</tt>
 * <br>
 * To compile the model <tt>model-name</tt> (if
 * the file <tt>models/src/model-name.cpp</tt> exists):
 * <br>
 * $ <tt>make model-name</tt>
 * <br>
 * To compile the library only:
 * <br>
 * $ <tt>make lib</tt>
 * <br><br>
 * The directory <b>include/</b> contains all headers 
 * and <b>lib/src/</b> all sources for the library.
 * The object files go into <b>lib/obj/</b> and the binaries
 * into <b>bin/</b>. Last, the directory <b>models/src/</b>
 * contains the models (examples).
 *
 * \section features_sec Features
 * 
 * Some features implemented in Mistral:
 * <ul>
 * <li> Neighborhood based heuristic (Bessiere, Chmeiss and Sas)
 * <li> Weighted Degree heuristic (Lecoutre et al.)
 * <li> Bound Consistency for the AllDifferent Constraint 
 * (Code from Lopez-Ortiz, Quimper, Tromp and van Beek)
 * <li> NValue Constraint (Beldiceanu) <!--<li> Singleton Arc Consistency (Bessiere and Debruyne) and Shaving (Torres and Lopez)-->
 * <li> EdgeFinder Algorithm implemented as a global Constraint (Carlier) <!--<li> Operation Resource Reliance Heuristic for job-shop scheduling (Sadeh)-->
 * <li> PCP branching algorithm for job-shop scheduling (Smith and Cheng)
 * <li> ...
 *
 *</ul>
 * \section tutorial_sec Tutorial
 *
 * This \ref tutorial shows step by step how to write a 
 * model for the N queens problem using Mistral.
 * It also shows how to declare new Constraints
 * and propagators.
 *
 * \section copyright_sec Copyright
 *
 * This software is under GNU 
 *  <a href="http://www.gnu.org/copyleft/gpl.html">General Public License</a>.
 *  <br><br>
 *  Mistral, a Constraint satisfaction and optimisation library
 *  Copyright (C) 2003-2005  <br> Emmanuel Hebrard
 *  <br><br>
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  <br><br>
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <br><br>
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * <br><br>
 * The author can be contacted electronically at 
 * ehebrard@cse.unsw.edu.au.
 */

/*! \page tutorial Tutorial

  <A NAME="top"></A>
  \section nqueen_sec NQueens Problem

  <ul>
  <li> <A HREF="#step1">Step 1: Prerequisites</A>
  <li> <A HREF="#step2">Step 2: CSP and variables</A>
  <li> <A HREF="#step3">Step 3: Predefined Constraints</A>
  <li> <A HREF="#step4">Step 4: Defining a new Constraint (the easy way)</A>
  <li> <A HREF="#step5">Step 5: Posting Constraints</A>
  <li> <A HREF="#step6">Step 6: Solving</A>
  <li> <A HREF="#step7">Step 7: Handling Solution and Statistics</A>
  <li> <A HREF="#step8">Step 8: The whole Model (1)</A>
  <li> <A HREF="#step9">Step 9: Defining a new Constraint (implementing the propagator)</A>
  <li> <A HREF="#step10">Step 10: The whole Model (2)</A>
  <li> <A HREF="#step11">Step 11: Result</A>
  </ul>
  <HR>

  <A NAME="step1"></A>
  <H3>Step 1: Prerequisites</H3>
  <HR>

  The following is a simple model for the N Queens problem.


  \code


int main(int argc, char *argv[]) { 
 
  // command line
  int N = ( argc > 1 ? atoi(argv[1]) : 40 );

  // the model
  VarArray queens(N,N);
  CSP model;

  // extra variables  
  VarArray first_diagonal( queens.size() );
  VarArray second_diagonal( queens.size() );
  for(int i=0; i<queens.size(); ++i) {
  first_diagonal[i]  = (queens[i] + i);
  second_diagonal[i] = (queens[i] - i);
  }
  
  // the constraints
  model.add( AllDifferent(first_diagonal) );
  model.add( AllDifferent(second_diagonal) );
  model.add( AllDifferent(queens) );
  
  // the solver
  MinDomMin heuristic;
  
  Solver s( model, queens, heuristic );
  s.solve_and_restart(Solver::GEOMETRIC, 32, 2);

  }

  \endcode


  <BR><BR>
  <A href="#top">Back to top</A>
  <HR>

  <A NAME="step2"></A>
  <H3>Step 2: CSP and variables</H3>

  <HR>

  <BR>
  <A href="#top">Back to top</A>
  <HR>

  <A NAME="step3"></A>
  <H3>Step 3: Predefined Constraints</H3>

  <HR>

  <BR>
  <A href="#top">Back to top</A>
  <HR>

  <A NAME="step4"></A>
  <H3>Step 4: Defining a new Constraint (the easy way)</H3>

  <HR>

  <BR>
  <A href="#top">Back to top</A>
  <HR>

  <A NAME="step5"></A>
  <H3>Step 5: Posting Constraints</H3>

  <HR>

  <BR>
  <A href="#top">Back to top</A>
  <HR>

  <A NAME="step6"></A>
  <H3>Step 6: Solving</H3>

  <HR>

  <BR>
  <A href="#top">Back to top</A>
  <HR>

  <A NAME="step7"></A>
  <H3>Step 7: Handling Solution and Statistics</H3>

  <HR>

  <BR>
  <A href="#top">Back to top</A>
  <HR>

  <A NAME="step8"></A>
  <H3>Step 8: The whole Model (1)</H3>

  <HR>

  <BR>
  <A href="#top">Back to top</A>
  <HR>

  <A NAME="step9"></A>
  <H3>Step 9: Defining a new Constraint (implementing the propagator)</H3>

  <HR>

  <BR>
  <A href="#top">Back to top</A>
  <HR>

  <A NAME="step10"></A>
  <H3>Step 10: The whole Model (2)</H3>

  <HR>

  <BR>
  <A href="#top">Back to top</A>
  <HR>

  <A NAME="step11"></A>
  <H3>Step 11: Result</H3>

  <HR>

  Here is an example of output ;-)

  <pre>

  100 queens:              SAT        10 BTS       106 NDS           51322 CKS      0.02 s
  
  Q...................................................................................................
  ..Q.................................................................................................
  ....Q...............................................................................................
  ............................................................Q.......................................
  ..................................................................Q.................................
  ...Q................................................................................................
  .............................................................................Q......................
  ......Q.............................................................................................
  ..........................................................Q.........................................
  ....................................................................Q...............................
  .........................................................Q..........................................
  ....................................................................................Q...............
  .....Q..............................................................................................
  .................................................................................Q..................
  ...................................................................Q................................
  .........................................................................................Q..........
  .......Q............................................................................................
  ................................................................................................Q...
  .............................................................Q......................................
  ...................................................................................Q................
  ......................................................................................Q.............
  ..................................................................................Q.................
  ........Q...........................................................................................
  ...................................................Q................................................
  ................................................Q...................................................
  ............................................................................Q.......................
  ..................................................................................................Q.
  ..............................................................Q.....................................
  ........................................................................................Q...........
  .........Q..........................................................................................
  ...............................................................................................Q....
  .........................Q..........................................................................
  ..............................................................................Q.....................
  ...............................................................Q....................................
  .........................................Q..........................................................
  ........................................................Q...........................................
  .................Q..................................................................................
  ..........Q.........................................................................................
  ..................................................Q.................................................
  .............................................Q......................................................
  ...........................................................Q........................................
  .................................................Q..................................................
  .......................................................Q............................................
  ...................................Q................................................................
  ................................Q...................................................................
  ................Q...................................................................................
  ...........Q........................................................................................
  ..........................................................................Q.........................
  ............................................Q.......................................................
  .................................................................Q..................................
  .......................................................................................Q............
  ....................................Q...............................................................
  ..................................Q.................................................................
  .........................................................................Q..........................
  .............................................................................................Q......
  ..................Q.................................................................................
  ............Q.......................................................................................
  ....................................................Q...............................................
  ........................................................................Q...........................
  ...........................................................................................Q........
  .....................................................................Q..............................
  .................................................................................................Q..
  ...............................................................................Q....................
  .....................................................................................Q..............
  ..........................................................................................Q.........
  ...................................................................................................Q
  ...................Q................................................................................
  .............Q......................................................................................
  .......................................................................Q............................
  ...............................Q....................................................................
  ...........................................................................Q........................
  ..............................................................................................Q.....
  .......................................Q............................................................
  ......................................................................Q.............................
  ................................................................Q...................................
  ...............................................Q....................................................
  ................................................................................Q...................
  ............................................................................................Q.......
  ....................Q...............................................................................
  ..............Q.....................................................................................
  ........................................Q...........................................................
  .Q..................................................................................................
  ..........................Q.........................................................................
  .....................................................Q..............................................
  .....................Q..............................................................................
  ..............................................Q.....................................................
  ......................................................Q.............................................
  ............................Q.......................................................................
  ........................Q...........................................................................
  .............................Q......................................................................
  .....................................Q..............................................................
  ...........................................Q........................................................
  ...............Q....................................................................................
  ..........................................Q.........................................................
  .......................Q............................................................................
  ......................................Q.............................................................
  ...........................Q........................................................................
  ......................Q.............................................................................
  ..............................Q.....................................................................
  .................................Q..................................................................

  </pre>

  <BR>
  <A href="#top">Back to top</A>
  <HR>

*/
