
/*
  Mistral is a constraint satisfaction and optimisation library
  Copyright (C) 2003-2005  Emmanuel Hebrard
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  The author can be contacted electronically at 
  ehebrard@cse.unsw.edu.au.
*/

/** \file dvo.h
    \brief Header for DVOs (Dynamic Variable/Value Ordering) and Goals (Objective functions).
*/

#ifndef _DVO_H
#define _DVO_H

#include <con.h>
#include <gen.h>
#include <var.h>


class Weighter
{
 public:

  /**@name Parameters*/
  //@{ 
  static const int NO=0;
  static const int WDG=1;
  static const int WLD=3;
  static const int IPT=8;
  static const int SAC=16;

  int& level;
  //@}

  /**@name Constructors*/
  //@{
  Weighter( Solver* );
  virtual ~Weighter() {}
  //@}  

  /**@name Utils*/
  //@{ 
  virtual void notifyFailure( Constraint *con ) {}
  virtual void notifySuccess() {}
  virtual void notifyChoice() {}  
  virtual int getType() { return NO; }
  virtual void print( ostream& o ) const {}
  //@}  
};

class WeighterDegree : public Weighter
{
 public:

  /**@name Constructors*/
  //@{
  WeighterDegree( Solver* s );  
  virtual ~WeighterDegree( ) {}
  //@}  

  /**@name Utils*/
  //@{ 
  virtual void notifyFailure( Constraint *con ) ;
  virtual int getType() { return WDG; }
  //@}  
};

class WeighterLevelDegree : public Weighter
{
 public:

  /**@name Parameters*/
  //@{ 
  /// Depth of the search tree
  int lmax;
  //@}

  /**@name Constructors*/
  //@{
  WeighterLevelDegree( Solver* s );  
  virtual ~WeighterLevelDegree( ) {}
  //@}  

  /**@name Utils*/
  //@{ 
  virtual void notifyFailure( Constraint *con ) ;
  virtual int getType() { return WLD; }
  //@}  
};

class WeighterImpact : public Weighter
{
 public:

  /**@name Parameters*/
  //@{ 
  // repetition of some of the dvo's params
  /// The list of decisions
  VariableInt**& decision;
  /// A reference to the beginning of the array of future variables
  VariableInt**& first;
  /// A reference to the end of the array of future variables
  VariableInt**&  last;

  // original length of the variable array
  unsigned int ilength;

  // let Sb be the size of the CSP before decision d,
  // and Sa be the size of the CSP after this decision.
  // I(d) = Sb/Sa, we store 1/I(d), that is Sa/Sb in [0..1].
  double **decision_impact;
  bool *isRange;

  // number of time this decision was taken (to compute the average)
  unsigned int **decision_count;

  // use to store the variables and domain size 
  // when exploring the future vars.
  // when we are notified that the filtering occurred, 
  // we can compare the size in order to update impacts
  VariableInt  **X;
  unsigned int *domain_size;
  unsigned int nbvars;

  int *varmin;
  bool needUpdate;

  VariableInt** variables;
  //@}

  /**@name Constructors*/
  //@{
  WeighterImpact( Solver* s );  
  virtual ~WeighterImpact( );
  //@}

  /**@name Utils*/
  //@{ 
  virtual void notifyFailure( Constraint *con ) ;
  virtual void notifySuccess() ;
  virtual void notifyChoice() ; 
  virtual int getType() { return IPT; }
  virtual void print( ostream& o ) const;
  //@}
};

class WeighterSAC : public Weighter
{
 public:

  /**@name Parameters*/
  //@{ 
  bool         complete;
  bool              ins;
  bool         *isRange;
  bool      needPruning;
  int           ilength;
  int           domsize;
  int           sacvals;
  Solver        *solver;
  MistralSet *SACdomain;
  //@}

  /**@name Constructors*/
  //@{
  WeighterSAC( Solver *s );
  virtual ~WeighterSAC( );
  //@}  

  /**@name Utils*/
  //@{ 
  bool isNotSac();

  virtual void notifyFailure( Constraint *con ) ;
  virtual void notifyChoice() ; 
  virtual int getType() { return SAC; }
  //@}  
};


class VarSelectorDomain 
{
public: 

  /**@name Parameters*/
  //@{ 
  int d_;
  //@}  

  /**@name Utils*/
  //@{ 
  inline bool operator<( VarSelectorDomain& x ) const { return d_ < x.d_; }
  inline void operator=( VarSelectorDomain& x ) { d_ = x.d_; }
  inline void operator=( VariableInt    *x ) { d_ = x->domsize(); }
  //@}  
};

class VarSelectorDomainMinVal 
{
public: 

  /**@name Parameters*/
  //@{ 
  int d_;
  int v_;
  //@}  

  /**@name Utils*/
  //@{ 
  inline bool operator<( VarSelectorDomainMinVal& x ) const { return d_ < x.d_  || ( d_ <= x.d_ && v_ < x.v_) ;}
  inline void operator=( VarSelectorDomainMinVal& x ) { d_ = x.d_; v_ = x.v_; }
  inline void operator=( VariableInt    *x ) { d_ = x->domsize(); v_ = x->min(); }
  //@}  
};

class VarSelectorDegree 
{
public: 

  /**@name Parameters*/
  //@{ 
  int d_;
  //@}    

  /**@name Utils*/
  //@{ 
  inline bool operator<( VarSelectorDegree& x ) const { return d_ > x.d_; }
  inline void operator=( VarSelectorDegree& x ) { d_ = x.d_; }
  inline void operator=( VariableInt    *x ) { d_ = x->degree; }
  //@}  
};

class VarSelectorDomainDegree 
{
public: 

  /**@name Parameters*/
  //@{ 
  int deg_;
  int dom_;
  //@}  
  
  /**@name Utils*/
  //@{ 
  inline bool operator<( VarSelectorDomainDegree& x ) const { 
    return dom_ < x.dom_  || ( dom_ <= x.dom_ && deg_ > x.deg_) ; }   
  inline void operator=( VarSelectorDomainDegree& x ) { deg_ = x.deg_; dom_ = x.dom_; }
  inline void operator=( VariableInt    *x ) { deg_ = x->degree; dom_ = x->domsize(); }
  //@}  
};

class VarSelectorDomainOverDegree 
{
public: 

  /**@name Parameters*/
  //@{ 
  int deg_;
  int dom_;
  //@}  
  
  /**@name Utils*/
  //@{ 
  inline bool operator<( VarSelectorDomainOverDegree& x ) const { return dom_ * x.deg_ < x.dom_ * deg_ ; }
  inline void operator=( VarSelectorDomainOverDegree& x ) { deg_ = x.deg_; dom_ = x.dom_; }
  inline void operator=( VariableInt    *x ) { deg_ = x->degree; dom_ = x->domsize(); }
  //@}  
};

class VarSelectorWeight 
{
public: 

  /**@name Parameters*/
  //@{ 
  int w_;
  //@}  

  /**@name Utils*/
  //@{ 
  inline bool operator<( VarSelectorWeight& x ) const { return w_ > x.w_; }
  inline void operator=( VarSelectorWeight& x ) { w_ = x.w_; }
  inline void operator=( VariableInt    *x ) { w_ = x->weight; }
  //@}  
};

class VarSelectorDomainOverWeight 
{
public: 

  /**@name Parameters*/
  //@{ 
  int wgt_;
  int dom_;
  //@}  
  
  /**@name Utils*/
  //@{ 
  inline bool operator<( VarSelectorDomainOverWeight& x ) const { return dom_ * x.wgt_ < x.dom_ * wgt_ ; }
  inline void operator=( VarSelectorDomainOverWeight& x ) { wgt_ = x.wgt_; dom_ = x.dom_; }
  inline void operator=( VariableInt    *x ) { 
    wgt_ = x->weight; 
    dom_ = x->domsize(); 
    //cout << dom_ << "/" << wgt_ ;
  }
  //@}  
};


class VarSelectorImpact 
{
public: 

  /**@name Parameters*/
  //@{ 
  double impact_;
  double **decision_impact;
  //@}  
  
  /**@name Utils*/
  //@{ 
  inline bool operator<( VarSelectorImpact& x ) const { return impact_ < x.impact_; }
  inline void operator=( VarSelectorImpact& x ) { impact_ = x.impact_; }
  inline void operator=( VariableInt    *x ) 
  { 
/*     x->printshort( cout ); */
/*     cout << "\t"; */
    impact_ = 0.0;
    int idx = x->id;
    int v = x->first();
    do {
      impact_ += decision_impact[idx][v]; 
      //      cout << setw(4) << setprecision(2) << decision_impact[idx][v];
    } while( x->setNext(v) );    
    //    cout << endl;
  }
  //@}  
};

class VarSelectorImpactOverDegree 
{
public: 

  /**@name Parameters*/
  //@{ 
  int deg_;
  double impact_;
  double **decision_impact;
  //@}  
  
  /**@name Utils*/
  //@{ 
  inline bool operator<( VarSelectorImpactOverDegree& x ) const { return impact_ * x.deg_ < x.impact_ * deg_; }
  inline void operator=( VarSelectorImpactOverDegree& x ) { impact_ = x.impact_; deg_ = x.deg_; }
  inline void operator=( VariableInt    *x ) 
  { 
    impact_ = 0.0;
    int idx = x->id;
    int v = x->first();
    do impact_ += decision_impact[idx][v]; 
    while( x->setNext(v) );
    deg_ = x->degree;
  }
  //@}  
};

class VarSelectorImpactOverWeight 
{
public: 

  /**@name Parameters*/
  //@{ 
  int wgt_;
  double impact_;
  double **decision_impact;
  //@}  
  
  /**@name Utils*/
  //@{ 
  inline bool operator<( VarSelectorImpactOverWeight& x ) const { return impact_ * x.wgt_ < x.impact_ * wgt_; }
  inline void operator=( VarSelectorImpactOverWeight& x ) { impact_ = x.impact_; wgt_ = x.wgt_; }
  inline void operator=( VariableInt    *x ) 
  { 
    impact_ = 0.0;
    int idx = x->id;
    int v = x->first();
    do impact_ += decision_impact[idx][v]; 
    while( x->setNext(v) );
    wgt_ = x->weight;
  }
  //@}
};

class VarSelectorNeighborDD 
{
 public: 
  
  /**@name Parameters*/
  //@{ 
  double val_;
  VariableInt **scp;
  int j;
  //@}  

  /**@name Utils*/
  //@{ 
  inline bool operator<( VarSelectorNeighborDD& x ) const { return val_ < x.val_ ; }
  inline void operator=( VarSelectorNeighborDD& x ) { val_ = x.val_; }
  inline void operator=( VariableInt    *x ) 
    { 
      val_ = 0.0;
      //MistralNode<Constraint*> *nd = &(x->valueTrigger.head);
      MistralNode<Constraint*> *nd = x->constraintsOnValue();
      while( nextNode(nd) ) {
	scp = nd->elt->scope;
	j=nd->elt->arity;
	while( j-- )
	  if( scp[j]->isLinked() && 
	      scp[j]->degree && 
	      scp[j] != x )
	    val_ += ( (double)(scp[j]->domsize()) / 
		      (double)(scp[j]->degree) );
      }
      val_ = ((double)(x->domsize() + val_) /
	      (double)(x->degree*x->degree));
    }
  //@}
};


class ValSelector
{
 public:
  
  VariableInt *_X;
  
 ValSelector( VariableInt *x ) : _X(x) {}

  /**@name Utils*/
  //@{ 
  virtual int getBest() = 0;
  virtual void getOrder( int* dtv, int* vtd ) {
    int i=0, v=_X->first();
    //    cout << endl;
    do {
      //      cout << setw(2) << v << " ";
      dtv[i] = v;
      vtd[v] = i;
      ++i;
    } while( _X->setNext(v) );
    //    cout << endl;
  }
  virtual void left() = 0;
  virtual bool right() = 0;
  //@}  

  virtual void printLeft(std::ostream& o) const {}
  virtual void printRight(std::ostream& o) const {}
};


class ValSelectorMin : public ValSelector
{
public:

 ValSelectorMin( VariableInt *x ) : ValSelector(x) {}

  /**@name Utils*/
  //@{ 
  inline int getBest() { return _X->min(); }
  inline void left() { _X->setDomain( _X->min() ); }
  inline bool right() { return _X->remove( _X->min() ); }
  //@}  

  virtual void printLeft(std::ostream& o) const { 
    _X->printshort(o);
    cout << " == " << _X->min();
  }
  virtual void printRight(std::ostream& o) const {
    _X->printshort(o);
    cout << " =/= " << _X->min();
  }
};

class ValSelectorMax : public ValSelector
{
public:

 ValSelectorMax( VariableInt *x ) : ValSelector(x) {}

  /**@name Utils*/
  //@{ 
  inline int getBest() { return _X->max(); }
  inline void left() { _X->setDomain( _X->max() ); }
  inline bool right() { return _X->remove( _X->max() ); }
  //@}  

  virtual void printLeft(std::ostream& o) const { 
    _X->printshort(o);
    cout << " == " << _X->max();
  }
  virtual void printRight(std::ostream& o) const {
    _X->printshort(o);
    cout << " =/= " << _X->max();
  }
};

/* class ValSelectorRand : public ValSelector */
/* { */
/* public: */

/*  ValSelectorRand( VariableInt *x ) : ValSelector(x) {} */

/*   int val; */

/*   /\**@name Utils*\/ */
/*   //@{  */
/*   inline void left() { _X->setDomain( _X->max() ); } */
/*   inline bool right() { _X->remove( _X->max() ); } */
/*   //@}   */
/* }; */

class ValSelectorSplit : public ValSelector
{
public:

 ValSelectorSplit( VariableInt *x ) : ValSelector(x) {}

  /**@name Utils*/
  //@{ 
  inline int getBest() { return _X->min(); }
  inline void left() { _X->setMax( ((_X->max() + _X->min()) >> 1) ); }
  inline bool right() { return _X->setMin( ((_X->max() + _X->min()) >> 1)+1 ); }
  //@}  

  virtual void printLeft(std::ostream& o) const { 
    _X->printshort(o);
    cout << " <= " << ((_X->max() + _X->min()) >> 1);
  }
  virtual void printRight(std::ostream& o) const {
    _X->printshort(o);
    cout << " > " << ((_X->max() + _X->min()) >> 1);
  }
};

class ValSelectorWeight : public ValSelector
{
 public:
  
  double *weight;
  int val;

 ValSelectorWeight( VariableInt *x, double *w ) : ValSelector(x) { weight = w; }

  /**@name Utils*/
  //@{ 
  inline int getBest() 
  { 
    int v = val = _X->first();
    double w = weight[val];
    while( _X->setNext(v) )
      if( weight[v] > w )
	{
	  w = weight[v];
	  val = v;
	}
    return val;
  }
  virtual void getOrder( int* dtv, int* vtd ) {
    ValSelector::getOrder( dtv, vtd );

    int i, j, aux, n=_X->domsize();

/*     _X->print(cout); */
/*     cout << endl; */
/*     for(i=0; i<n; ++i) { */
/*       cout << setw(2) << dtv[i] << " "; */
/*     } */
/*     cout << endl; */
/*     for(i=0; i<n; ++i) { */
/*       cout << setw(2) << vtd[i] << " "; */
/*     } */
/*     cout << endl; */


    for(i=1; i<n; ++i) {
      j = i;
      while( j && weight[dtv[j]] > weight[dtv[j-1]] ) {
	aux = dtv[j] ;
	dtv[j] = dtv[j-1] ;
	dtv[j-1] = aux;
	//vtd[dtv[j]] = j;
	//vtd[dtv[j-1]] = j-1;
	--j;
      }
    }
    for(i=0; i<n; ++i) {
      vtd[dtv[i]] = i;

      //      cout << i << " " << dtv[i] << " " << vtd[dtv[i]] << endl;
      //      assert( vtd[dtv[i]] == i );
    }

/*     _X->print(cout); */
/*     cout << " " << endl; */
/*     for(i=0; i<n; ++i) { */
/*       //cout << dtv[i] << ":" << weight[dtv[i]] << " "; */
/*       //cout.flush(); */
/*       cout << i << " " << dtv[i] << " " << vtd[dtv[i]] << endl; */

/*       assert( vtd[dtv[i]] == i ); */
/*     } */
/*     cout << endl; */
  }
  inline int getBestConst() const 
  { 
    int v = _X->first();
    int v_ = v;
    double w = weight[val];
    while( _X->setNext(v) )
      if( weight[v] > w )
	{
	  w = weight[v];
	  v_ = v;
	}
    return v_;
  }
  inline void left() 
  { 
    _X->setDomain( getBest() );    
  }
  inline bool right() { return _X->remove( val ); }

  virtual void printLeft(std::ostream& o) const { 
    _X->printshort(o);
    cout << " == " << getBestConst();
  }
  virtual void printRight(std::ostream& o) const {
    _X->printshort(o);
    cout << " =/= " << getBestConst();
  }
  //@}  

};

class ValSelectorSac : public ValSelector
{
 public:
  
  MistralSet *visited;
  int val;

 ValSelectorSac( VariableInt *x, MistralSet *s ) : ValSelector(x) { visited = s; }

  /**@name Utils*/
  //@{ 
  inline int getBest() 
  {
    int j = val = _X->first();
    bool isin = true;
    while( visited->member( j ) && ( isin = _X->setNext( j ) ) );
    if( isin )
      val = j;
    return val;
  }
  inline void left() 
  { 
    _X->setDomain( getBest() );
  }
  inline bool right() { return _X->remove( val ); }
  virtual void printLeft(std::ostream& o) const { 
    _X->printshort(o);
    
    int j = _X->first();
    bool isin = true;
    while( visited->member( j ) && ( isin = _X->setNext( j ) ) );

    cout << " == " << j;
    cout << " : ";
    visited->print( cout );
  }
  virtual void printRight(std::ostream& o) const {
    _X->printshort(o);
    cout << " =/= " << val;
  }
  //@}  
};

class ValSelectorSacRange : public ValSelector
{
 public:
  
  MistralSet *visited;
  int val;

 ValSelectorSacRange( VariableInt *x, MistralSet *s ) : ValSelector(x) { visited = s; }

  /**@name Utils*/
  //@{ 
  inline int getBest() { return _X->min(); }
  inline void left() 
  { 
    val = ( visited->member(1) );

    //cout << "pivot = " << (((_X->max() + _X->min()) >> 1) + ( val ? 0 : 1 )) << endl;


    if( val ) 
      //_X->setMax( ((_X->max() + _X->min()) >> 1) );
      _X->setMax( _X->min() );
    else
      _X->setMin( _X->max() );
      //_X->setMin( ((_X->max() + _X->min()) >> 1)+1 );

  }
  inline bool right() 
  {
    bool c = true;
    if( val )
      _X->setMin( _X->min()+1 );
      //c = _X->setMin( ((_X->max() + _X->min()) >> 1)+1 );
    else
      _X->setMax( _X->max()-1 );
      //c = _X->setMax( ((_X->max() + _X->min()) >> 1) );
    return c;
  }
  virtual void printLeft(std::ostream& o) const { 
    _X->printshort(o);
    int v = ( visited->member(1) );
    if( v )
      cout << " >= " << _X->max();
    else 
      cout << " <= " << _X->min();
    cout << " : ";
    visited->print( cout );
  }
  virtual void printRight(std::ostream& o) const {
    _X->printshort(o);
    int v = ( visited->member(1) );
    if( v )
      cout << " < " << _X->max();
    else 
      cout << " > " << _X->min();
  }
  //@}  
};


/**********************************************
 * DVOs Dynamic Variable Ordering
/*! \class DVO
  \brief  Representation of Dynamic Variable Ordering.

  "Future variables", that is, those that are not
  yet assigned are stored between the pointers 'first'
  and 'last' The function select() return the  
  next variable to branch on.
*/
class  DVO {
 public:

  /**@name Parameters*/
  //@{ 
  /// A reference to the variables in lex static order
  VariableInt**& variables;
  /// A reference to the beginning of the array of future variables
  VariableInt**& first;
  /// A reference to the end of the array of future variables
  VariableInt**&  last;
  /// The list of decisions
  VariableInt**& decision;
  /// The number of decisions;
  int& level;
  /// Structure used to learn weights
  //Weighter *learner;
  //@}

  /**@name Constructors*/
  //@{
  DVO(Solver*);
  virtual ~DVO() {}
  //@}  

  /**@name Utils*/
  //@{ 
  /*!
    The variable selection method.
    Returns the next variable to pick,
    a value may also be selected by assigning the parameter, if no
    value is chosen then the default value ordering will be used
    (depending on the Variable's implementation).
  */
  virtual VariableInt* select() = 0;
  //@} 
};



/**********************************************
 * Generic heuristic
 **********************************************/

template <class T>
class GenericDVO : public DVO
{
 public: 

  /**@name Parameters*/
  //@{ 
  T best;
  T current;
  //@}

  /**@name Constructors*/
  //@{
 GenericDVO(Solver* s) : DVO(s) {}
  //@}

  /**@name Utils*/
  //@{ 
  inline VariableInt* select()
  {
    VariableInt *var=*first, **var_iterator;
    best = var; 
    for( var_iterator = first+1; var_iterator != last; ++var_iterator ) 
      {  
	current = (*var_iterator);
	if( current < best ) 
	  {
	    best = current;
	    var = (*var_iterator);
	  } 
      }
    
    return var;
  }
  //@}
};



/*! \class VariableOrdering 
  \brief  Wrapper for DVO's
*/
class VariableOrdering {
 public:

  /**@name Utils*/
  //@{
  virtual DVO* extract( Solver* ) = 0 ;
  //@}
};


/**********************************************
 * No Ordering
 **********************************************/

/*! \class DVONoOrder
  \brief  No variable ordering

  Variables are explored in no order 
*/
class DVONoOrder : public DVO {
 public:
  /**@name Constructors*/
  //@{
 DVONoOrder(Solver* s) : DVO(s) {}
  //@} 

  /**@name Utils*/
  //@{    
  /**!
      The variable selection method.
      Return the first unassigned variable
  */
  VariableInt* select();
  //@} 
};

/*! \class No
  \brief  Wrapper for DVONo
*/
class NoOrder : public VariableOrdering {
 public:

  /**@name Utils*/
  //@{  
  DVO* extract( Solver* s )
  {
    return new DVONoOrder(s);
  }
  //@}
};


/**********************************************
 * Lexicographic Ordering
 **********************************************/

/*! \class DVOLexicographic
  \brief  Lexico variable ordering

  Variables are explored in lexicographic order 
*/
class DVOLexicographic : public DVO {
 public:
  /**@name Constructors*/
  //@{
 DVOLexicographic(Solver* s) : DVO(s) {}
  //@} 

  /**@name Utils*/
  //@{    
  /**!
      The variable selection method.
      Return the first unassigned variable
      in lexico order
  */
  VariableInt* select();
  //@} 
};

/*! \class Lexicographic
  \brief  Wrapper for DVOLexicographic
*/
class Lexicographic : public VariableOrdering {
 public:

  /**@name Utils*/
  //@{  
  DVO* extract( Solver* s )
  {
    return new DVOLexicographic(s);
  }
  //@}
};


/**********************************************
 * Random Ordering
 **********************************************/

/*! \class DVORandom
  \brief  Random variable ordering

  Variables are explored in lexicographic order
*/
class DVORandom : public DVO {
 public:
  /**@name Constructors*/
  //@{
  DVORandom(Solver* s);//, unsigned int seed) ;
  //@}

  /**@name Utils*/
  //@{
  /**!
      The variable selection method.
      Return the first unassigned variable
  */
  virtual VariableInt* select();
  //@}
};

/*! \class Random
  \brief  Wrapper for DVORandom
*/
class Random : public VariableOrdering {
 public:

  /**@name Parameters*/
  //@{ 
  //unsigned int seed;
  //@}

/*   /\**@name Constructors*\/ */
/*   //@{ */
/*   Random( unsigned int graine=11041979 ) */
/*     { */
/*       seed = graine; */
/*     } */
/*   //@} */

  /**@name Utils*/
  //@{
  DVO* extract( Solver* s )
  {
    return new DVORandom(s);//, seed);
  }
  //@}
};


/**********************************************
 * Min Domain
 **********************************************/

/*! \class MinDomain
  \brief  Wrapper for DVOMinDomain
*/
class MinDomain : public VariableOrdering {
 public:
  
  /**@name Utils*/
  //@{
  DVO* extract( Solver* s )
  {
    return new GenericDVO<VarSelectorDomain>(s);
  }
  //@}
};


/**********************************************
 * Min Domain, Min Value
 **********************************************/

/*! \class MinDomMin
  \brief  Wrapper for DVOMinDomMin
*/
class MinDomMin : public VariableOrdering {
 public:
  
  /**@name Utils*/
  //@{
  DVO* extract( Solver* s )
  {
    return new GenericDVO<VarSelectorDomainMinVal>(s);
  }
  //@}
};


/**********************************************
 * Max Degree
 **********************************************/


/*! \class MaxDegree
  \brief  Wrapper for DVOMaxDegree
*/
class MaxDegree : public VariableOrdering {
 public:
  
  /**@name Utils*/
  //@{
  DVO* extract( Solver* s )
  {
    return new GenericDVO<VarSelectorDegree>(s);
  }
  //@}
};


/**********************************************
 * Min Domain , Max Degree
 **********************************************/

/*! \class MinDomMaxDeg
  \brief  Wrapper for DVOMinDomMaxDeg
*/
class MinDomMaxDeg : public VariableOrdering {
 public:
  
  /**@name Utils*/
  //@{
  DVO* extract( Solver* s )
  {
    return new GenericDVO<VarSelectorDomainDegree>(s);
  }
  //@}
};


/**********************************************
 * Min domain/degree
 **********************************************/

/*! \class DomOverDeg
  \brief  Wrapper for DVODomOverDeg
*/
class DomOverDeg : public VariableOrdering {
 public:
  
  /**@name Utils*/
  //@{
  DVO* extract( Solver* s )
  {
    return new GenericDVO<VarSelectorDomainOverDegree>(s);
  }
  //@}
};


/**********************************************
 * Min Domain/degree on neighbors
 **********************************************/

/*! \class Neighbor
  \brief  Wrapper for DVONeighbor
*/
class Neighbor : public VariableOrdering {
 public:
  
  /**@name Utils*/
  //@{
  DVO* extract( Solver* s )
  {
    return new GenericDVO<VarSelectorNeighborDD>(s);
  }
  //@}
};



/**********************************************
 * Min domain/weighted degree (level)
 **********************************************/

/*! \class DomOverWLDeg
  \brief  Wrapper for DVODomOverWLDeg
*/
class DomOverWLDeg : public VariableOrdering {
 public:

  /**@name Utils*/
  //@{
  DVO* extract( Solver* s );
  //@}
};


/**********************************************
 * Min domain/weighted degree
 **********************************************/

/*! \class DomOverWDeg
  \brief  Wrapper for DVODomOverWDeg
*/
class DomOverWDeg : public VariableOrdering {
 public:

  /**@name Utils*/
  //@{
  DVO* extract( Solver* s );
  //@}
};


/**********************************************
 * Max impact
 **********************************************/

/*! \class Impact
  \brief  Wrapper for DVOImpact
*/
class Impact : public VariableOrdering {
 public:
  
  /**@name Utils*/
  //@{
  DVO* extract( Solver* s );
  //@}
};


/**********************************************
 * Max impact * degree
 **********************************************/

/*! \class ImpactOverDeg
  \brief  Wrapper for DVOImpactOverDeg
*/
class ImpactOverDeg : public VariableOrdering {
 public:

  /**@name Utils*/
  //@{
  DVO* extract( Solver* s );
  //@}
};


/**********************************************
 * Max impact * weighted degree
 **********************************************/

/*! \class ImpactOverWDeg
  \brief  Wrapper for DVOImpactOverWDeg
*/
class ImpactOverWDeg : public VariableOrdering {
 public:

  /**@name Utils*/
  //@{
  DVO* extract( Solver* s );
  //@}
};


/**********************************************
 * Max impact * weighted degree (level)
 **********************************************/

/*! \class ImpactOverWLDeg
  \brief  Wrapper for DVOImpactOverWLDeg
*/
class ImpactOverWLDeg : public VariableOrdering {
 public:

  /**@name Utils*/
  //@{  
  DVO* extract( Solver* s );
  //@}
};


/**********************************************
 * Probing heuristic
 **********************************************/

/*! \class Probe
  \brief  Wrapper for Probing heuristic
*/
class Probe : public VariableOrdering {
 public:

  /**@name Parameters*/
  //@{ 
  //unsigned int seed;  
  int ltype;
  int htype;
  //@}

  /**@name Constructors*/
  //@{
  Probe( /* unsigned int graine=11041979, */ int lt=Weighter::WDG, int ht=0 )
    {
      //seed = graine;
      ltype = lt;
      htype = ht;
    }

  Probe( /* unsigned int graine=11041979, */ bool ulw, bool uip, int ht )
    {
      //seed = graine;
      htype = ht;

      ltype = Weighter::WDG;
      if( ulw ) 
	ltype |= Weighter::WLD;
      if( uip )
	ltype |= Weighter::IPT;
    }
  //@}

  /**@name Utils*/
  //@{
  virtual DVO* extract( Solver* s );
  //@}
};


/**********************************************
 * Singleton Arc Consistency
 **********************************************/

/*! \class DVOSingletonAC
  \brief SAC variable ordering

  Choose first the variables and values not
  yet SAC.
*/
class DVOSingletonAC : public DVO {
 public:

  /**@name Parameters*/
  //@{ 
  MistralSet *SACdomain;
  bool *isRange;
  //  unsigned int *sacsize;
  //@}

  /**@name Constructors*/
  //@{
  DVOSingletonAC(Solver*, WeighterSAC*) ;//: DVO(s) {}
  //@}

  /**@name Utils*/
  //@{
  VariableInt* select();
  //@}
};


/**********************************************
 * Objective Functions 
 **********************************************/

/*! \class ObjectiveFunction 
  \brief  Abstract objective function class

  Always minimise a criterion. The upper bound is 
  updated when a solution is found. 
*/
class ObjectiveFunction {
 public:
  /**@name Parameters*/
  //@{ 
  /// The current upper bound on the objective to minimise
  int upper_bound;
  //@}

  /**@name Utils*/
  //@{   
  /// Return an optimistic evaluation of the best objective value currently achievable
  virtual int score() = 0;
  /// Called by the solver whenever a solution is found 
  virtual int update();
  //@}
};


/*! \class MaximiseVar
  \brief  Maximise a variable.
*/
class MaximiseVar : public ObjectiveFunction {

 public:
  /**@name Parameters*/
  //@{ 
  /// The unary constraint used to bound search
  UnaryConstraintMore *umore;
  /// The best achievable value, used to terminate early
  int maxX;
  //@}  

  /**@name Constructors*/
  //@{ 
  MaximiseVar(Solver*, VariableInt*);
  virtual ~MaximiseVar();
  //@}  

  /**@name Utils*/
  //@{ 
  /*!
    returns an optimistic estimate of the bound
  */
  virtual int score();
  /*!
     The score of the solution is computed, then
     the unary constraint is updated consequently.
  */
  virtual int update();
  //@}  
};


/*! \class MinimiseVar
  \brief  Minimise a variable.
*/
class MinimiseVar : public ObjectiveFunction {

 public:
  /**@name Parameters*/
  //@{ 
  /// The unary constraint used to bound search
  UnaryConstraintLess *uless;
  /// The best achievable value, used to terminate early
  int minX;
  //@}  

  /**@name Constructors*/
  //@{ 
  MinimiseVar(Solver*, VariableInt*);
  virtual ~MinimiseVar();
  //@}  

  /**@name Utils*/
  //@{ 
  /*!
    returns an optimistic estimate of the bound
  */
  virtual int score();
  /*!
     The score of the solution is computed, then
     the unary constraint is updated consequently.
  */
  virtual int update();
  //@}  
};



#endif // _DVO_H
