
/*
  Mistral is a constraint satisfaction and optimisation library
  Copyright (C) 2003-2005  Emmanuel Hebrard
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  The author can be contacted electronically at 
  ehebrard@cse.unsw.edu.au.
*/



/** \file mod.h
    \brief Header for the variables and constraints wrappers and build objects.
*/

/**********************************************
 * Wrappers & Build object
 **********************************************/

#ifndef _Wrapper_H
#define _Wrapper_H

#include <dvo.h>


class CSP;
class VarArray;

/**********************************************
 * MistralGacStack
 **********************************************/

/// Stack of Variable for computing the GAC closure.
template <class T>
class MistralGacStack {
 public:
  T *MistralGacStack_;
  int        *trigger;
  int            size;
  
  /**@name Constructors*/
  //@{
  ///
  MistralGacStack()
    {
      size=0;
      MistralGacStack_=NULL;
      trigger=NULL;
    }
  ///
  MistralGacStack(int n)
    {
      initStack(n);
    }
  ///
  ~MistralGacStack()
    {
      delete [] trigger;
      delete [] MistralGacStack_;
    }
  ///
  void initStack(int n)
  { 
    trigger = new int[n];
    std::fill( trigger, trigger+n, 0 );
    MistralGacStack_ = new T[n];
    size = 0;
  }
  //@}
  
  void push(T v, int event)
  {
    if( !trigger[v->id] )
      MistralGacStack_[size++] = v;    
    trigger[v->id] |= event;
  }

  void push(T v, VariableInt* x)
  {
    if( !trigger[v->id] )
      MistralGacStack_[size++] = v;    
    trigger[v->id] = x->id;
  }


  T pop( int& event )
  {
    T last = MistralGacStack_[--size];
    event = trigger[last->id];
    trigger[last->id] = 0;
    return last;
  }

  T back( int& event )
  {
    T last = MistralGacStack_[size-1];
    event = trigger[last->id];
    return last;
  }

  bool empty() const
  {
    return (!size);
  }

  void clear()
  {
    while( size )
      trigger[MistralGacStack_[--size]->id] = 0; 
  }
};


////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
//////                                                                            //////
//////  There are 3 layers:                                                       //////
//////  1/ Wrappers are simply syntactic sugar that allow to                      //////
//////  a/ hide 'new' operations in models                                        //////
//////  b/ allow to overload operator (+, [], ...)                                //////
//////  2/ Build objects are 'temporary' variables and/or constraints.             //////
//////  They contain only the necessary information (domain, scope, ...)          //////
//////  and the method 'build' actually create the real object                    //////
//////  used during search. They allow to do some re-modelling                    //////
//////  before the solving phase.                                                 //////
//////  For instance, if a variable is an interval and is                         //////
//////  constrained only by BC constraints, the RangeVar                          //////
//////  implementation will be used instead of BitVar.                            //////
//////   + BuildObject stand for variables. They keep a representation            //////
//////     of their domain, as well as a number of informations in the            //////
//////     struct state.                                                          //////
//////   + BuildObjectConstraints stand for constraints' *relation*,              //////
//////     that is, at most one of each is created, no matter how many            //////
//////     such constraints are added to the model.                               //////
//////   + BuildObjectPredicate stand for constraint scope. They ihnerit          //////
//////     from BuildObject and therefore can also represent variables            //////
//////     themselves.                                                            //////
//////     - When a constraint, or variable wrapper is declared,                  //////
//////       the corresponding BuildObject is created                             //////
//////     - When a (tree of) objects are added to the model                      //////
//////       the top-level one is marked as constraint only, whilst               //////
//////       all descendants are marked as variables.                             //////


//////     - During the prebuilding phase, two things happen                      //////
//////       * The method propagateInfo() is called on all                        //////
//////         top-level objects, which recursively call their                    //////
//////         children.                                                          //////
//////         . Constraints that become incorrect when reasoning on bounds       //////
//////         . Constraints that can profit from holes                           //////
//////       * The method prebuild() is called on all objects from                //////
//////         top to bottom to perform rewriting when needed.                    //////

/*
+ transformations: (limited propagation from top to bottom + rewriting thereof)
  > general:
    - if all arguments are constants, delete the declaration (check consistency?)
      unsetRel(): the relation part should not be build, because it is 
                  a totology
      deReference() all arguments. Each argument loses a reference 
  > Not
    - if set to true or at toplevel (!isVar()):
      unsetRel(), set argument to false (0).
    - if set to false:
      unsetRel(), set argument to false (0).
  > And 
    - if set to true or at toplevel (!isVar()):
      unsetRel(), set arguments to true (1).
  > Or
    - if set to false:
      unsetRel(), set arguments to false (0).
  > Add 
    - if an argument is  a constant:
      turn into an offset predicate
  > Sub
    - if an argument is  a constant:
      turn into an offset predicate
  > EqualConstant
    - if set to true or at toplevel (!isVar()):
      unsetRel(), set argument to val.
    - if set to false:
      unsetRel(), remove val in arg's domain.
  > NotEqualConstant
    - if set to true or at toplevel (!isVar()):
      unsetRel(), remove val in arg's domain.
    - if set to false:
      unsetRel(), set argument to val.
  > Equal
    - if set to true or at toplevel (!isVar()):
      unsetRel(), merge arguments.
    - if an argument is  a constant:
      turn into an EqualConstant predicate
  > NotEqual
    - if set to true or at toplevel (!isVar()), and if one of the agrs is a constant
      unsetRel(), and remove the constant from the other agr's domain.
    - if an argument is  a constant:
      turn into an NotEqualConstant predicate
  > SumEq
    - absorb constants.
  > Sum
    - if the last arg is a constant and is not referenced otherwise:
      turns into a SumEq constraint
    - absorb constants then according to the arity,
      change to offset/factor/add/sub/sumeq
  > if N is constant:
    - turns into an equal predicate

*/

//////                                                                            //////
//////  3/ Actual variables and constraints (csp.h / var.h / con.h)               //////
//////                                                                            //////
////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
/* 

   propagate info

   * needed because when rewriting, the state of an object can change,
   the change must be propagated downstream.


   mechanism to decide if the representation should be predicate 
   or constraint.
   * if not at top level -> always variable
   * some object are variables even at the top level 
     (actual variables, element, sum, etc)
   * top level known when adding, but it can change 
     (for instance with a AND constraint)
 */


class BuildObjectConstraint;
/**********************************************
 * ConstraintStore
 **********************************************/
/*! \class ConstraintStore
  \brief A pool of constraint build objects

  We need only one instance of each constraint
  buld object.
*/
class ConstraintStore
{
 public: 
  // LOGIC
  static const int AND         =  2 ;
  static const int OR          = 21 ;
  static const int NOT         = 19 ;

  // ARITHMETIC
  static const int SUB         = 31 ;
  static const int ADD         = 32 ;
  static const int ABS         =  0 ;
  static const int DIV         =  5 ;
  static const int MOD         = 16 ;
  static const int MODCONSTANT = 38 ;
  static const int MUL         = 17 ;
  static const int NEG         = 18 ;
  static const int POW         = 22 ;
  static const int FACTOR      =  9 ;
  static const int OFFSET      = 20 ;

  // RELATION
  static const int DISJUNCTIVE =  4 ;
  static const int EQUAL       =  8 ;
  static const int EQCONSTANT  = 36 ;
  static const int NEQ         = 30 ;
  static const int NEQCONSTANT = 37 ;
  static const int PRECEDENCE  = 23 ;
  static const int LOWERBOUND  = 34 ;
  static const int UPPERBOUND  = 35 ;
  static const int NEQW        = 27 ;

  // GLOBAL
  static const int ALLDIFF     =  1 ;
  static const int CLAUSE      =  3 ;
  static const int EDGEFINDER  =  6 ;
  static const int ELEMENT     =  7 ;
  static const int GCC         = 10 ;
  static const int LEX         = 11 ;
  static const int MAX         = 12 ;
  static const int MIN         = 13 ;
  static const int MAXOCCUR    = 14 ;
  static const int MINOCCUR    = 15 ;
  static const int SUM         = 24 ;
  static const int SUMEQ       = 33 ;
  static const int MAXDIFF     = 25 ;
  static const int MINDIFF     = 26 ;
  static const int SLIDE       = 28 ;
  static const int REGULAR     = 29 ;
  static const int IFTHENELSE  = 39 ;

  // COUNT
  static const int NUMCONS     = 40 ;


 private:
  MistralStack<BuildObjectConstraint *> store;
  BuildObjectConstraint *getConstraint(const int c);

 public:
  ConstraintStore() ;
  virtual ~ConstraintStore() ;
  BuildObjectConstraint *operator[]( const int c ) ;
  int addConstraint( BuildObjectConstraint *c ) ;

};


static ConstraintStore ENV;

/********************************************************
 *
 *  BuildObject for Variables, Constraints 
 *  and Predicates.
 *
 ********************************************************/

class BuildObjectPredicate;
/**********************************************
 * BuildObject
 **********************************************/
class BuildObject
{

 protected:
 
  CSP *model;
  int size_;
  int min_;
  int max_;
  BuildObject *veqptr_;
  int refs;

  int state;


 public:

  MistralSet *domain_;  
  VariableInt *varptr_;

  BuildObject *getBuildObject();
  const BuildObject *getConstBuildObject() const ;

  MistralStack<BuildObjectPredicate*> parent;
  int id;

  int getVarId() const { return getConstBuildObject()->id; } 
  int getState() { return state; }
  void setState( int nstate ) { state = nstate; }

  static const int VRANGE=   1;
  static const int CRANGE=1024;
  static const int NOT   =   2;
  static const int NEG   =   4;
  static const int VAR   =   8;
  static const int CONST =  16;
  static const int AWV   =  32;
  static const int CONVEX=  64;
  static const int VLIST = 128;
  static const int CLIST =2048;
  static const int REL   = 256;
  static const int PAR   = 512;
  static const int NEQ   =4096;
  static const int FIXED =8192;

  static const int VALUETRIGGER =7;
  static const int RANGETRIGGER =3;
  static const int DOMAINTRIGGER=1;
  static const int LISTTRIGGER  =2;
  static const int CONTTRIGGER =4;
  void releaseEvent();
  void releaseEvent( const int );

  /*!
   * set to 1 iff it is an interval and consider only intervals
   */
  virtual bool isConvex() const { return (state & CONVEX); }
  virtual void setConvex() { if( !(state & CONVEX) ) state |= CONVEX; }
  virtual void unsetConvex() { if( state & CONVEX ) state &= (0xffffffff ^ CONVEX); }

  /*!
   * set to 1 iff it preserves intervals
   */
  virtual bool isCRange() const { return (state & CRANGE); }
  virtual void setCRange() { if( !(state & CRANGE) ) state |= CRANGE; }
  virtual void unsetCRange() { if(state & CRANGE) state &= (0xffffffff ^ CRANGE); }

  /*!
   * set to 1 iff it can be implemented as an interval
   */
  virtual bool isVRange() const { return (state & VRANGE); }
  virtual void setVRange() { if( !(state & VRANGE) ) state |= VRANGE; }
  virtual void unsetVRange() { if(state & VRANGE) state &= (0xffffffff ^ VRANGE); }

  /*!
   * set to 1 iff the predicate can absord a "not" predicate
   */  
  virtual bool isNot() const { return (state & NOT); }
  virtual void setNot() { if( !(state & NOT) ) state |= NOT; }
  virtual void unsetNot() { if( state & NOT ) state &= (0xffffffff ^ NOT); }

  /*!
   * set to 1 iff the predicate can absord a "-" predicate
   */
  virtual bool isNeg() const { return (state & NEG); }
  virtual void setNeg() { if( !(state & NEG) ) state |= NEG; }
  virtual void unsetNeg() { if( state & NEG ) state &= (0xffffffff ^ NEG); }

  /*!
   * set to 1 iff it has a domain, and will be a variable
   */
  virtual bool isVar() const { return (state & VAR); }
  virtual void setVar() { if( !(state & VAR) ) state |= VAR; }
  virtual void unsetVar() { if( (state & VAR) && !(state & AWV) ) state &= (0xffffffff ^ VAR); }

  /*!
   * set to 1 iff it has a domain, and will be a variable
   */
  virtual bool isAwv() const { return (state & AWV); }
  virtual void setAwv() { if( !(state & AWV) ) state |= (AWV | VAR); }
  virtual void unsetAwv() { if( state & AWV ) state &= (0xffffffff ^ AWV); }

  /*!
   * set to 1 iff it is a constant
   */
  virtual bool isConst() const { return (state & CONST); }
  virtual void setConst() { if( !(state & CONST) ) state |= CONST; }
  virtual void unsetConst() { if( state & CONST ) state &= (0xffffffff ^ CONST); }

  /*!
   * set to 1 iff it can be a list
   */
  virtual bool isVList() const { return (state & VLIST); }
  virtual void setVList() { if( !(state & VLIST) ) state |= VLIST; }
  virtual void unsetVList() { if( state & VLIST ) state &= (0xffffffff ^ VLIST); }

  /*!
   * set to 1 iff its args can be lists
   */
  virtual bool isCList() const { return (state & CLIST); }
  virtual void setCList() { if( !(state & CLIST) ) state |= CLIST; }
  virtual void unsetCList() { if( state & CLIST ) state &= (0xffffffff ^ CLIST); }

  /*!
   * set to 1 iff it is a constraint
   */
  virtual bool isRel() const { return (state & REL); }
  virtual void setRel() { if( !(state & REL) ) state |= REL; }
  virtual void unsetRel() { if( state & REL ) state &= (0xffffffff ^ REL); }

  /*!
   * set to 1 iff it can be replaced by a parameter 
   */
  virtual bool isPar() const { return (state & PAR); }
  virtual void setPar() { if( !(state & PAR) ) state |= PAR; }
  virtual void unsetPar() { if( state & PAR ) state &= (0xffffffff ^ PAR); }

  /*!
   * set to 1 iff it is a top level neq constraint 
   */
  virtual bool isNeq() const { return (state & NEQ); }
  virtual void setNeq() { if( !(state & NEQ) ) state |= NEQ; }
  virtual void unsetNeq() { if( state & NEQ ) state &= (0xffffffff ^ NEQ); }

  /*!
   * set to 1 iff it cannot be simplified anymore
   */
  virtual bool isFixed() const { return (state & FIXED); }
  virtual void setFixed() 
  { 
    if( !(state & FIXED) ) {
      state |= FIXED; 
      //setDomain( relation->computeValue( this ) );
    }
  }
  virtual void unsetFixed() { if( state & FIXED ) state &= (0xffffffff ^ FIXED); }


 public:

  //  inline void deReference() { if( !--refs ) unsetVar(); }
  inline void deReference() { --refs; }
  inline int isReferenced() { return refs; }

  virtual void add( CSP *model, const int l, const int inf );
  virtual bool preBuild() { return true; }


  inline bool notIn()
  {
    return (id < 0);
  }
  
  virtual bool isBoolean()
  {
    return (size() == 2 && min() == 0 && max() == 1);
  }

  virtual bool isFDomain()
  {
    return (size() > 1 && !isBoolean() && !isRange());
  }

  virtual bool isRange()
  {
    return (size() == (max() - min() + 1) && isVRange() /* && size_ > 64 */);
  }
  
  virtual bool isSearchable()
  {
    return (size_ > 1 && veqptr_ == NULL);
  }

  virtual bool isConstraint()
  {
    return false;
  }
  
  virtual VariableInt *_getVariable()
  {
    return varptr_;
  }
  
  virtual VariableInt *getVariable()
  {
    return getBuildObject()->_getVariable();
  }
  
  void initVar(const int l, const int u);
  BuildObject() ;
  BuildObject( const int lo, const int up ) ;
  BuildObject( const int* array, const int length ) ;
  BuildObject( const int* array, const int length, const int lo, const int up ) ;
  virtual ~BuildObject() 
    {
      if( domain_ ) {
	domain_->neg_words = 0;
	domain_->pos_words = 0;
	domain_->table = NULL;
	delete domain_;
      }
    }

  virtual void build( Solver *s );


  virtual int size() const ;
  virtual int domsize() const ;
  virtual int value() const ;
  virtual int first() const ;
  virtual int last() const ;
  virtual int min() const ;
  virtual int max() const ;
  virtual int minCapacity() const ;
  virtual int maxCapacity() const ;
  virtual int minPosAbs() const ;
  virtual int minNegAbs() const ;
  virtual bool remove(const int v);
  virtual bool setDomain(const int v);
  virtual bool setMin(const int l);
  virtual bool setMax(const int u);
  virtual bool setDomain(const MistralSet& s);
  virtual void setEqual(BuildObject *v); 
  virtual bool removeSet(const MistralSet& s);
  virtual bool removeRange(const int l, const int u);

  virtual int getNext(const int v) const ;
  virtual bool good(const int v) const ;
  virtual bool setNext(int& v) const ;

  virtual bool isRange() const ;
  virtual bool isGround() const ;

  virtual int _size() const ;
  virtual int _domsize() const ;
  virtual int _value() const ;
  virtual int _first() const ;
  virtual int _last() const ;
  virtual int _minCapacity() const ;
  virtual int _maxCapacity() const ;
  virtual int _min() const ;
  virtual int _max() const ;
  virtual int _minPosAbs() const ; 
  virtual int _minNegAbs() const ; 
  virtual bool _remove(const int v);
  virtual bool _setDomain(const int v);
  virtual bool _setMin(const int l);
  virtual bool _setMax(const int u);
  virtual bool _setDomain(const MistralSet& s);
  virtual void _setEqual(BuildObject *v); 
  virtual bool _removeSet(const MistralSet& s);
  virtual bool _removeRange(const int l, const int u);

  virtual int _getNext(const int v) const ;
  virtual bool _good(const int v) const ;
  virtual bool _setNext(int& v) const ;

  virtual bool _isRange() const ;
  virtual bool _isGround() const ;


  virtual bool setDomain(const int* a, const int l) { return true; }
  virtual bool setDomain(VariableInt* x) const { return true; }
  virtual void makeDecision(int&) {}
  virtual bool makeComplementary(int&) { return true; }


  virtual bool contain(const int v) const ;
  virtual bool intersect(const int lo, const int up) const { return true; } // TODO
  virtual bool included(const int lo, const int up) const { return true; } // TODO
  virtual bool intersect(const VariableInt* x) const { return true; } // TODO
  virtual bool included(const VariableInt* x) const { return true; } // TODO
  virtual bool intersect(const MistralSet* s) const { return true; } // TODO
  virtual bool included(const MistralSet* s) const { return true; } // TODO
  virtual bool intersect(const MistralSet& s) const { return true; } // TODO
  virtual bool intersect(const MistralSet& s, int& r) const { return true; } // TODO
  virtual bool included(const MistralSet& s) const { return true; } // TODO
  virtual void intersectTo( MistralSet* s ) const {} // TODO
  virtual void unionTo( MistralSet* s ) const {} // TODO
  virtual void intersectTo( MistralSet& s ) const {} // TODO
  virtual void unionTo( MistralSet& s ) const {} // TODO

  virtual bool _contain(const int v) const ;
/*   virtual bool _intersect(const int lo, const int up) const ; */
/*   virtual bool _included(const int lo, const int up) const ; */
/*   virtual bool _intersect(const VariableInt* x) const ; */
/*   virtual bool _included(const VariableInt* x) const ; */
/*   virtual bool _intersect(const MistralSet* s) const ; */
/*   virtual bool _included(const MistralSet* s) const ; */
/*   virtual bool _intersect(const MistralSet& s) const ; */
/*   virtual bool _intersect(const MistralSet& s, int& r) const ; */
/*   virtual bool _included(const MistralSet& s) const ; */
/*   virtual void _intersectTo( MistralSet* s ) const ; */
/*   virtual void _unionTo( MistralSet* s ) const ; */
/*   virtual void _intersectTo( MistralSet& s ) const ; */
/*   virtual void _unionTo( MistralSet& s ) const ; */

  virtual void printshort(std::ostream& o) const   
  {
    if( veqptr_ ) {
      o << "*";
      veqptr_->BuildObject::printshort( o );
    } else if( size_ == 1 ) {
      o << "c" << getVarId() << "(" << min_ << ")";
    } else {
      o << "x" << getVarId();
    } 
  }

  virtual void print(std::ostream& o) const 
  {
    if( varptr_ )
      varptr_->print(o);
    else {
      BuildObject::printshort( o );
      const BuildObject *self = getConstBuildObject();
      if( self->domain_ != NULL ) {
	o << " in " ;
	self->domain_->print( o );
      } else if( self->size_ != 1 ) {
	o << " in {" << self->min_ << ",.." << self->max_ << "}";
      } //else o << "(" << self->min_ << ")";
    }
  }

  virtual string xmlPred( int& idx, int level, int *sub ) const 
  {
    string s = ("X" + int2string(idx++)); 
    return s;
  }
  virtual string xmlCon( MistralSet& scope, int *index ) const 
  {
    int i=getVarId();
    if( size() > 1 ) scope.insert( index[i] );
    string s = (size() > 1 ? ("V" + int2string(index[i])) : (int2string(min_)) ); 
    return s;
  }
  virtual string xmlDom( ) const 
  {

/*     cout << (getVarId()) << " " << (size()) << endl; */

/*     print( cout ); */
/*     cout << endl; */

    string s;
    if( domain_ != NULL ) {
      if( getConstBuildObject()->varptr_ )
	s += ((VariableDomain*)(getConstBuildObject()->varptr_))->values.toString();
      else
	s += getConstBuildObject()->domain_->toString();
    } else if( size() > 2 ) {
      s += ((int2string(min())) + ".." + (int2string(max()))); 
    } else if( size() == 2 ) {
      s += ((int2string(min())) + " " + (int2string(max())));
    } else {
      s += (int2string(min()));
    }

/*     cout << s << endl <<  endl; */
    

    return s;
  }

  virtual void printstate( std::ostream& o ) const 
  {
    o << "\t" << ((state & VRANGE) ? ( state & CONVEX ? " [] " : " <> ") : " }{ " )
      << (isNot() ? " (yes/no) " : "." )
      << (isNeg() ? " (+/-) " : "." )
      << (isConst() ? " const " : "." )
      << (isVList() ? " list " : " bit " );
    if( isVar() && isRel() ) o << " pred";
    else if( isVar() ) o << " var";
    else if( isRel() ) o << " cons";
    o << ( isPar() ? " par " : " !par" ); 
  }

  bool operator==( BuildObject& x ) const
  {
    return (getConstBuildObject() == x.getConstBuildObject());
  }

};


class BuildObjectPredicate;
/**********************************************
 * Constraint BuildObject
 **********************************************/
class BuildObjectConstraint 
{

 public:

  BuildObjectConstraint() {}
  virtual ~BuildObjectConstraint() {}

  virtual void add( CSP *model, const int l, BuildObjectPredicate *pred ) ;
  virtual void build( Solver *s, VariableInt **tmp, 
		      BuildObjectPredicate *pred ) = 0;
  virtual bool preBuild( BuildObjectPredicate *pred ) ;
  //virtual bool check   ( BuildObjectPredicate *pred ) const { return true; }
  virtual int computeValue( BuildObjectPredicate *pred ) const { return -1; }

  virtual BuildObjectPredicate *declare(BuildObject*,
					const int,
					BuildObject*,
					const int)
  {
    return NULL;
  }
  virtual BuildObjectPredicate *declare(BuildObject*,
					const int,
					BuildObject*)
  {
    return NULL;
  }
  virtual BuildObjectPredicate *declare(BuildObject*,
					BuildObject*)
  {
    return NULL;
  }
  virtual BuildObjectPredicate *declare(BuildObject*,
					BuildObject*,
					BuildObject*)
  {
    return NULL;
  }
  virtual BuildObjectPredicate *declare(BuildObject*)
  {
    return NULL;
  }
  virtual BuildObjectPredicate *declare(BuildObject*,
					const int)
  {
    return NULL;
  }
  virtual BuildObjectPredicate *declare(BuildObject**,
					const int)
  {
    return NULL;
  }
  virtual BuildObjectPredicate *declare(BuildObject**,
					const int, 
					const int*)
  {
    return NULL;
  }

  virtual void print(std::ostream&, const BuildObjectPredicate*) const {}
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const
  {string s; return s;}

};



/**********************************************
 * Predicate BuildObject
 **********************************************/
class BuildObjectPredicate : public BuildObject
{

 public:
 
  BuildObjectConstraint *relation;
  BuildObject **scope;
  int *params;

  // includes itself? NO
  int arity;

  int nestedArity() const
  {
    return arity+isVar();
  }
  
  virtual void unsetCRange() { 
    if(state & CRANGE) {
      state &= (0xffffffff ^ CRANGE); 
      //print( cout );
      //cout << " ========== " << arity << endl;
      int i = arity+isVar();
      //for(int i=0; i<arity; ++i)
      while( i-- ) 
	scope[i]->unsetVRange();
    }
  }

  virtual void unsetCList() { 
    if( state & CLIST ) {
      state &= (0xffffffff ^ CLIST); 
      int n=arity+isVar();
      for(int i=0; i<n; ++i)
	scope[i]->unsetVList();
    }
  }

  virtual void setFixed() 
  { 
    if( !(state & FIXED) ) {
      state |= FIXED; 
      setDomain( relation->computeValue( this ) );
    }
  }
  
  BuildObjectPredicate(BuildObject **x, 
		       const int, const int, const int, 
		       BuildObjectConstraint*, int*, 
		       int istate=0);
  virtual ~BuildObjectPredicate() 
    {
      delete [] scope;
      delete [] params;
    }

  virtual void build( Solver *s ) ;
  virtual void add( CSP *model, const int l, const int inf ) ;  
  virtual bool preBuild( ) ; 

  virtual bool isConstraint()
  {
    // BUGGY
    return ( !isVar() || scope[arity]->size() == 1 || !isPar() );
  }

  virtual void printshort(std::ostream& o) const ;
  virtual void print(std::ostream& o) const ;
  virtual string xmlPred( int& idx, int level, int *sub ) const ;
  virtual string xmlCon( MistralSet& scp, int *index ) const
  {
    string s;
    if( arity > 3 )
      s = BuildObject::xmlCon( scp, index );
    else {
      int n=arity-1;
      for(int i=0; i<n; ++i)
	s += (scope[i]->xmlCon( scp, index ) + " ");
      s += scope[n]->xmlCon( scp, index );
    }
    return s;
  }

};


/**********************************************
 * Wrappers
 **********************************************/

/**********************************************
 * Variable
 **********************************************/
/*! \class Variable
  \brief Syntactic sugar used to hide the implementation of variables

  Arithmetic and logic C++ operators are overloaded
  so that an Ilog-like syntax is possible. 
*/
class Variable
{

 public:

  //int val;
  BuildObject *var_ptr_;
  
  Variable() {/*  val=INT_MAX; */ var_ptr_=NULL; }
  Variable( const Variable& v );
  Variable( BuildObject *v );
  Variable( const int nvals ) ;
  Variable( const int lo, const int up ) ;
  Variable( const int* array, const int length ) ;
  Variable( const int* array, const int length, const int lo, const int up ) ;
  virtual ~Variable() {}


  int value()
  {
    //int v = val;
    //if( v == INT_MAX ) {
    int v = INT_MAX;
    if( var_ptr_->getVariable() != NULL ) {
      //var_ptr_->getVariable()->print( cout );
      //cout << endl;
      v = var_ptr_->getVariable()->value();
    } 
    //else v = var_ptr_->min();
    //}
    return v;
  }

  int size()
  {
    int s = INT_MAX;
    if( var_ptr_->getVariable() != NULL )
      s = var_ptr_->getVariable()->domsize();
    else s = var_ptr_->size();
    return s;
  }

  int min()
  {
    int m = var_ptr_->min();
    if( var_ptr_->getVariable() != NULL )
      m = var_ptr_->getVariable()->min();
    else m = var_ptr_->min();
    return m;
  }

  int max()
  {
    int m = var_ptr_->max();
    if( var_ptr_->getVariable() != NULL )
      m = var_ptr_->getVariable()->max();
    else m = var_ptr_->max();
    return m;
  }

  /*!@name Operators*/
  Variable& operator= ( const Variable& x )
    {
      var_ptr_ = x.var_ptr_;
      return *this;
    }
  Variable operator+ ( Variable x ) ;
  Variable operator- ( Variable x ) ;
  Variable operator* ( Variable x ) ; 
  Variable operator/ ( Variable x ) ; 
  Variable operator% ( Variable x ) ; 
  Variable operator^ ( Variable x ) ;
  Variable operator+ ( const int k ) ;
  Variable operator- ( const int k ) ;
  Variable operator* ( const int k ) ; 
  Variable operator/ ( const int k ) ; 
  Variable operator% ( const int k ) ; 
  Variable operator^ ( const int k ) ; 
  Variable operator- ( ) ;
  Variable operator&&( Variable x ) ;
  Variable operator||( Variable x ) ; 
  Variable operator! ( ) ; 
  Variable operator==( Variable x ) ;
  Variable operator!=( Variable x ) ;
  Variable operator<=( Variable x ) ;
  Variable operator< ( Variable x ) ;
  Variable operator>=( Variable x ) ;
  Variable operator> ( Variable x ) ;
  Variable operator==( const MistralSet& s ) ;
  Variable operator!=( const MistralSet& s ) ;
  Variable operator==( const int k ) ;
  Variable operator!=( const int k ) ;
  Variable operator<=( const int k ) ;
  Variable operator< ( const int k ) ;
  Variable operator>=( const int k ) ;
  Variable operator> ( const int k ) ;
  VarArray operator,( Variable x ) ;

  VariableInt *getVariable() 
  {
    return var_ptr_->getVariable();
  }

  virtual void printshort(std::ostream& o) const 
  {
/*     if( val != INT_MAX ) o << val; */
/*     else { */
      if( var_ptr_->getVariable() != NULL )
	var_ptr_->getVariable()->printshort( o );
      else var_ptr_->printshort( o );
      //    }
  }

  virtual void print(std::ostream& o) const 
  {
/*     if( val != INT_MAX ) o << val; */
/*     else { */
      if( var_ptr_->getVariable() != NULL )
	var_ptr_->getVariable()->print( o );
      else var_ptr_->print( o );
      //    }
  }

};


/**********************************************
 * VarArray
 **********************************************/
/*! \class VarArray
  \brief An array of variables (no wonder there)

*/
class VarArray 
{

 public:
  
  MistralStack<Variable> array_;
  
  VarArray( ) {}
  VarArray( const int nvars ) ;
  VarArray( const int nvars, const int nvals ) ;
  VarArray( const int nvars, const int lo, const int up ) ;
  VarArray( const int nvars, const int* array, const int length ) ;
  VarArray( const int nvars, const int* array, const int length, const int lo, const int up ) ;

  virtual ~VarArray() {}

  BuildObject **getArgs( ) ;

  virtual void build( Solver *s ) 
  {
    int n=array_.size();
    for(int i=0; i<n; ++i)
      array_[i].var_ptr_->build( s );
  }  

  void resize( const int n )
  {
    array_.resize( n );
  }

  void clear()
  {
    array_.clear();
  }

  int size() const
  {
    return array_.size();
  }

  inline void add( const int k )
  {
    array_.push_back( Variable(k, k) );
  }

  inline void add( Variable x )
  {
    array_.push_back( x );
  }

  inline void add( VarArray& a )
  {
    int i=0, n=a.size();
    for(i=0; i<n; ++i)
      add( a[i] );
  }
  
  /*!@name Operators*/
  Variable& operator[]( const int i ) 
    {
      return array_[i]; 
    }
  
  Variable operator[]( Variable x );  
  VarArray operator,( Variable x );
  VarArray operator,( VarArray& x );

  Variable operator< ( VarArray& x ); 
  Variable operator> ( VarArray& x ); 
  Variable operator<=( VarArray& x ); 
  Variable operator>=( VarArray& x );

  VarArray& operator=( VarArray& x )
    {
      array_ = x.array_;
      return *this;
    }

  virtual void printshort(std::ostream& o) const 
  {
    int i=0, n=size()-1;
    if( n >= 0 ) {
      o << "[";
      o.flush();
      for(i=0; i<n; ++i)
	{
	  array_[i].printshort( o );
	  o << ", ";
	  o.flush();
	}
      array_[n].printshort( o );
      o << "]";
      o.flush();
    }
  }
  
  virtual void print(std::ostream& o) const 
  {
    int i=0, n=size()-1;
    if( n >= 0 ) {
      o << "[";
      o.flush();
      for(i=0; i<n; ++i)
	{
	  array_[i].print( o );
	  o << ", ";
	  o.flush();
	}
      array_[n].print( o );
      o << "]";
      o.flush();
    }
  }

};


/********************************************************
 *
 *  Unary Predicates BuildObjects & Wrappers
 *
 ********************************************************/

/**********************************************
 *  Offset Predicate BuildObject
 **********************************************/ 
class BuildObjectOffset : public BuildObjectConstraint {
 
 public:

  BuildObjectPredicate *declare( BuildObject *x_ , int o ) ;
  virtual bool  preBuild( BuildObjectPredicate *pred ) ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ; 
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ; 
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ; 
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 *  Offset Predicate Wrapper
 **********************************************/ 
class Offset : public Variable {

 public:
 
  Offset( Variable& x, int o )
    {
      var_ptr_ = ENV[ConstraintStore::OFFSET]->declare( x.var_ptr_, o ) ;
    }

};


/**********************************************
 *  Factor Predicate BuildObject
 **********************************************/ 
class BuildObjectFactor : public BuildObjectConstraint {
 
 public:

  BuildObjectPredicate* declare( BuildObject *x_, int o ) ;
  virtual bool  preBuild( BuildObjectPredicate *pred ) ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate *) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 *  Factor Predicate Wrapper
 **********************************************/ 
class Factor : public Variable {

 public:

  Factor( Variable& x_, int o )
    {
      var_ptr_ = ENV[ConstraintStore::FACTOR]->declare( x_.var_ptr_, o ) ;
    }

};


/**********************************************
 *  Negation Predicate BuildObject
 **********************************************/ 
class BuildObjectNegation : public BuildObjectConstraint {
 
 public:

  virtual bool  preBuild( BuildObjectPredicate *pred ) ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 *  Negation Predicate Wrapper
 **********************************************/ 
class Negation : public Variable {

 public:
 
  Negation( Variable& x_ )
    {
      BuildObject **x = new BuildObject*[2];
      x[0] = x_.var_ptr_;
      var_ptr_ = new BuildObjectPredicate( x, 1, 
					   -x[0]->max(), 
					   -x[0]->min(), 
					   ENV[ConstraintStore::NEG],
					   NULL, (BuildObject::AWV | BuildObject::CRANGE) ) ;
    }

};


/**********************************************
 *  Not Predicate BuildObject
 **********************************************/ 
class BuildObjectNot : public BuildObjectConstraint {
 
 public:

  virtual bool  preBuild( BuildObjectPredicate *pred ) ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 *  Not Predicate Wrapper
 **********************************************/ 
class Not : public Variable {
 
 public:
 
  Not( Variable& x_ )
    {
      BuildObject **x = new BuildObject*[2];
      x[0] = x_.var_ptr_;
      var_ptr_ = new BuildObjectPredicate( x, 1, 0, 1, 
					   ENV[ConstraintStore::NOT],
					   NULL, (BuildObject::CRANGE | BuildObject::CONVEX) ) ;
    }

};


/**********************************************
 *  Abs Predicate BuildObject
 **********************************************/ 
class BuildObjectAbs : public BuildObjectConstraint {
 
 public:

  BuildObjectPredicate* declare( BuildObject *x_ ) ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 *  Abs Predicate Wrapper
 **********************************************/ 
class Abs : public Variable {
 
 public:
 
  Abs( Variable x )
    {
      var_ptr_ = ENV[ConstraintStore::ABS]->declare( x.var_ptr_ ) ;
    }

};


/********************************************************
 *
 *  Binary Predicates BuildObjects & Wrappers
 *
 ********************************************************/

//////////// LOGIC PREDICATES /////////////////
/**********************************************
 * And Constraint BuildObject
 **********************************************/
class BuildObjectAnd : public BuildObjectConstraint {

 public:

  virtual bool  preBuild( BuildObjectPredicate *pred ) ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 *  And Predicate Wrapper
 **********************************************/ 
class And : public Variable {

 public:

  And( Variable& x_, Variable& y_ )
    {
      BuildObject **x = new BuildObject*[3];
      x[0] = x_.var_ptr_;
      x[1] = y_.var_ptr_;
      var_ptr_ = new BuildObjectPredicate(x, 2, 0, 1, 
					  ENV[ConstraintStore::AND],
					  NULL, (BuildObject::CRANGE | BuildObject::CONVEX)) ;
    }

};


/**********************************************
 * Or Constraint BuildObject
 **********************************************/
class BuildObjectOr : public BuildObjectConstraint {

 public:

  virtual bool  preBuild( BuildObjectPredicate *pred ) ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 *  Or Predicate Wrapper
 **********************************************/ 
class Or : public Variable {

 public:
  
  Or( Variable& x_, Variable& y_ )
    {
      BuildObject **x = new BuildObject*[3];
      x[0] = x_.var_ptr_;
      x[1] = y_.var_ptr_;
      var_ptr_ = new BuildObjectPredicate(x, 2, 0, 1, 
					  ENV[ConstraintStore::OR],
					  NULL, (BuildObject::CRANGE | BuildObject::CONVEX) ) ;
    }

};

/**********************************************
 * IfThenElse Constraint BuildObject
 **********************************************/
class BuildObjectIfThenElse : public BuildObjectConstraint {

 public:

  virtual bool  preBuild( BuildObjectPredicate *pred ) ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 *  IfThenElse Predicate Wrapper
 **********************************************/ 
class IfThenElse : public Variable {

 public:

  IfThenElse( Variable& x_, Variable& y_, Variable& z_ )
    {
      BuildObject **x = new BuildObject*[4];
      x[0] = x_.var_ptr_;
      x[1] = y_.var_ptr_;
      x[2] = z_.var_ptr_;
      var_ptr_ = new BuildObjectPredicate(x, 3, std::min(x[1]->min(), x[2]->min()), 
					  std::max(x[1]->max(), x[2]->max()), 
					  ENV[ConstraintStore::IFTHENELSE],
					  NULL, (BuildObject::CRANGE | BuildObject::AWV)) ;
    }

};


//////////// COMPARISON PREDICATES ////////////
/**********************************************
 *  Disjunctive Predicate BuildObject
 **********************************************/ 
class BuildObjectDisjunctive : public BuildObjectConstraint {
 
 public:
  
  BuildObjectPredicate *declare( BuildObject *x_, int dx, BuildObject *y_, int dy ) ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 *  Disjunctive Predicate Wrapper
 **********************************************/ 
class Disjunctive : public Variable {

 public:

  Disjunctive( Variable& x, int dx, Variable& y, int dy )
    {
      var_ptr_ = ENV[ConstraintStore::DISJUNCTIVE]->declare( x.var_ptr_, dx, y.var_ptr_, dy ) ;
    }

};


/**********************************************
 *  Precedence Predicate BuildObject
 **********************************************/ 
class BuildObjectPrecedence : public BuildObjectConstraint {
 
 public:

  BuildObjectPredicate *declare( BuildObject *x_, int d, BuildObject *y_ ) ;
  virtual bool  preBuild( BuildObjectPredicate *pred ) ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 *  LowerBound Predicate BuildObject
 **********************************************/ 
class BuildObjectLowerBound : public BuildObjectConstraint {
 
 public:

  BuildObjectPredicate *declare( BuildObject *x_, int b ) ;
  virtual bool  preBuild( BuildObjectPredicate *pred ) ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 *  UpperBound Predicate BuildObject
 **********************************************/ 
class BuildObjectUpperBound : public BuildObjectConstraint {
 
 public:

  BuildObjectPredicate *declare( BuildObject *x_, int b ) ;
  virtual bool  preBuild( BuildObjectPredicate *pred ) ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 *  Precedence Predicate Wrapper
 **********************************************/ 
class Precedence : public Variable {

 public:

  Precedence( Variable& x, int d, Variable& y )
    {
      var_ptr_ = ENV[ConstraintStore::PRECEDENCE]->declare( x.var_ptr_, d, y.var_ptr_ ) ;
    }

  Precedence( Variable& x, int b )
    {
      var_ptr_ = ENV[ConstraintStore::UPPERBOUND]->declare( x.var_ptr_, b ) ;
    }

  Precedence( int b, Variable& x )
    {
      var_ptr_ = ENV[ConstraintStore::LOWERBOUND]->declare( x.var_ptr_, b ) ;
    }

};


/**********************************************
 *  Add Predicate BuildObject
 **********************************************/
class BuildObjectAdd : public BuildObjectConstraint {
 
 public:

  BuildObjectPredicate *declare( BuildObject *x_, BuildObject *y_ ) ;
  virtual bool  preBuild( BuildObjectPredicate *pred ) ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 *  Add Predicate Wrapper
 **********************************************/
class Add : public Variable {

 public:

  Add( Variable& x, Variable& y )
    {
      var_ptr_ = ENV[ConstraintStore::ADD]->declare( x.var_ptr_, y.var_ptr_ ) ;
    }

};


/**********************************************
 *  Sub Predicate BuildObject
 **********************************************/
class BuildObjectSub : public BuildObjectConstraint {
 
 public:

  BuildObjectPredicate *declare( BuildObject *x_, BuildObject *y_ ) ;
  virtual bool  preBuild( BuildObjectPredicate *pred ) ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 *  Sub Predicate Wrapper
 **********************************************/
class Sub : public Variable {

 public:

  Sub( Variable& x, Variable& y )
    {
      var_ptr_ = ENV[ConstraintStore::SUB]->declare( x.var_ptr_, y.var_ptr_ ) ;
    }

};


/**********************************************
 *  Mul Predicate BuildObject
 **********************************************/
class BuildObjectMul : public BuildObjectConstraint {
 
 public:

  BuildObjectPredicate *declare( BuildObject *x_, BuildObject *y_ ) ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 *  Mul Predicate Wrapper
 **********************************************/
class Mul : public Variable {

 public:

  Mul( Variable& x, Variable& y )
    {
      var_ptr_ = ENV[ConstraintStore::MUL]->declare( x.var_ptr_, y.var_ptr_ ) ;
    }

};


/**********************************************
 *  Div Predicate BuildObject
 **********************************************/
class BuildObjectDiv : public BuildObjectConstraint {
 
 public:

  BuildObjectPredicate *declare( BuildObject *x_, BuildObject *y_ ) ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 *  Div Predicate Wrapper
 **********************************************/
class Div : public Variable {

 public:

  Div( Variable& x, Variable& y )
    {
      var_ptr_ = ENV[ConstraintStore::DIV]->declare( x.var_ptr_, y.var_ptr_ ) ;
    }

};


/**********************************************
 *  Pow Predicate BuildObject
 **********************************************/
class BuildObjectPow : public BuildObjectConstraint {
 
 public:

  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 *  Pow Predicate Wrapper
 **********************************************/
class Pow : public Variable {

 public:

  Pow( Variable& x_, Variable& y_ )
    {
    }

};


/**********************************************
 *  Mod Predicate BuildObject
 **********************************************/
class BuildObjectMod : public BuildObjectConstraint {
 
 public:

  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 *  ModConstant Predicate BuildObject
 **********************************************/
class BuildObjectModConstant : public BuildObjectConstraint {
 
 public:

  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 *  Mod Predicate Wrapper
 **********************************************/
class Mod : public Variable {

 public:

  Mod( Variable& x_, Variable& y_ )
    {
      BuildObject **x = new BuildObject*[3];
      x[0] = x_.var_ptr_;
      x[1] = y_.var_ptr_;
      int lb = ( x[0]->min() < 0 ? -x[1]->max()+1 : 0 );
      int ub = ( x[0]->max() > 0 ? x[1]->max()-1 : 0 );
      var_ptr_ = new BuildObjectPredicate(x, 2, lb, ub,
					  ENV[ConstraintStore::MOD],
					  NULL, BuildObject::AWV) ;
    }

  Mod( Variable& x_, const int k )
    {
      int *param = new int[1];
      param[0] = k;
      BuildObject **x = new BuildObject*[2];
      x[0] = x_.var_ptr_;
      int lb = ( x[0]->min() < 0 ? 1-k : 0 );
      int ub = ( x[0]->max() > 0 ? k-1 : 0 );
      var_ptr_ = new BuildObjectPredicate(x, 1, lb, ub,
					  ENV[ConstraintStore::MODCONSTANT],
					  param, BuildObject::AWV) ;
    }
  
};


/**********************************************
 *  Equal Predicate BuildObject
 **********************************************/ 
class BuildObjectEqual : public BuildObjectConstraint {
 
 public:

  virtual bool  preBuild( BuildObjectPredicate *pred ) ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 *  Equal Constant Predicate BuildObject
 **********************************************/ 
class BuildObjectEqualConstant : public BuildObjectConstraint {
 
 public:

  virtual bool  preBuild( BuildObjectPredicate *pred ) ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 *  Equal Predicate Wrapper
 **********************************************/ 
class Equal : public Variable {
 
 public:

  Equal( Variable& x_, Variable& y_ )
    {
      BuildObject **x = new BuildObject*[3];
      x[0] = x_.var_ptr_;
      x[1] = y_.var_ptr_;
      var_ptr_ = new BuildObjectPredicate(x, 2, 0, 1, 
					  ENV[ConstraintStore::EQUAL],
					  NULL, BuildObject::CRANGE) ;
    }

  Equal( Variable& x_, int k )
    {
      int *param = new int[1];
      param[0] = k;
      BuildObject **x = new BuildObject*[2];
      x[0] = x_.var_ptr_;
      var_ptr_ = new BuildObjectPredicate(x, 1, 0, 1, 
					  ENV[ConstraintStore::EQCONSTANT],
					  param, (BuildObject::CRANGE | BuildObject::CONVEX)) ;
    }

};


/**********************************************
 *  NotEqual Predicate BuildObject
 **********************************************/ 
class BuildObjectNotEqual : public BuildObjectConstraint {
 
 public:

  virtual bool  preBuild( BuildObjectPredicate *pred ) ;
  //virtual bool  check   ( BuildObjectPredicate *pred ) const ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 *  NotEqual Predicate BuildObject
 **********************************************/ 
class BuildObjectNotEqualConstant : public BuildObjectConstraint {
 
 public:

  virtual bool  preBuild( BuildObjectPredicate *pred ) ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 *  NotEqual Predicate Wrapper
 **********************************************/ 
class NotEqual : public Variable {
 
 public:

  NotEqual( Variable& x_, Variable& y_ )
    {
      BuildObject **x = new BuildObject*[3];
      x[0] = x_.var_ptr_;
      x[1] = y_.var_ptr_;
      var_ptr_ = new BuildObjectPredicate(x, 2, 0, 1, 
					  ENV[ConstraintStore::NEQ],
					  NULL, BuildObject::CLIST) ;
    }

  NotEqual( Variable& x_, int k )
    {
      int *param = new int[1];
      param[0] = k;
      BuildObject **x = new BuildObject*[2];
      x[0] = x_.var_ptr_;
      var_ptr_ = new BuildObjectPredicate(x, 1, 0, 1, 
					  ENV[ConstraintStore::NEQCONSTANT],
					  param, BuildObject::CLIST) ;
    }

};


/**********************************************
 *  NotEqualWildcard Predicate BuildObject
 **********************************************/ 
class BuildObjectNotEqualWildcard : public BuildObjectConstraint {
 
 public:

  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ; 
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 *  NotEqualWildcard Predicate Wrapper
 **********************************************/ 
class NotEqualWildcard : public Variable {
 
 public:

  NotEqualWildcard( Variable& x_, Variable& y_, const int w )
    {
      BuildObject **x = new BuildObject*[3];
      x[0] = x_.var_ptr_;
      x[1] = y_.var_ptr_;
      int *params = new int[1];
      *params = w;
      var_ptr_ = new BuildObjectPredicate(x, 2, 0, 1, 
					  ENV[ConstraintStore::NEQW],
					  params) ;
    }

};



/********************************************************
 *
 *  N-ary Predicates BuildObjects & Wrappers
 *
 ********************************************************/

/**********************************************
 * SumEq Constraint BuildObject
 **********************************************/
class BuildObjectSumEq : public BuildObjectConstraint {

 public:

  virtual bool  preBuild(BuildObjectPredicate *) ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ; 
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
};
/**********************************************
 * Sum Constraint BuildObject
 **********************************************/
class BuildObjectSum : public BuildObjectConstraint {

 public:

  BuildObjectPredicate *declare( BuildObject** scope, const int n, const int *os ) ;
  virtual bool  preBuild(BuildObjectPredicate *) ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ; 
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 * Sum Predicate Wrapper
 **********************************************/
class Sum : public Variable {

 public:

  Sum( VarArray& x, const int *os=NULL )
    {
      var_ptr_ = ENV[ConstraintStore::SUM]->declare( x.getArgs(), x.size(), os ) ;
    }

};


/**********************************************
 * Min Constraint BuildObject
 **********************************************/
class BuildObjectMin : public BuildObjectConstraint {

 public:

  BuildObjectPredicate *declare( BuildObject **scope, const int n ) ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ; 
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 * Min Predicate Wrapper
 **********************************************/
class Min : public Variable {

 public:

  Min( VarArray& x )
    {
      var_ptr_ = ENV[ConstraintStore::MIN]->declare( x.getArgs(), x.size() ) ;
    }

};

/**********************************************
 * Max Constraint BuildObject
 **********************************************/
class BuildObjectMax : public BuildObjectConstraint {

 public:

  BuildObjectPredicate *declare( BuildObject **scope, const int n ) ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ; 
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 * Max Predicate Wrapper
 **********************************************/
class Max : public Variable {

 public:

  Max( VarArray& x )
    {
      var_ptr_ = ENV[ConstraintStore::MAX]->declare( x.getArgs(), x.size() ) ;
    }

};


#ifdef _CPLEX

/**********************************************
 * CPlexMinDifference Constraint BuildObject
 **********************************************/
class BuildObjectCPlexMinDifference : public BuildObjectConstraint {

 public:

  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ; 
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
};
/**********************************************
 * CPlexMinDifference Predicate Wrapper
 **********************************************/
class CPlexMinDifference : public Variable {

 public:

  CPlexMinDifference( VarArray& x, const int ft )
    {
      int *params = new int[1];
      params[0] = ft;
      int n = x.size() ;
      var_ptr_ = new BuildObjectPredicate( x.getArgs(), n, 0, (n*(n-1)/2), 
					   new BuildObjectCPlexMinDifference(), 
					   params, (BuildObject::AWV | BuildObject::CRANGE) ) ;
    }

};

#endif


#ifdef _COINLP

/**********************************************
 * MaxDifference Constraint BuildObject
 **********************************************/
class BuildObjectMaxDifference : public BuildObjectConstraint {

 public:

  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ; 
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
};
/**********************************************
 * MaxDifference Predicate Wrapper
 **********************************************/
class MaxDifference : public Variable {

 public:

  MaxDifference( VarArray& x, const int mt, const int ft )
    {
      int *params = new int[2];
      params[0] = ft;
      params[1] = mt;
      int n = x.size() ;
      var_ptr_ = new BuildObjectPredicate( x.getArgs(), n, 0, (n*(n-1)/2), 
					   ENV[ConstraintStore::MAXDIFF],
					   params, (BuildObject::AWV | BuildObject::CRANGE) ) ;
    }

};


/**********************************************
 * MinDifference Constraint BuildObject
 **********************************************/
class BuildObjectMinDifference : public BuildObjectConstraint {

 public:

  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ; 
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
};
/**********************************************
 * MinDifference Predicate Wrapper
 **********************************************/
class MinDifference : public Variable {

 public:

  MinDifference( VarArray& x, const int mt, const int ft )
    {
      int *params = new int[2];
      params[0] = ft;
      params[1] = mt;
      int n = x.size() ;
      var_ptr_ = new BuildObjectPredicate( x.getArgs(), n, 0, (n*(n-1)/2), 
					   ENV[ConstraintStore::MINDIFF],
					   params, (BuildObject::AWV | BuildObject::CRANGE) ) ;
    }

};

#endif

/**********************************************
 * Element Constraint BuildObject
 **********************************************/
class BuildObjectElement : public BuildObjectConstraint {

 public:
  virtual bool  preBuild(BuildObjectPredicate *) ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ; 
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ;
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
  virtual int computeValue( BuildObjectPredicate *pred ) const;
  //  virtual void   add    (CSP *, const int, BuildObjectPredicate *) ;
};
/**********************************************
 * Element Predicate Wrapper
 **********************************************/
class Element : public Variable {

 public:

  Element( VarArray& x, Variable& y )
    {
      int n = x.size() ;
      BuildObject **scope = new BuildObject*[n+2];
      int lb = INT_MAX, ub = -INT_MAX;
      for(int i=0; i<n; ++i) {
	scope[i] = x[i].var_ptr_;
	if(lb > scope[i]->min()) lb = scope[i]->min() ;
	if(ub < scope[i]->max()) ub = scope[i]->max() ;
      }
      scope[n] = y.var_ptr_;
      var_ptr_ = new BuildObjectPredicate( scope, n+1, lb, ub, 
					   ENV[ConstraintStore::ELEMENT],
					   NULL, BuildObject::AWV ) ;
      for(int i=0; i<n+2; ++i) 
	scope[i]->unsetPar() ;      
    }

  Element( Variable& y, VarArray& x, Variable& z )
    {
      int n = x.size() ;
      BuildObject **scope = new BuildObject*[n+2];
      int lb = z.var_ptr_->min(), ub = z.var_ptr_->max() ;
      for(int i=0; i<n; ++i) 
	scope[i] = x[i].var_ptr_;
      scope[n] = y.var_ptr_;
      var_ptr_ = new BuildObjectPredicate( scope, n+1, lb, ub, 
					   ENV[ConstraintStore::ELEMENT],
					   NULL, BuildObject::AWV ) ;
      for(int i=0; i<n+2; ++i) 
	scope[i]->unsetPar() ;
      z.var_ptr_->setEqual( var_ptr_ ) ;
    }

  Element( VarArray& x )
    {
      int n = x.size() ;
      BuildObject **scope = new BuildObject*[n];
      int lb = x[n-1].var_ptr_->min(), ub = x[n-1].var_ptr_->max() ;
      for(int i=0; i<n-2; ++i) 
	scope[i] = x[i+1].var_ptr_;
      scope[n-2] = x[0].var_ptr_;
      var_ptr_ = new BuildObjectPredicate( scope, n-1, lb, ub, 
					   ENV[ConstraintStore::ELEMENT],
					   NULL, BuildObject::AWV ) ;
      for(int i=0; i<n; ++i) 
	scope[i]->unsetPar() ;
      x[n-1].var_ptr_->setEqual( var_ptr_ ) ;
      //var_ptr_->setEqual( x[n-1].var_ptr_ ) ;

      //var_ptr_->print( cout );
      //cout << endl;

    }
  
};



/********************************************************
 *
 *  Pure Constraints BuildObjects & Wrappers
 *
 ********************************************************/

/**********************************************
 * Clause BuildObject
 **********************************************/ 
class BuildObjectClause : public BuildObjectConstraint {
 
 public:

  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ; 
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ; 
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
};
/**********************************************
 *  Clause Wrapper
 **********************************************/ 
class Clause : public Variable {
 
 public:

  Clause( VarArray& x, const int* c )
    {
      int n = x.size() ;
      int *params = new int[n];
      for(int i=0; i<n; ++i) 
	params[i] = c[i];
      var_ptr_ = new BuildObjectPredicate( x.getArgs(), n, 0, -1, 
					   ENV[ENV.addConstraint( new BuildObjectClause() )],
					   params, BuildObject::CLIST ) ;
      //var_ptr_->setCList() ;
    }

};


/**********************************************
 *  Extensional Constraint BuildObject
 **********************************************/ 
class BuildObjectTable : public BuildObjectConstraint {
 
 public:

  int spin;
  int arity;
  MistralStack<int*> tuples;

  MistralSet* supportList[2];
  int *mindomain;
  int *maxdomain;

  MistralSet matrix;
  int *var_sizes;
  int matrixmin;

  int *supports;
  MistralStack<int*>** supportTuples;


  BuildObjectTable(const int a) ;
  BuildObjectTable(const int a, const int nbt) ;
  virtual ~BuildObjectTable() ;
  void setMinDom(const int i, const int m) ;
  void setMaxDom(const int i, const int m) ;

  virtual bool  preBuild( BuildObjectPredicate *pred ) ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ; 
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ; 

  void add( const int* sol ) ;
};
/**********************************************
 *  Extensional Constraint Wrapper
 **********************************************/ 
class Table : public Variable {

 private:

  BuildObjectTable *tabptr_;
 
 public:

  Table() {}
  Table( VarArray& x, const bool sp=true )
    {
      int n=x.size() ;
      
      tabptr_ = (BuildObjectTable*)
	(ENV[ENV.addConstraint( new BuildObjectTable(n) )]) ;
      tabptr_->spin = sp;
      BuildObject **temp = x.getArgs() ;
      var_ptr_ = new BuildObjectPredicate( temp, 
					   n, 0, -1, 
					   tabptr_, 
					   NULL, BuildObject::CLIST ) ;
      //var_ptr_->setCList() ;
      //((BuildObjectTable*)var_ptr_)->spin = sp;
      for(int i=0; i<n; ++i)
	temp[i]->unsetPar() ;

      for(int i=0; i<n; ++i) {
	tabptr_->setMinDom(i, temp[i]->min()) ;
	tabptr_->setMaxDom(i, temp[i]->max()+1) ;
      }
    }
  Table( VarArray& x, const int tidx, const bool sp )
    {
      int n=x.size() ;
      tabptr_ = (BuildObjectTable*)(ENV[tidx]) ;
      tabptr_->spin = sp;
      BuildObject **temp = x.getArgs() ;
      var_ptr_ = new BuildObjectPredicate( temp, 
					   n, 0, -1, 
					   tabptr_, 
					   NULL, BuildObject::CLIST ) ;
      //var_ptr_->setCList() ;
      //((BuildObjectTable*)var_ptr_)->spin = sp;
      for(int i=0; i<n; ++i)
	temp[i]->unsetPar() ;
      
      for(int i=0; i<n; ++i) {
	tabptr_->setMinDom(i, temp[i]->min()) ;
	tabptr_->setMaxDom(i, temp[i]->max()+1) ;
      }
    }
  
  void add( const int* sol )
  {
    tabptr_->add( sol ) ;
  }

};


/**********************************************
 * AllDiff Constraint BuildObject
 **********************************************/
class BuildObjectAllDiff : public BuildObjectConstraint {

 public:

  virtual bool  preBuild(BuildObjectPredicate *) ;
  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ; 
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ; 
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ; 
};
/**********************************************
 * AllDifferent Constraint Wrapper
 **********************************************/
class AllDifferent : public Variable {
 
 public:
 
  AllDifferent( VarArray& x )
    {
      BuildObject **temp = x.getArgs() ;
      int n = x.size() ;
      var_ptr_ = new BuildObjectPredicate( temp, n, 0, -1, 
					   ENV[ConstraintStore::ALLDIFF],
					   NULL ) ;
      for(int i=0; i<n; ++i)
	temp[i]->unsetPar() ;
      var_ptr_->unsetPar() ;
    }
 
};
/**********************************************
 * BoundAllDifferent Constraint Wrapper
 **********************************************/
class BoundAllDifferent : public Variable {
 
 public:
 
  BoundAllDifferent( VarArray& x )
    {
      int *par = new int[1];
      var_ptr_ = new BuildObjectPredicate( x.getArgs(), x.size(), 0, -1, 
					   ENV[ConstraintStore::ALLDIFF],
					   par, (BuildObject::CRANGE | BuildObject::CONVEX) ) ;
      var_ptr_->unsetPar() ;   
    }
 
};

 
/**********************************************
 * Gcc Constraint BuildObject
 **********************************************/
class BuildObjectGcc : public BuildObjectConstraint {

 public:

  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ; 
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ; 
};
/**********************************************
 * Gcc Constraint Wrapper
 **********************************************/
class Gcc : public Variable {

 public:
 
  Gcc( VarArray& x,
       const int firstDomainValue, 
       const int lastDomainValue,
       const int* minOccurrences,
       const int* maxOccurrences )
    {
      int n = (lastDomainValue - firstDomainValue + 1) ;
      int *params = new int[((n+1) << 1)];
      params[0] = firstDomainValue;
      params[1] = lastDomainValue;
      std::memcpy(params+2, minOccurrences, n*sizeof(int)) ;
      std::memcpy(params+2+n, maxOccurrences, n*sizeof(int)) ;

      var_ptr_ = new BuildObjectPredicate( x.getArgs(), x.size(), 0, -1, 
					   ENV[ConstraintStore::GCC],
					   params ) ;
      var_ptr_->unsetPar() ;
    }
  
};


/**********************************************
 * EdgeFinder Constraint BuildObject
 **********************************************/
class BuildObjectEdgeFinder : public BuildObjectConstraint {

 public:

  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ; 
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ; 
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ;
};
/**********************************************
 * EdgeFinder Constraint Wrapper
 **********************************************/
class EdgeFinder : public Variable {

 public:
 
  EdgeFinder( VarArray& x,
	      const int *d )
    {
      int n = x.size() ;
      int *params = new int[n];
      for(int i=0; i<n; ++i) 
	params[i] = d[i];
      var_ptr_ = new BuildObjectPredicate( x.getArgs(), n, 0, -1, 
					   ENV[ConstraintStore::EDGEFINDER],
					   params, BuildObject::CRANGE ) ;
      var_ptr_->unsetPar() ;
    }
  
};


/**********************************************
 * Lex Constraint BuildObject
 **********************************************/
class BuildObjectLex : public BuildObjectConstraint {
 
 public:

  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ; 
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ; 
};
/**********************************************
 * Lex < Constraint Wrapper
 **********************************************/
class LexOrder : public Variable {
 
 public:
  LexOrder( VarArray& x, VarArray& y, int eq=1 )
    {
      int *params = new int[1];
      *params = eq;
      int n = x.size() ;
      BuildObject **scope = new BuildObject*[(n << 1)+1];
      for(int i=0; i<n; ++i) {
	scope[i] = x[i].var_ptr_;
	scope[i+n] = y[i].var_ptr_;
      }
      var_ptr_ = new BuildObjectPredicate( scope, (n << 1), 0, -1, 
					   ENV[ConstraintStore::LEX],
					   params, (BuildObject::CRANGE | BuildObject::CONVEX) ) ;
    }

};


/**********************************************
 * MaxOccurrence Constraint BuildObject
 **********************************************/
class BuildObjectMaxOccurrence : public BuildObjectConstraint {
 
 public:

  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ; 
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ; 
  //virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 * MaxOccurrence < Constraint Wrapper
 **********************************************/
class MaxOccurrence : public Variable {
 
 public:
  MaxOccurrence( VarArray& x, const int k, const int ub  )
    {
      int *params = new int[2];
      params[0] = k;
      params[1] = ub;
      int n = x.size() ;
      var_ptr_ = new BuildObjectPredicate( x.getArgs(), n, 0, -1, 
					   ENV[ConstraintStore::MAXOCCUR],
					   params ) ;//, false, false, false ) ;
    }

};


/**********************************************
 * MinOccurrence Constraint BuildObject
 **********************************************/
class BuildObjectMinOccurrence : public BuildObjectConstraint {
 
 public:

  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ; 
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ; 
  //virtual int computeValue( BuildObjectPredicate *pred ) const;
};
/**********************************************
 * MinOccurrence < Constraint Wrapper
 **********************************************/
class MinOccurrence : public Variable {
 
 public:
  MinOccurrence( VarArray& x, const int k, const int lb  )
    {
      int *params = new int[2];
      params[0] = k;
      params[1] = lb;
      int n = x.size() ;
      var_ptr_ = new BuildObjectPredicate( x.getArgs(), n, 0, -1, 
					   ENV[ConstraintStore::MINOCCUR],
					   params ) ;//, false, false, false ) ;
    }

};


/**********************************************
 * Slide Constraint BuildObject
 **********************************************/
class BuildObjectSlide : public BuildObjectConstraint {

 public:

  CSP* sub_problem;

  void tuple2int(int& x, int*, int, int*) ;

  virtual void   build  (Solver *, VariableInt **, BuildObjectPredicate *) ;
  virtual void   print  (std::ostream&, const BuildObjectPredicate*) const ; 
  virtual string xmlPred(int&, int, const BuildObjectPredicate*, int*) const ; 
};
/**********************************************
 * Slide Constraint Wrapper
 **********************************************/
class Slide : public Variable {
 
 public:
 
  Slide( VarArray& x, const int n, const int m, CSP* p )
    {
      int *params = new int[2];
      params[0] = n;
      params[1] = m;
      BuildObjectSlide *rel = (BuildObjectSlide*)(ENV[ConstraintStore::SLIDE]) ;
      var_ptr_ = new BuildObjectPredicate( x.getArgs(), x.size(), 0, -1, 
					   rel, params ) ; //false, false, false ) ;
      rel->sub_problem = p;
    }
 
};


/**********************************************
 *  Objectives declaration & wrappers
 **********************************************/


class BuildObjectObjective {
 public:
  BuildObject* X;
  virtual void build( Solver *s ) = 0;

 BuildObjectObjective(BuildObject *x) : X(x) {}

  virtual void print( std::ostream& o ) const {}
};


/// Maximisation declaration
class BuildObjectMaximiseVar : public BuildObjectObjective {
 public:
 BuildObjectMaximiseVar(BuildObject *x) 
   : BuildObjectObjective(x)
    {
    }
  
  virtual void build( Solver *s ) ;

  virtual void print( std::ostream& o ) const 
  {
    o << "Maximise ";
    X->printshort( o ) ;
  }
};

/// Minimisation declaration
class BuildObjectMinimiseVar : public BuildObjectObjective {
 public:
 BuildObjectMinimiseVar(BuildObject *x) 
   : BuildObjectObjective(x)
  {
  }
  
  virtual void build( Solver *s ) ;

  virtual void print( std::ostream& o ) const 
  {
    o << "Minimise ";
    X->printshort( o ) ;
  }
};


class Objective {
 public:
  BuildObject *X;
  BuildObjectObjective *obj;
  
 Objective(BuildObject *x) : X(x) { obj = NULL; }
};

/// Maximisation wrapper
class Maximise : public Objective {
 public:
 Maximise(Variable x) : Objective( x.var_ptr_ ) 
    {
      obj = new BuildObjectMaximiseVar( X ) ; 
    }
};

/// Minimisation wrapper
class Minimise : public Objective {
 public:
 Minimise(Variable x) : Objective( x.var_ptr_ ) 
    {
      obj = new BuildObjectMinimiseVar( X ) ;
    }
};


/**********************************************
 * CSP
 **********************************************/
/*! \class CSP
  \brief Implementation of a Constraint network.

  This implementation is composed
  of an array of pointers to Variables, and
  a list of pointers to Constraints
*/
class CSP {

 public:
  /*!@name Parameters*/
  //@{

  // stats
  unsigned int nList;
  unsigned int eList;
  unsigned int nBit;
  unsigned int eBit;
  unsigned int nRange;
  unsigned int eRange;
  unsigned int nBoolean;
  unsigned int eBoolean;
  unsigned int nConstant;
  unsigned int eConstant;
  unsigned int nValues;
  unsigned int eValues;
  unsigned int nConstraint;
  unsigned int eConstraint;
  unsigned int nNeq;
 

  int numVars() const
  {
    return (nList + nBit + nBoolean + nRange) ;
  }

  int numEVars() const
  {
    return (eList + eBit + eBoolean + eRange) ;
  }
  

  // Model declaration
  MistralStack<int> first;
  MistralStack<int> last;
  MistralStack<int> next;

  MistralStack<BuildObject*> declarations; 
  MistralGacStack<BuildObject*> buildstack;

  MistralStack<BuildObject*> notequals;
  int inferAllDiffs( const int all, const int limit );

  inline void triggerEvent(BuildObject* x, int e) ;

  void close()
  {
    buildstack.initStack( declarations.size() ) ;
  }
  

  inline void add( BuildObject *d, const int lvl )
  {
    while( first.size() <= lvl ) {
      first.push( -1 ) ;
      last.push( -1 ) ;
    }
    if( first[lvl] < 0 ) {
      first[lvl] = next.size() ;
      last[lvl] = next.size() ;
    } else {
      next[last[lvl]] = next.size() ;
      last[lvl] = next.size() ;
    }

    d->id = declarations.size() ;
    declarations.push( d ) ;
    next.push( -1 ) ;
  }

  BuildObjectObjective *goald;
  //@}

  /*!@name Constructors*/
  //@{
  CSP() ;
  virtual ~CSP() ;
  void cleanUp() ;
  //@}
  
  /*!@name Accessors*/
  //@{
  int size() 
  {
    return declarations.size() ;
  }
  //@}

  /*!@name Accessors*/
  //@{
  /// Add a single variable x 
  void add( Variable x ) ;
  /// Add an array of variables 
  void add( VarArray& x ) ;
  /// Add an objective function g 
  void add( Objective g ) ;
  /// Build the data structure used during search;
  void build( Solver *s ) ;
  void preBuild( ) ;
  //@}
  
  /*!@name Miscellaneous*/
  //@{
  /// Printshort the CSP
  void print(std::ostream&) const ;
  void printXML(std::ostream&) const ;
  //@}
};


ostream& operator<< (ostream& os, const MistralGraph& g) ;
ostream& operator<< (ostream& os, const Variable& x) ;
ostream& operator<< (ostream& os, const VarArray& x) ;
ostream& operator<< (ostream& os, const CSP& p) ;


#endif // __Wrapper_H



