
/*
  Mistral is a constraint satisfaction and optimisation library
  Copyright (C) 2003-2005  Emmanuel Hebrard
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  The author can be contacted electronically at 
  ehebrard@cse.unsw.edu.au.
*/

/** \file set.h
    \brief Header for the class MistralSet.
*/

//-------------Representation of sets as vectors of bits--------------------------
//
//    the set has to be bounded in size (N)
//
//--------------------------------------------------------------------------------
// Emmanuel Hebrard 29st Mar 2005


#ifndef _MistralSet_h
#define _MistralSet_h

#include <stdlib.h> 
#include <iostream>
//#include <string>
#include <sstream>


using namespace std;

typedef unsigned int uint;

void showUint(uint n, std::ostream & os);



/**********************************************
 * MistralStack
 **********************************************/

/// An extendable array.
/**
   Could be represented as a vector, though
   this class is simpler and faster.
*/
template <class T>
class MistralStack {
 public:
  static const int MIN_CAPACITY = 16;

  T *   stack_;
  int capacity;
  int    size_;
  
  MistralStack( const MistralStack<T>& s )
    {
      capacity = s.capacity;
      stack_ = (T*) malloc(capacity*sizeof(T));
      for(size_ = 0; size_<s.size_; ++size_)
	stack_[size_] = s[size_];
    }

  MistralStack()
    {
      capacity = MIN_CAPACITY;
      size_ = 0;
      stack_ = (T*) malloc(capacity*sizeof(T));
    }

  MistralStack(const int n)
    {
      capacity = n;
      size_ = 0;
      stack_ = (T*) malloc(capacity*sizeof(T));
    }
  
  ~MistralStack()
    {
      free( stack_ );
    }

  void init(const int s, const int c)
  {
    size_ = s;
    if( capacity < c ) {
      capacity = c;
      T* new_stack = (T*) malloc(capacity*sizeof(T));
      free( stack_ );
      stack_ = new_stack;
    }
  }

  void extendStack()
  {
    capacity = capacity << 1;
    T* new_stack = (T*) malloc(capacity*sizeof(T));
    for(int i=0; i<size_; ++i)
      new_stack[i] = stack_[i];
    free( stack_ );
    stack_ = new_stack;
  }


  void extendStack( int l )
  {
    capacity += l;
    T* new_stack = (T*) malloc(capacity*sizeof(T));
    for(int i=0; i<size_; ++i)
      new_stack[i] = stack_[i];
    free( stack_ );
    stack_ = new_stack;
  }

  void resize( int l )
  {
    if( capacity < l ) {
      capacity = l;
      T* new_stack = (T*) malloc(capacity*sizeof(T));
      for(int i=0; i<size_; ++i)
	new_stack[i] = stack_[i];
      free( stack_ );
      stack_ = new_stack;
    }
    size_ = l;
  }

  int size() const
  {
    return size_;
  }
  int empty() const
  {
    return !size_;
  }
  
  void push(T x)
  {
    if( capacity == size_ )
      extendStack();
    stack_[size_++] = x;
  }
  void push_back(T x)
  {
    if( capacity == size_ )
      extendStack();
    stack_[size_++] = x;
  }
  void pop()
  {
    --size_;
  }
  void pop(T& x)
  {
    x = stack_[--size_];
  }
  void clear()
  {
    size_ = 0;
  }

  
  void erase(int i)
  {  
    stack_[i] = stack_[--size_];
  }

  void setBack(const T& x)
  {
    stack_[size_-1] = x;
  }

  T& back()
    {
      return stack_[size_-1];
    }
  T& operator[](int i)
    {
      return stack_[i];
    }
  T operator[](int i) const
    {
      return stack_[i];
    }

  MistralStack<T>& operator=( MistralStack<T>& s )
    {
      if( capacity < s.capacity ) {
	capacity = s.capacity;
	T *new_stack_ = (T*) malloc(capacity*sizeof(T));
	free( stack_ );
	stack_ = new_stack_;
      }
      for(size_ = 0; size_<s.size_; ++size_)
	stack_[size_] = s[size_];
      return *this;
    }
};


template <class T>
class MistralNode {
 public:

  int type;

  T elt;  
  MistralNode< T > *next;
  MistralNode< T > *pred;

  MistralNode( )
    {
      type = 0;

      next = NULL;
      pred = NULL;
      elt = NULL;
    }

  ~MistralNode() {}

  inline void erase( )
  {
    pred->next = next;
    if( next )
      next->pred = pred;
  }

};


template <class T>
inline MistralNode< T >* nextNode( MistralNode< T >*& nd )
{
  do nd = nd->next;
  while( nd && !nd->elt );
  return nd;
}





template <class T>
class MistralList {
 public:
  MistralNode< T > head;

  MistralList( ) {}
  ~MistralList( ) {}

  void checkConsistency()
  {
    MistralNode< T > *nd;
    nd = &head;
    while( nextNode(nd) ) 
      assert( nd->pred->next == nd );
  }

  inline void insert( MistralNode< T >& x )
  {
    x.next = head.next;
    x.pred = &head;
    if( head.next ) head.next->pred = &x;
    head.next = &x;
  }

  inline void clear( )
  {
    head.next = NULL;
  }

  void print(std::ostream & os) 
  {
    MistralNode<T> *nd = &head;
    while( nextNode(nd) ) {
      os << endl << "\t\t<<[";
      nd->elt->print( os );
      os << "]>>";
    }
    os << endl;
  }

};


template<typename T, typename F> T StrmConvert( F from )
{
  std::stringstream temp;
  temp << from;
  T to = T();
  temp >to;
  return to;
}

template<typename F> std::string StrmConvert( F from )
{
  return StrmConvert<std::string>( from );
}

inline std::string int2string(int x)
{
  std::ostringstream o;
  o << x;
  return o.str();
}


/* class MistralTuple */
/* { */
/*   int */
/* } */


class ReversibleStruct;
/**********************************************
 * Set
 **********************************************/

///   A representation of sets using a vector of bits.
/*!
   The sets are bounded. Beware that the space complexity
   depends only on the bound, and not on the current
   number of elements. As 32 elements can be 
   handled at once, the usual set operations (union,
   inetersectiuon, difference...) are very efficient.
   However iterating on the elements or counting
   them may not be so.
*/
class MistralSet
{ 
 public:

  static const uint idx1 = 0xff;
  static const uint idx2 = 0xff00;
  static const uint idx3 = 0xff0000;
  static const uint idx4 = 0xff000000;

  /// 32 bits words
  static const uint EXP = 5;

  /*   /// 64 bits words */
  /*   static const uint EXP = 6; */

  static const uint size_word_bit = (1 << EXP);
  static const uint size_word_byte = (size_word_bit >> 3);
  static const uint CACHE = (size_word_bit - 1);
  static const uint LASTCHAR = (size_word_bit - 8);

  void clone(const MistralSet& s);

 public:
  static const uint empt = 0;
  static const uint full = 0xffffffff;

  /// A vector of bits
  uint* table;
  /// Number of 4-bytes words used to represent the set
  int pos_words;
  int neg_words;
  /*!
     Return the cardinality of the set [O(N/8)].
  */
  uint size( const int i )const;
  uint size()const;
  uint wordSize()const;
  uint sizeb()const;

  /*!
     Minimum element in the set [O(N/32)]. (however, inefficient)
  */
  int min()const;
  int wordMin()const;
  /*!
     Maximum element in the set [O(N/32)]. (however, not very efficient)
  */
  int max()const;
  int wordMax()const;

  MistralSet()
    {
      pos_words = 0;
      neg_words = 0;
      table = NULL;
    }
  //  MistralSet(const int ub, const uint p);
  MistralSet(const int lb, const int ub, const uint p);
  MistralSet(const MistralSet&);
  void pointTo(MistralSet& s);
  void pointTo(unsigned int *);

  /// Copy the set s [O(N/32)].
  void copy(const MistralSet& s);

  ~MistralSet();

  void init(const int sz, const uint p);
  void init(const int lb, const int ub, const uint p);
  void reinit(const int lb, const int ub, const uint p);
  bool isBuilt()
  {
    return (table != NULL);
  }

  /*! 
      Flip all bits [O(N/32)].
  */
  void invert     (const MistralSet& s); 
  void fastInvert (const MistralSet& s); 
  /*! 
      Do the union with s [O(N/32)]. The result is put into the calling object 
  */
  void unionWith    (const MistralSet& s); 
  void unionTo    (MistralSet& s) const; 
  /*! 
      Do the intersection with s [O(N/32)]. The result is put into the calling object
  */
  void intersectWith(const MistralSet& s); 
  void intersectTo(MistralSet& s) const; 
  void intersectWith(const MistralSet&  s,
		     ReversibleStruct&  r,
		     const int offset ); 
  /*! 
      Do the set difference (x-s) [O(N/32)]. The result is put into the calling object
  */
  void setminusWith (const MistralSet& s); 
  void setminusTo (MistralSet& s) const; 
  void setminusWith(const MistralSet&  s,
		    ReversibleStruct&  r,
		    const int offset ); 
  /*! 
      Flip all bits [O(N/32)].
  */
  void invert     (const MistralSet* s); 
  /*! 
      Do the union with s [O(N/32)]. The result is put into the calling object 
  */
  void unionWith    (const MistralSet* s); 
  void unionTo    (MistralSet* s) const; 
  /*! 
      Do the intersection with s [O(N/32)]. The result is put into the calling object
  */
  void intersectWith(const MistralSet* s); 
  void intersectTo(MistralSet* s) const; 
  /*! 
      Do the set difference (x-s) [O(N/32)]. The result is put into the calling object
  */
  void setminusWith (const MistralSet* s); 
  void setminusTo (MistralSet* s) const; 
  /*!
     Check if element elt belong to the set [O(1)]
  */
  bool member(const int elt)const ; 
  bool fastMember(const int elt)const ;
  bool wordMember(const int elt)const ; 
  /*!
     Insert element elt into the set [O(1)]
  */
  void insert(const int elt);
  void fastInsert(const int elt);
  void wordInsert(const int elt);

  /*!
     Insert element elt into the set or erase it if it is already member [O(1)]
  */
  void invert(const int elt);
  void fastInvert(const int elt);
  void wordInvert(const int elt);

  /*!
     Erase element elt from the set [O(1)]
  */
  void erase(const int elt);
  void fastErase(const int elt);
  void wordErase(const int elt);
  void erase(const int elt,
	     ReversibleStruct&  r,
	     const int offset ); 

  /*!
     Return the element following elt in the set [O(N/32)] (however inefficient)
  */
  int next(const int elt) const; 
  /*!
     Return the element preceding elt in the set [O(N/32)] (however inefficient)
  */
  int prev(const int elt) const;

  /*!
     Return true iff the set is empty [O(N/32)]
  */
  bool empty()const;

  /*!
     Return true iff the calling object is included into s [O(N/32)]
  */
  bool included (const MistralSet& s)const;
  bool included (const MistralSet& s, int)const;
  /*!
     Return true iff the calling object intersect s [O(N/32)]
  */
  bool intersect(const MistralSet& s)const;
  bool wordIntersect(const MistralSet& s)const;
  bool fastIntersect(const MistralSet& s, int& idx)const;
  /*!
     Return true iff the calling object intersect [lo..up] [O(N/32)]
  */
  bool intersect(const int lb, const int ub)const; 
  /*!
     Return true iff the calling object is included into s [O(N/32)]
  */
  bool included (const MistralSet* s)const; 
  /*!
     Return true iff the calling object intersect s [O(N/32)]
  */
  bool intersect(const MistralSet* s)const;

  /*!
     Increment by x all elements in the set.
     Any element greater or equal than the capacity 
     of the set minus x is erased [O(N/32)]
  */
  void increment(const int x);
  /*!
     Decrement by x all elements in the set.
     Any element lower than x is erased [O(N/32)]
  */
  void decrement(const int x);
  /*!
     Changes every value to its arythmetic negation
   */
  void negate( MistralSet& r );
  /*!
     Insert all elements between 0 to capacity [O(N/32)]
  */
  void fill();
  /*!
     Erase all elements [O(N/32)]
  */
  void clear();
  /*!
     flip all elements [O(N/32)]
  */
  void flip();

  /*!
     Erase all elements strictly lower than l [O(N/32)]
  */
  void setMin(const int l);
  void setMin(const int l,
	      ReversibleStruct&  r,
	      const int offset ); 
  /*!
     Erase all elements strictly greater than u [O(N/32)]
  */
  void setMax(const int u);
  void setMax(const int l,
	      ReversibleStruct&  r,
	      const int offset ); 
  /*!
     Erase all elements in the interval [l..u] [O(N/32)]
  */
  void removeInterval(const int l, const int u);
  void removeInterval(const int l,
		      const int u,
		      ReversibleStruct&  r,
		      const int offset); 
  /*!
     Add all elements in the interval [l..u] [O(N/32)]
  */
  void addInterval(int l, int u);

  /*
    Print the set
  */
  string toString() const;
  void print(std::ostream & os)const;
  void print(std::ostream & os, const char* d)const;
  void printBits(std::ostream & os)const;
};


/********************************************
 * Reversible Structure
 ********************************************/
/*! \class ReversibleStruct
    \brief Backtrackable data structures.

   All structures that need to be restored to previous
   state during search.
*/
class Solver;
class ReversibleStruct {

 private:
  MistralSet        *mem_;
  MistralSet        *elt_;
  MistralStack<int> *lvl_;
  MistralStack<int>  dlt_;
  MistralStack<int>  chg_;
  int& level;

 protected:
  Solver *solver;

 public:
  MistralSet memoryStore;

  //ReversibleStruct() { solver = NULL; }
  void init(Solver *s);
  ~ReversibleStruct()
    {
      delete [] lvl_;
      delete [] mem_;
    }

  void restore();
  void save();
  void save(const int j, int i);

  void push(const int x)
  {
    if( !elt_[level].member(x) ) {
      chg_.push(x);
      elt_[level].fastInsert( x );
    }
  }
};



bool operator<=(const MistralSet&,const MistralSet&);
bool operator>=(const MistralSet&,const MistralSet&);
bool operator==(const MistralSet&,const MistralSet&);
bool operator!=(const MistralSet&,const MistralSet&);
     
#endif
