
/*
  Mistral is a constraint satisfaction and optimisation library
  Copyright (C) 2003-2005  Emmanuel Hebrard
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  The author can be contacted electronically at 
  ehebrard@cse.unsw.edu.au.
*/

/** \file sol.h
    \brief Header for the generic Solver.
*/


#ifndef _SOLVER_H
#define _SOLVER_H

#include <mod.h>
#include <string>

using namespace std;

/**********************************************************************/




/**********************************************
 * MistralGacVarStack
 **********************************************/

/// Stack of Variable for computing the GAC closure.
class MistralGacVarStack {
 public:
  VariableInt **MistralGacVarStack_;
  int *trigger;
  int     size;
  
  /**@name Constructors*/
  //@{
  ///
  MistralGacVarStack()
    {
      size=0;
    }
  ///
  MistralGacVarStack(int n)
    {
      initStack(n);
    }
  ///
  ~MistralGacVarStack()
    {
      delete [] trigger;
      delete [] MistralGacVarStack_;
    }
  ///
  void initStack(int n)
  { 
    trigger = new int[n];
    std::fill( trigger, trigger+n, 0 );
    MistralGacVarStack_ = new VariableInt*[n];
    size = 0;
  }
  //@}
  
  void push(VariableInt* v, int event)
  {
    if( !trigger[v->id] )
      MistralGacVarStack_[size++] = v;    
    trigger[v->id] |= event;
  }

  void push(VariableInt* v, VariableInt* x)
  {
    if( !trigger[v->id] )
      MistralGacVarStack_[size++] = v;    
    trigger[v->id] = x->id;
  }


  VariableInt* pop( int& event )
  {
    int j, i = size, mindom = INT_MAX;
    VariableInt *best, *last = MistralGacVarStack_[--size];
    while( i-- )
      if( mindom > MistralGacVarStack_[i]->domsize() ) 
	{
	  j = i;
	  //best = MistralGacVarStack_[i];
	  mindom = MistralGacVarStack_[i]->domsize();
	}
    event = trigger[MistralGacVarStack_[j]->id];
    trigger[MistralGacVarStack_[j]->id] = 0;
    best = MistralGacVarStack_[j];
    MistralGacVarStack_[j] = last;
    return best;
  }

  VariableInt* back( int& event )
  {
    VariableInt* last = MistralGacVarStack_[size-1];
    event = trigger[last->id];
    return last;
  }

  bool empty() const
  {
    return (!size);
  }

  void clear()
  {
    while( size )
      trigger[MistralGacVarStack_[--size]->id] = 0; 
  }
};



/**********************************************
 * MistralGacList
 **********************************************/

/// List of Variable for computing the GAC closure.
class MistralGacList {
 public:
  VariableInt **X;
  int *next;
  int *trigger;
  int head;
  int tail;
  
  /**@name Constructors*/
  //@{
  ///
  MistralGacList()
    {
    }
  ///
  ~MistralGacList()
    {
      delete [] trigger;
      delete [] next;
    }
  ///
  void initList(int n, VariableInt **x)
  {
    X = x;
    trigger = new int[n];
    std::fill( trigger, trigger+n, 0 );
    next = new int[n+2];
    std::fill( next, next+n, INT_MAX );
    head = tail = n;
    next[n] = n;
  }
  //@}
  
  void push(VariableInt* v, int event)
  {
    int i=v->id;
    if( !trigger[i] ) {
      next[tail] = i;
      tail = i;
      next[tail] = head;
    }
    trigger[i] |= event;
  }

  VariableInt* pop( int& event )
  {
    int i=next[head];
    VariableInt *x = X[i];
    next[head] = next[i];
    event = trigger[i];
    trigger[i] = 0;

    if( next[head] == head )
      tail = head;
    return x;
  }

  bool empty() const
  {
    return (next[head] == head);
  }

  void clear()
  {
    while( next[head] != head ) {
      trigger[next[head]] = 0;
      next[head] = next[next[head]];
    }
    tail = head;
  }

  void print(std::ostream& o) const
  {
    int i;
    cout << "{";
    i = next[head];
    while( i != head ) {
      X[i]->printshort( cout );
      cout << " ";
      cout.flush();
      i = next[i];
    }
    cout << "}";
  }

};


/**********************************************
 * Solver
 **********************************************/

/// A representation of the search process.
/**
   Maintains data structures 
   and Implements the generic backtracking method
*/
class Solver {

 public:

  static const int UNSAT     =  0;
  static const int SAT       =  1;
  static const int LIMITOUT  =  2;
  static const int LUBY      =  0;
  static const int GEOMETRIC =  1;
  static const int GLUBY     =  2;  
  static const int PROBE     =  3;  
  static const int RUNTIME   =  1;
  static const int BTS       =  2;
  static const int NDS       =  4;
  static const int FLS       =  8;
  static const int SPEED     = 16;
  static const int CKS       = 32;
  static const int OUTCOME   = 64;
  static const int TOTALTIME = 128;
  static const int ALL       = 255;

  bool randomizedRestart;
  bool domainSplitting;
  void setDomainSplitting() { domainSplitting = true; }
  void randomizeSequence();
  void setRandomized() { randomizedRestart = true; }
  void setRandomSeed( const unsigned int seed ) { 
    srand(seed); 
  }

  /// The objective function, for optimisation problem. Should be NULL for decision problems.
  ObjectiveFunction *goal;
  /// The dynamic variable ordering heuristic
  DVO  *heuristic;
  //Weighter *learner;
  MistralStack<Weighter*> learners;

  Weighter* setLearner( int wtype );

  inline void learnFailure( Constraint *con )
  {
    ++FAILURES;
    int i=learners.size();
    while( i-- )
      learners[i]->notifyFailure( con );
  }

  inline void learnSuccess( )
  {
    int i=learners.size();
    while( i-- ) {
      learners[i]->notifySuccess( );
    }
  }

  inline void newNode( VariableInt* x )
  {
    decision.push( x  );
    past.push( future );
    auxv.push( auxilliary  );
    lvl_.push( store.size_ );
    
    ++NODES;
    int i=learners.size();
    while( i-- )
      learners[i]->notifyChoice( );
  }

  /// A pointer to the CSP
  CSP *itsCSP;

  int status;
  /// The current level in the search tree (number of left branches from the root)
  int      level;
  /// The number of search variables
  int     length;
  /// The number of variables
  int    numvars;

  /// The instantiation sequence, an array containing all variables, ordered until the index level-1
  VariableInt   **sequence;
  /// A pointer to the first unassigned variable in sequence
  VariableInt     **future;
  /// A pointer to the end of the array 'sequence'
  VariableInt      **empty;
  VariableInt **auxilliary;
  

  /// An array of decisions
  MistralStack< VariableInt* > decision;
  /// An array to store the last solution found
  int     *solution;


  /// The list of Variables
  MistralStack < VariableInt* > variables;
  /// The list of Constraints
  MistralStack < Constraint* > constraints;
  /// The list of unary Constraints
  MistralStack < UnaryConstraint* > unaryCons;
  /// The list of constants used in the model
  std::map<int, VariableInt*> constants;


/*   /// used for LDS */
/*   MistralSet gc; */
/*   int threshold; */
/*   int  maxDepth; */

/*   ReversibleSet goodChoices; */
/*   ReversibleInt nBadChoices; */


  /******************************************
   * Backtracking structures
   ******************************************/
  void printSearch();

 private:

  MistralStack<ReversibleObj*> store;
  MistralStack<int> lvl_;

 public:
  
  void save( ReversibleObj* );

  MistralStack<VariableInt **> past;
  MistralStack<VariableInt **> auxv;


  /// GAC Stack
  MistralGacList gacvarstack;
  /// GAC Stack
  MistralGacStack<Constraint*> gacconstack;
  /// List of unary constraints

  /// Number of solutions to find, if equal to -1, then all solutions are listed
  int FIND_ALL;
  unsigned long int NODES, BACKTRACKS, CHECKS, SOLUTIONS, 
    FAILURES, FAILLIMIT, BTSLIMIT, NDSLIMIT;      //DG Change
  double INITTIME, STARTTIME, TIME, TIMELIMIT;
  int verbosity;

#ifdef _MORESTATS

  static const int pbarlength = 50;

  int LEVELMAX;
  double getProgress();
  void printProgressBar();
  int progressbar;

  unsigned int  *DOMAINSIZE;

  unsigned int  *FSTBRANCHING;
  unsigned int  *STOBRANCHING;
  double        *AVGBRANCHING;

  unsigned int  *STONODES;
  double        *AVGNODES;
  unsigned int  *LSTNODES;

  double        *STOTIME;
  double        *AVGTIME;

  unsigned int  *COUNT;

#endif 

  /**@name Constructors*/
  //@{
  Solver(CSP& c, VarArray& x, VariableOrdering& v);
  Solver(CSP& c, VariableOrdering& v);
  Solver(CSP& c, VarArray& x);
  Solver(CSP& c);
  Solver();

  void init();
  void prebuild(CSP&);
  void build(CSP&, const int n=INT_MAX);
  void initSearch( VarArray& X, const int s=0 );
  void initSearch( const int n=0, 
		   const int s=INT_MAX, 
		   const bool r=true );

  void add(VariableOrdering& g);
  virtual ~Solver();
  //@}

  

  /**@name Solving methods*/
  //@{
  /// Returns UNSAT/SAT/LIMITOUT
  /*!
     Generic backtracking method.
     Assignment to a variable works in two phases. First we select      
     the next variable using the method select() of
     the heuristic.
     We then have a binary choice point, the left branch
     (makeDecision()) is first explored.
     If this fails, then the domain is restored 
     and and the complementary right branch is explored 
     (makeComplementary()).
  */
  int depthFirstSearch();
  int startNewSearch();
  int getNextSolution();
  //  int limitedDiscrepancySearch();
  /*!
     Preprocessing step (closure computed with filtering())
  */
  bool preprocess();
  
  /// Returns UNSAT/SAT/LIMITOUT
  /*!
    presolving method
   */
  bool presolve();
  /*!
    Simple solving method
   */
  int solve();
  /*!
    LDS solving method
   */
  int ldSolve();
  /// Returns UNSAT/SAT/LIMITOUT
  /*!
    Solve using restarts. There is no randomization besides
    that induced by the variables and values ordering.
   */
  int solve_and_restart( const int policy = GEOMETRIC, 
			 const unsigned int base = 32, 
			 const double factor = 1.3333333 );
  /*!
    Diarmuid's random probes
   */
  int random_probe( const unsigned int iterations = 100, 
		    const unsigned int limit = 30 );
  /*!
    singleton arc consistency
   */
  int sacPreprocess( const bool complete = false, 
		     const int wtype = (Weighter::IPT | Weighter::WLD) );
  /// GAC closure
  /*!
    Implementation of the AC3 algorithm
   */
  bool filtering();
  /*!
     Called when a solution is found. 
  */
  int solutionFound();
  /*!
     Check the solution for consistency
  */
  int checkSolution();
  /*! 
      Call the individual restore() methods of Variable
      and Constraint
  */
  inline void restore()
  {

    //#ifdef _DEBUGSEARCH
    //    for(int k=0; k<(level >> 1); ++k) cout << " ";
    //    cout << "restore at level " << level << ": " << endl;
    //#endif

    decision.pop();
    VariableInt **vp, **va;
    int lvl;
    past.pop( vp );
    auxv.pop( va );
    lvl_.pop( lvl );

    while( store.size_ > lvl ) {
      store.back()->restore();
      store.pop();
    }

    while(vp != future) {

/* #ifdef _DEBUGSEARCH */
/*       for(int k=0; k<level; ++k) cout << "  "; */
/*       future[-1]->print( cout ); */
/*       cout << endl; */
/* #endif */

      future[-1]->link( );
    }
    while(va != auxilliary) {

/* #ifdef _DEBUGSEARCH */
/*       for(int k=0; k<level; ++k) cout << "  "; */
/*       auxilliary[-1]->print( cout ); */
/*       cout << endl; */
/* #endif */

      auxilliary[-1]->link( );
    }
      
  }
  
  inline void reInit()
  {
    SOLUTIONS = BACKTRACKS = CHECKS = 0;
    level = decision.size()-1;
    while( level ) {
      restore();
      --level;
    }
  }

  /// Triggers an event e for the variable x
  inline void triggerEvent(VariableInt* x, int e) 
  {
    gacvarstack.push( x, e );
  }

  bool limitsExpired();
  //@}

  /**@name Helpers*/
  //@{
  /// Cutoff in number of backtracks     
  void setNodeLimit     (unsigned long nl){ NDSLIMIT  = nl; }
  /// Cutoff in number of backtracks     
  void setBacktrackLimit(unsigned long nl){ BTSLIMIT  = nl; }
  /// Cutoff in number of failures      //DG Change
  void setFailureLimit  (unsigned long nl){ FAILLIMIT = nl; }      //DG Change
  /// Cutoff in cpu time (s)
  void setTimeLimit     (double tl)       { TIMELIMIT = tl; }  
  /// Print satisfiability, number of backtracks and cpu time
  void setVerbosity     (const int v)     { verbosity =  v; }  
  void printStatistics(std::ostream &log, const int stats=ALL);
  unsigned long int getBacktracks() const {return BACKTRACKS;}
  unsigned long int getNodes     () const {return NODES     ;}
  unsigned long int getFailures  () const {return FAILURES  ;}
  unsigned long int getChecks    () const {return CHECKS    ;}
  double getTime() const { return (getRunTime() - STARTTIME); }

  void print(std::ostream& o) const;
  void printXML(std::ostream& o) const;
  void printWeightProfile(std::ostream& o, int limit) const;


#ifdef _MORESTATS

  void updateMoreStats();
  void storeMoreStats();

#endif

  //@}

};

#endif // _SOLVER_H
