
/*
  Mistral is a constraint satisfaction and optimisation library
  Copyright (C) 2003-2005  Emmanuel Hebrard
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  The author can be contacted electronically at 
  ehebrard@cse.unsw.edu.au.
*/

/*! \file var.h
    \brief Header for alternative implementations of Variable.
*/

/**********************************************
 * Some alternative variable implementations 
 **********************************************/

#ifndef _VAR_H
#define _VAR_H

#include <csp.h>



/********************************************
 * Reversible Objects
 ********************************************/
/*! \class ReversibleObj
    \brief Backtrackable data structures.

   All structures that need to be restored to previous
   state during search implement the method
   restore(). 
*/
class ReversibleObj {
 protected:
  Solver *solver;

 public:
  
  ReversibleObj() { solver = NULL; }
  virtual void init(Solver *s);

  ~ReversibleObj();

  virtual void restore() = 0; 

  /*!@name Miscellaneous*/
  //@{
  /// Print on out stream
  virtual void print(std::ostream& o) const {}
};


/********************************************
 * Reversible Boolean
 ********************************************/
/*! \class ReversibleBool
    \brief Backtrackable Boolean
*/
class ReversibleBool : public ReversibleObj
{
 public:
  int state;

 ReversibleBool() : ReversibleObj()
    {
      state = 3;
    };
  
  inline operator const int() const
  {
    return state;
  }
  
  void operator= ( const int nstat );

  inline void restore()
  {
    state = 3;
  }

  /*!@name Miscellaneous*/
  //@{
  /// Print on out stream
  virtual void print(std::ostream& o) const
  {
  }
};



/* /\******************************************** */
/*  * Reversible Word */
/*  ********************************************\/ */
/* /\*! \class ReversibleWord */
/*     \brief Backtrackable Word */
/* *\/ */
/* class ReversibleWordDomain : public ReversibleObj */
/* { */
/*  public: */
/*   unsigned int *word; */
/*   unsigned int *level; */
/*   unsigned int *min; */
/*   unsigned int *max; */
/*   unsigned int *size; */
/*   unsigned int cur; */

/*  ReversibleWord() : ReversibleObj() */
/*     { */
/*       word  = new unsigned int[32]; */
/*       level = new unsigned int[32]; */
/*       min   = new unsigned int[32]; */
/*       max   = new unsigned int[32]; */
/*       size  = new unsigned int[32]; */
/*     }; */
  
  
/*   inline bool remove() */
/*   { */

/*   } */

/*   inline void restore() */
/*   { */
/*     --cur; */
/*   } */

/*   /\*!@name Miscellaneous*\/ */
/*   //@{ */
/*   /// Print on out stream */
/*   virtual void print(std::ostream& o) const */
/*   { */
/*   } */
/* }; */


/********************************************
 * Reversible Integer
 ********************************************/
/*! \class ReversibleInt
    \brief Backtrackable Integer
*/
class ReversibleInt : public ReversibleObj
{
 public:

  MistralStack<int> val_;
  MistralStack<int> lvl_;

  int value;

 ReversibleInt() : ReversibleObj() {};
  ~ReversibleInt() {};
  
  inline operator const int() const
  {
    return value;
  }
  
  inline void setValue( const int x )
  {
    if( !(val_.size_ || lvl_.size_) ) {
      value = x;
      val_.push(value);
      lvl_.push(0);
    }
  }
  
  void operator= ( const int x );
  void operator+= ( const int x );
  void operator-= ( const int x );
  void operator*= ( const int x );
  void operator/= ( const int x );
  void operator|= ( const int x );
  void operator&= ( const int x );
  void operator^= ( const int x );
  void operator++ ();
  void operator-- ();

  inline void restore()
  {
    val_.pop( value );
    lvl_.pop();
  }

  /*!@name Miscellaneous*/
  //@{
  /// Print on out stream
  virtual void print(std::ostream& o) const 
  {
    for(int i=0; i<lvl_.size_; ++i)
      o << setw(3) << val_[i] << "(" << setw(2) << lvl_[i] << ")";
    o << endl;
  }
};


/********************************************
 * Reversible Double
 ********************************************/
/*! \class ReversibleDouble
    \brief Backtrackable Double
*/
class ReversibleDouble : public ReversibleObj
{
 public:

  MistralStack<double> val_;
  MistralStack<int> lvl_;

  double value;

 ReversibleDouble() : ReversibleObj() {};
  ~ReversibleDouble() {};
  
  inline operator const double() const
  {
    return value;
  }
  
  inline void setValue( const double x )
  {
    if( !(val_.size_ || lvl_.size_) ) {
      value = x;
      val_.push(value);
      lvl_.push(0);
    }
  }
  
  void operator= ( const double x );
  void operator++ ();
  void operator-- ();

  inline void restore()
  {
    val_.pop( value );
    lvl_.pop();
  }

  /*!@name Miscellaneous*/
  //@{
  /// Print on out stream
  virtual void print(std::ostream& o) const 
  {
    for(int i=0; i<lvl_.size_; ++i)
      o << setw(3) << val_[i] << "(" << setw(2) << lvl_[i] << ")";
    o << endl;
  }
};


class ReversibleList : public ReversibleObj
{
 private:
  unsigned int *absidx;
  
 public:
  MistralStack<unsigned int> size_;
  MistralStack<int> lvl_;

  unsigned int size;

  int *list_;
  unsigned int *index_;

  MistralSet value_;

 ReversibleList() : ReversibleObj() {}
  virtual ~ReversibleList();

  virtual void init(Solver*, MistralSet&, const int, const int, const unsigned int);

  inline operator const MistralSet() const
  {
    return value_;
  }

  inline int min()const {return value_.min();}
  inline int max()const {return value_.max();}
  inline void unionTo    (MistralSet* s) const {value_.unionTo(s);} 
  inline void intersectTo(MistralSet* s) const {value_.intersectTo(s);} 
  inline void setminusTo (MistralSet* s) const {value_.setminusTo(s);} 
  inline bool member(const int elt)const {return value_.member(elt);} 
  inline bool empty()const {return value_.empty();} 
  inline bool included (const MistralSet& s)const {return value_.included(s);} 
  inline bool include (const MistralSet& s)const {return s.included(value_);} 
  inline bool intersect(const MistralSet& s)const {return value_.intersect(s);} 
  inline bool intersect(const int lb, const int ub)const {return value_.intersect(lb,ub);} 
  inline bool included (const MistralSet* s)const {return value_.included(s);} 
  inline bool include (const MistralSet* s)const {return s->included(value_);} 
  inline bool intersect(const MistralSet* s)const {return value_.intersect(s);}

  void save();
  void restore();

  void setTo(int v);
  void intersectWith(const MistralSet& s);
  void setminusWith(const MistralSet& s);
  void erase(const int elt);
  void setMin(int l);
  void setMax(int u);
  void removeInterval(int l, int u);

  int next(const int v) const;

  /*!@name Miscellaneous*/
  //@{
  /// Print on out stream
  void checkConsistency();
  virtual void print(std::ostream& o) const ;

};

/********************************************
 * Reversible Set
 ********************************************/
/*! \class ReversibleSet
    \brief Backtrackable Set
*/
class ReversibleSet : public ReversibleObj
{
 private:
  MistralSet *val_;
  MistralStack<int> *lvl_;
  int glvl;
  bool clone;

 public:
  MistralSet value;

 ReversibleSet() : ReversibleObj() 
    {
      lvl_ = NULL;
    }

  virtual void init(Solver* s, MistralSet& e, const int l);
  virtual void init(Solver* s, const int lb, const int ub);

  virtual ~ReversibleSet() 
    {
      if( lvl_ ) {
	lvl_ += value.neg_words;
	delete [] lvl_;

	val_ += value.neg_words;
	delete [] val_;
      }

      if( clone ) {
	value.neg_words = 0;
	value.table = NULL;
      }
    }
  
  inline operator const MistralSet() const
  {
    return value;
  }


  inline uint size()const {return value.size();}
  inline int min()const {return value.min();}
  inline int max()const {return value.max();}
  inline void unionTo    (MistralSet* s) const {value.unionTo(s);} 
  inline void intersectTo(MistralSet* s) const {value.intersectTo(s);} 
  inline void setminusTo (MistralSet* s) const {value.setminusTo(s);} 
  inline bool member(const int elt)const {return value.member(elt);} 
  inline int next(const int elt) const {return value.next(elt);} 
  inline int prev(const int elt) const {return value.prev(elt);} 
  inline bool empty()const {return value.empty();} 
  inline bool included (const MistralSet& s)const {return value.included(s);} 
  inline bool include (const MistralSet& s)const {return s.included(value);} 
  inline bool intersect(const MistralSet& s)const {return value.intersect(s);} 
  inline bool intersect(const int lb, const int ub)const {return value.intersect(lb,ub);} 
  inline bool included (const MistralSet* s)const {return value.included(s);} 
  inline bool include (const MistralSet* s)const {return s->included(value);} 
  inline bool intersect(const MistralSet* s)const {return value.intersect(s);}


  void save();
  void restore();

  void setTo(int v)
  {
    save();
    value.clear();
    value.insert(v);
  }
  void intersectWith(const MistralSet& s)
  {
    save();
    value.intersectWith(s);
  }
  void setminusWith(const MistralSet& s)
  {
    save();
    value.setminusWith(s);
  }
  void erase(const int elt)
  {
    save();
    value.fastInvert(elt);
  }
  void setMin(int l)
  {
    save();
    value.setMin(l);
  }
  void setMax(int u)
  {
    save();
    value.setMax(u);
  }
  void removeInterval(int l, int u)
  {
    save();
    value.removeInterval(l,u);
  }

  /*!@name Miscellaneous*/
  //@{
  /// Print on out stream
  virtual void print(std::ostream& o) const 
  {
/*     int i = value.pos_words; */
/*     int j = value.neg_words; */

    value.print( o );
/*     o << endl; */
/*     while( i-- > j )  */
/*       { */
/* 	for(int k=0; k<lvl_[i].size_; ++k) */
/* 	  o << setw(3) << lvl_[i][k]; */
/* 	o << endl; */
/* 	for(int k=0; k<lvl_[i].size_; ++k) */
/* 	  o << setw(3) << val_[i].table[k]; */
/* 	o << endl; */
/*       } */
/*     o << endl; */
  }
};


/********************************************
 * Finite Domain Integer Variable
 ********************************************/
/*! \class VariableDomain
    \brief Finite domain integer variable
*/
class VariableDomain : public VariableInt {

 protected:

  inline bool boundchanged(const int v) 
  {
    if( v == vmin ) {
      vmin = values.next( v );
      return true;
    }
    if( v == vmax ) {
      vmax = values.prev( v );
      return true;
    }
    return false;
  }

  inline bool boundchanged(const MistralSet& s, const bool spin) 
  {
    bool bc = false;
    if( spin ^ s.member(vmin) ) {
      vmin = values.min();
      bc = true;
    }
    if( spin ^ s.member(vmax) ) {
      vmax = values.max();
      bc = true;
    }

    return bc;
  }

  inline void triggerEvent( const bool bc );


 public:

  /*!@name Parameters*/
  //@{
  /// Minimum value in the domain
  ReversibleInt vmin;
  /// Maximum value in the domain
  ReversibleInt vmax;
  /// Minimum value
  int negval;
  /// Maximum value + 1
  int posval;
  /// The current domain
  MistralSet values;
  //@}


  /*!@name Constructors*/
  //@{
 VariableDomain( Solver *s ) : VariableInt(s) 
    {
      vmin.init(s); 
      vmax.init(s); 
      //size.init(s); 
    }
  virtual ~VariableDomain() {}

  void setVariable(const int low, const int up);
  void setVariable(MistralSet *s, const int length, 
		   const int low, const int up);
  //@}

  /*!@name Accessors and Iterators*/
  //@{
  /// Returns the assigned value if it exists
  inline int value() const 
  {
    return ( vmin );    
  }

  /// Returns the minimum value in the domain
  inline int min() const { return vmin; }

  /// Returns the maximum value in the domain
  inline int max() const { return vmax; }

  /// Returns the minimum value that could belong to the domain
  inline int minCapacity() const {return negval;}

  /// Returns 1 + the maximum value that could belong to the domain
  inline int maxCapacity() const {return posval;}
  //@}

  /*!@name Query methods*/
  //@{
  /// Whether or not the Variable is bound to a ground value
  inline bool isGround() const { return (vmin == vmax); }
  /*!
     Whether "v" is currently contained in the domain
  */
  inline bool contain(const int v) const 
  { 
    return ( values.member(v) ); 
  }
  inline bool fastContain(const int v) const 
  { 
    return ( values.fastMember(v) ); 
  }

  /*!
     Whether the domain has a nonempty intersection
     with the interval [l..u]
  */
  inline bool intersect(const int lo, const int up) const
  {
    return ( vmin <= up && vmax >= lo );
  }
  /*!
     Whether the domain is included in
     the interval [l..u]
  */
  inline bool included(const int lo, const int up) const
  {
    return ( vmin >= lo && vmax <= up );
  }
/*   /\*! */
/*      Whether the interval [l..u] is included in  */
/*      the domain */
/*   *\/ */
/*   inline bool include(const int lo, const int up) const */
/*   { */
/*     return ( vmin >= lo && vmax <= up ); */
/*   } */
  /*!
     Whether the domain has a nonempty intersection
     with the Variable x
  */
  inline bool intersect(const VariableInt* x) const 
  {
    return x->intersect( values );
  }
  /*!
     Whether the domain is included
     in the Variable x 
  */
  inline bool included(const VariableInt* x) const 
  {
    return x->included( values );
  }
  /*!
     Whether the domain is included
     in the Variable x 
  */
  inline bool include(const VariableInt* x) const 
  {
    return x->included( this );
  }
  /*!
     Whether the domain has a nonempty intersection
     with the set s 
  */
  inline bool intersect(const MistralSet* s) const
  {
    return values.intersect(s);
  }
  /*!
     Whether the domain is included
     in the set s
  */
  inline bool included(const MistralSet* s) const
  {
    return values.included(s);
  }
  inline bool include(const MistralSet* s) const
  {
    return s->included(values);
  }
  /*!
     Whether the domain has a nonempty intersection
     with the set s 
  */
  inline bool intersect(const MistralSet& s) const
  {
    return values.intersect(s);
    //return values.fastIntersect(s);
  }
  /*!
     Whether the domain has a nonempty intersection
     with the set s
  */
  inline bool intersect(const MistralSet& s, int& r) const
  {
    return values.fastIntersect(s,r);
    //return values.wordIntersect(s);
  }
  inline bool wordIntersect(const MistralSet& s) const
  {
    return values.wordIntersect(s);
  }
  /*!
     Whether the domain is included
     in the set s
  */
  inline bool included(const MistralSet& s) const
  {
    return values.included(s);
  }
  inline bool include(const MistralSet& s) const
  {
    return s.included(values);
  }
  /*!
     Intersect its domain with a set s
  */
  inline void intersectTo( MistralSet* s ) const
  {
    s->intersectWith(values);
  }
  /*!
     Do the union of its domain with a set s
  */
  inline void unionTo( MistralSet* s ) const
  {
    s->unionWith(values);
  }
  /*!
     Intersect its domain with a set s
  */
  inline void intersectTo( MistralSet& s ) const
  {
    s.intersectWith(values);
  }
  /*!
     Do the union of its domain with a set s
  */
  inline void unionTo( MistralSet& s ) const
  {
    s.unionWith(values);
  }
  //@}

};


class BitsetIterator : public ValIterator {

 public: 

  BitsetIterator( MistralSet& vals ) //: domain(vals) 
    {
      domain.pointTo( vals );
    }

  virtual ~BitsetIterator() 
    {
      domain.table = NULL;
      domain.neg_words = 0;
    }

  MistralSet domain;
  int curval;

  inline bool next() { curval=domain.next(curval); return curval != INT_MAX; }
  inline operator const int() const { return curval; }
};


//BitsetIterator* static_domain_it = NULL;//  = new BitsetIterator();q

/*********************************************************
 * VariableBit || An implementation of Variable using a bit vector
 *********************************************************/
/*! \class VariableBit
    \brief Implementation of integer domain Variable

  This implementation uses a revesible vector of bits for 
  representing the domain.
*/
class VariableBit : public VariableDomain {

  // protected:
  //static const int bidule= 5;
  //BitsetIterator *domain_it;

 public:

  //  static BitsetIterator* static_bit_domain_it = NULL;//  = new BitsetIterator();

  /*!@name Parameters*/
  //@{
  /// Current domain size
  ReversibleInt size;
  /// Domain
  ReversibleSet domain;
  //@}

  /*!@name Constructors*/
  //@{
  /*! 
    Constructor, create a domain containing all values
    in the interval [0,-,s-1]
  */
  VariableBit(Solver *s, const int v);// : VariableDomain(s) 
/*   {  */
/*     size.init(s); */
/*     size.setValue(v); */
/*     setVariable(0, v-1);  */
/*     domain.init(s, values, size); */

/*     cout <<  VariableBit::static_domain_it << endl; */

/*     //if(!domain_it) domain_it = new BitsetIterator(); */
/*     domain_it = new BitsetIterator( values ); */
/*   } */
  /*! 
    Constructor, create a domain containing all values
    in the interval [l,-,u]
  */
  VariableBit(Solver *s, const int l, const int u);// : VariableDomain(s) 
/*   {  */
/*     size.init(s); */
/*     size.setValue( u-l+1 ); */
/*     setVariable(l, u);  */
/*     domain.init(s, values, size); */

/*     cout << VariableBit::static_domain_it << endl; */
    
/*     //if(!domain_it) domain_it = new BitsetIterator(); */
/*     domain_it = new BitsetIterator( values ); */
/*   } */
  /*! 
    Constructor, create a domain containing all values 
    in the set e
  */
 VariableBit(Solver *s, MistralSet *e, 
	     const int lg, 
	     const int lo, 
	     const int up);// : VariableDomain(s)
/*   {  */
/*     size.init(s); */
/*     size.setValue( lg ); */
/*     setVariable(e, lg, lo, up);  */
/*     domain.init(s, values, size); */

/*     cout << VariableBit::static_domain_it << endl; */

/*     //if(!domain_it) domain_it = new BitsetIterator(); */
/*     domain_it = new BitsetIterator( values ); */
/*   } */

  VariableBit(Solver *s) : VariableDomain(s) {}
  virtual ~VariableBit();
  //@}

  /*!@name Accessors and Iterators*/
  //@{
  /// access to the domain iterator
  inline ValIterator *getDomainIterator();
/*   { */
/*     domain_it->curval = vmin; */
/*     //domain_it->domain.pointTo( values ); */
/*     return domain_it; */
/*   } */

  /// Return the domain size
  inline int domsize() const 
  {
    return size;
  }
  /// Whether or not the Variable is currently an interval
  inline bool isRange() const 
  { 
    return (size == (vmax - vmin + 1)); 
  }
  /// Return the first (equal to min) value in the domain
  inline int first() const 
  { 
    return vmin; 
  }
  /// Return the last (equal to max) value in the domain
  inline int  last() const 
  { 
    return vmax; 
  }
  /// Return the smallest value currently in the domain that is strictly greater than "v"
  inline int getNext(const int v) const 
  { 
    return ( values.next(v) ); 
  }
  /*! Used when looping on all values in the domain, return    
    true if "v" is valid value 
  */
  inline bool good(const int v) const 
  { 
      return ( v != NOVAL ); 
  }
  /*! Set "v" to the smallest value currently in the domain that is strictly greater than "v"
   */
  inline bool setNext(int& v) const 
  { 
    return ( (v = values.next(v)) != NOVAL ); 
  }
  
  /// Returns the minimum absolute value in [0..max] \inter domain
  inline int minPosAbs() const 
  {
    if( vmax < 0 ) return vmax;
    if( vmin > 0 ) return vmin;
    if( !values.table || values.member( 0 ) ) return 0;
    return values.next( 0 );
  }
  
  /// Returns the minimum absolute value in [min..0] \inter domain
  inline int minNegAbs() const  
  {
    if( vmin > 0 ) return vmin;
    if( vmax < 0 ) return vmax;
    if( !values.table || values.member( 0 ) ) return 0;
    return values.prev( 0 );
  }
  //@}
  

  /*!@name Domain handling*/
  //@{
  /*! 
      Remove the value "v"
  */
  bool remove(const int v);
  /*! 
      Remove all values but "v"
  */
  bool setDomain(const int v);
  /*! 
      Remove all values strictly lower than l
  */
  bool setMin(const int l);
  /*! 
      Remove all values strictly greater than u
  */
  bool setMax(const int u);
  /*! 
      Remove all values that do not appear in the array "a" of length "l"
  */
  bool setDomain(const int* a, const int l);
  /*! 
      Remove all values that do not appear in the set "s"
  */
  bool setDomain(const MistralSet& s);
  /*! 
      Remove all values that do not appear in the current domain of the Variable "x"
  */
  bool setDomain( VariableInt* x ) const;
  /*! 
      Remove all values that belong to the set "s"
  */
  bool removeSet(const MistralSet& s);
  /*! 
      Remove all values in the interval [l..u]
  */
  bool removeRange(const int l, const int u);
  //@}

  /*!@name Propagation method*/
  //@{
  bool revise( MistralSet*, int*, VariableInt* );
  bool revise( Constraint*, const int );
  //@}

  /*!@name Miscellaneous*/
  //@{
  void check_domain(int x) const ;
  /// Print on out stream
  virtual void print(std::ostream& o) const;
  virtual void printshort(std::ostream&) const ;
  virtual int getType() const { return VariableInt::BIT; };
  //@}
};


class ListIterator : public ValIterator {

 public: 

  int* vals;
  int  size;

  inline bool next() { return size--; }
  inline operator const int() const { return vals[size]; }
};

/*********************************************************
 * VariableList || An implementation of Variable using a list
 *********************************************************/
/*! \class VariableList
    \brief Implementation of integer domain Variable

  This implementation uses a revesible list
  representing the domain.
*/
class VariableList : public VariableDomain {

  // protected:
  //ListIterator *domain_it;

 public:

  //  static ListIterator* static_list_domain_it = NULL;

  /*!@name Parameters*/
  //@{
  /// Domain
  ReversibleList domain;
  //@}

  /*!@name Constructors*/
  //@{
  /*! 
    Constructor, create a domain containing all values
    in the interval [0,-,s-1]
  */
  VariableList(Solver *s, const int v); /* : VariableDomain(s)  */
/*   {  */
/*     setVariable(0, v-1);  */
/*     domain.init(s, values, 0, v-1, v); */
/*     domain_it = new ListIterator(); */
/*   } */
  /*! 
    Constructor, create a domain containing all values
    in the interval [l,-,u]
  */
  VariableList(Solver *s, const int l, const int u); /* : VariableDomain(s)  */
/*   {  */
/*     setVariable(l, u);  */
/*     domain.init(s, values, l, u, u-l+1); */
/*     domain_it = new ListIterator(); */
/*   } */
  /*! 
    Constructor, create a domain containing all values 
    in the set e
  */
 VariableList(Solver *s, 
	      MistralSet *e, 
	      const int lg, 
	      const int lo, 
	      const int up); /* : VariableDomain(s) */
/*   {  */
/*     setVariable(e, lg, lo, up);  */
/*     domain.init(s, values, lo, up, lg); */
/*     domain_it = new ListIterator(); */
/*   } */

  VariableList(Solver *s) : VariableDomain(s) {}
  virtual ~VariableList();
  //@}

  /*!@name Accessors and Iterators*/
  //@{
  /// access to the domain iterator
  inline ValIterator *getDomainIterator();
/*   { */
/*     domain_it->size = domain.size-1; */
/*     domain_it->vals = domain.list_; */
    
/*     //    cout << "* " << (domain_it->size) << " " << (domain_it->vals[domain_it->size]) << endl;  */

/*     return domain_it; */
/*   } */

  /// Return the domain size
  inline int domsize() const 
  {
    return domain.size;
  }
  /// Whether or not the Variable is currently an interval
  inline bool isRange() const 
  { 
    return (domain.size == (vmax - vmin + 1)); 
  }
  /// Return the first (equal to min) value in the domain
  inline int first() const 
  { 
    //return vmin; 
    return *(domain.list_);
  }
  /// Return the last (equal to max) value in the domain
  inline int  last() const 
  { 
    //return vmax; 
    return domain.list_[domain.size-1];
  }
  /// Return the smallest value currently in the domain that is strictly greater than "v"
  inline int getNext(const int v) const 
  { 
    //TODO
    return ( domain.next(v) ); 
  }
  /*! Used when looping on all values in the domain, return    
    true if "v" is valid value 
  */
  inline bool good(const int v) const 
  { 
    return ( v != NOVAL ); 
  }
  /*! Set "v" to the smallest value currently in the domain that is strictly greater than "v"
   */
  inline bool setNext(int& v) const 
  { 
    return ( (v = domain.next(v)) != NOVAL ); 
  }
  
  /// Returns the minimum absolute value in [0..max] \inter domain
  inline int minPosAbs() const 
  {
    if( vmax < 0 ) return vmax;
    if( vmin > 0 ) return vmin;
    if( !values.table || values.member( 0 ) ) return 0;
    return values.next( 0 );
  }
  
  /// Returns the minimum absolute value in [min..0] \inter domain
  inline int minNegAbs() const  
  {
    if( vmin > 0 ) return vmin;
    if( vmax < 0 ) return vmax;
    if( !values.table || values.member( 0 ) ) return 0;
    return values.prev( 0 );
  }
  //@}
  

  /*!@name Domain handling*/
  //@{
  /*! 
      Remove the value "v"
  */
  bool remove(const int v);
  /*! 
      Remove all values but "v"
  */
  bool setDomain(const int v);
  /*! 
      Remove all values strictly lower than l
  */
  bool setMin(const int l);
  /*! 
      Remove all values strictly greater than u
  */
  bool setMax(const int u);
  /*! 
      Remove all values that do not appear in the array "a" of length "l"
  */
  bool setDomain(const int* a, const int l);
  /*! 
      Remove all values that do not appear in the set "s"
  */
  bool setDomain(const MistralSet& s);
  /*! 
      Remove all values that do not appear in the current domain of the Variable "x"
  */
  bool setDomain( VariableInt* x ) const;
  /*! 
      Remove all values that belong to the set "s"
  */
  bool removeSet(const MistralSet& s);
  /*! 
      Remove all values in the interval [l..u]
  */
  bool removeRange(const int l, const int u);
  //@}

  /*!@name Propagation method*/
  //@{
  bool revise( MistralSet*, int*, VariableInt* );
  bool revise( Constraint*, const int );
  //@}

  /*!@name Miscellaneous*/
  //@{
  void check_domain(int x) const ;
  /// Print on out stream
  virtual void print(std::ostream& o) const;
  virtual void printshort(std::ostream&) const ;
  virtual int getType() const { return VariableInt::LIST; };
  //@}
};


class BoolIterator : public ValIterator {

 public: 

  int curval;

  inline bool next() { return curval--; }
  inline operator const int() const { return curval; }
};

/*********************************************************
 * VariableBool || Boolean variable
 *********************************************************/
/*! \class VariableBool
    \brief Implementation of Boolean Variable
*/
class VariableBool : public VariableInt {

 protected:
  //BoolIterator *domain_it;
  inline bool setState( const int nstat ) ;

 public:

  //static BoolIterator* static_bool_domain_it = NULL;


  /*!@name Parameters*/
  //@{
  ReversibleBool domain;
  //@}

  /*!@name Constructors*/
  //@{
  VariableBool(Solver *s); /* : VariableInt(s)  */
/*   { */
/*     domain.init(s); */
/*     domain_it = new BoolIterator(); */
/*   } */
  virtual ~VariableBool();
  //@}

  /*!@name Accessors and Iterators*/
  //@{
  /// access to the domain iterator
  inline ValIterator *getDomainIterator();
/*   { */
/*     domain_it->curval = !(domain.state & 1); */
/*     return domain_it; */
/*   } */

  /// Return the first (equal to min) value in the domain
  inline int first() const
  {
    return !(domain.state & 1);
  }
  /// Return the last (equal to max) value in the domain
  inline int  last() const
  {
    return (domain.state >> 1);
  }
  /// Return the smallest value currently in the domain that is strictly greater than "v"
  inline int getNext(const int v) const
  {
    return ((domain.state >> 1) & (1-v));
  }
  /*! Used when looping on all values in the domain, return
      true if "v" is valid value
  */
  inline bool good(const int v) const
    {
      return ( v );
    }
  /*! Set "v" to the smallest value currently in the domain that is strictly greater than "v"
   */
  inline bool setNext(int& v) const
    {
      return ( v = ((domain.state >> 1) & (1-v)) );
    }

  /// Returns the minimum absolute value in [0..max] \inter domain
  inline int minPosAbs() const
  {
    if( domain.state & 1 ) return 0;
    return 1;
  }

  /// Returns the minimum absolute value in [min..0] \inter domain
  inline int minNegAbs() const
  {
    if( domain.state & 1 ) return 0;
    return 1;
  }

  /*!
     Whether "v" is currently contained in the domain
  */
  inline bool contain(const int v) const
  {
    return ( !(v >> 1) && (domain.state & (v+1)) );
  }
  inline bool fastContain(const int v) const
  {
    return ( domain.state & (v+1) );
  }

  /// Returns the assigned value if it exists
  inline int value() const
  {
    return ( domain.state == 3 ? NOVAL : domain.state-1 );
  }

  /// Returns the minimum value in the domain
  inline int min() const
  {
    return !(domain.state & 1);
  }

  /// Returns the maximum value in the domain
  inline int max() const
  {
    return ( domain.state >> 1 );
  }

  /// Returns the minimum value that could belong to the domain
  inline int minCapacity() const
  {
    return 0;
  }

  /// Returns 1 + the maximum value that could belong to the domain
  inline int maxCapacity() const
  {
    return 2;
  }

  /// Returns the domain size
  inline int domsize() const
  {
    return ( (domain.state & 1) + ((domain.state & 2) >> 1) );
  }
  //@}

  /*!@name Query methods*/
  //@{
  /// Whether or not the Variable is currently an interval
  inline bool isRange() const
  {
    return true;
  }
  /// Whether or not the Variable is bound to a ground value
  inline bool isGround() const
  {
    return ( domain.state != 3 );
  }
  /*!
     Whether the domain has a nonempty intersection
     with the interval [l..u]
  */
  inline bool intersect(int lo, int up) const
  {
    if( lo > up || lo > 1 || up < 0 ) return false;
    int nstat = 3;
    if(lo < 1) nstat ^= 1;
    if(up > 0) nstat ^= 2;
    return ( nstat & domain.state );
  }
  inline bool included(int lo, int up) const
  {
    //return ( !(domain.state ^ lo) && !(domain.state ^ up) );
    return ( !(domain.state ^ lo) && !(domain.state ^ up) );
  }
  /*!
     Whether the domain has a nonempty intersection
     with the Variable x
  */
  inline bool intersect(const VariableInt* v) const
  {
    return (v->contain( !(domain.state & 1) ) || v->contain( (domain.state >> 1) )) ;
  }
  /*!
     Whether the domain is included
     in the Variable x
  */
  inline bool included(const VariableInt* v) const
  {
    return (v->min() >= !(domain.state & 1) && v->max() <= (domain.state >> 1));
  }
  inline bool include(const VariableInt* v) const
  {
    return v->contain( min() ) && v->contain( max() );
  }
  /*!
     Whether the domain has a nonempty intersection
     with the set s
  */
  inline bool intersect(const MistralSet* s) const
  {
    return ( s->pos_words > 0 && s->neg_words < 1 &&
	     (domain.state & s->table[0]) );
  }
  /*!
     Whether the domain is included
     in the set s
  */
  inline bool included(const MistralSet* s) const
  {
    return (
	    s->pos_words > 0 &&
	    s->neg_words <= 0 &&
	    (s->table[0] & domain.state) == domain.state 
	    );
  }
  inline bool include(const MistralSet* s) const
  {
    return (
	    s->size() <= 2 &&//change that
	    s->pos_words > 0 &&
	    s->neg_words <= 0 &&
	    (s->table[0] & domain.state) == s->table[0]
	    );
  }
  /*!
     Whether the domain has a nonempty intersection
     with the set s
  */
  inline bool intersect(const MistralSet& s) const
  {
    return ( s.pos_words > 0 && s.neg_words < 1 &&
	     (domain.state & s.table[0]) );
  }
  /*!
     Whether the domain has a nonempty intersection
     with the set s
  */
  inline bool intersect(const MistralSet& s, int& r) const
  {
    return ( s.pos_words > 0 && s.neg_words < 1 &&
	     (domain.state & s.table[0]) );
  }
  inline bool wordIntersect(const MistralSet& s) const
  {
    return ( (domain.state & s.table[0]) );
  }
  /*!
     Whether the domain is included
     in the set s
  */
  inline bool included(const MistralSet& s) const
  {
    return (
	    s.pos_words > 0 &&
	    s.neg_words <= 0 &&
	    (s.table[0] & domain.state) == domain.state 
	    );
  }
  inline bool include(const MistralSet& s) const
  {
    return (
	    s.size() <= 2 &&//change that
	    s.pos_words > 0 &&
	    s.neg_words <= 0 &&
	    (s.table[0] & domain.state) == s.table[0]
	    );
  }
  /*!
     Intersect its domain with a set s
  */
  inline void intersectTo( MistralSet* s ) const
  {
    int i=s->pos_words;
    int j=s->neg_words;
    while( i-- > j )
      if( !i ) s->table[i] &= domain.state;
      else s->table[i] = 0;
  }
  /*!
     Do the union of its domain with a set s
  */
  inline void unionTo( MistralSet* s ) const
  {
    if( s->pos_words > 0 && s->neg_words < 1 )
      s->table[0] |= domain.state;
  }
  /*!
     Intersect its domain with a set s
  */
  inline void intersectTo( MistralSet& s ) const
  {
    int i=s.pos_words;
    int j=s.neg_words;
    while( i-- > j )
      if( !i ) s.table[i] &= domain.state;
      else s.table[i] = 0;
  }
  /*!
     Do the union of its domain with a set s
  */
  inline void unionTo( MistralSet& s ) const
  {
    if( s.pos_words > 0 && s.neg_words < 1 )
      s.table[0] |= domain.state;
  }
  //@}
    

  /*!@name Domain handling*/
  //@{
  /*!
      Remove the value "v"
  */
  bool remove(const int v);
  /*!
      Remove all values but "v"
  */
  bool setDomain(const int v);
  /*!
      Remove all values strictly lower than l
  */
  bool setMin(const int l);
  /*!
      Remove all values strictly greater than u
  */
  bool setMax(const int u);
  /*!
      Remove all values that do not appear in the array "a" of length "l"
  */
  bool setDomain(const int* a, const int l);
  /*!
      Remove all values that do not appear in the {@link MistralSet set} "s"
  */
  bool setDomain(const MistralSet& s);
  /*!
      Remove all values that do not appear in the current domain of the variable "x"
  */
  bool setDomain(VariableInt*) const;
  /*!
      Remove all values that belong to the {@link MistralSet set} "s"
  */
  bool removeSet(const MistralSet& s);
  /*!
      Remove all values in the interval [l..u]
  */
  bool removeRange(const int l, const int u);
  //@}

  /// The mandatory methods
  /*!@name Decision methods*/
  //@{
  bool revise( MistralSet*, int*, VariableInt* );
  bool revise( Constraint*, const int );
  //@}

  /*!@name Miscellaneous*/
  //@{
  /// Print on out stream
  virtual void print(std::ostream& o) const;
  virtual void printshort(std::ostream&) const ;
  virtual int getType() const { return VariableInt::BOOL; };
  //@}
};


class ConstantIterator : public ValIterator {

 public: 

  ConstantIterator( const int v )
    {
      curval = v;
    }
  ConstantIterator( )
    {
    }

  int curval;

  inline bool next() { return false; }
  inline operator const int() const { return curval; }
};

/*********************************************************
 * Constant 
 *********************************************************/
/*! \class Constant
    \brief Implementation of a Constant
*/
class Constant : public VariableInt {

 protected:

  //ConstantIterator *domain_it;
  int val;

 public:
  

  //static ConstantIterator* static_constant_domain_it = NULL;

  Constant(Solver *s, const int v );
  virtual ~Constant() {}

  /*!@name Accessors and Iterators*/
  //@{
  /// access to the domain iterator
  ValIterator *getDomainIterator();

  /// Return the first (equal to min) value in the domain
  inline int first() const 
  {
    return val;
  }
  /// Return the last (equal to max) value in the domain
  inline int  last() const 
  { 
    return val; 
  }
  /// Return the smallest value currently in the domain that is strictly greater than "v"
  inline int getNext(const int v) const 
  {
    return NOVAL;
  }
  /*! Used when looping on all values in the domain, return    
      true if "v" is valid value 
  */
  inline bool good(const int v) const 
    { 
      return ( v != NOVAL );
    }
  /*! Set "v" to the smallest value currently in the domain that is strictly greater than "v"
   */
  inline bool setNext(int& v) const 
    { 
      return false; 
    }
  inline int minPosAbs() const 
  {
    return val;
  }

  /// Returns the minimum absolute value in [min..0] \inter domain
  inline int minNegAbs() const  
  {
    return val;
  }
  //@}


  /*!@name Utilities*/
  //@{
  /*!
     Whether "v" is currently contained in the domain
  */
  inline bool contain(const int v) const 
  { 
    return ( v == val ); 
  }
  inline bool fastContain(const int v) const 
  { 
    return ( v == val ); 
  }
  /// Returns the assigned value if it exists
  inline int value() const
  {
    return val;
  }    
  /// Returns the minimum value in the domain
  inline int min() const
  {
    return val;
  }
  /// Returns the maximum value in the domain
  inline int max() const
  {
    return val;
  }  
  /// Returns the minimum value that could belong to the domain
  inline int minCapacity() const 
  {
    return val;
  }
  /// Returns 1 + the maximum value that could belong to the domain
  inline int maxCapacity() const 
  {
    return val+1;
  }
  /// Returns the domain size
  inline int domsize() const
  {
    return 1;
  }
  /// Whether or not the Variable is currently an interval
  inline bool isRange() const 
  { 
    return true; 
  }
  /// Whether or not the Variable is bound to a ground value
  inline bool isGround() const
  {
    return true;   
  }
  /*!
     Whether the domain has a nonempty intersection
     with the interval [l..u]
  */
  inline bool intersect(int lo, int up) const
  {
    return ( lo <= val && val <= up );
  }
  /*!
     Whether the domain is included in
     the interval [l..u]
  */
  inline bool included(int lo, int up) const
  {
    return ( lo == up && up == val );
  }
  /*!
     Whether the domain has a nonempty intersection
     with the Variable x 
  */
  inline bool intersect(const VariableInt* v) const 
  {  
    return v->contain( val ) ;
  }
  /*!
     Whether the domain is included
     in the Variable x 
  */
  inline bool included(const VariableInt* v) const 
  {
    return (v->min() == val && v->max() == val);
  }
  inline bool include(const VariableInt* v) const 
  {
    return (v->contain( val));
  }
  /*!
     Whether the domain has a nonempty intersection
     with the set s 
  */
  inline bool intersect(const MistralSet* s) const
  {
    return s->member( val );
  }
  /*!
     Whether the domain is included
     in the set s
  */
  inline bool included(const MistralSet* s) const
  {
    return (s->member(val));
  }
  inline bool include(const MistralSet* s) const
  {
    return (s->size() == 1 && s->member(val));
  }
  /*!
     Whether the domain has a nonempty intersection
     with the set s 
  */
  inline bool intersect(const MistralSet& s) const
  {
    return s.member( val );
  }
  /*!
     Whether the domain has a nonempty intersection
     with the set s
  */
  inline bool intersect(const MistralSet& s, int& r) const
  {
    return s.member( val );
  }
  inline bool wordIntersect(const MistralSet& s) const
  {
    return s.wordMember( val );
  }
  /*!
     Whether the domain is included
     in the set s
  */
  inline bool included(const MistralSet& s) const
  {
    return (s.member(val));
  }
  inline bool include(const MistralSet& s) const
  {
    return (s.size() == 1 && s.member(val));
  }
  /*!
     Intersect its domain with a set s
  */
  inline void intersectTo( MistralSet* s ) const
  {
    if( s->member( val ) ) {
      s->clear();
      s->insert( val );
    } else s->clear();
  }
  /*!
     Do the union of its domain with a set s
  */
  inline void unionTo( MistralSet* s ) const
  {
    s->insert( val );
  }
  /*!
     Intersect its domain with a set s
  */
  inline void intersectTo( MistralSet& s ) const
  {
    if( s.member( val ) ) {
      s.clear();
      s.insert( val );
    } else s.clear();
  }
  /*!
     Do the union of its domain with a set s
  */
  inline void unionTo( MistralSet& s ) const
  {
    s.insert( val );
  }
  //@}
    

  /*!@name Domain handling*/
  //@{
  /*! 
      Remove the val "v"
  */
  bool remove(const int v) 
  {
    return ( v != val );
  }
  /*! 
      Remove all vals but "v"
  */
  bool setDomain(const int v)
  {
    return ( v == val );
  }
  /*! 
      Remove all vals strictly lower than l
  */
  bool setMin(const int l) 
  {
    return ( l <= val );
  }
  /*! 
      Remove all values strictly greater than u
  */
  bool setMax(const int u)
  {
    return ( u >= val );
  }
  /*! 
      Remove all values that do not appear in the array "a" of length "l"
  */
  bool setDomain(const int* a, const int l) 
  {
    int i = 0;
    while( i < l )
      if (a[i++] == val) return true;
    return false;
  }
  /*! 
      Remove all values that do not appear in the {@link MistralSet set} "s"
  */
  bool setDomain(const MistralSet& s)
  {
    return s.member( val ) ;
  }
  /*! 
      Remove all values that do not appear in the current domain of the variable "x"
  */
  bool setDomain(VariableInt* x) const
  {
    //return x->contain( val );
    return x->setDomain( val );
  }
  /*! 
      Remove all values that belong to the {@link MistralSet set} "s"
  */
  bool removeSet(const MistralSet& s)
  {
    return !s.member( val ) ;
  }
  /*! 
      Remove all values in the interval [l..u]
  */
  bool removeRange(const int l, const int u)
  {
    return (l > val || u < val );
  }
  //@}

  inline bool revise( MistralSet* supports, 
		      int* residues, 
		      VariableInt* X)
  {
    return X->intersect( supports[val], residues[val] );
  }
  bool revise( Constraint* c, const int i)
  {
    return true;
  }

  /*!@name Miscellaneous*/
  //@{
  /// Print on out stream
  virtual void print(std::ostream& o) const
  {
    o << val;
  }
  virtual void printshort(std::ostream& o) const 
  {
    o << val;
  }
  virtual int getType() const { return VariableInt::CONST; };
  //@}
};


class RangeIterator : public ValIterator {

 public: 

  int curval;
  int ubound;

  inline bool next() { return ++curval <= ubound; }
  inline operator const int() const { return curval; }
};

/********************************************************
 * VariableRange || An implementation of Variable using an interval
 ********************************************************/
/*! \class VariableRange
    \brief Implementation of interval variable

   Only the bounds are stored, as reversible int
*/
class VariableRange : public VariableInt {

 protected:
  /*!@name Parameters*/
  //@{
  //RangeIterator *domain_it;
  ReversibleInt vmin;
  ReversibleInt vmax;
  //@}

 public:

  //static RangeIterator* static_range_domain_it = NULL;

  /*!@name Constructors*/
  //@{
  VariableRange(Solver *s, const int v); /* : VariableInt(s)  */
/*   {  */
/*     setVariable(0, v-1);  */
/*     vmin.init(s); */
/*     vmax.init(s); */
/*     domain_it = new RangeIterator(); */
/*   } */
  VariableRange(Solver *s, const int l, const int u); /* : VariableInt(s)  */
/*   {  */
/*     setVariable(l, u);  */
/*     vmin.init(s); */
/*     vmax.init(s); */
/*     domain_it = new RangeIterator(); */
/*   } */
  //Initialize the variable to have id num and domain size "size"
  void setVariable(const int low, const int up);
  VariableRange(Solver *s); /* : VariableInt(s)  */
/*   { */
/*     domain_it = new RangeIterator(); */
/*   } */
  virtual ~VariableRange();
  //@}

 /*!@name Accessors and Iterators*/
  //@{
  /// access to the domain iterator
  inline ValIterator *getDomainIterator();
/*   { */
/*     domain_it->curval = vmin; */
/*     domain_it->ubound = vmax; */
/*     return domain_it; */
/*   } */

  /// Return 0 if it was reduced to its lower bound, 1 for upper bound
  inline int bound()
  {
    return( vmin.lvl_.back() > vmax.lvl_.back() );
  }
  /// Return the first (equal to min) value in the domain
  inline int first() const
  {
    return vmin;
  }
  /// Return the last (equal to max) value in the domain
  inline int  last() const
  {
    return vmax;
  }
  /// Return the smallest value currently in the domain that is strictly greater than "v"
  inline int getNext(const int v) const
  {
    return ( v+1 );
  }
  /*! Used when looping on all values in the domain, return
      true if "v" is valid value
  */
  inline bool good(const int v) const
  {
    return ( v <= vmax );
  }
  /*! Set "v" to the smallest value currently in the domain that is strictly greater than "v"
   */
  inline bool setNext(int& v) const
  {
    return ( ++v <= vmax );
  }
  inline int minPosAbs() const 
  {
    return ( 0 < vmin ? (int)vmin : 0 );
  }

  /// Returns the minimum absolute value in [min..0] \inter domain
  inline int minNegAbs() const  
  {
    return ( 0 > vmax ? (int)vmax : 0);
  }
  //@}


  /*!@name Utilities*/
  //@{
  inline bool contain(const int v) const
  {
    return ( v >= vmin && v <= vmax );
  }
  inline bool fastContain(const int v) const
  {
    return ( v >= vmin && v <= vmax );
  }
  /// Returns the assigned value if it exists
  inline int value() const
  {
    return ( (int)vmin == (int)vmax ? vmin : INT_MAX );    
  }    
  /// Returns the minimum value in the domain
  inline int min() const
  {
    return vmin;
  }
  /// Returns the maximum value in the domain
  inline int max() const
  {
    return vmax;
  }  
  /// Returns the minimum value that could belong to the domain
  inline int minCapacity() const 
  {
    return vmin.val_[0];
  }
  /// Returns 1 + the maximum value that could belong to the domain
  inline int maxCapacity() const 
  {
    return vmax.val_[0]+1;
  }
  /// Returns the domain size
  inline int domsize() const
  {
    return ( 1 + vmax - vmin );
  }
  /// Whether or not the Variable is currently an interval
  inline bool isRange() const 
  { 
    return true; 
  }
  /// Whether or not the Variable is bound to a ground value
  inline bool isGround() const
  {
    return ( (int)vmin == (int)vmax );   
  }
  /*!
     Whether the domain has a nonempty intersection
     with the interval [l..u]
  */
  inline bool intersect(const int lo, const int up) const
  {
    return ( up >= vmin && lo <= vmax );
  }
  /*!
     Whether the domain is included in
     the interval [l..u]
  */
  inline bool included(const int lo, const int up) const
  {
    return ( lo >= vmin && up <= vmax );
  }
  /*!
     Whether the domain has a nonempty intersection
     with the Variable x
  */
  inline bool intersect(const VariableInt* v) const
  {
    return intersect(v->min(), v->max());
  }
  inline bool included(const VariableInt* v) const
  {
    return included(v->min(), v->max());
  }
  inline bool include(const VariableInt* v) const
  {
    return v->min() >= vmin && v->max() <= vmax;
  }
  inline bool intersect(const MistralSet* s) const
  {
    return ( s->max() >= vmin && s->min() <= vmax );
  }
  inline bool included(const MistralSet* s) const
  {
    return ( s->min() >= vmin && s->max() <= vmax );
  }
  inline bool include(const MistralSet* s) const
  {
    return ( true );
  }
  inline bool intersect(const MistralSet& s) const
  {
    return ( s.max() >= vmin && s.min() <= vmax );
  }
  inline bool intersect(const MistralSet& s, int& r) const
  {
    return ( s.max() >= vmin && s.min() <= vmax );
  }
  inline bool wordIntersect(const MistralSet& s) const
  {
    return ( s.max() >= vmin && s.min() <= vmax );
  }
  inline bool included(const MistralSet& s) const
  {
    return ( s.min() >= vmin && s.max() <= vmax );
  }
  inline bool include(const MistralSet& s) const
  {
    return ( true );
  }
  /*!
     Intersect its domain with a set s
  */
  inline void intersectTo( MistralSet* s ) const
  {
    s->setMin( vmin );
    s->setMax( vmax );
  }
  /*!
     Do the union of its domain with a set s
  */
  inline void unionTo( MistralSet* s ) const
  {
    s->addInterval(vmin, vmax);
  }
  /*!
     Intersect its domain with a set s
  */
  inline void intersectTo( MistralSet& s ) const
  {
    s.setMin( vmin );
    s.setMax( vmax );
  }
  /*!
     Do the union of its domain with a set s
  */
  inline void unionTo( MistralSet& s ) const
  {
    s.addInterval(vmin, vmax);
  }
  //@}

  /*!@name Domain handling*/
  //@{
  /*!
      Remove the value "v"
  */
  bool remove(const int v);
  /*!
      Remove all values but "v"
  */
  bool setDomain(const int v);
  /*!
      Remove all values strictly lower than l
  */
  bool setMin(const int l);
  /*!
      Remove all values strictly greater than u
  */
  bool setMax(const int u);
  /*!
      Remove all values that do not appear in the array "a" of length "l"
  */
  bool setDomain(const int* a, const int l);
  /*!
      Remove all values that do not appear in the {@link MistralSet set} "s"
  */
  bool setDomain(const MistralSet& s);
  /*!
      Remove all values that do not appear in the current domain of the variable "x"
  */
  bool setDomain(VariableInt*) const;
  /*!
      Remove all values that belong to the {@link MistralSet set} "s"
  */
  bool removeSet(const MistralSet& s);
  /*!
      Remove all values in the interval [l..u]
  */
  bool removeRange(const int l, const int u);
  //@}

  /// The mandatory methods
  /*!@name Decision methods*/
  //@{
  bool revise( MistralSet*, int*, VariableInt* );
  bool revise( Constraint*, const int );
  //@}

  /*!@name Miscellaneous*/
  //@{
  /// Print on out stream
  virtual void print(std::ostream& o) const;
  virtual void printshort(std::ostream&) const ;
  virtual int getType() const { return VariableInt::RANGE; };
  //@}
};

#endif // __VAR_H
