
/*
  Mistral is a constraint satisfaction and optimisation library
  Copyright (C) 2003-2005  Emmanuel Hebrard
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  The author can be contacted electronically at 
  ehebrard@cse.unsw.edu.au.
*/

#include <csp.h>
#include <sol.h>
#include <con.h>
#include <set.h>

void checkCons( MistralList<Constraint*>& l, VariableInt* x )
{
  Constraint* con;
  int i = 100;
  MistralNode<Constraint*>* nd = &(l.head);
  while( nextNode(nd) && --i ) {
    con = nd->elt;
    bool isIn = false;
    for(int i=0; !isIn && i<con->arity; ++i)
      isIn = (con->scope[i] == x);
    assert( isIn );
  }
  assert( i );
}


/**********************************************
 * Constraint
 **********************************************/

void Constraint::initScope(VariableInt ** scp, const int n, const int t, const int w) 
{
  weight = w;
  arity = n;
  scope = new VariableInt*[arity];
  unasgnVars = arity;
  triggers = new MistralList<Constraint*>*[arity];
  elements = new MistralNode<Constraint*>[arity+2];
  
  int i = arity+2;
  while( i-- )
    elements[i].elt = this;
  i = arity;
  while( i-- ) {
    scope[i] = scp[i];
    switch(t) {
    case DOMAINTRIGGER:
      { 
	triggers[i] = scope[i]->triggerOnDomain();
	break;
      }
    case RANGETRIGGER: 
      {
	triggers[i] = scope[i]->triggerOnRange();
	break;
      }
    case VALUETRIGGER: 
      {
	triggers[i] = scope[i]->triggerOnValue();
	break;
      }
    }
  }

  /// this is in case some variables are duplicated in scope
  /// the watches should always be different variables
  watch[1] = scope[arity-1];
  i = arity-1;
  while( i-- && scope[i] == watch[1] );
  if( i < 0 ) i=0;
  watch[0] = scope[i];

  /// keep that test for now. If we couldn't find two different 
  /// variables to use as watchers, then we hope that the
  /// sole variable is a constant (apparently it is not a problem in that case)
  //assert( watch[0] != watch[1] || watch[0]->isGround() );


  lastlinked = arity-1;

  elements[arity].type = 1;
  elements[arity+1].type = 1;
  
  if( elements[arity].elt ) 
    watch[0]->watchedConstraints.insert( elements[arity] );
  if( elements[arity+1].elt ) 
    watch[1]->watchedConstraints.insert( elements[arity+1] );
  
  i = 0;
  while( i < arity ) {
    if(triggers[i]) {
      triggers[i]->insert( elements[i] );
      ++scope[i]->degree;
      scope[i]->weight += weight;
    }
    ++i;
  }
}

Constraint::Constraint(Solver *s,
		       VariableInt ** scp, 
		       const int n,
		       const int t,
		       const int w) 
{
  solver = s;
  delayed = false;
  initScope( scp, n, t, w );
  supp = new int[n];
  supports = NULL;
  id = solver->constraints.size();
  solver->constraints.push( this );
}

Constraint::Constraint(Solver *s) 
{
  solver = s;
  delayed = false;
  supp = NULL;
  supports = NULL;
  id = solver->constraints.size();
  solver->constraints.push( this );
}

Constraint::~Constraint() 
{
  if( supports ) {
    for(int x=0; x<arity; ++x) {
      for(int v=scope[x]->minCapacity(); v<scope[x]->maxCapacity(); ++v)
	delete [] supports[x][v];
      supports[x] += scope[x]->minCapacity();
      delete [] supports[x];
    }
    delete [] supports;
  }
  delete [] supp;
  delete [] scope;
  delete [] elements;
  delete [] triggers;
}

void Constraint::useResidualSupports()
{  
  int x, v, vfill = VariableInt::NOVAL;
  supports = new int**[arity];
  for(x=0; x<arity; ++x) {
    supports[x] = new int*[scope[x]->maxCapacity()-scope[x]->minCapacity()];
    supports[x] -= scope[x]->minCapacity();
    for(v=scope[x]->minCapacity(); v<scope[x]->maxCapacity(); ++v) {
      if( scope[x]->contain(v) ) {
	supports[x][v] = new int[arity];
	std::fill(supports[x][v],
		  supports[x][v]+arity,
		  vfill);
      } else supports[x][v] = NULL;
    }
  }
}

 bool Constraint::firstSupport(const int vri, 
				     const int vli) 
{
  int j;
  if( supports && supports[vri][vli][0] != VariableInt::NOVAL ) {
    j=arity;
    while( j-- ) 
      if( vri != j )
	if (!scope[j]->contain( supports[vri][vli][j] )) break;
    if( j < 0 ) return true;
  } 
  j=arity;
  while( j-- )
    supp[j] = scope[j]->first();
  supp[vri] = vli; 
  return false;
}

bool Constraint::findSupport(const int vri, const int vli) 
{
  int i=arity, vali;
  bool found=false;
  // sol is initialized: either to the value 
  // a variable is already assigned to
  // or to the first value in its domain
  while(i >= 0) {
    // check this assignment
    if( !check( supp ) ) {
      found=true;
      if( supports ) {
	vali = arity;
	while( vali-- )
	  supports[vri][vli][vali] = supp[vali];
      }
      break;
    }
 
    // try to assign more things
    // find the last var whose domain we have not exhausted
    --i;
    while( i >= 0 ) {
      if( i == vri || scope[i]->isGround() ) {
	--i;
	continue;
      }
      if( scope[i]->setNext( supp[i] ) )
      	break;
      else
	supp[i] = scope[i]->first();
      --i;
    }
    if( i >= 0 )
      i = arity;
  }
  return found;
}

bool Constraint::propagate()
{
  int i, consistent=1, evt = ( Constraint::RANGETRIGGER );
  //int i, consistent=1, evt = ( Constraint::VALUETRIGGER );
  for( i=0; consistent && i<arity; ++i )
    consistent = propagate( scope[i], evt );
  return consistent;
} 

bool Constraint::propagate(VariableInt* pvi, const int e) 
{
  bool consistent = true;
  for( int i=0; consistent && i<arity; ++i )
    if( scope[i] != pvi && scope[i]->isLinked() ) { 
      consistent = scope[i]->revise( this, i );
    }
  
  return consistent;
}

bool Constraint::isLinked()
{
  return ( watch[0]->isLinked() && watch[1]->isLinked() );
}

void Constraint::stopWatching( VariableInt* x ) 
{
  // x became unlinked.
  // we thus need to find another variable to replace x
  // lastlinked is set to the index of the new watcher if
  // one could be found
  // or to the index of the last remaining watcher otherwise

  // x == watch[idx]
  int i=arity, idx = (watch[0] != x);
  // find a new variable for replacing the first one (x)
  while( i-- ) {
    if( scope[i] == watch[1-idx] )
      lastlinked = i;
    else if( scope[i] != watch[idx] &&
	     scope[i]->isLinked()
	     ) {

      // stop being watched by that variable
      elements[arity+idx].erase();
      watch[idx] = scope[i];
      scope[i]->watchedConstraints.insert( elements[arity+idx] );

      lastlinked = i;
      break;
    }
  }
  // if a new variable could not be found we remove this constraint
  // from the list of the last linked variable
  if( i < 0 ) {
    oldweight = weight;
    scope[lastlinked]->weight -= weight;
    --scope[lastlinked]->degree;
    elements[lastlinked].erase();
    elements[arity+1-idx].erase();
  } 
}

void Constraint::resumeWatching( VariableInt* x ) 
{
  if( x != scope[lastlinked] ) {
    int idx = (watch[0] != x);
    watch[1-idx]->watchedConstraints.insert( elements[arity+1-idx] );
    triggers[lastlinked]->insert( elements[lastlinked] );
    ++scope[lastlinked]->degree;
    scope[lastlinked]->weight += oldweight;
  }
}  

void Constraint::print(std::ostream& o) const 
{
  o << "(" ;
  for(int i=0; i<arity-1; ++i) {
    scope[i]->printshort( o );
    o << ", ";
  }
  scope[arity-1]->printshort( o );
  o << ")";
}

bool Constraint::constrains( VariableInt *x ) const 
{
  bool cstr = false;
  for(int i=0; !cstr && i<arity; ++i) 
    cstr = ( scope[i] == x ) ;
  return cstr;
}

/**********************************************
 * VariableInt
 **********************************************/

VariableInt::VariableInt( Solver *s ) 
{
  isWord = false;
  id = -1;
  degree = 0;
  weight = 0;
  branch = NULL;
  valueTrigger.head.next  = &(rangeTrigger.head);
  rangeTrigger.head.next  = &(domainTrigger.head);
  domainTrigger.head.pred = &(rangeTrigger.head);
  rangeTrigger.head.pred  = &(valueTrigger.head);
  solver = s;
  id = solver->variables.size();
  solver->variables.push( this );
}

VariableInt::~VariableInt() 
{
  delete branch;
}

void VariableInt::unLink( ) 
//void DecisionVariable::unLink( ) 
{
  *seqIdx = **seqBeg;
  (*seqIdx)->seqIdx = seqIdx;
  seqIdx = *seqBeg;
  **seqBeg = this;
  ++(*seqBeg);

  MistralNode<Constraint*> *nd, *nxt = &(watchedConstraints.head);
  nd = nextNode(nxt);
  while( nd ) {
    nextNode(nxt);
    nd->elt->stopWatching( this );
    nd = nxt;
  }
}

void VariableInt::link( ) 
//void DecisionVariable::link( ) 
{
  --(*seqBeg);

  MistralNode<Constraint*> *nd = &(watchedConstraints.head);
  while( nextNode(nd) ) 
    nd->elt->resumeWatching( this );
  
}

void VariableInt::printConsList( std::ostream& o ) 
{
  o << "x" ;
  if(id >= 0) o << id << "'s neighbors : ";
  
  o.flush();

  MistralNode<Constraint*> *nd = &(valueTrigger.head);

  while( nextNode(nd) ) {
    nd->elt->print( o );
    o << " - ";
    o.flush();
  }

  o  << "X" << endl << "x" ;
  if(id >= 0) o << id << "'s watch list: ";

  o.flush();

  nd = &(watchedConstraints.head);

  while( nextNode(nd) ) {
    nd->elt->print( o );
    o << " - ";
    o.flush();
  }

  o << "X";

}


/**********************************************
 * MistralGraph
 **********************************************/

/// A directed graph implementation.
MistralGraph::MistralGraph( VariableInt **vars, const int l ) 
{
  int i=l, j, lb=INT_MAX, ub=-1;
  while( i-- ) {
    if( vars[i]->min() < lb ) lb = vars[i]->min();
    if( vars[i]->max() > ub ) ub = vars[i]->max();
  }
  n = l+ub-lb+1;

  first = new int *[n];
  last  = new int *[n];    
  node  = new int**[n];

  for(i=0; i<n; ++i) {
    first[i] = new int[n];
    last[i]  = first[i];
    node [i] = new int*[n];
    j = n;
    while( j-- ) node[i][j] = NULL;
  }
  for(i=0; i<l; ++i) {
    j = vars[i]->first();
    do add(i, j);
    while( vars[i]->setNext(j) );
  }  
}

void MistralGraph::getAllCliques( vector< vector< int > >& cliques, const int limit )
{
  MistralSet cand  (0, n-1, MistralSet::full);
  MistralSet K     (0, n-1, MistralSet::empt);
  MistralSet gammaK(0, n-1, MistralSet::full);
  for(int i=0; i<n; ++i)
    {
      if( !degree(i) )
	{
	  cand.erase(i);
	  gammaK.erase(i);
	}
    }
  BronKerbosch( cand, K, gammaK, cliques, limit );
}

int MistralGraph::getAndRemoveMaxCand(MistralSet& cand) {
  int index = -1; 
  if (!cand.empty()) {
    int i = cand.min(), max = 0; 
    do {
      if(degree(i)>max) {
	max = degree(i); 
	index = i; 
      } 
      i = cand.next(i);
    } while( i != INT_MAX );
    cand.erase(index);
  }
  return index; 
}

void MistralGraph::removeNeighbours(const int x, MistralSet& neighbours) 
{
  int *varit;
  for(varit=first[x]; varit!=last[x]; ++varit) 
    neighbours.erase( *varit );
}

void MistralGraph::updateGammaK(const int x, MistralSet& res) 
{
  int i=res.min(), *varit, isntIn;
  do {
    isntIn=true;
    for(varit=first[x]; isntIn && varit!=last[x]; ++varit)
      isntIn = (*varit != i);
    if( isntIn ) res.erase(i);
    i = res.next(i);
  } while( i != INT_MAX );
  res.erase(x);
}

void MistralGraph::BronKerbosch( MistralSet& cand,
				 MistralSet& K,
				 MistralSet& gammaK,
				 vector< vector< int > >& cliques, 
				 const int limit )
{
  if( cliques.size() < limit ) {
    MistralSet updatedCand(cand);
    MistralSet cloneK(K);
    MistralSet cloneGammaK(gammaK);
    
    int x = getAndRemoveMaxCand(updatedCand);	
    removeNeighbours(x, updatedCand);
    if( !updatedCand.empty() )
      BronKerbosch(updatedCand, cloneK, cloneGammaK, cliques, limit);
    
    if( cliques.size() < limit ) {
      MistralSet updatedGammaK(gammaK);  
    
      K.insert(x);
      updateGammaK(x, updatedGammaK);
      if( !updatedGammaK.empty() )
	BronKerbosch(updatedGammaK, K, updatedGammaK, cliques, limit);
      else storeClique( K, cliques );
    }
  }
}

void MistralGraph::storeClique(MistralSet& K, vector< vector< int > >& cliques) 
{
  vector<int> clique;
  int i=K.min();
  do { 
    //    cout << i << " ";
    clique.push_back(i);
    i = K.next(i);
  } while( i != INT_MAX );
  //  cout << endl;
  cliques.push_back( clique );
}


void MistralGraph::getSomeCliques( vector< vector< int > >& cliques )
{
  int x, *varit, isNeighbor, i;

  MistralSet visited( 0, n-1, MistralSet::empt );
  MistralStack<int> vertice;
  for(x=0; x<n; ++x)
    vertice.push( x );

  while( !vertice.empty() )
    {
      // arbitrarily pick a node x
      vertice.pop( x );

      //cout << x << " ";

      if( !visited.member( x ) ) {

	//cout << "start a new clique" << endl;

	visited.insert( x );
	vector<int> neighbors;
	neighbors.push_back( x );

	// add all neighbors as long as they form a clique
	for(varit=first[x]; varit!=last[x]; ++varit) {

	  //cout << "\t" << (*varit) ;

	  isNeighbor = true;
	  i = neighbors.size();
	  while( isNeighbor && --i )
	    isNeighbor = isIn( *varit, neighbors[i]);
	  if( isNeighbor ) {
	    neighbors.push_back( *varit );
	    visited.insert( *varit );
	    //cout << " in" << endl;
	  } //else cout << " out" << endl;
	}
	
	cliques.push_back( neighbors );

//  	for(int y=0; y<neighbors.size(); ++y)
//  	  cout << neighbors[y] << " ";
//  	cout << endl;
// 	visited.print( cout );
// 	cout << endl;

      } //else cout << " already in a clique" << endl;
    }
}

MistralGraph::MistralGraph( const int l, vector<int> E ) 
{
  n=l;
  int i, j;

  //  cout << n << " " << (E.size()) << endl;

  first = new int *[n];
  last  = new int *[n];    
  node  = new int**[n];

  for(i=0; i<n; ++i) {
    first[i] = new int[n];
    last[i]  = first[i];
    node [i] = new int*[n];
    j = n;
    while( j-- ) node[i][j] = NULL;
  }

  i=E.size();
  while( i ) {
    i-=2;
    add(E[i], E[i+1]);
    add(E[i+1], E[i]);
  }
}

MistralGraph::MistralGraph( const int l ) 
{
  init(l);
}

void MistralGraph::init( const int l )
{
  n = l;

  first = new int*[n];
  last  = new int*[n];
  node  = new int**[n];
  int i = n, j;
  while( i-- ) {
    first[i] = new int[n];
    last[i]  = first[i];
    node[i] = new int*[n];
    j = n;
    while( j-- ) node[i][j] = NULL;
  }
}

MistralGraph::~MistralGraph()
{
  cleanUp();
}

void MistralGraph::cleanUp()
{
  for(int i=0; i<n; ++i) {
    delete [] first[i];
    delete [] node[i];
  }
  delete [] first; 
  delete [] last; 
  delete [] node;
}

void MistralGraph::add(const int x, const int y)
{
  if( !node[x][y] ) {
    *last[x] = y;
    node[x][y] = last[x]++;
  }
}

void MistralGraph::remove(const int x, const int y)
{
  if( node[x][y] ) {
    --last[x];
    *node[x][y] = *last[x];
    node[x][*last[x]] = last[x];
    node[x][y] = NULL;
  }
}

int MistralGraph::degree(const int x) const
{
  return(last[x] - first[x]);
}

void MistralGraph::dfs( const int v ) const
{
  MistralSet visited( 0, n-1, MistralSet::empt );
  MistralStack<int> branches( n );
  int x, *varit;
  branches.push( v );

  while( branches.size_ ) {
    branches.pop( x );
    visited.insert( x );

    cout << x << endl;

    for(varit=first[x]; varit!=last[x]; ++varit) {
      if( !visited.member( *varit ) )
	branches.push( *varit );
    }
  }
}

/*!
 * return the maximum cardinality of a matching in a
 * bipartite graph
 */
int MistralGraph::AltBlumMehlorn( int nv )
{
//   int m=0, i=n, j, *k;
//   while( i-- )
//     m+=degree(i);

//   int LMAX = (int)(sqrt(n) * sqrt(m * log(n)) / n)+4;

//   int layer[n], isfree[n];
//   int next_free[n*(n+1)], pred_free[n*(n+1)];
//   int cardinality=0, l, vertex, path[n], length=0;

//   std::fill(isfree, isfree+n, 1);
//   std::fill(next_free, next_free+(n*(n+1)), INT_MAX);
//   std::fill(pred_free, pred_free+(n*(n+1)), INT_MAX);
 
//   // values are on layer 0
//   std::fill(layer+nv, layer+n, 0);
//   j = n;
//   for(i=nv; i<n; ++i) {
//     next_free[j] = i;
//     pred_free[i] = j;
//     j = i;
//   }

//   // variables are on layer 1
//   std::fill(layer, layer+nv, 1);
//   j = n;
//   for(i=0; i<nv; ++i) {
//     next_free[n+1+j] = i;
//     pred_free[n+1+i] = j;
//     j = i;
//   }

//   //g.invariant(layer);

//   l = 1;
//   while( l < n && cardinality < nv ) {

//     //cout << "layer " << l << endl;
//     //g.invariant(layer);

//     while( (vertex = next_free[l*(n+1) + n]) != INT_MAX ) {

//       //g.invariant(layer);

//       path[length++] = vertex;
//       while( length ) {
// 	if( !layer[path[length-1]] && isfree[path[length-1]] ) {
// 	  // we found an augmenting path
// 	  ++cardinality;

// 	  // Both ends of the path are now matched 
// 	  isfree[path[0]] = 0;
// 	  isfree[path[length-1]] = 0;

// 	  for(j=0; j<length; j+=(length-1)) {
// 	      i = (layer[path[j]] * (n+1)) + path[j];
// 	      next_free[(layer[path[j]] * (n+1)) + pred_free[i]] = next_free[i];
// 	      pred_free[(layer[path[j]] * (n+1)) + next_free[i]] = pred_free[i];
// 	      pred_free[i] = INT_MAX;
// 	      next_free[i] = INT_MAX;
// 	      if( length < 2 ) break;
// 	  }

// 	  // reverse all edges in path;
// 	  i=0;
// 	  while( i < length-1 ) {
// 	    swap(path[i],path[i+1]);
// 	    ++i;
// 	  }

// 	  // clear the path
// 	  length = 0;

// 	  //g.invariant(layer);

// 	} else {
	  
// 	  j = path[length-1];
// 	  //do if( (j = next[path[length-1]][j]) == INT_MAX ) break;
// 	  do {
// 	    k = node[path[length-1]][j]+1;
// 	    if( k = last[path[length-1]] ) break;
// 	    j = *k;
// 	  }
// 	  while( layer[path[length-1]] != layer[j]+1 );
	  
// 	  if( j != INT_MAX ) {

// 	    // add jnxt to the path
// 	    path[length++] = j;
	    
// 	  }
// 	  else {

// 	    --length;
// 	    if( isfree[path[length]] ) {
// 	      i = (layer[path[length]] * (n+1)) + path[length];
// 	      next_free[(layer[path[length]] * (n+1)) + pred_free[i]] = next_free[i];
// 	      pred_free[(layer[path[length]] * (n+1)) + next_free[i]] = pred_free[i];
// 	      pred_free[i] = INT_MAX;
// 	      next_free[i] = INT_MAX;
// 	    }
// 	    layer[path[length]] += 2;

// 	    if( isfree[path[length]] ) {
// 	      i += 2*(n+1);
// 	      j = (layer[path[length]] * (n+1)) + n;
// 	      pred_free[(layer[path[length]] * (n+1)) + next_free[j]] = path[length];
// 	      next_free[i] = next_free[j];
// 	      next_free[j] = path[length];
// 	      pred_free[i] = n;
// 	    }

// 	    //g.invariant(layer);
// 	  }
// 	}
//       }
//     }
//     l += 2;
//   }

// //   g.print( cout );
// //   cout << endl;

// //   for(int x=0; x<n; ++x) {
// //     bool empty = true;
// //     for(int y=0; y<=n; ++y) 
// //       if( layer[y] == x ) empty = false;
// //     if( empty ) continue;

// //     cout << "layer" << x << ":";
// //     for(int y=0; y<=n; ++y) 
// //       if( layer[y] == x )
// // 	cout << setw(3) << y;
// //     cout << endl;
// //   }

// //   for(int x=0; x<n; ++x) {
// //     bool empty = ((j = next_free[x*(n+1)+n]) == INT_MAX);
// //     if(empty) continue;
// //     cout << "free" << x << ":";
// //     do {
// //       cout << setw(3) << j ;
// //       cout.flush();
// //       assert(isfree[j]);
// //       assert(layer[j] == x);
// //       j = next_free[x*(n+1)+j];
// //     } while( j != INT_MAX );
// //     cout << endl;
// //   }

// //   cout << cardinality << endl;

//   return cardinality;  
}


void MistralGraph::print(std::ostream& o) const
{
  int i, *varit;
  for(i=0; i<n; ++i) {
    cout << i << ": ";
    for(varit=first[i]; varit!=last[i]; ++varit)
      cout << " ->" << setw(3) << (*varit);
    cout << endl;
  }
}

ostream& operator<< (ostream& os, const MistralGraph& g)
{
  g.print( os );
  return os;
}


