
/*
  Mistral is a constraint satisfaction and optimisation library
  Copyright (C) 2003-2005  Emmanuel Hebrard
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  The author can be contacted electronically at 
  ehebrard@cse.unsw.edu.au.
*/

#include <mod.h>
#include <sol.h>


void BuildObject::add( CSP *m, const int l, const int info )
{
  ++refs;

  if( l ) setVar();
  else if(isRel()) unsetVar();
  
  if( notIn() ) {
    model = m;
    model->add( this, l );
   }
}


/**********************************************
 * Variable BuildObject
 **********************************************/

void BuildObject::initVar(const int lb, const int ub) 
{
  refs = 0;
  id = -1;
  min_ = lb;
  max_ = ub;
  size_ = (ub-lb+1);
  state = (CRANGE | CLIST | PAR | VLIST | ((size_ > 0 ? VRANGE : 0) | (size_ == 1 ? CONST : 0)));
  veqptr_ = NULL;
  varptr_ = NULL;
  domain_ = NULL;
  model = NULL;
}

BuildObject::BuildObject() 
{
  initVar(0,-1);
}

BuildObject::BuildObject( const int lo, const int up ) 
{
  initVar(lo, up);
}

BuildObject::BuildObject( const int* array, const int length ) 
{
  min_ = INT_MAX;
  max_ = -INT_MAX;
  for(int i=0; i<length; ++i) {
    if(min_ > array[i]) min_ = array[i];
    if(max_ < array[i]) max_ = array[i];
  }
  initVar(min_, max_);
  size_ = length;
  if( length < (max_ - min_ + 1) ) {
    state |= (size_ == 1 ? CONST : 0); 
    domain_ = new MistralSet(min_, max_, MistralSet::empt);
    for(int i=0; i<length; ++i) domain_->insert( array[i] );
  } else state |= (VRANGE | (size_ == 1 ? CONST : 0));
}

BuildObject::BuildObject( const int* array, 
			  const int length, 
			  const int lo, 
			  const int up ) 
{
  initVar(lo, up);
  size_ = length;
  if( length < (max_ - min_ + 1) ) {
    state |= (size_ == 1 ? BuildObject::CONST : 0);
    domain_ = new MistralSet(min_, max_, MistralSet::empt);
    for(int i=0; i<length; ++i) domain_->insert( array[i] );
  } else state |= (VRANGE | (size_ == 1 ? CONST : 0));
}

BuildObject *BuildObject::getBuildObject()
{
  return (veqptr_ ? veqptr_->getBuildObject() : this);
}

const BuildObject *BuildObject::getConstBuildObject() const
{
  return (veqptr_ ? veqptr_->getConstBuildObject() : this);
}

void BuildObject::build(Solver *s)
{
  if( !veqptr_ && (isRel() || isReferenced()) ) {
    if( size_ == 1) {
      map<int,VariableInt*>::iterator iter = s->constants.find( (int)min_ );
      if( iter != s->constants.end() ) {
	varptr_ = iter->second;
      } else {// this constant as not yet been created
	varptr_ = new Constant( s, min_ );
	s->constants[(int)min_] = varptr_;
      }
    } else if( size_ == 2 && min_ == 0 && max_ == 1 ) {
      varptr_ = new VariableBool(s);
    } else if( size_ == (max_ - min_ + 1) ) {
      if( isVRange() && size_ > 64 ) {
	//cout << "range ";
	//cout.flush();
	varptr_ = new VariableRange( s, min_, max_ );
      } else {
	if( isVList() ) {
	  varptr_ = new VariableList( s, min_, max_ );
	} else {
	  //cout << "bit ";
	  //cout.flush();
	  varptr_ = new VariableBit( s, min_, max_ );
	}
      }
    } else {
      //assert( domain_ != NULL );
      if( isVList() ) {
	varptr_ = new VariableList( s, domain_, size_, min_, max_ );
      } else {
	varptr_ = new VariableBit( s, domain_, size_, min_, max_ );
      }
    }
  }//  else {
//     varptr_ = veqptr_->varptr_;
//   }
}

void BuildObject::releaseEvent( const int sb )
{
  int sn = size(), evt;
  if( sb != sn ) {
    evt = (sn == 1 ? 7 : 1);
    model->triggerEvent( this, evt );
    if( veqptr_ ) model->triggerEvent( getBuildObject(), evt );
  }
}

void BuildObject::releaseEvent( )
{
  model->triggerEvent( this, VALUETRIGGER );
  if( veqptr_ ) model->triggerEvent( getBuildObject(), VALUETRIGGER );
}


int BuildObject::_size() const
{
  return size_;
}

int BuildObject::size() const
{
  return getConstBuildObject()->_size();
}

int BuildObject::_min() const
{
  return min_;
}

int BuildObject::min() const
{
  return getConstBuildObject()->_min();
}

int BuildObject::_max() const
{
  return max_;
}

int BuildObject::max() const
{
  return getConstBuildObject()->_max();
}

int BuildObject::_domsize() const
{
  return size_;
}

int BuildObject::domsize() const
{
  return getConstBuildObject()->_domsize();
}

int BuildObject::_value() const
{
  return min_;
}

int BuildObject::value() const
{
  return getConstBuildObject()->_value();
}

int BuildObject::_first() const
{
  return min_;
}

int BuildObject::first() const
{
  return getConstBuildObject()->_first();
}

int BuildObject::_last() const
{
  return max_;
}

int BuildObject::last() const
{
  return getConstBuildObject()->_last();
}

int BuildObject::_minCapacity() const
{
  return min_;
}

int BuildObject::minCapacity() const
{
  return getConstBuildObject()->_minCapacity();
}

int BuildObject::_maxCapacity() const
{
  return max_;
}

int BuildObject::maxCapacity() const
{
  return getConstBuildObject()->_maxCapacity();
}

int BuildObject::_minPosAbs() const 
{
  if( max_ < 0 ) return max_;
  if( min_ > 0 ) return min_;
  if( !domain_ || domain_->member( 0 ) ) return 0;
  return domain_->next( 0 );
}

int BuildObject::minPosAbs() const
{
  return getConstBuildObject()->_minPosAbs();
}

int BuildObject::_minNegAbs() const 
{
  if( min_ > 0 ) return min_;
  if( max_ < 0 ) return max_;
  if( !domain_ || domain_->member( 0 ) ) return 0;
  return domain_->prev( 0 );
}

int BuildObject::minNegAbs() const
{
  return getConstBuildObject()->_minNegAbs();
}

bool BuildObject::_remove(const int v) 
{
  if( min_ <= v && max_ >= v ) {
    if( domain_ != NULL ) {
      if(domain_->member(v)) {
	domain_->erase( v );
	--size_;
	if( v == min_ )
	  min_ = domain_->min();
	if( v == max_ )
	  max_ = domain_->max();
      }
    } else {
      if( v == min_ ) {
	--size_;
	++min_;
      } else if( v == max_ ) {
	--size_;
	--max_;
      } else {
	domain_ = new MistralSet(min_, max_, MistralSet::full);
	domain_->erase( v );
	--size_;
      }
    }
  }
 
  return size_ > 0;
}

bool BuildObject::remove(const int v) 
{
  int sb = size();
  bool rval = getBuildObject()->_remove( v );
  releaseEvent( sb );
  return rval;
}

bool BuildObject::_setDomain(const int v) 
{
  if( domain_ != NULL )
    delete domain_;
  size_ = 1;
  min_ = v;
  max_ = v;
  return size_ > 0;
}

bool BuildObject::setDomain(const int v) 
{
  if( size() > 1 ) {
    releaseEvent();
    setConst();
  }
  return getBuildObject()->_setDomain( v );
}

void BuildObject::_setEqual(BuildObject *v) 
{

//   cout << "SET EQUAL " << endl;
//   print( cout );
//   cout << endl;
//   v->print( cout );
//   cout << endl;

  if( !v->isVRange() )
    unsetVRange();
  if( v->domain_ ) {
    if( domain_ == NULL ) 
      domain_ = new MistralSet( min_, max_, MistralSet::full );
    domain_->intersectWith( v->domain_ );
    min_ = domain_->min();
    max_ = domain_->max();
    size_ = domain_->size();
  } else {
    _setMin( v->min_ );
    _setMax( v->max_ );
  }
  v->veqptr_ = this;

//   print( cout );
//   cout << endl;
//   v->print( cout );
//   cout << endl << endl;
}

void BuildObject::setEqual(BuildObject *v) 
{
  if( getConstBuildObject() != v->getConstBuildObject() )
    getBuildObject()->_setEqual( v->getBuildObject() );
}

bool BuildObject::_setMin(const int l) 
{
  if( l > min_ ) { 
    if( domain_ != NULL ) {
      domain_->setMin(l);
      size_ = domain_->size();
      min_ = domain_->min();
    } else {
      size_ -= ( l - min_ );
      min_ = l;
    }
  }
  return size_ > 0;
}

bool BuildObject::setMin(const int l) 
{
  return getBuildObject()->_setMin( l );
}

bool BuildObject::_setMax(const int u)
{
  if( u < max_ ) { 
    if( domain_ != NULL ) {
      domain_->setMax(u);
      size_ = domain_->size();
      max_ = domain_->max();
    } else {
      size_ -= ( max_ - u);
      max_ = u;
    }
  }
  return size_ > 0;
}

bool BuildObject::setMax(const int u) 
{
  return getBuildObject()->_setMax( u );
}

bool BuildObject::_setDomain(const MistralSet& s) 
{
  if( domain_ == NULL ) 
    domain_ = new MistralSet(min_, max_, MistralSet::full);
  domain_->intersectWith( s );
  size_ = domain_->size();
  min_ = domain_->min();
  max_ = domain_->max();
  return size_ > 0;
}

bool BuildObject::setDomain(const MistralSet& s) 
{
  return getBuildObject()->_setDomain( s );
}

bool BuildObject::_removeSet(const MistralSet& s) 
{
  if( domain_ == NULL ) 
    domain_ = new MistralSet(min_, max_, MistralSet::full);
  domain_->setminusWith( s );
  size_ = domain_->size();
  min_ = domain_->min();
  max_ = domain_->max();
  return size_ > 0;
}

bool BuildObject::removeSet(const MistralSet& s) 
{
  return getBuildObject()->_removeSet( s );
}

bool BuildObject::_removeRange(const int l, const int u) 
{
  if( l <= min_ )
    setMin( u+1 );
  else if( u >= max_ )
    setMax( l-1 );
  else {
    if( domain_ == NULL ) 
      domain_ = new MistralSet( min_, max_, MistralSet::full );
    domain_->removeInterval( l, u );
    size_ = domain_->size();
  }
  return size_ > 0;
}

bool BuildObject::removeRange(const int l, const int u) 
{
  return getBuildObject()->_removeRange(l,u);
}

bool BuildObject::_contain(const int v) const 
{ 
  return (v >= min_ &&
	  v <= max_ &&
	  ( !domain_ || domain_->member(v) ));
}

bool BuildObject::contain(const int v) const
{
  return getConstBuildObject()->_contain( v );
}

int BuildObject::_getNext(const int v) const 
{ 
  if( domain_ != NULL )
    return ( domain_->next(v) ); 
  else return ( v < max_ ? v+1 : INT_MAX );
}

int BuildObject::getNext(const int v) const
{
  return getConstBuildObject()->_getNext( v );
}

bool BuildObject::_good(const int v) const 
{ 
  return ( v != INT_MAX ); 
}

bool BuildObject::good(const int v) const
{
  return getConstBuildObject()->_good(v);
}

bool BuildObject::_setNext(int& v) const 
{ 
  if( domain_ != NULL )
    return ( (v = domain_->next(v)) != INT_MAX ); 
  else return ( ++v <= max_ );
}

bool BuildObject::setNext(int& v) const
{
  return getConstBuildObject()->_setNext(v);
}

bool BuildObject::_isRange() const
{
  return (domain_ == NULL);
}

bool BuildObject::_isGround() const
{
  return (size_ <= 1);
}

bool BuildObject::isRange() const
{
  return getConstBuildObject()->_isRange();
}

bool BuildObject::isGround() const
{
  return getConstBuildObject()->_isGround();
}


/**********************************************
 * Constraint BuildObject
 **********************************************/ 

void BuildObjectConstraint::add( CSP *m, const int l, 
				 BuildObjectPredicate* pred ) 
{
  int i, n=pred->arity;
  for(i=0; i<n; ++i) {
    pred->scope[i]->add( m, l+1, pred->getState() );
  }
}

bool BuildObjectConstraint::preBuild( BuildObjectPredicate *pred )
{  
  BuildObject **x = pred->scope;
  bool allconstant = true;
  int n=pred->arity, i=n;
  while( allconstant && i-- )
    allconstant = (x[i]->size() == 1);  
  if( allconstant )
    {
      pred->unsetRel();
      i=n;
      while( i-- ) {
	x[i]->deReference();
      }
      //pred->setFixed();
    }
  return !allconstant;
}


/**********************************************
 * Predicate BuildObject
 **********************************************/
BuildObjectPredicate::BuildObjectPredicate(BuildObject **x, const int n, 
					   const int lb, const int ub, 
					   BuildObjectConstraint *r, int *p, 
					   int istate)
  : BuildObject(lb, ub)
{
  arity = n;
  params = p;
  scope = x;
  scope[n] = this;
  relation = r;
  setRel();

  // should never been a top level constraint
  if( istate & AWV ) setAwv();

  // preserve ranges
  if( istate & CRANGE ) setCRange();
  else unsetCRange();

  // reasons only about bounds
  if( istate & CONVEX ) setConvex();
  else unsetConvex();

  if( istate & CLIST ) setCList();
  else unsetCList();

//   print( cout );
//   cout << endl;

}

void BuildObjectPredicate::add( CSP *m, const int l, const int inf )
{
  for(int i=0; i<arity; ++i)
    scope[i]->parent.push( this );
  relation->add( m, l, this );
  BuildObject::add( m, l, inf );
}

bool BuildObjectPredicate::preBuild( ) 
{
  bool res = true;
  if( !isFixed() ) {
    res = relation->preBuild( this );
    if( !res ) setFixed();
  }
  return res;
}


void BuildObjectPredicate::build(Solver *s) 
{

//   cout << "build ";
//   print( cout );
//   cout << " " << isReferenced() << " " << isVar() << endl;

  if( isVar() ) {
    //cout << "build var" << endl;
    BuildObject::build( s );
  }

  if( isRel() ) {
    int n=arity+isVar();
    VariableInt *tmpvars[n];
    for(int i=0; i<n; ++i) { 
      tmpvars[i] = scope[i]->getVariable();
//       cout << i << " ";
//       cout << (scope[i]->getBuildObject()) << " ";
//       cout << (scope[i]->getBuildObject())->varptr_ << " ";
//       cout << tmpvars[i] << endl;
//       cout <<  scope[i]->getVarId() << endl;
//       scope[i]->print( cout );
//       cout << endl << endl;
    }
    relation->build( s, tmpvars, this );
  }
}

void BuildObjectPredicate::printshort(std::ostream& o) const
{
  if( isRel() )
    relation->print( o, this );
  else if( isVar() )
    BuildObject::printshort( o );
}
 
void BuildObjectPredicate::print(std::ostream& o) const
{

  if( getConstBuildObject()->varptr_ ) {
    getConstBuildObject()->varptr_->print( o );    
    o << " <-> "; 
  } else {
    if( isVar() ) {
      BuildObject::print( o );
      o << " <-> "; 
    }
  }
  if( isRel() ) {
    o << "(" << arity << ") "; 
    
    //cout << endl << scope[0]->getVarId() << " " 
    //<< scope[1]->getVarId() << endl;
    
    relation->print( o, this );
//     if( isVar() ) {
//       if( domain_ != NULL ) {
// 	o << " in ";
// 	domain_->print( o );
//       } else if( size_ != 1 ) 
// 	o << " in {" << min_ << ",.." << max_ << "}";
//     }
  }
}

string BuildObjectPredicate::xmlPred( int& idx, int level, int *sub ) const
{
  //if( relation != ENV[ConstraintStore::ELEMENT] ) sub[getVarId()] = 1;
  return relation->xmlPred( idx, level, this, sub );
}


/**********************************************
 * Variable Wrapper
 **********************************************/
BuildObject **VarArray::getArgs() 
{
  int n = array_.size();
  BuildObject **tmp = new BuildObject*[n+1];
  while( n-- ) tmp[n] = array_[n].var_ptr_;  
  return tmp;
}

VarArray Variable::operator,( Variable x )
{
  VarArray table(2);
  table[0] = *this;
  table[1] = x;
  return table;
}

Variable::Variable( const int nvals ) 
{
  //val = INT_MAX;
  var_ptr_ = new BuildObject( 0, nvals-1 );
  var_ptr_->setVar();
}

Variable::Variable( const Variable& v )
{
  //val = v.val;
  var_ptr_ = v.var_ptr_;
}

Variable::Variable( BuildObject *v )
{
  // val = INT_MAX;
  var_ptr_ = v;
}

Variable::Variable( const int lo, const int up ) 
{
  //  val = INT_MAX;
  var_ptr_ = new BuildObject(lo, up);
  var_ptr_->setVar();
}

Variable::Variable( const int* array, const int length ) 
{
  //  val = INT_MAX;
  var_ptr_ = new BuildObject(array, length);
  var_ptr_->setVar();
}

Variable::Variable( const int* array, const int length, 
		    const int lo, const int up ) 
{
  //  val = INT_MAX;
  var_ptr_ = new BuildObject(array, length, lo, up);
  var_ptr_->setVar();
}


/**********************************************
 * VarArray Wrapper
 **********************************************/

VarArray::VarArray( const int nvars )
{
  array_.resize( nvars );
}

VarArray::VarArray( const int nvars, const int nvals )
{
  array_.resize( nvars );
  int i=nvars;
  while( i-- ) {
    array_[i] = Variable( nvals );
  }
}

VarArray::VarArray( const int nvars, const int lo, const int up ) 
{
  array_.resize( nvars );
  int i=nvars;
  while( i-- ) {
    array_[i] = Variable( lo, up );
  }
}

VarArray::VarArray( const int nvars, const int* array, const int length ) 
{
  array_.resize( nvars );
  int i=nvars;
  while( i-- ) {
    array_[i] = Variable( array, length );
  }
}

VarArray::VarArray( const int nvars, const int* array, const int length, const int lo, const int up )
{
  {
    array_.resize( nvars );
    int i=nvars;
    while( i-- ) {
      array_[i] = Variable( array, length, lo, up );
    }
  }
}

VarArray VarArray::operator,( Variable x )
{
  int i=size();
  VarArray table(i+1);
  table[i] = x;
  table[i] = x;
  while( i-- ) 
    table[i] = array_[i];
  return table;
}

VarArray VarArray::operator,( VarArray& x )
{
  int i=size(), j=x.size();
  VarArray table(i+j);
  while( j-- )
    table[i+j] = x.array_[j];
  while( i-- )
    table[i] = array_[i];
  return table;
}


/**********************************************
 *  Offset Predicate BuildObject
 **********************************************/ 
void BuildObjectOffset::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  if( pred->isRel() ) {
    Constraint *con = new PredicateOffset( s, tmp, pred->params[0] );
  }
}

BuildObjectPredicate* BuildObjectOffset::declare( BuildObject *x_ , int o )
{
  BuildObject **x = new BuildObject*[2];
  x[0] = x_;
  int *params = new int[1];
  *params = o;
  return new BuildObjectPredicate( x, 1, x[0]->min()+o, x[0]->max()+o, 
				   this, params, (BuildObject::AWV | BuildObject::CRANGE) );
}

bool BuildObjectOffset::preBuild( BuildObjectPredicate *pred )
{
  int res = BuildObjectConstraint::preBuild( pred );
  //if( BuildObjectConstraint::preBuild( pred ) ) {
  if( res ) {
    BuildObject **scope=pred->scope;
    if( !pred->params[0] ) {

      //cout << "equal0" << endl;

      pred->unsetRel();
//       //pred->unsetVar();
//       scope[0]->print( cout );
//       cout << " == " ;
//       scope[1]->print( cout );
//       //cout << endl;
//       cout << " == ";
//       scope[2]->print( cout );
//       cout << endl;

      int i = ( scope[0]->getVarId() < scope[1]->getVarId() );
      scope[1-i]->setEqual( scope[i] );
      //scope[i]->setEqual( scope[1-i] );
      //res = false;
      //pred->setFixed();
    }
  }
  return res;
}

int BuildObjectOffset::computeValue( BuildObjectPredicate *pred ) const
{
  return (pred->scope[0]->value() + pred->params[0]);
}

void BuildObjectOffset::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "(";
  pred->scope[0]->printshort( o );
  if( pred->params[0] < 0 )
    o << " - " << (-1 * pred->params[0]);
  else 
    o << " + " << pred->params[0];
  o << ")";
}

string BuildObjectOffset::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string x1 = pred->scope[0]->xmlPred(idx, level+1, sub);
  string s = ( "add(" + x1 + "," + int2string(pred->params[0]) + ")" );
  return s;
}
  

/**********************************************
 *  Factor Predicate BuildObject
 **********************************************/ 
void BuildObjectFactor::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  if( pred->isRel() )
    Constraint *con = new PredicateFactor( s, tmp, pred->params[0] );
}

BuildObjectPredicate* BuildObjectFactor::declare( BuildObject *x_, int o )
{
  BuildObject **x = new BuildObject*[2];
  x[0] = x_;
  int *params = new int[1];
  *params = o;
  return new BuildObjectPredicate( x, 1,
				   (o < 0 ? x[0]->max()*o : x[0]->min()*o), 
				   (o < 0 ? x[0]->min()*o : x[0]->max()*o), 
				   this, params, (BuildObject::AWV | BuildObject::CRANGE) );
}

void BuildObjectFactor::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "(";
  o << pred->params[0] << " * ";
  pred->scope[0]->printshort( o );
  o << ")";
}

bool BuildObjectFactor::preBuild( BuildObjectPredicate *pred )
{
  //if( BuildObjectConstraint::preBuild( pred ) ) {
  bool res = BuildObjectConstraint::preBuild( pred );
  if( res ) {
    BuildObject **scope=pred->scope;
    if( pred->params[0] == 1 ) {
      
      //cout << "equal1" << endl;

      pred->unsetRel();
      //pred->unsetVar();

      int i = ( scope[0]->getVarId() < scope[1]->getVarId() );
      scope[1-i]->setEqual( scope[i] );
      //scope[i]->setEqual( scope[1-i] );

      //      int i = ( scope[0]->id < scope[1]->id );
      //      scope[1-i]->setEqual( scope[i] );
      //res = false;
      //pred->setFixed();
    }
  }
  return res;
}

int BuildObjectFactor::computeValue( BuildObjectPredicate *pred ) const
{
  return (pred->scope[0]->value() * pred->params[0]);
}

string BuildObjectFactor::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string x1 = pred->scope[0]->xmlPred(idx, level+1, sub);
  string s = ( "mul(" + int2string(pred->params[0]) + "," +  x1 + ")" );
  return s;
}


/**********************************************
 *  Negation Predicate BuildObject
 **********************************************/ 
void BuildObjectNegation::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred )
{
  Constraint *con = new PredicateNegation( s, tmp );
} 

bool BuildObjectNegation::preBuild( BuildObjectPredicate *pred )
{
  return true;
//   BuildObject **x = pred->scope;
//   bool allconstant = true;
//   int n=pred->arity, i=n;
//   while( allconstant && i-- )
//     allconstant = (x[i]->size() == 1);
//   if( allconstant )
//     {
//       //pred->unsetRel();
//       i=n;
//       while( i-- )
// 	x[i]->deReference();

//       assert( x[0]->min() == (-1*x[1]->min()) );

//     }
//   assert( pred->isReferenced() );
}

int BuildObjectNegation::computeValue( BuildObjectPredicate *pred ) const
{
  return (-pred->scope[0]->value());
}

void BuildObjectNegation::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "-";
  pred->scope[0]->printshort( o );
}

string BuildObjectNegation::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string x1 = pred->scope[0]->xmlPred(idx, level+1, sub);
  string s = ( "neg(" + x1 + ")" );
  return s;
}


/**********************************************
 *  Not Predicate BuildObject
 **********************************************/ 
void BuildObjectNot::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  if( pred->isVar() )
    Constraint *con = new PredicateNot( s, tmp );
}

bool BuildObjectNot::preBuild( BuildObjectPredicate *pred )
{
  bool res = BuildObjectConstraint::preBuild( pred ); 
  if( res ) {
    int pmax = !pred->max();
    //if( pmax || pred->min() == 1 || !pred->isVar() ) {
    res = ( !pmax && pred->min() != 1 && pred->isVar() );
    if( !res ) {
      pred->unsetRel();
      BuildObject *x=pred->scope[0];
      if( pmax ) {
	x->remove( 0 );
      }
      else {
	x->setDomain( 0 );
      }
    } 
  }
  return res;
}

int BuildObjectNot::computeValue( BuildObjectPredicate *pred ) const
{
  return (!pred->scope[0]->value());
}

void BuildObjectNot::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "!";
  pred->scope[0]->printshort( o );
}

string BuildObjectNot::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string x1 = pred->scope[0]->xmlPred(idx, level+1, sub);
  string s = ( "not(" + x1 + ")" );
  return s;
}


/**********************************************
 *  LowerBound Predicate BuildObject
 **********************************************/ 
void BuildObjectLowerBound::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  if( pred->isVar() )
    Constraint *con = new PredicateLowerBound( s, tmp, pred->params[0] );
}

bool BuildObjectLowerBound::preBuild( BuildObjectPredicate *pred )
{
  bool res = BuildObjectConstraint::preBuild( pred );
  if( res ) {
    int pmax = !pred->max();
    if( pmax || pred->min() )
      pred->unsetVar();
    if( !pred->isVar() ) {

      //cout << "set lb" << endl;

      pred->unsetRel();
      if( pmax ) pred->scope[0]->setMax( pred->params[0]-1 );
      else pred->scope[0]->setMin( pred->params[0] );
      res = false;
    }

    //if( pred->scope[0]->size() == 1 ) {
    //cout << pred->id << " HERE" << endl;
    //}

  } // else {
//     if( pred->scope[0]->min() < pred->params[0] )
//       pred->setDomain(0);
//     else
//       pred->setDomain(1);
//   }
  return res;
}

int BuildObjectLowerBound::computeValue( BuildObjectPredicate *pred ) const
{
  return (pred->scope[0]->value() >= pred->params[0]);
}

BuildObjectPredicate* BuildObjectLowerBound::declare( BuildObject *x_ , int b )
{
  BuildObject **x = new BuildObject*[2];
  x[0] = x_;
  int *params = new int[1];
  *params = b;
  return new BuildObjectPredicate( x, 1, 0, 1, this, params, 				   
				   (BuildObject::CRANGE | BuildObject::CONVEX) );
}



void BuildObjectLowerBound::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "(";
  pred->scope[0]->printshort( o );
  o << " >= " << (pred->params[0]) << ")";
}

string BuildObjectLowerBound::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string x1 = pred->scope[0]->xmlPred(idx, level+1, sub);
  string s = ( "ge(" + x1 + "," + int2string(pred->params[0]) + ")" );
  return s;
}


/**********************************************
 *  UpperBound Predicate BuildObject
 **********************************************/ 
void BuildObjectUpperBound::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  if( pred->isVar() )
    Constraint *con = new PredicateUpperBound( s, tmp, pred->params[0] );
}

bool BuildObjectUpperBound::preBuild( BuildObjectPredicate *pred )
{
  bool res = BuildObjectConstraint::preBuild( pred );
  if( res ) {
    int pmax = !pred->max();
    if( pmax || pred->min() )
      pred->unsetVar();
    if( !pred->isVar() ) {

      //cout << "set ub" << endl;

      pred->unsetRel();
      if( pmax ) pred->scope[0]->setMin( pred->params[0]+1 );
      else pred->scope[0]->setMax( pred->params[0] );
      res = false;
    }
  }
  return res;
}

int BuildObjectUpperBound::computeValue( BuildObjectPredicate *pred ) const
{
  return (pred->scope[0]->value() <= pred->params[0]);
}

BuildObjectPredicate* BuildObjectUpperBound::declare( BuildObject *x_ , int b )
{
  BuildObject **x = new BuildObject*[2];
  x[0] = x_;
  int *params = new int[1];
  *params = b;
  return new BuildObjectPredicate( x, 1, 0, 1, 
				   this, params, (BuildObject::CRANGE | BuildObject::CONVEX) );
}

void BuildObjectUpperBound::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "(";
  pred->scope[0]->printshort( o );
  o << " <= " << (pred->params[0]) << ")";
}

string BuildObjectUpperBound::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string x1 = pred->scope[0]->xmlPred(idx, level+1, sub);
  string s = ( "le(" + x1 + "," + int2string(pred->params[0]) + ")" );
  return s;
}



/**********************************************
 *  Abs Predicate BuildObject
 **********************************************/ 
void BuildObjectAbs::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  Constraint *con = new PredicateAbs( s, tmp );
}

int BuildObjectAbs::computeValue( BuildObjectPredicate *pred ) const
{
  return (abs(pred->scope[0]->value()));
}

BuildObjectPredicate* BuildObjectAbs::declare( BuildObject *x_ )
{
  BuildObject **x = new BuildObject*[2];
  x[0] = x_;
  int lb, ub, aux;     
  if( x[0]->min() >= 0 ) {
    lb = x[0]->min();
    ub = x[0]->max();
  } else if( x[0]->max() <= 0 ) {
    ub = -1*x[0]->min();
    lb = -1*x[0]->max();
  } else {
    aux = -1*(x[0]->minNegAbs());
    lb = x[0]->minPosAbs();
    if( aux < lb ) lb = aux;
    
    aux = -1*(x[0]->min());
    ub = x[0]->max();
    if( aux > ub ) ub = aux;
  }
  return new BuildObjectPredicate( x, 1, lb, ub, this, NULL, BuildObject::AWV );
}

void BuildObjectAbs::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "|";
  pred->scope[0]->printshort( o );
  o << "|";
}

string BuildObjectAbs::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string x1 = pred->scope[0]->xmlPred(idx, level+1, sub);
  string s = ( "abs(" + x1 + ")" );
  return s;
}

/**********************************************
 * And Constraint BuildObject
 **********************************************/
void BuildObjectAnd::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  if( pred->isVar() )
    Constraint *con = new PredicateAnd( s, tmp );
}

bool BuildObjectAnd::preBuild( BuildObjectPredicate *pred )
{
  bool res = BuildObjectConstraint::preBuild( pred );
  if( res ) {
    if( pred->min() == 1 || ( pred->max() && !pred->isVar() ) ) {
      pred->unsetRel();
      for( int i=0; i<2; ++i ) {
	pred->scope[i]->deReference();
	pred->scope[i]->remove( 0 );
      }
      res = false;
    }
  }
  return res;
}

int BuildObjectAnd::computeValue( BuildObjectPredicate *pred ) const
{
  return (pred->scope[0]->value() && pred->scope[1]->value());
}

void BuildObjectAnd::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "(";
  pred->scope[0]->printshort( o );
  o << " & ";
  pred->scope[1]->printshort( o );
  o << ")";
}

string BuildObjectAnd::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string x1 = pred->scope[0]->xmlPred(idx, level+1, sub);
  string x2 = pred->scope[1]->xmlPred(idx, level+1, sub);
  string s = ( "and(" + x1 + "," + x2 + ")" );
  return s;
}


/**********************************************
 * Or Constraint BuildObject
 **********************************************/
void BuildObjectOr::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  if( pred->isVar() ) {
    Constraint *con = new PredicateOr( s, tmp );
  } else {
    Constraint *con = new ConstraintOr( s, tmp );
  }
}

bool BuildObjectOr::preBuild( BuildObjectPredicate *pred )
{
  bool res = BuildObjectConstraint::preBuild( pred );
  int i;
  if( res && !pred->max() ) {
    pred->unsetRel();
    for( i=0; i<2; ++i ) {
      pred->scope[i]->deReference();
      pred->scope[i]->setDomain( 0 );
    }
    res = false;
  } else {
    for( i=0; i<2; ++i ) 
      if( !pred->scope[i]->max() ) {
	pred->unsetRel();
	pred->scope[0]->deReference();
	pred->scope[1]->deReference();
	pred->scope[1-i]->remove( 0 );
	pred->scope[1-i]->preBuild();
	res = false;
      }
  } 
  return res;
}

int BuildObjectOr::computeValue( BuildObjectPredicate *pred ) const
{
  return (pred->scope[0]->value() || pred->scope[1]->value());
}

void BuildObjectOr::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "(";
  pred->scope[0]->printshort( o );
  o << " || ";
  pred->scope[1]->printshort( o );
  o << ")";
}

string BuildObjectOr::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string x1 = pred->scope[0]->xmlPred(idx, level+1, sub);
  string x2 = pred->scope[1]->xmlPred(idx, level+1, sub);
  string s = ( "or(" + x1 + "," + x2 + ")" );
  return s;
}


/**********************************************
 * IfThenElse Constraint BuildObject
 **********************************************/
void BuildObjectIfThenElse::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  if( pred->isVar() )
    Constraint *con = new PredicateIfThenElse( s, tmp );
}

bool BuildObjectIfThenElse::preBuild( BuildObjectPredicate *pred )
{
  bool res = BuildObjectConstraint::preBuild( pred );
  
  //  cout << res << endl;
  //  assert( res );

  return res;
}

int BuildObjectIfThenElse::computeValue( BuildObjectPredicate *pred ) const
{
  if( pred->scope[0]->value() )
    return pred->scope[1]->value();
  else 
    return pred->scope[2]->value();
}

void BuildObjectIfThenElse::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "(";
  pred->scope[0]->printshort( o );
  o << " ? ";
  pred->scope[1]->printshort( o );
  o << " : ";
  pred->scope[2]->printshort( o );
  o << ")";
}

string BuildObjectIfThenElse::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string x1 = pred->scope[0]->xmlPred(idx, level+1, sub);
  string x2 = pred->scope[1]->xmlPred(idx, level+1, sub);
  string x3 = pred->scope[2]->xmlPred(idx, level+1, sub);
  string s = ( "if(" + x1 + "," + x2 + "," + x3 + ")" );
  return s;
}


/**********************************************
 *  Disjunctive Predicate BuildObject
 **********************************************/ 
void BuildObjectDisjunctive::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  Constraint *con = new PredicateDisjunctive( s, tmp, pred->params );
}

BuildObjectPredicate* BuildObjectDisjunctive::declare( BuildObject *x_, int dx, BuildObject *y_, int dy )
{
  int *params = new int[2];
  params[0] = dx;
  params[1] = dy;
  BuildObject **x = new BuildObject*[3];
  x[0] = x_;
  x[1] = y_;
  return new BuildObjectPredicate(x, 2, 0, 1, this, params, 
				  (BuildObject::AWV | BuildObject::CRANGE) );
}

int BuildObjectDisjunctive::computeValue( BuildObjectPredicate *pred ) const
{
  return ( pred->scope[0]->value() + pred->params[0] <= pred->scope[1]->value() );
}

void BuildObjectDisjunctive::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "(";
  pred->scope[0]->printshort( o );
  o << " + " << pred->params[0] << " <= ";
  pred->scope[1]->printshort( o );
  o << " || ";
  pred->scope[1]->printshort( o );
  o << " + " << pred->params[1] << " <= ";
  pred->scope[0]->printshort( o );
  o << ")";
}
string BuildObjectDisjunctive::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string x1 = pred->scope[0]->xmlPred(idx, level+1, sub);
  string x2 = pred->scope[1]->xmlPred(idx, level+1, sub);
  string s = ( "or(le(add(" + x1 + "," + int2string(pred->params[0]) + ")," + x2 + ")," +
	       "le(add(" + x2 + "," + int2string(pred->params[1]) + ")," + x1 + "))" );
  return s;
}


/**********************************************
 *  Precedence Predicate BuildObject
 **********************************************/ 
void BuildObjectPrecedence::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  if( !(pred->isVar()) ) { 
    if( !pred->max() )
      {
	VariableInt *aux = tmp[0];
	tmp[0] = tmp[1];
	tmp[1] = aux;
	pred->params[0] = (1-pred->params[0]);
      }
    Constraint *con = new ConstraintLess( s, tmp, pred->params[0] );
  } else {
    Constraint *con = new PredicateLess( s, tmp, pred->params[0] );
  }
}

BuildObjectPredicate* BuildObjectPrecedence::declare( BuildObject *x_, int d, BuildObject *y_ )
{
  BuildObject **x = new BuildObject*[3];
  x[0] = x_;
  x[1] = y_;
  int *params = new int[1];
  params[0] = d;
  return new BuildObjectPredicate(x, 2, 0, 1, this, params, 
				  (BuildObject::CRANGE | BuildObject::CONVEX));
}


bool BuildObjectPrecedence::preBuild( BuildObjectPredicate *pred )
{
  bool res = BuildObjectConstraint::preBuild( pred );

  if( res ) {
    BuildObject **scope = pred->scope;
    if( scope[0]->size() == 1 )
      {

	//cout << "low bound" << endl;
	//pred->print( cout );
	//cout << endl;

	pred->relation = ENV[ConstraintStore::LOWERBOUND];
	pred->params[0] += scope[0]->min();
	scope[0]->deReference();
	scope[0] = scope[1];
	scope[1] = scope[2];
	--pred->arity;

	//pred->print( cout );
	//cout << endl;

	res = pred->relation->preBuild( pred );

	//pred->print( cout );
	//cout << endl;
	//cout << "[" << scope[0]->min() << ".." << scope[0]->max() << "]" << endl;

	//cout << endl;

      }
    else if( scope[1]->size() == 1 )
      {

	//cout << "up bound" << endl;

	pred->relation = ENV[ConstraintStore::UPPERBOUND];
	pred->params[0] = (scope[1]->min() - pred->params[0]);
	scope[1]->deReference();
	scope[1] = scope[2];
	--pred->arity;
	res = pred->relation->preBuild( pred );
      }
    else if( !pred->max() || pred->min() ) {
      pred->unsetVar();
      //cout << "unset" << endl;
    }
    if( !pred->isVar() && pred->params[0] > 0 )
      {
	pred->setNeq();
      }
  } // else {
//     if( pred->scope[0]->min() + pred->params[0] > pred->scope[1]->max() )
//       pred->setDomain(0);
//     else
//       pred->setDomain(1);

//     //cout << pred->isVar() << " " << pred->isRel() << endl;

//   }
  return res;
}

int BuildObjectPrecedence::computeValue( BuildObjectPredicate *pred ) const
{
  return (pred->scope[0]->value() + pred->params[0] <= pred->scope[1]->value());
}

void BuildObjectPrecedence::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "(";
  pred->scope[0]->printshort( o );
  if( pred->params[0] == 1 )
    o << " < ";
  else {
    if( pred->params[0] )
      o << " + " << pred->params[0] ;
    o << " <= ";
  }
  pred->scope[1]->printshort( o );
  o << ")";
}    

string BuildObjectPrecedence::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string x1 = pred->scope[0]->xmlPred(idx, level+1, sub);
  string x2 = pred->scope[1]->xmlPred(idx, level+1, sub);
  string s;
  if(pred->params[0] == 0 )
    s = ( "le(" + x1 + "," + x2 + ")" );
  else if(pred->params[0] == 1)
    s = ( "lt(" + x1 + "," + x2 + ")" );
  else
    s = ( "le(add(" + x1 + "," + int2string(pred->params[0]) + ")," + x2 + ")" );
  return s;
}


/**********************************************
 *  Add Predicate BuildObject
 **********************************************/
void BuildObjectAdd::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  Constraint *con = new PredicateAdd( s, tmp );
}

BuildObjectPredicate* BuildObjectAdd::declare( BuildObject *x_, BuildObject *y_ )
{
  BuildObject **x = new BuildObject*[3];
  x[0] = x_;
  x[1] = y_;
  return new BuildObjectPredicate(x, 2, (x[0]->min()+x[1]->min()), 
				  (x[0]->max()+x[1]->max()), this, 
				  NULL, (BuildObject::AWV | BuildObject::CRANGE));
}

bool BuildObjectAdd::preBuild( BuildObjectPredicate *pred )
{
  bool res = BuildObjectConstraint::preBuild( pred );

  if( res ) {
    BuildObject **scope = pred->scope;
    if( scope[0]->size() == 1 ) {   

      //cout << "offset0" << endl;

      pred->relation = ENV[ConstraintStore::OFFSET];
      pred->params = new int[1];
      pred->params[0] = scope[0]->min();
      scope[0]->deReference();
      scope[0] = scope[1];
      scope[1] = scope[2];
      --pred->arity;

//       pred->print( cout );
//       cout << endl;

      //res = pred->relation->preBuild( pred );
      res = pred->preBuild( );

//       pred->print( cout );
//       cout << endl;

    } else if( scope[1]->size() == 1 ) {

      //cout << "offset1" << endl;

      pred->relation = ENV[ConstraintStore::OFFSET];
      pred->params = new int[1];
      pred->params[0] = scope[1]->min();
      scope[1]->deReference();
      scope[1] = scope[2];
      --pred->arity;
      res = pred->relation->preBuild( pred );
    }
  }

  //cout << res << endl;

  return res;
}

int BuildObjectAdd::computeValue( BuildObjectPredicate *pred ) const
{
  return (pred->scope[0]->value() + pred->scope[1]->value());
}

void BuildObjectAdd::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "(";
  pred->scope[0]->printshort( o );
  o << " + ";
  pred->scope[1]->printshort( o );
  o << ")";
}

string BuildObjectAdd::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string x1 = pred->scope[0]->xmlPred(idx, level+1, sub);
  string x2 = pred->scope[1]->xmlPred(idx, level+1, sub);
  string s = ( "add(" + x1 + "," + x2 + ")" );
  return s;
}


/**********************************************
 *  Sub Predicate BuildObject
 **********************************************/
void BuildObjectSub::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  Constraint *con = new PredicateSub( s, tmp );
}

BuildObjectPredicate* BuildObjectSub::declare( BuildObject *x_, BuildObject *y_ )
{
  BuildObject **x = new BuildObject*[3];
  x[0] = x_;
  x[1] = y_;
  return new BuildObjectPredicate(x, 2, (x[0]->min()-x[1]->max()), 
				  (x[0]->max()-x[1]->min()), this, 
				  NULL, (BuildObject::AWV | BuildObject::CRANGE));
}

bool BuildObjectSub::preBuild( BuildObjectPredicate *pred )
{
  bool res = BuildObjectConstraint::preBuild( pred );

  if( res ) {
    BuildObject **scope = pred->scope;
    if( scope[1]->size() == 1 ) {
      pred->relation = ENV[ConstraintStore::OFFSET];
      pred->params = new int[1];
      pred->params[0] = -scope[1]->min();
      scope[1]->deReference();
      scope[1] = scope[2];
      --pred->arity;
      res = pred->relation->preBuild( pred );
    }
  }
  return res;
}

int BuildObjectSub::computeValue( BuildObjectPredicate *pred ) const
{
  return (pred->scope[0]->value() - pred->scope[1]->value());
}

void BuildObjectSub::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "(";
  pred->scope[0]->printshort( o );
  o << " - ";
  pred->scope[1]->printshort( o );
  o << ")";
}

string BuildObjectSub::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string x1 = pred->scope[0]->xmlPred(idx, level+1, sub);
  string x2 = pred->scope[1]->xmlPred(idx, level+1, sub);
  string s = ( "sub(" + x1 + "," + x2 + ")" );
  return s;
}


/**********************************************
 *  Mul Predicate BuildObject
 **********************************************/
void BuildObjectMul::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  Constraint *con = new PredicateMul( s, tmp );
}

BuildObjectPredicate* BuildObjectMul::declare( BuildObject *x_, BuildObject *y_ )
{
  BuildObject **x = new BuildObject*[3];
  x[0] = x_;
  x[1] = y_;
  int z[4], lb = INT_MAX, ub = -1*INT_MAX, i = 4;
  z[0] = x[0]->min() * x[1]->min();
  z[1] = x[0]->min() * x[1]->max();
  z[2] = x[0]->max() * x[1]->min();
  z[3] = x[0]->max() * x[1]->max();  
  while( i-- ) {
    if( z[i] > ub ) ub = z[i];
    if( z[i] < lb ) lb = z[i];
  }
  return new BuildObjectPredicate(x, 2, lb, ub, this, NULL, (BuildObject::AWV | BuildObject::CRANGE));
}

int BuildObjectMul::computeValue( BuildObjectPredicate *pred ) const
{
  return (pred->scope[0]->value() * pred->scope[1]->value());
}

void BuildObjectMul::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "(";
  pred->scope[0]->printshort( o );
  o << " * ";
  pred->scope[1]->printshort( o );
  o << ")";
}

string BuildObjectMul::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string x1 = pred->scope[0]->xmlPred(idx, level+1, sub);
  string x2 = pred->scope[1]->xmlPred(idx, level+1, sub);
  string s = ( "mul(" + x1 + "," + x2 + ")" );
  return s;
}


/**********************************************
 *  Div Predicate BuildObject
 **********************************************/
void BuildObjectDiv::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  Constraint *con = new PredicateDiv( s, tmp );
}

BuildObjectPredicate* BuildObjectDiv::declare( BuildObject *x_, BuildObject *y_ )
{
  BuildObject **x = new BuildObject*[3];
  x[0] = x_;
  x[1] = y_;
  int lb, ub, aux;
      
  int mp1 = x[1]->minPosAbs();
  int mn1 = x[1]->minNegAbs();
  int mp0 = x[0]->minPosAbs();
  int mn0 = x[0]->minNegAbs();
      
  int m1 = x[1]->min();
  int x1 = x[1]->max();
  int m0 = x[0]->min();
  int x0 = x[0]->max();
      
  if( !mp1 ) {
    if( x[1]->max() )
      mp1 = 1;
    else mp1 = -1;
  }
  if( !mn1 ) {
    if( x[1]->min() )
      mn1 = -1;
    else mn1 = 1;
  }
  if( !mp0 ) {
    if( x[0]->max() )
      mp0 = 1;
    else mp0 = -1;
  }
  if( !mn0 ) {
    if( x[0]->min() )
      mn0 = -1;
    else mn0 = 1;
  }
      
  lb = x0/mn1;
  aux = m0/mp1;
  if( lb > aux ) lb = aux;
      
  ub = x0/mp1;
  aux = m0/mn1;
  if( ub < aux ) ub = aux;
      
  return new BuildObjectPredicate(x, 2, lb, ub, this, NULL, 
				  (BuildObject::AWV | BuildObject::CRANGE));
}

int BuildObjectDiv::computeValue( BuildObjectPredicate *pred ) const
{
  return (pred->scope[0]->value() / pred->scope[1]->value());
}

void BuildObjectDiv::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "(";
  pred->scope[0]->printshort( o );
  o << " / ";
  pred->scope[1]->printshort( o );
  o << ")";
}

string BuildObjectDiv::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string x1 = pred->scope[0]->xmlPred(idx, level+1, sub);
  string x2 = pred->scope[1]->xmlPred(idx, level+1, sub);
  string s = ( "div(" + x1 + "," + x2 + ")" );
  return s;
}


/**********************************************
 *  Pow Predicate BuildObject
 **********************************************/
void BuildObjectPow::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
}

int BuildObjectPow::computeValue( BuildObjectPredicate *pred ) const
{
  return (int)(pow((double)(pred->scope[0]->value()), (double)(pred->scope[1]->value())));
}

void BuildObjectPow::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "(";
  pred->scope[0]->printshort( o );
  o << " ^ ";
  pred->scope[1]->printshort( o );
  o << ")";
}

string BuildObjectPow::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string x1 = pred->scope[0]->xmlPred(idx, level+1, sub);
  string x2 = pred->scope[1]->xmlPred(idx, level+1, sub);
  string s = ( "pow(" + x1 + "," + x2 + ")" );
  return s;
}


/**********************************************
 *  Mod Predicate BuildObject
 **********************************************/
void BuildObjectMod::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  //cout << " +++++++ " << (pred->isCRange()) << endl;
  Constraint *con = new PredicateMod( s, tmp );
}

int BuildObjectMod::computeValue( BuildObjectPredicate *pred ) const
{
  return (pred->scope[0]->value() % pred->scope[1]->value());
}

void BuildObjectMod::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "(";
  pred->scope[0]->printshort( o );
  o << " % ";
  pred->scope[1]->printshort( o );
  o << ")";
}

string BuildObjectMod::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string x1 = pred->scope[0]->xmlPred(idx, level+1, sub);
  string x2 = pred->scope[1]->xmlPred(idx, level+1, sub);
  string s = ( "mod(" + x1 + "," + x2 + ")" );
  return s;
}


/**********************************************
 *  ModConstant Predicate BuildObject
 **********************************************/
void BuildObjectModConstant::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  Constraint *con = new PredicateModConstant( s, tmp, pred->params[0] );
}

int BuildObjectModConstant::computeValue( BuildObjectPredicate *pred ) const
{
  return (pred->scope[0]->value() % pred->params[0]);
}

void BuildObjectModConstant::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "(";
  pred->scope[0]->printshort( o );
  o << " % ";
  pred->params[0];
  o << ")";
}

string BuildObjectModConstant::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string x1 = pred->scope[0]->xmlPred(idx, level+1, sub);
  string s = ( "mod(" + x1 + "," + int2string(pred->params[0]) + ")" );
  return s;
}


/**********************************************
 *  EqualConstant Predicate BuildObject
 **********************************************/ 
void BuildObjectEqualConstant::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  if( pred->isVar() )
    Constraint *con = new PredicateEqualConstant( s, tmp, pred->params[0], 1 );
}

bool BuildObjectEqualConstant::preBuild( BuildObjectPredicate *pred )
{
  bool res = BuildObjectConstraint::preBuild( pred );

  if( res ) {
    int val = pred->params[0];
    BuildObject *x=pred->scope[0];
    int pmax = !pred->max();
    if( pmax || pred->min() )
      pred->unsetVar();
    if( !pred->isVar() ) {
      x->deReference();
      pred->unsetRel();
      if( pmax ) x->remove( val );
      else x->setDomain( val );
      res = false;
    } else if( val > x->min() && val < x->max() ) 
      pred->unsetCRange();
  }
  return res;
}

int BuildObjectEqualConstant::computeValue( BuildObjectPredicate *pred ) const
{
  return (pred->scope[0]->value() == pred->params[0]);
}

void BuildObjectEqualConstant::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "(";
  pred->scope[0]->printshort( o );
  o << " == " << (pred->params[0]) << ")";
}

string BuildObjectEqualConstant::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string x1 = pred->scope[0]->xmlPred(idx, level+1, sub);
  string s = ( "eq(" + x1 + "," + int2string(pred->params[0]) + ")" );
  return s;
}


/**********************************************
 *  NotEqualConstant Predicate BuildObject
 **********************************************/ 
void BuildObjectNotEqualConstant::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  if( pred->isVar() ) 
    Constraint *con = new PredicateEqualConstant( s, tmp, pred->params[0], 0 );
}

bool BuildObjectNotEqualConstant::preBuild( BuildObjectPredicate *pred )
{
  bool res = BuildObjectConstraint::preBuild( pred );

  if( res ) {
    int pmax = !pred->max();
    if( pmax || pred->min() == 1 )
      pred->unsetVar();
    if( !pred->isVar() ) {
      pred->scope[0]->deReference();
      pred->unsetRel();
      if( pmax ) pred->scope[0]->setDomain( pred->params[0] );
      else pred->scope[0]->remove( pred->params[0] );
      res = false;
    } else pred->unsetCList();
  } 
  return res;
}

int BuildObjectNotEqualConstant::computeValue( BuildObjectPredicate *pred ) const
{
  return (pred->scope[0]->value() != pred->params[0]);
}

void BuildObjectNotEqualConstant::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "(";
  pred->scope[0]->printshort( o );
  o << " =/= " << (pred->params[0]) << ")";
}

string BuildObjectNotEqualConstant::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string x1 = pred->scope[0]->xmlPred(idx, level+1, sub);
  string s = ( "ne(" + x1 + "," + int2string(pred->params[0]) + ")" );
  return s;
}

/**********************************************
 *  Equal Predicate BuildObject
 **********************************************/ 
void BuildObjectEqual::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  if( pred->isVar() )
    Constraint *con = new PredicateEqual( s, tmp, 1 );
}

bool BuildObjectEqual::preBuild( BuildObjectPredicate *pred )
{
  bool res = BuildObjectConstraint::preBuild( pred );

  if( res ) {
    BuildObject **scope = pred->scope;
    if( pred->min() == 1 )
      pred->unsetVar();
    if( !pred->isVar() ) {
      pred->unsetRel();
      int i = ( scope[0]->getVarId() < scope[1]->getVarId() );

      //cout << "equal" << endl;

      scope[1-i]->setEqual( scope[i] );

      res = false;
    } else {
      if( pred->isVar() )
	pred->unsetCRange();
      if( scope[0]->size() == 1 ) {    
	pred->relation = ENV[ConstraintStore::EQCONSTANT];
	pred->params = new int[1];
	pred->params[0] = scope[0]->min();
	scope[0]->deReference();
	scope[0] = scope[1];
	scope[1] = scope[2];
	--pred->arity;
	res = pred->relation->preBuild( pred );
      } else if( scope[1]->size() == 1 ) {
	pred->relation = ENV[ConstraintStore::EQCONSTANT];
	pred->params = new int[1];
	pred->params[0] = scope[1]->min();
	scope[1]->deReference();
	scope[1] = scope[2];
	--pred->arity;
	res = pred->relation->preBuild( pred );
      }
    }
  }
  return res;
}

int BuildObjectEqual::computeValue( BuildObjectPredicate *pred ) const
{
  return (pred->scope[0]->value() == pred->scope[1]->value());
}

void BuildObjectEqual::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "(";
  pred->scope[0]->printshort( o );
  o << " == " ;
  pred->scope[1]->printshort( o );
  o << ")";
}

string BuildObjectEqual::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string x1 = pred->scope[0]->xmlPred(idx, level+1, sub);
  string x2 = pred->scope[1]->xmlPred(idx, level+1, sub);
  string s = ( "eq(" + x1 + "," + x2 + ")" );
  return s;
}


/**********************************************
 *  NotEqual Predicate BuildObject
 **********************************************/ 
void BuildObjectNotEqual::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  if( !(pred->isVar()) ) 
    Constraint *con = new ConstraintNotEqual( s, tmp );
  else 
    Constraint *con = new PredicateEqual( s, tmp, 0 );
}

// bool BuildObjectNotEqual::check( BuildObjectPredicate *pred ) 
// {
//   BuildObject **scope = pred->scope;
//   return( //scope[0]->size() > 1 || 
// 	  //scope[1]->size() > 1 || 
// 	  scope[0]->min() != scope[1]->min() );
// }


bool BuildObjectNotEqual::preBuild( BuildObjectPredicate *pred )
{
  bool res = BuildObjectConstraint::preBuild( pred );

  if( res ) {
    BuildObject **scope = pred->scope;
    int i, constant=-1;
    if( pred->min() == 1 )
      pred->unsetVar();
    if( !pred->isVar() ) {
      for(i=0; i<2; ++i)
	if( pred->scope[i]->size() == 1 )
	  constant = i;
      if( constant >= 0 ) {
	pred->scope[1-constant]->remove( pred->scope[constant]->min() );
	pred->unsetRel();
	res = false;
      } else {
	pred->setNeq();
      }
    } else {
      pred->unsetCList();
      if( scope[0]->size() == 1 ) {    
	pred->relation = ENV[ConstraintStore::NEQCONSTANT];
	pred->params = new int[1];
	pred->params[0] = scope[0]->min();
	scope[0]->deReference();
	scope[0] = scope[1];
	scope[1] = scope[2];
	--pred->arity;
	res = pred->relation->preBuild( pred );
      } else if( scope[1]->size() == 1 ) {
	pred->relation = ENV[ConstraintStore::NEQCONSTANT];
	pred->params = new int[1];
	pred->params[0] = scope[1]->min();
	scope[1]->deReference();
	scope[1] = scope[2];
	--pred->arity;
	res = pred->relation->preBuild( pred );
      }
    }
  } 
//   else {
//     if( !check( pred ) ) 
//       pred->remove( 1 );
//   }
  return res;
}

int BuildObjectNotEqual::computeValue( BuildObjectPredicate *pred ) const
{
  return (pred->scope[0]->value() != pred->scope[1]->value());
}

void BuildObjectNotEqual::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "(";
  pred->scope[0]->printshort( o );
  o << " =/= ";
  pred->scope[1]->printshort( o );
  o << ")";
}

string BuildObjectNotEqual::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string x1 = pred->scope[0]->xmlPred(idx, level+1, sub);
  string x2 = pred->scope[1]->xmlPred(idx, level+1, sub);
  string s = ( "ne(" + x1 + "," + x2 + ")" );
  return s;
}


/**********************************************
 *  NotEqualWildcard Predicate BuildObject
 **********************************************/ 
void BuildObjectNotEqualWildcard::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  if( !(pred->isVar()) ) {
    Constraint *con = new ConstraintNotEqualWildcard( s, tmp, pred->params[0] );
  } else {
    // TODO
  }
}

int BuildObjectNotEqualWildcard::computeValue( BuildObjectPredicate *pred ) const
{
  int x = pred->scope[0]->value();
  return (x == pred->params[0] || x != pred->scope[1]->value());
}

void BuildObjectNotEqualWildcard::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "(";
  pred->scope[0]->printshort( o );
  o << " =/= ";
  pred->scope[1]->printshort( o );
  o << " or ";
  pred->scope[0]->printshort( o );
  o << " = ";
  o << pred->params[0] ;
  o << ")";
}

string BuildObjectNotEqualWildcard::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string x1 = pred->scope[0]->xmlPred(idx, level+1, sub);
  string x2 = pred->scope[1]->xmlPred(idx, level+1, sub);
  string s = ( "or(ne(" + x1 +  "," + x2 + "), eq(" + x1 + "," + int2string(pred->params[0]) + "))" );
  return s;
}


/**********************************************
 * SumEq Constraint BuildObject
 **********************************************/
void BuildObjectSumEq::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  Constraint *con;
  bool isBoolean = true, isWeighted=false;
  int i=pred->arity;
  while( i-- ) {
    isBoolean &= ( tmp[i]->getType() == VariableInt::BOOL );
    isWeighted |= pred->params[i] != 1;
  }

  if( isWeighted ) {
    con = new ConstraintWeightedSumEqual( s, tmp, pred->arity, 
					  pred->params );
  } else if( isBoolean ) {
    con = new ConstraintBoolSumEqual( s, tmp, pred->arity, 
				      pred->params[pred->arity] );
  } else {
    con = new ConstraintSumEqual( s, tmp, pred->arity, 
				  pred->params[pred->arity] );
  }
}

bool BuildObjectSumEq::preBuild( BuildObjectPredicate *pred )
{
  int j, i, n=pred->arity, *params=pred->params;
  BuildObject **scope=pred->scope, *aux;

  i=n;
  // absorb constants
  while( i-- ) {
    if( scope[i]->size() == 1 ) {
      scope[i]->deReference();
      params[n] -= (params[i] * scope[i]->min());
      params[i] = params[n-1];
      scope[i] = scope[n-1];
      params[n-1] = params[n];
      scope[n-1] = scope[n];
      --n;
    } else {
      j=-1;
      while( ++j < i )
	if( *(scope[i]) == *(scope[j]) ) {
	  break;
	}

      if( j < i ) {
	++params[j];
	scope[i]->deReference();
	params[i] = params[n-1];
	scope[i] = scope[n-1];
	params[n-1] = params[n];
	scope[n-1] = scope[n];
	--n;
      }
    }
  }
  pred->arity = n;

  return n>0;
}

void BuildObjectSumEq::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "(";
  pred->scope[0]->printshort( o );
  for(int i=1; i<pred->arity; ++i) {
    if( pred->params ) {
      if( pred->params[i] == -1 )
	o << " - ";
      else if( pred->params[i] < 0 )
	o << " - " << (-1 * pred->params[i]) << "*";
      else if( pred->params[i] > 1 )
	o << " + " << pred->params[i] << "*";
      else o << " + " ;
    } else 
      o << " + " ;
    pred->scope[i]->printshort( o );  
  }
  o << ") = " << pred->params[pred->arity]; 
}


string BuildObjectSumEq::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string s = "global:sumeq";
  return s;
}

/**********************************************
 * Sum Constraint BuildObject
 **********************************************/
void BuildObjectSum::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  Constraint *con;
  if( pred->params ) 
    con = new PredicateWeightedSum( s, tmp, pred->arity+1, pred->params );
  else 
    con = new PredicateSimpleSum( s, tmp, pred->arity+1 );
}

BuildObjectPredicate *BuildObjectSum::declare( BuildObject **scope, 
					       const int n, const int *os )
{
  int lb = 0, ub = 0, realsize=0, *params = NULL;
  bool simple = true;
  if( os ) {
    //bool simple = false;
    for(int i=0; simple && i<n; ++i) 
      simple = ( os[i] == 1 );
    if( !simple ) {
      params = new int[n+1];
      //params[n] = 0;
    }
  }
  for(int i=0; i<n; ++i) {
    if( params ) {
      lb+=( os[i] * ( os[i] < 0 ? scope[i]->max() : scope[i]->min() ) );
      ub+=( os[i] * ( os[i] < 0 ? scope[i]->min() : scope[i]->max() ) );
      params[realsize] = os[i];
      if( os[i] ) {
	if( realsize < i )
	  scope[realsize] = scope[i];
	++realsize;
      }
    } else {      
      lb+=scope[realsize]->min();
      ub+=scope[realsize]->max();
      ++realsize;
    }
  }

  if( !simple ) {
    params[realsize] = 0;
  }

  //cout << " **** ";
  BuildObjectPredicate *pred = new BuildObjectPredicate( scope, realsize, 
							 lb, ub, this, params, 
							 (BuildObject::AWV | 
							  BuildObject::CRANGE | 
							  BuildObject::CONVEX) );
  //pred->print( cout );
  //cout << endl;

  return pred;
//   return new BuildObjectPredicate( scope, n, lb, ub, this, params, 
// 				   (BuildObject::AWV | BuildObject::CRANGE | BuildObject::CONVEX) );
}

bool BuildObjectSum::preBuild( BuildObjectPredicate *pred )
{
  int n=pred->arity, i=n, *params=pred->params;
  BuildObject **scope=pred->scope, *aux;

  // check if it should be a sum equality constraint
  if( scope[i]->size() == 1 ) {
    if( !params ) {
      pred->params = new int[n+1];
      std::fill(pred->params, pred->params+n, 1);
      pred->params[n] = 0;
    }
    pred->params[i] = scope[i]->min();
    pred->relation = ENV[ConstraintStore::SUMEQ];
    //--pred->arity;
    scope[i]->deReference();

    pred->unsetAwv();
    pred->unsetVar();
    pred->relation->preBuild( pred );
  } else {
    
    // absorb constants
    while( i-- ) {
      if( scope[i]->size() == 1 ) {
	scope[i]->deReference();
	params[n] -= (params[i] * scope[i]->min());
	params[i] = params[n-1];
	scope[i] = scope[n-1];
	params[n-1] = params[n];
	scope[n-1] = scope[n];
	--n;
      }
    }
    pred->arity = n;

    // change to equal, factor or offset?
    if( n == 1 )
      {
	if( params && params[0] == 1 ) {
	  pred->relation = ENV[ConstraintStore::OFFSET];
	  params[0] = -params[1];
	  params[1] = 0;
	  pred->relation->preBuild( pred );
	} else if( params && params[1] == 0 ) { 
	  pred->relation = ENV[ConstraintStore::FACTOR];
	  pred->relation->preBuild( pred );
	}
      }
    // change to add or sub?
    else if( n == 2 && params && params[n] == 0 ) 
      {
	if( !params || (params[0] == 1 && params[1] == 1) ) {
	  pred->relation = ENV[ConstraintStore::ADD];
	  pred->relation->preBuild( pred );
	} else if( params && params[0] == 1 && params[1] == -1 ) {
	  pred->relation = ENV[ConstraintStore::SUB];
	  pred->relation->preBuild( pred );
	} else if( params && params[0] == -1 && params[1] == 1 ) {
	  pred->relation = ENV[ConstraintStore::SUB];
	  aux = scope[0];
	  scope[0] = scope[1];
	  scope[1] = aux;
	  pred->relation->preBuild( pred );
	}
      }
 //    // change to sumeq?
//     if( params && params[n] != 0 )
//       {
// 	pred->print( cout );
// 	//cout << endl;
// 	// TODO 
// 	//assert( false );
//       }
  }
  return pred->arity>0;
}

int BuildObjectSum::computeValue( BuildObjectPredicate *pred ) const
{
  int i = pred->arity, total = 0;
  while( i-- )
    total += (pred->params ? pred->params[i] * pred->scope[i]->value() : pred->scope[i]->value());
  return total;
}

void BuildObjectSum::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "(";

  if( pred->params ){
    if( pred->params[0] == -1 )
      o << "-";
    else if( pred->params[0] != 1 )
      o << pred->params[0] << "*";
  }
  for(int i=0; i<pred->arity-1; ++i) {
    pred->scope[i]->printshort( o );

    if( pred->params ){
      if( pred->params[i+1] == -1 )
	o << " - ";
      else if( pred->params[i+1] < 0 )
	o << " - " << (-1 * pred->params[i+1]) << "*";
      else if( pred->params[i+1] > 1 )
	o << " + " << pred->params[i+1] << "*";
      else o << " + " ;
    } else 
      o << " + " ;
  }
  pred->scope[pred->arity-1]->printshort( o );
  o << ")";
}


string BuildObjectSum::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  
  string s;
  if( pred->arity > 2 )
    s = "global:sum";
  else {
    string x1 = pred->scope[0]->xmlPred(idx, level+1, sub);
    string x2 = pred->scope[1]->xmlPred(idx, level+1, sub);
    if( pred->params && pred->params[0] == 1 && pred->params[1] == -1 )
      s = ("sub(" + x1 + "," + x2 + ")");
    else
      s = ("add(" + ((pred->params && pred->params[0] != 1) ? 
		     ("mul(" + int2string(pred->params[0]) + "," + x1 + ")")  
		     : x1 )
	   + "," + ((pred->params && pred->params[1] != 1) ? 
		    ("mul(" + int2string(pred->params[1]) + "," + x2 + ")")  
		    : x2 )
	   + ")" );
  }
  return s;
}


/**********************************************
 * Min Constraint BuildObject
 **********************************************/
void BuildObjectMin::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  Constraint *con = new PredicateMin( s, tmp, pred->arity+1 );
}

BuildObjectPredicate *BuildObjectMin::declare( BuildObject **scope, const int n )
{
  int lb = INT_MAX, ub = -INT_MAX, i=n;
  while( i-- ) {
    if(lb > scope[i]->min()) lb = scope[i]->min();
    if(ub < scope[i]->max()) ub = scope[i]->max();
  }
  return new BuildObjectPredicate( scope, n, lb, ub, this, NULL, 
				   (BuildObject::AWV | BuildObject::CRANGE | BuildObject::CONVEX) );
}

int BuildObjectMin::computeValue( BuildObjectPredicate *pred ) const
{
  int i = pred->arity, themin = pred->scope[0]->value();
  while( --i )
    if( themin > pred->scope[i]->value() )
      themin = pred->scope[i]->value();
  return themin;
}

void BuildObjectMin::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "min(";
    for(int i=0; i<pred->arity-1; ++i) {
    pred->scope[i]->printshort( o );
    o << " " ;
  }
  pred->scope[pred->arity-1]->printshort( o );
  o << ")";
}

string BuildObjectMin::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string x[pred->arity];
  for(int i=0; i<pred->arity; ++i) 
    x[i] = pred->scope[i]->xmlPred( idx, level+1, sub );
  string s;
  for(int i=0; i<pred->arity-1; ++i) 
    s += ("min(" + x[i] + ",");
  s += x[pred->arity-1];
  for(int i=0; i<pred->arity-1; ++i) 
    s += ")";
  return s;
}


/**********************************************
 * Max Constraint BuildObject
 **********************************************/
void BuildObjectMax::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  Constraint *con = new PredicateMax( s, tmp, pred->arity+1 );
}

BuildObjectPredicate *BuildObjectMax::declare( BuildObject **scope, const int n )
{
  int lb = INT_MAX, ub = -INT_MAX, i=n;
  while( i-- ) {
    if(lb > scope[i]->min()) lb = scope[i]->min();
    if(ub < scope[i]->max()) ub = scope[i]->max();
  }
  return new BuildObjectPredicate( scope, n, lb, ub, this, NULL, 
				   (BuildObject::AWV | BuildObject::CRANGE | BuildObject::CONVEX) );
}

int BuildObjectMax::computeValue( BuildObjectPredicate *pred ) const
{
  int i = pred->arity, themax = pred->scope[0]->value();
  while( --i )
    if( themax < pred->scope[i]->value() )
      themax = pred->scope[i]->value();
  return themax;
}

void BuildObjectMax::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "max(";
  for(int i=0; i<pred->arity-1; ++i) {
    pred->scope[i]->printshort( o );
    o << " " ;
  }
  pred->scope[pred->arity-1]->printshort( o );
  o << ")";
}

string BuildObjectMax::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string x[pred->arity];
  for(int i=0; i<pred->arity; ++i) 
    x[i] = pred->scope[i]->xmlPred( idx, level+1, sub );
  string s;
  for(int i=0; i<pred->arity-1; ++i) 
    s += ("max(" + x[i] + ",");
  s += x[pred->arity-1];
  for(int i=0; i<pred->arity-1; ++i) 
    s += ")";
  return s;
}


#ifdef _CPLEX

/**********************************************
 * CPlexMinDifference Constraint BuildObject
 **********************************************/
void BuildObjectCPlexMinDifference::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  Constraint *con = new PredicateCPlexMinDiff( s, pred->params[0], tmp, pred->arity+1 );
}

void BuildObjectCPlexMinDifference::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "#[=/=](";
  for(int i=0; i<pred->arity-1; ++i) {
    pred->scope[i]->printshort( o );
    o << " ";
  }
  pred->scope[pred->arity-1]->printshort( o );
  o << ")";
}

#endif


#ifdef _COINLP

/**********************************************
 * MaxDifference Constraint BuildObject
 **********************************************/
void BuildObjectMaxDifference::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  Constraint *con = new PredicateMaxDiff( s, pred->params[0], tmp, 
					  pred->arity+1, pred->params[1] );
}

void BuildObjectMaxDifference::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "#[=/=](";
  for(int i=0; i<pred->arity-1; ++i) {
    pred->scope[i]->printshort( o );
    o << " ";
  }
  pred->scope[pred->arity-1]->printshort( o );
  o << ")";
}


/**********************************************
 * MinDifference Constraint BuildObject
 **********************************************/
void BuildObjectMinDifference::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  Constraint *con = new PredicateMinDiff( s, pred->params[0], tmp, 
					  pred->arity+1, pred->params[1] );
}

void BuildObjectMinDifference::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "#[=/=](";
  for(int i=0; i<pred->arity-1; ++i) {
    pred->scope[i]->printshort( o );
    o << " ";
  }
  pred->scope[pred->arity-1]->printshort( o );
  o << ")";
}

#endif

/**********************************************
 * Element Constraint BuildObject
 **********************************************/
void BuildObjectElement::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  Constraint *con = new PredicateElement( s, tmp, pred->arity+1 );
}

bool BuildObjectElement::preBuild( BuildObjectPredicate *pred )
{
  bool res = true;
  int v=pred->arity, n=v-1, i;
  BuildObject **scope=pred->scope;
  if( scope[n]->size() == 1 )
    {
      i = scope[n]->min();
      while( n-- ) 
	if( n != i )
	  scope[n]->deReference();
      scope[0] = scope[i];
      scope[1] = scope[v];

      pred->unsetRel();
      i = ( scope[0]->getVarId() < scope[1]->getVarId() );
      scope[1-i]->setEqual( scope[i] );
    }
  else res = BuildObjectConstraint::preBuild( pred );
  return res;
}

int BuildObjectElement::computeValue( BuildObjectPredicate *pred ) const
{
  return pred->scope[pred->scope[pred->arity-1]->value()]->value();
}

void BuildObjectElement::print(std::ostream& o, const BuildObjectPredicate *pred) const
{

  //assert( pred->scope[pred->arity] == pred );

  //pred->scope[pred->arity]->BuildObject::printshort( cout );  
  o << "T[";
  pred->scope[pred->arity-1]->printshort( o );
  o << "], T={ ";
  for(int i=0; i<pred->arity-1; ++i) {
    pred->scope[i]->printshort( cout );
    cout << " ";
  }
  o << "}";
}

string BuildObjectElement::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string s;
  if( !level ) 
    s = "global:element";
  else
    s = ("X" + int2string(idx++)); 
  return s;
}


/**********************************************
 * Clause BuildObject
 **********************************************/ 
void BuildObjectClause::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  Constraint *con = new ConstraintClause( s, tmp, pred->arity, pred->params );
}

void BuildObjectClause::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  for(int i=0; i<pred->arity-1; ++i) {
    pred->scope[i]->printshort( o );
    o << " =/= ";
    o << pred->params[i];
    o << " or ";
  }
  pred->scope[pred->arity-1]->printshort( o );
  o << " =/= ";
  o << pred->params[pred->arity-1];
}

string BuildObjectClause::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string s = "table";
  return s;
}


/**********************************************
 *  Extensional Constraint BuildObject
 **********************************************/ 
BuildObjectTable::BuildObjectTable(const int a) 
{ 
  mindomain = new int[a];
  maxdomain = new int[a];
  for(int i=0; i<a; ++i) {
    mindomain[i] = INT_MAX;
    maxdomain[i] = -INT_MAX;
  }
  arity = a; 
  supports = NULL;
  supportTuples = NULL;
  supportList[0] = NULL;
  supportList[1] = NULL;
  var_sizes = NULL;
  spin=0;
}
BuildObjectTable::BuildObjectTable(const int a, const int nbt) 
{ 
  mindomain = new int[a];
  maxdomain = new int[a];
  for(int i=0; i<a; ++i) {
    mindomain[i] = INT_MAX;
    maxdomain[i] = -INT_MAX;
  }
  arity = a; 
  supports = NULL;
  supportTuples = NULL;
  supportList[0] = NULL;
  supportList[1] = NULL;
  var_sizes = NULL;
  tuples.init(0, nbt);
  spin=0;
}

void BuildObjectTable::setMinDom(const int i, const int m)
{
  if(mindomain[i] > m) mindomain[i] = m;
}
void BuildObjectTable::setMaxDom(const int i, const int m)
{
  if(maxdomain[i] < m) maxdomain[i] = m;
}
bool BuildObjectTable::preBuild( BuildObjectPredicate *pred )
{
  int i, j, k;
  bool res = BuildObjectConstraint::preBuild( pred );
  if( res ) {
    if( !spin && !tuples.size() ) {
      pred->unsetRel();
      res = false;
    } else if( !spin && arity == 2 ) {
      for(k=0; k<2; ++k) 
	if( pred->scope[k]->size() == 1 )
	  {
	    i=tuples.size();
	    j = pred->scope[k]->min();
	    pred->scope[k]->deReference();
	    pred->unsetRel();
	    while( i-- )
	      if( tuples[i][k] == j )
		pred->scope[1-k]->remove( tuples[i][1-k] );
	    res = false;
	  }
    }
    
    //  else {
    if( res ) {  
      res = BuildObjectConstraint::preBuild( pred );
      if( res ) {
	if( arity == 2 ) {
	  BuildObject **scope = pred->scope;
	  i=tuples.size();
	  if( spin ) {
	    while( i-- )
	      if(tuples[i][0] == tuples[i][1]) break;
	    if( i < 0 )
	      pred->setNeq();
	  } else {
	    int borneinf = std::max( scope[0]->min(), scope[1]->min() );
	    int bornesup = std::min( scope[0]->max(), scope[1]->max() );
	    k=(bornesup - borneinf + 1);
	    int values[k];
	    int *vals = values-borneinf;
	    for(j=borneinf; j<=bornesup; ++j)
	      if( scope[0]->contain(j) && scope[1]->contain(j) )
		vals[j] = 1;
	      else {
		vals[j] = 0;
		--k;
	      }
	    while( k && i-- ) {
	      j = tuples[i][0];
	      if(j == tuples[i][1] && j >= borneinf && j <= bornesup && vals[j]) 
		{
		  --k;
		  vals[j] = 0;
		}
	    }
	    if( !k )
	      pred->setNeq();
	  }
	}
      }
    }
  }
  return res;
}
void BuildObjectTable::build( Solver *s, VariableInt **tmp, 
			      BuildObjectPredicate *pred ) 
{
  long int matrixsize = 1;
  int i, j, n=pred->arity, tightness;
  for(i=0; matrixsize > 0 && i<n; ++i)
    matrixsize *= (maxdomain[i] - mindomain[i]);
  if( matrixsize < 0 ) tightness = INT_MAX;
  else {
    if( spin )
      tightness = (matrixsize / tuples.size());
    else
      tightness = (matrixsize / (matrixsize - tuples.size()));
  }
  //cout << spin << " " << n << " " << tuples.size() << " " << matrixsize << " " << tightness << " " ;

  if( spin || tuples.size() ) {    
//     i=tuples.size();
//     while( i-- )
//       {	
// 	for(j=0; j<n; ++j) 
// 	  cout << tuples[i][j] << " ";
// 	cout << endl;
//       }
//     cout << endl;
        

    if( !spin && tuples.size() == 1 ) { // CLAUSE 
      ConstraintClause *con = new ConstraintClause( s, tmp, n, tuples[0] );
      //cout << "clause" << endl;
    } else if( tightness < 1000 && n == 2 ) { // SUPPORT BITSETS
      //cout << "supports" << endl;
      if( supportList[0] == NULL ) {  
	i=2;
	while( i-- ) { 
	  supportList[i] = new MistralSet[maxdomain[i] - mindomain[i]];
	  supportList[i]-= mindomain[i];
	  j = maxdomain[i];
	  while( j-- > mindomain[i] )
	    supportList[i][j].init( mindomain[1-i],
				    maxdomain[1-i]-1,
				    (spin ? 0 : 0xffffffff ) );
	}
 	i=tuples.size();
	while( i-- ) {
	  j=2;
	  while( j-- ) {
	    if( spin )
	      supportList[j][tuples[i][j]].insert( tuples[i][1-j] );
	    else 
	      supportList[j][tuples[i][j]].erase( tuples[i][1-j] );
	  }
	}
      }
      ConstraintBinaryExtensional *con = 
	new ConstraintBinaryExtensional( s, tmp, supportList );
      con->init();
    } else if( !spin || (matrixsize < 32000000 && tightness < 64) ) { // N-ARY MATRIX
      //cout << "matrix" << endl;
      if( var_sizes == NULL ) {
	var_sizes = new int[n-1];
	var_sizes[n-2] = (maxdomain[n-1] - mindomain[n-1]);
	for(i=n-3; i >= 0; --i)
	  var_sizes[i] = var_sizes[i+1] * 
	    (maxdomain[i+1] - mindomain[i+1]);
	matrixmin = mindomain[n-1];
	int matrixmax = maxdomain[n-1]-1;
	
	i = n-1;
	while( i-- ) {
	  matrixmax += (maxdomain[i])*var_sizes[i]; 
	  matrixmin += (mindomain[i])*var_sizes[i]; 
	}

	matrix.init(matrixmin, matrixmax, MistralSet::empt);
      }

      ConstraintExtensional *con = new ConstraintExtensional( s, tmp, n, 
							      matrix, 
							      var_sizes );
     
      con->useResidualSupports();
      
      if( spin ) {
	con->fillNogood();
	i=tuples.size();
	while( i-- )
	  con->addSupport( tuples[i] );
      } else {
	con->fillSupport();
	i=tuples.size();
	while( i-- )
	  con->addNogood( tuples[i] );
      }

//       con->print( cout );
//       cout << endl;
      
    } else { // LIST OF TUPLES
      //      cout << "tuples" << endl;
//       if( !supports ) {
// 	supportTuples = new MistralStack<int*> *[n];
// 	for(i=0; i<n; ++i) {
// 	  supportTuples[i] = new MistralStack<int*> [maxdomain[i] - mindomain[i]];
// 	  supportTuples[i] -= mindomain[i];
// 	  j = maxdomain[i];
// 	}
// 	i=tuples.size();
// 	supports = new int[ i*n*(n-1) ];
// 	int k=0, l;
// 	while( i-- )
// 	  {
// 	    for(j=0; j<n; ++j) {
// 	      supportTuples[j][tuples[i][j]].push( &(supports[k-1]) );
// 	      for(l=0; l<n; ++l)
// 		if( j != l ) {
// 		  supports[k++] = tuples[i][l];
// 		}
// 	    }
// 	  }
//       }
//       ConstraintLargeExtensional *con = 
// 	new ConstraintLargeExtensional( s, tmp, n, supportTuples );

      if( !supportTuples ) {
	//cout << "-----------------" << endl;
	supportTuples = new MistralStack<int*>*[n];
	for(i=0; i<n; ++i) {
	  supportTuples[i] = new MistralStack<int*> [maxdomain[i] - mindomain[i]];
	  supportTuples[i] -= mindomain[i];
	  for(j=mindomain[i]; j<maxdomain[i]; ++j)
	    assert( supportTuples[i][j].size_ == 0 );
	}
	i=tuples.size();
	while( i-- )
	  for(j=0; j<n; ++j) {
	    supportTuples[j][tuples[i][j]].push( tuples[i] );
	    //if( j == 5 && tuples[i][j] == 22 )
	    //cout << "XX" << endl;
	  }
      }
      ConstraintLargeExtensional *con = 
	new ConstraintLargeExtensional( s, tmp, n, supportTuples );
      
      //      con->printlong( cout );
      //      cout << endl;

    }
  }
}

BuildObjectTable::~BuildObjectTable() 
{
  int i;
  i=tuples.size();
  while( i-- ) {
    delete [] tuples[i];
  }
  i=arity;  
  while( i-- ) {
    if( i<2 && supportList[i] ) {
      supportList[i] += mindomain[i];
      delete [] supportList[i];
    }
    if( supportTuples ) {
      supportTuples[i] += mindomain[i];
      delete [] supportTuples[i];
    }
  }

  if( var_sizes ) {
    delete [] var_sizes;
  }
  delete [] mindomain;
  delete [] maxdomain;
  delete [] supports;
  delete [] supportTuples;
}

void BuildObjectTable::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "table(";
  for(int i=0; i<pred->arity-1; ++i) {
    pred->scope[i]->printshort( o );
    o << " " ;
  }
  pred->scope[pred->arity-1]->printshort( o );
  o << ")";
}

string BuildObjectTable::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string s = "table";
  return s;
}

void BuildObjectTable::add( const int* sol )
{
  int i = arity, *ngd = new int[arity];
  while( i-- ) {
    ngd[i] = sol[i];
  }
  tuples.push_back( ngd );
}


/**********************************************
 * AllDiff Constraint BuildObject
 **********************************************/
void BuildObjectAllDiff::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  Constraint *con = new ConstraintAllDiff( s, tmp, pred->arity );
  if( !pred->params ) {
    int i, j;
    VariableInt *scp[2];
    for(i=1; i<pred->arity; ++i)
      for(j=0; j<i; ++j) {
	scp[0] = tmp[j];
	scp[1] = tmp[i];
	con = new ConstraintNotEqual( s, scp );
      }
  }
}

bool BuildObjectAllDiff::preBuild( BuildObjectPredicate *pred )
{
  bool res = true;
  int n=pred->arity, i=n, j, v;
  if( n == 2 ) {
    pred->relation = ENV[ConstraintStore::NEQ];
    res = pred->relation->preBuild( pred );
  } else {
    BuildObject **scope=pred->scope;
    while( i-- )
      if( scope[i]->size() == 1 ) 
	{
	  j = n;
	  v = scope[i]->min();
	  while( j-- ) if( j != i ) scope[j]->remove( v );
	  scope[i]->deReference();
	  scope[i] = scope[--n];
	}
    pred->arity = n;
  }
  if( !pred->arity ) {
    pred->unsetRel();
    res = false;
  }
  return res;
}

void BuildObjectAllDiff::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "AllDifferent(";
  for(int i=0; i<pred->arity-1; ++i) {
    pred->scope[i]->printshort( o );
    o << " " ;
  }
  pred->scope[pred->arity-1]->printshort( o );
  o << ")";
}
string BuildObjectAllDiff::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string s;
  if( !level ) 
    s = "global:allDifferent";
  else
    s = ("X" + int2string(idx++)); 
  return s;
}


/**********************************************
 * Gcc Constraint BuildObject
 **********************************************/
void BuildObjectGcc::build( Solver *s, VariableInt **tmp, 
			    BuildObjectPredicate *pred )
{
  int n=pred->arity, m = (pred->params[1]-pred->params[0]+1);
  Constraint *con = 
    new ConstraintGlobalCardinality( s, tmp, n,
				     pred->params[0],
				     pred->params[1],
				     &(pred->params[2]),
				     &(pred->params[2+m]) );
  for(int j=0; j<m; ++j) {
    if( pred->params[2+j] > 0 ) {
      VariableInt *B[n];
      VariableInt *scp[2];
      for(int i=0; i<n; ++i) {
	B[i] = new VariableBool(s);
	scp[0] = tmp[i];
	scp[1] = B[i];
	con = new PredicateEqualConstant(s, scp, pred->params[0]+j, 1);
      }
      con = new ConstraintBoolSumMore(s, B, n, pred->params[2+j]);
    }
    if( pred->params[2+m+j] < n ) {
      VariableInt *B[n];
      VariableInt *scp[2];
      for(int i=0; i<n; ++i) {
	B[i] = new VariableBool(s);
	scp[0] = tmp[i];
	scp[1] = B[i];
	con = new PredicateEqualConstant(s, scp, pred->params[0]+j, 1);
      }
      con = new ConstraintBoolSumLess(s, B, n, pred->params[2+m+j]);
    }
  }
}

void BuildObjectGcc::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "Gcc(";
  for(int i=0; i<pred->arity-1; ++i) {
    pred->scope[i]->printshort( o );
    o << " " ;
  }
  pred->scope[pred->arity-1]->printshort( o );
  o << ", [" << pred->params[0] << ".." 
    << pred->params[1] << "] >={";
  for(int i=0; i<(pred->params[1]-pred->params[0]+1); ++i) {
    o << " " << pred->params[2+i];
  }
  o << " }, <={";
  for(int i=0; i<(pred->params[1]-pred->params[0]+1); ++i) {
    o << " " << pred->params[3+pred->params[1]-pred->params[0]+i];
  }
  o << " })";
}
string BuildObjectGcc::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string s = "global:gcc";
  return s;
}


/**********************************************
 * EdgeFinder Constraint BuildObject
 **********************************************/
void BuildObjectEdgeFinder::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  Constraint *con = new ConstraintEdgeFinder( s, tmp, pred->arity, pred->params );
}

void BuildObjectEdgeFinder::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "EdgeFinder(";
  for(int i=0; i<pred->arity-1; ++i) {
    pred->scope[i]->printshort( o );
    o << " " ;
  }
  pred->scope[pred->arity-1]->printshort( o );
  o << ")";
}

string BuildObjectEdgeFinder::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string s = "global:unary";
  return s;
}


/**********************************************
 * Lex Constraint BuildObject
 **********************************************/
void BuildObjectLex::build( Solver *s, VariableInt **tmp, 
			    BuildObjectPredicate *pred ) 
{
  Constraint *con;
  int i, m = (pred->arity >> 1);
  VariableInt *B[m+1];
  for(i=1; i<m; ++i)
    B[i] = new VariableBool(s);
  B[0] = new Constant(s,0);
  if( pred->params[0] ) {
    B[m] = new Constant(s,1);
  } else 
    B[m] = new VariableBool(s);
  
  for(i=0; i<m; ++i) {
    VariableInt *scp[4];
    scp[0] = tmp[i];
    scp[1] = tmp[i+m];
    scp[2] = B[i];
    scp[3] = B[i+1];
    con = new ConstraintLex(s, scp);
  }
}

void BuildObjectLex::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  int i;
  o << "[ ";
  for(i=0; i<(pred->arity/2-1); ++i) {
    pred->scope[i]->printshort( o );
    o << " " ;
  }
  pred->scope[i]->printshort( o );
  o << "]"; 
  if(pred->params[0])
    o << " <= ";
  else
    o << " < ";
  o << "[ ";
  for(i=(pred->arity/2); i<pred->arity-1; ++i) {
    pred->scope[i]->printshort( o );
    o << " " ;
  }
  pred->scope[i]->printshort( o );
  o << "]";
}

string BuildObjectLex::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string s = "global:lex";
  return s;
}


/**********************************************
 * MaxOccurrence Constraint BuildObject
 **********************************************/
void BuildObjectMaxOccurrence::build( Solver *s, VariableInt **tmp, 
				      BuildObjectPredicate *pred ) 
{
  int n=pred->arity;
  Constraint *con;
  VariableInt *B[n];
  VariableInt *scp[2];
  for(int i=0; i<n; ++i) {
    B[i] = new VariableBool(s);
    scp[0] = tmp[i];
    scp[1] = B[i];
    con = new PredicateEqualConstant(s, scp, pred->params[0], 1);
  }
  con = new ConstraintBoolSumLess(s, B, n, pred->params[1]);
}

void BuildObjectMaxOccurrence::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  int i;
  o << "occ(" << pred->params[0] << ") <= " << pred->params[1] << " in [";
  for(i=0; i<pred->arity-1; ++i) {
    pred->scope[i]->printshort( o );
    o << "," ;
  }
  pred->scope[pred->arity-1]->printshort( o );
  o << "]";
}

string BuildObjectMaxOccurrence::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string s = "global:maxOccurence";
  return s;
}


/**********************************************
 * MinOccurrence Constraint BuildObject
 **********************************************/
void BuildObjectMinOccurrence::build( Solver *s, VariableInt **tmp, 
				      BuildObjectPredicate *pred ) 
{
  int n=pred->arity;
  Constraint *con;
  VariableInt *B[n];
  VariableInt *scp[2];
  for(int i=0; i<n; ++i) {
    B[i] = new VariableBool(s);
    scp[0] = tmp[i];
    scp[1] = B[i];
    con = new PredicateEqualConstant(s, scp, pred->params[0], 1);
  }
  con = new ConstraintBoolSumMore(s, B, n, pred->params[1]);
}

void BuildObjectMinOccurrence::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  int i;
  o << "occ(" << pred->params[0] << ") <= " << pred->params[1] << " in [";
  for(i=0; i<pred->arity-1; ++i) {
    pred->scope[i]->printshort( o );
    o << "," ;
  }
  pred->scope[pred->arity-1]->printshort( o );
  o << "]";
}

string BuildObjectMinOccurrence::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string s = "global:minOccurrence";
  return s;
}


/**********************************************
 * Slide Constraint BuildObject
 **********************************************/
inline void BuildObjectSlide::tuple2int(int& x, int* tuple, int n, int* exposant) {
  x = 0;
  while( n-- )
    x += (tuple[n] * exposant[n]);  
}

void BuildObjectSlide::build( Solver *s, VariableInt **tmp, BuildObjectPredicate *pred ) 
{
  Solver sub_solver( *sub_problem );

  int i, j, k, left, right, aux;

  int sequenceWidth = sub_solver.length;
  int nbVals = sub_solver.sequence[0]->domsize();
  int x, tuple[sequenceWidth];

  int *exposant = new int[sequenceWidth];
  for(i=0; i<sequenceWidth; ++i)
    exposant[i] = (int)pow(nbVals, i); 

  int* int2ind = new int[exposant[sequenceWidth-1]];
  int* ind2int = new int[exposant[sequenceWidth-1]];
  int* conflicts = new int[nbVals*exposant[sequenceWidth-1]]; 
  std::fill( ind2int, ind2int+exposant[sequenceWidth-1], -1 );
  std::fill( int2ind, int2ind+exposant[sequenceWidth-1], -1 );
  std::fill( conflicts, conflicts+(nbVals*exposant[sequenceWidth-1]), 1);


  sub_solver.startNewSearch();

  j=0; k=0;
  while( sub_solver.getNextSolution() ) {
    for(i=0; i<sequenceWidth; ++i) 
      tuple[i] = sub_solver.variables[i]->value();

    tuple2int(x, tuple, sequenceWidth, exposant);

    left = (x%exposant[sequenceWidth-1]);
    right = (x/nbVals);

    if( int2ind[left] < 0 ) {
      int2ind[left] = k;
      ind2int[k++] = left;
    }

    if( int2ind[right] < 0 ) {
      int2ind[right] = k;
      ind2int[k++] = right;
    }

    //ind2int[left]%nbVals + ind2int[right]*nbVals

    conflicts[(left%nbVals + right*nbVals)] = 0;

    //conflicts[ktuple(int2ind[left], int2ind[right])] = 0;

    cout << setw(6) <<  int2ind[left] << setw(6) << int2ind[right] ;
    for(i=0; i<sequenceWidth; ++i) {
      cout << setw(3) << tuple[i] ;
    }
    cout << endl;

  }

  Constraint *decomposition = new ConstraintSlideDecomp( s, tmp, pred->params[0], pred->params[1], 
							 nbVals, sequenceWidth, k,  
							 conflicts, int2ind, ind2int, exposant );

  //   int nSlideVars = m-sequenceWidth+2;
  //   VariableInt *slideVars[n*nSlideVars];
  //   for(i=0; i<n; ++i)
  //     for(j=0; j<nSlideVars; ++j)
  //       slideVars[i*nSlideVars+j] = new VariableBit( solver, 0, k-1 );
  
  //   VariableInt *scp[sequenceWidth];
  //   for(i=0; i<n; ++i) {
  //     for(j=0; j<nSlideVars; ++j) {
  //       for(k=0; k<sequenceWidth-1; ++k)
  // 	scp[k] = X[i*m+j+k];
  //       scp[sequenceWidth-1] = slideVars[i*nSlideVars+j];
  //       new PredicateTuple( solver, scp,
  // 			  sequenceWidth,
  // 			  nbVals,
  // 			  int2ind,
  // 			  ind2int,
  // 			  exposant ) );
    
  //     for(j=0; j<nSlideVars-1; ++j) {
  //       scp[0] = slideVars[i*nSlideVars+j];
  //       scp[1] = slideVars[i*nSlideVars+j+1];
  //       new SlideI( solver, scp, decomposition );
  //     }
  //   }
}

void BuildObjectSlide::print(std::ostream& o, const BuildObjectPredicate *pred) const
{
  o << "Slide(";
  for(int i=0; i<pred->arity-1; ++i) {
    pred->scope[i]->printshort( o );
    o << " " ;
  }
  pred->scope[pred->arity-1]->printshort( o );
  o << ")";
}

string BuildObjectSlide::xmlPred(int& idx, int level, const BuildObjectPredicate *pred, int *sub) const 
{
  string s = "global:slide";
  return s;
}




/**********************************************
 * Adding variables/constraints to the model
 **********************************************/

inline void CSP::triggerEvent(BuildObject* x, int e) 
{
  buildstack.push( x, e );
}

void CSP::add( Variable x )
{
  x.var_ptr_->add( this, 0, (BuildObject::CRANGE | BuildObject::CLIST) );
}

void CSP::add( VarArray& x )
{
  int n=x.size();
  for(int i=0; i<n; ++i) {
    x[i].var_ptr_->add( this, 0, (BuildObject::CRANGE | BuildObject::CLIST) );
  }
}

void CSP::add( Objective g )
{
  g.X->add( this, 1, (BuildObject::CRANGE | BuildObject::CLIST) );
  goald = g.obj;
}

int CSP::inferAllDiffs( const int all, const int limit )
{
  int i, n=declarations.size(), // m=notequals.size(),
    x, y, ncliques=0, maxsize;

  // compute nodes and edges of the graph
  vector< int > edges;
  vector< int > vertice;
  int vertice_index[n];
  std::fill(vertice_index, vertice_index+n, -1);
  
  i = notequals.size();
  while( i-- )
    {
      x = ((BuildObjectPredicate*)notequals[i])->scope[0]->id;
      y = ((BuildObjectPredicate*)notequals[i])->scope[1]->id;
      
      if( vertice_index[x] < 0  ) {
	vertice_index[x] = vertice.size();
	vertice.push_back( x );
      }
      if( vertice_index[y] < 0  ) {
	vertice_index[y] = vertice.size();
	vertice.push_back( y );
      }
      
      edges.push_back( vertice_index[x] );
      edges.push_back( vertice_index[y] );
    }
    
  if( vertice.size() < 4096 && edges.size() > (vertice.size() * 2) )
    //if( true )
    {
      
      MistralGraph G( vertice.size(), edges );      
      //cout << G << endl;
      
      vector< vector< int > > cliques;      
      if( all ) G.getAllCliques( cliques, limit );
      else G.getSomeCliques( cliques );
      
      x = cliques.size();
      while( x-- ) {
	i = n = cliques[x].size();
	maxsize = 0;
	while( i-- ) {
	  y = declarations[vertice[cliques[x][i]]]->size();
	  if(maxsize < y) maxsize = y;
	}

	//cout << n << " " << maxsize << endl;

	if( n > 3 && maxsize <= n+1 ) {
	  i=n;
	  ++ncliques;
	  BuildObject **tmp = new BuildObject*[n+1];
	  while( i-- ) tmp[n-i-1] = declarations[vertice[cliques[x][i]]];
	  int *par = new int[1];
	  BuildObjectPredicate *cliqueAllDiff = 
	    new BuildObjectPredicate( tmp, n, 0, -1, 
				      ENV[ConstraintStore::ALLDIFF],
				      par, (BuildObject::CRANGE | BuildObject::CONVEX) ) ;
	  cliqueAllDiff->add( this, 0, (BuildObject::CRANGE | BuildObject::CLIST) );
	  

	  //	  cout << cliques[x].size() << endl;
// 	  for(y=0; y<cliques[x].size(); ++y)
// 	    {
// 	  // 	  cout << " " << cliques[x][y];
// 	      declarations[vertice[cliques[x][y]]]->print( cout );
// 	      cout << " ";
// 	    }
// 	  cout << endl;
	}
      }
    }
  return ncliques;
}
  
void CSP::preBuild()
{
  close();
  
//   print( cout );
//   cout << endl;

  int n=declarations.size(), i=n;

  while( i-- ) {

//     cout << "preBuild: " ;
//     declarations[i]->print( cout );
//     cout << endl;

     declarations[i]->preBuild( );

//     cout << "result:  " ;
//     declarations[i]->print( cout );
//     // cout << endl;
// //     print( cout );
//     cout << endl << endl;

  }

  while( !buildstack.empty() ) {
    int event;
    BuildObject* x = buildstack.pop( event );
    i=x->parent.size();
    while( i-- ) {
      if( x->parent[i]->isRel() ) {

// 	cout << "+preBuild: " ;
// 	x->parent[i]->print( cout );
// 	cout << endl;

	x->parent[i]->preBuild( );

// 	cout << "result:    " << endl;
// 	//	print( cout );
// 	x->parent[i]->print( cout );
// 	cout << endl << endl;

      }
    }
  }


  i=n;
  while( i-- ) {

    if( declarations[i]->isNeq() ) {
      ++nNeq;
      notequals.push( declarations[i] );
    }


    if( declarations[i]->isVar() ) {
      if( declarations[i]->isSearchable() ) {
	nValues += declarations[i]->size();
	if( declarations[i]->isBoolean() ) {
	  if( declarations[i]->isRel() )
	    ++eBoolean;
	  else
	    ++nBoolean;
	  //cout << "Boolean " ;
	} else if( declarations[i]->isRange() ) {
	  if( declarations[i]->isRel() )
	    ++eRange;
	  else
	    ++nRange;
	  //cout << "Range " ;
	} else if( declarations[i]->isVList() ) {
	  if( declarations[i]->isRel() )
	    ++eList;
	  else
	    ++nList;
	  //cout << "Domain " ;
	} else {
	  if( declarations[i]->isRel() )
	    ++eBit;
	  else
	    ++nBit;
	  //cout << "Domain " ;
	}
      } else {
	//cout << "Constant " ;
	if( declarations[i]->isRel() )
	  ++eConstant;
	else
	  ++nConstant;
      }
    }
    //if( declarations[i]->isConstraint() ) {
    if( declarations[i]->isRel() ) {

      //cout << "Constraint " ;
      if( declarations[i]->isVar() )
	++eConstraint;
      else
	++nConstraint;
    }
    //    cout << endl;
  }
  //  if( nNeq > 3 )
  //    greedyCliques();

//   print( cout );
//   cout << endl;


}


void CSP::build( Solver *s )
{
  
//   print( cout );
//   cout << endl;

  int i, n=declarations.size();
  for(i=0; i<n; ++i) {

//      cout << "build ";
//      declarations[i]->print( cout );
//      cout << " " << declarations[i]->isReferenced() << " " << declarations[i]->isVar() << endl;


    if( 
       declarations[i]->isVar() || 
       declarations[i]->isRel() ) {

      declarations[i]->build( s );

    }
  }
  
  if( goald )
    goald->build( s );
}



/**********************************************
 * Operator overloading
 **********************************************/

Variable VarArray::operator[]( Variable x ) 
{
  return Element( *this, x );
}

Variable VarArray::operator< ( VarArray& x ) 
{
  return LexOrder( *this, x, 1 );
}
Variable VarArray::operator> ( VarArray& x ) 
{
  return LexOrder( x, *this, 1 );
}
Variable VarArray::operator<=( VarArray& x ) 
{
  return LexOrder( *this, x, 0 );
}
Variable VarArray::operator>=( VarArray& x )
{
  return LexOrder( x, *this, 0 );
}

Variable Variable::operator==( const MistralSet& s ) 
{
  // UNSOUND
  var_ptr_->setDomain( s );
  return *this;
}

Variable Variable::operator!=( const MistralSet& s ) 
{
  // UNSOUND
  var_ptr_->removeSet( s );
  return *this;
}

Variable Variable::operator==( const int k ) 
{
  return Equal( *this, k );
}

Variable Variable::operator%( const int k ) 
{
  return Mod( *this, k );
}

Variable Variable::operator!=( const int k ) 
{
  return NotEqual( *this, k );
}

Variable Variable::operator<=( const int k ) 
{
  return Precedence( *this, k );
}

Variable Variable::operator< ( const int k ) 
{
  return Precedence( *this, k+1 );
}

Variable Variable::operator>=( const int k ) 
{
  return Precedence( k, *this );
}

Variable Variable::operator> ( const int k ) 
{
  return Precedence( k+1, *this );
}

Variable Variable::operator+ ( Variable x )
{
  return Add( *this, x );
}

Variable Variable::operator- ( Variable x )
{
  return Sub( *this, x );
}

Variable Variable::operator* ( Variable x )
{
  return Mul( *this, x );
}

Variable Variable::operator/ ( Variable x )
{
  return Div( *this, x );
}

Variable Variable::operator^ ( Variable x )
{
  return Pow( *this, x );
}

Variable Variable::operator%( Variable x )
{
  return Mod( *this, x );
}

Variable Variable::operator&&( Variable x )
{
  return And( *this, x );
}

Variable Variable::operator||( Variable x )
{
  return Or( *this, x );
}

Variable Variable::operator==( Variable x ) 
{
  return Equal( *this, x );
}
 
Variable Variable::operator!=( Variable x ) 
{
  return NotEqual( *this, x );
}

Variable Variable::operator<=( Variable x ) 
{
  return Precedence( *this, 0, x );
}

Variable Variable::operator>=( Variable x ) 
{
  return Precedence( x, 0, *this );
}

Variable Variable::operator< ( Variable x ) 
{
  return Precedence( *this, 1, x );
}

Variable Variable::operator> ( Variable x ) 
{
  return Precedence( x, 1, *this );
}

Variable Variable::operator-( )
{
  return Negation( *this );
}

Variable Variable::operator!( )
{
  return Not( *this );
}

Variable Variable::operator*( const int k )
{
  if( k != 1 )
    return Factor( *this, k );
  else return *this;
}

Variable Variable::operator-( const int k )
{
  if( k )
    return Offset( *this, -k );
  else return *this;
}

Variable Variable::operator+( const int k )
{
  if( k )
    return Offset( *this, k );
  else return *this;
}


/**********************************************
 * Printing routines
 **********************************************/

ostream& operator<< (ostream& os, const Variable& x)
{
  x.print( os );
  return os;
}

ostream& operator<< (ostream& os, const VarArray& x)
{
  //  x.printshort( os );
  x.print( os );
  return os;
}

ostream& operator<< (ostream& os, const CSP& p)
{
  p.print( os );
  return os;
}



/**********************************************
 * Objective Function BuildObjects
 **********************************************/

void BuildObjectMinimiseVar::build( Solver *s )
{
  s->goal = new MinimiseVar( s, X->getVariable() );
}

void BuildObjectMaximiseVar::build( Solver *s )
{
  s->goal = new MaximiseVar( s, X->getVariable() );
}


/**********************************************
 * CSP
 **********************************************/

CSP::CSP() 
{
  nValues     = 0;
  nList       = 0;
  nBit        = 0;
  nBoolean    = 0;
  nRange      = 0;
  nConstant   = 0;
  nConstraint = 0;
  eValues     = 0;
  eList       = 0;
  eBit        = 0;
  eBoolean    = 0;
  eRange      = 0;
  eConstant   = 0;
  eConstraint = 0;
  nNeq        = 0;
  goald = NULL;
}

void CSP::cleanUp()
{
  nValues     = 0;
  nList       = 0;
  nBit        = 0;
  nBoolean    = 0;
  nRange      = 0;
  nConstant   = 0;
  nConstraint = 0;
  eValues     = 0;
  eList       = 0;
  eBit        = 0;
  eBoolean    = 0;
  eRange      = 0;
  eConstant   = 0;
  eConstraint = 0;
  int i = declarations.size();
  while( i-- ) {
    delete declarations[i];
  }
}

CSP::~CSP() 
{  
  cleanUp();
  
  delete goald;
  
  declarations.clear();
  goald = NULL;
}


void CSP::print(std::ostream& o) const
{

  int i, n=declarations.size();
  for(i=0; i<n; ++i) {

    cout << "- ";
    cout << declarations[i]->isReferenced() << " ";
    cout << declarations[i]->id << " ";
    cout << declarations[i]->getVarId() << " ";
    declarations[i]->print( cout );
    cout << endl;

  }
  


 //  int i, j, n;
//   if( goald ) {
//     goald->print( o );
//     o << endl << "with respect to:" << endl;
//   } else {
//     o << "Satisfy: " << endl;
//   }
//   j = first[0], n=declarations.size();
//   while( j >= 0 )
//     {
//       o << "\t" ;
//       declarations[j]->print( o );
//       //o << "\t" ;
//       //declarations[j]->printstate( o );
//       o << endl;
//       j = next[j];
//     }
//   o << endl << "where:";
//   i = first.size();
//   while( i-- > 1 ) 
//     {
//       j = first[i];
//       while( j >= 0 )
// 	{
// 	  //if( declarations[j]->isRel() ) {
// 	  o << "\t" ;	  
// 	  declarations[j]->print( o );
// 	  //o << "\t" ;
// 	  //declarations[j]->printstate( o );
// 	  o << endl;
// 	  //}
// 	  j = next[j];
// 	}
//     }
//   //   i = first.size();
//   //   o << endl;
//   //   while( i-- > 1 ) 
//   //     {
//   //       j = first[i];
//   //       while( j >= 0 )
//   // 	{
//   // 	  if( !declarations[j]->isRel() && declarations[j]->isSearchable() ) {
//   // 	    o << "\t" ;	  
//   // 	    declarations[j]->print( o );
//   // 	    //o << "\t" ;
//   // 	    //declarations[j]->printstate( o );
//   // 	    o << endl;
//   // 	  }
//   // 	  j = next[j];
//   // 	}
//   //     }
}

void CSP::printXML(std::ostream& o) const
{
  int nVariables = 0;
  int nConstraints = 0;
  int skippedConstraints = 0;
  int skippedVariables = 0;
  int maxarity = 0;

  int i, j=declarations.size(), k, l;

  int var2dec[j], con2dec[j], dec2var[j], subconstraint[j];
  for( i=0; i<j; ++i )
    { 
      subconstraint[i] = 0;
      if( declarations[i]->isRel() ) {
	if( declarations[i]->isConstraint() ) {
	  con2dec[nConstraints++] = i;	
	  if( declarations[i]->isVar() ) {
	    dec2var[declarations[i]->getVarId()] = nVariables;
	    var2dec[nVariables++] = i;
	  }
	}
	k = ((BuildObjectPredicate*)(declarations[i]))->arity;
	if( k > maxarity )
	  maxarity = k;
      } else if( declarations[i]->isVar() && (!declarations[i]->isPar() || declarations[i]->size() > 1) ) {
	dec2var[declarations[i]->getVarId()] = nVariables;
	var2dec[nVariables++] = i;
      }
    }

//    i=nVariables;
//    cout << "variables:" << endl;
//    while( i-- ) {
//      cout << i << " " << var2dec[i] << " " << dec2var[declarations[var2dec[i]]->getVarId()] << " ";
//      declarations[var2dec[i]]->print( cout );
//      cout << endl;
//    }

//   i=nConstraints;
//   cout << "constraints:" << endl;
//   while( i-- ) {
//     cout << i << " " ;
//     declarations[con2dec[i]]->print( cout );
//     cout << endl;
//   }


  
  int var2dom[nVariables];
  int con2pre[nConstraints];
  int con_type[nConstraints];
  map<string, int> domains_map;
  map<string, int> predicates_map;
  vector<string> domains;
  vector<string> predicates;
  vector<int> domainsize;
  vector<int> predicatesize;

  vector<string> global;
  vector<int> table;

  int nDomains=0;
  int nPredicates=0;
  int nparams=0;

  //for(i=0; i<nConstraints; ++i) {
  i=nConstraints;
  while( i-- ) {

//     cout << endl << con2dec[i] << " " << subconstraint[con2dec[i]] << " " ;
//     declarations[con2dec[i]]->print(cout) ;
//     cout << endl;

    if( !subconstraint[con2dec[i]] ) {

      //cout << "*********" << endl;

      nparams=0;
      string pred = declarations[con2dec[i]]->xmlPred( nparams, 0, subconstraint );           
      subconstraint[con2dec[i]] = 0;
      
      if( pred.substr(0,6) == "global" ) {
	//cout << "GLOBAL: " << pred << endl;
	global.push_back( pred );
	con_type[i] = 2;
      } else if( pred == "table" ) {
	//cout << "TABLE: " << pred << endl;
	table.push_back( i );
	con_type[i] = 1;
      } else {
	//cout << "PREDICATE: " << pred << endl;
	if( declarations[con2dec[i]]->isVar() )
	  pred = "eq(" + pred + ",X" + int2string(nparams++) + ")";
	con_type[i] = 0;
	if( predicates_map.find(pred) != predicates_map.end() ) {
	  con2pre[i] = predicates_map[pred];
	} else {
	  predicates_map[pred] = nPredicates;
	  predicates.push_back(pred);
	  predicatesize.push_back(nparams);
	  con2pre[i] = nPredicates++;
	}
      }
    } else {
      ++skippedConstraints;
      //      if( declarations[con2dec[i]]->isVar() )
      //	++skippedVariables;
    }
  }



  for(i=0; i<nVariables; ++i) {
    string dom = declarations[var2dec[i]]->xmlDom();
    if( domains_map.find( dom ) != domains_map.end() ) {
      var2dom[i] = domains_map[dom];      
    } else {
      domains_map[dom] = nDomains;
      domains.push_back( dom );
      domainsize.push_back( declarations[var2dec[i]]->size() );

//       cout << (declarations[var2dec[i]]->getVarId()) << " ";
//       cout << (declarations[var2dec[i]]->size()) << " ";
//       declarations[var2dec[i]]->print( cout );
//       cout << endl;

      var2dom[i] = nDomains++;
    }
  }


  // header	  
  o << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" 
    << endl << endl << "<instance>" << endl 
    << "<presentation maxConstraintArity=\""
    << maxarity <<"\" format=\"XCSP 2.1\"/>" 
    << endl << endl;
	
  // domains
  o << "  <domains nbDomains=\"" << nDomains << "\">" << endl;
  for(i=0; i<nDomains; ++i)
    {
      o << "    <domain name=\"D" << i 
	<< "\" nbValues=\"" << domainsize[i] << "\">" 
	<< domains[i]
	<< "</domain>" << endl; 
    }
  o << "  </domains>" << endl << endl;

  // variables
  i=nVariables;
  while( i-- )
    if( subconstraint[var2dec[i]] ) {
      var2dec[i] = var2dec[--nVariables];
      var2dom[i] = var2dom[nVariables];
      dec2var[var2dec[i]] = i;
      //++skippedVariables;
    }
  k=0;
  o << "  <variables nbVariables=\"" << (nVariables-skippedVariables) << "\">" << endl;
  for(i=0; i<nVariables; ++i)
    {
      //if( !subconstraint[var2dec[i]] )
      o << "    <variable name=\"V" << i
	<< "\" domain=\"D" << var2dom[i] << "\"/>" << endl;
    }
  o << "  </variables>" << endl << endl;
  
  // predicates
  if( nPredicates ) {
    o << "  <predicates nbPredicates=\"" 
      << (nPredicates) << "\">" << endl;
    for(i=0; i<nPredicates; ++i)
      {
	o << "    <predicate name=\"P"<< i << "\">" << endl
	  << "      <parameters>";
	for(j=0; j<predicatesize[i]; ++j)
	  o << " int X" << j;
	o << "</parameters>" << endl
	  << "      <expression> " << endl
	  << "        <functional>" 
	  << predicates[i] 
	  << "</functional>" << endl
	  << "      </expression>" << endl
	  << "    </predicate>" << endl;
      }
    o << "  </predicates>" << endl << endl;
  }

  // constraints
  int arity;
  BuildObjectPredicate *bcon;
  MistralSet scope(0, nVariables-1, MistralSet::empt);
  o << "  <constraints nbConstraints=\"" 
    << nConstraints-skippedConstraints << "\">" << endl;
  k=0;
  l=0;

//   for(i=0; i<nConstraints; ++i)
//     {
// 	bcon = ((BuildObjectPredicate*)(declarations[con2dec[i]]));
// 	bcon->print( cout );
// 	cout << endl;
//     }

  //for(i=0; i<nConstraints; ++i)

  i=nConstraints;
  while( i-- ) 
    if( !subconstraint[con2dec[i]] )
      //if( !subconstraint[con2dec[i]] )
      {
	bcon = ((BuildObjectPredicate*)(declarations[con2dec[i]]));
	
	/// PREDICATES
	if( con_type[i] == 0 ) {
	  scope.clear();
	  
	  string params = declarations[con2dec[i]]->xmlCon( scope, dec2var ); 
	  if( declarations[con2dec[i]]->isVar() )
	    {
	      if( declarations[con2dec[i]]->size() > 1 ) 
		{
		  j = dec2var[declarations[con2dec[i]]->getVarId()];
		  scope.insert(j);
		  params += (" V" + int2string(j));
		}
	      else
		{
		  params += (" " + int2string(declarations[con2dec[i]]->min()));
		}
	    }
	  arity = scope.size();
	  if( arity ) {
	    o << "    <constraint name=\"C" << l++ <<"\" arity=\""
	      << arity << "\" scope=\"";
	    j=scope.min();
	    while(true) {
	      o << "V" << j;//dec2var[j] ;
	      j = scope.next( j );
	      if( j != INT_MAX )
		o << " ";
	      else break;
	    } 
	  o << "\" reference=\"P" << con2pre[i] << "\">" << endl
	    << "      <parameters>" 
	    << params
	    << "</parameters> " << endl
	    << "    </constraint>" << endl; 
	  }
	}
	
	/// RELATIONS
	else if( con_type[i] == 1 ) {
	  ;
	}
      
	/// GLOBAL CONSTRAINTS
	else if( con_type[i] == 2 ) {

// 	  cout << i << " -- " << bcon->arity << " ";
// 	  bcon->print( cout );
// 	  cout << " " << global[k] << endl;

	  scope.clear();
	  arity = bcon->arity;
	  for(j=0; j<arity; ++j) {
	    scope.insert(dec2var[bcon->scope[j]->getVarId()]);
	  }
	  if( bcon->isVar() ) scope.insert( dec2var[bcon->scope[arity]->getVarId()] );
	  arity = scope.size();
	  o << "    <constraint name=\"C" << l++ <<"\" arity=\""
	    << arity << "\" scope=\"";
	  j = scope.min();
	  do { 
	    o << "V" << j << " ";
	    j = scope.next( j );
	  } while(j != INT_MAX);

// 	  for(j=0; j<arity-1; ++j) {
// 	    o << "V" << dec2var[bcon->scope[j]->getVarId()] << " ";
// 	  }
//	  o << "V" << dec2var[bcon->scope[j]->getVarId()];
	  o << "\" reference=\"" << global[k] ;
	  
	  if( global[k] == "global:element" )
	    {

// 	      bcon->print( cout );
// 	      cout << endl;	      
// 	      cout << "\">" << endl
// 		   << "      <parameters>" << endl
// 		   << "        V" << bcon->scope[bcon->arity-1]->getVarId() 
// 		   << endl << "        [ ";
// 	      for(j=0; j<bcon->arity-1; ++j)
// 		cout << "V" << bcon->scope[j]->getVarId() << " ";
// 	      cout << "]\n        V" 
// 		   << bcon->scope[bcon->arity]->getVarId() 
// 		   << endl << "      </parameters>"
// 		   << endl << "    </constraint>" << endl;
// 	      cout << endl;


	      o << "\">" << endl
		<< "      <parameters>" << endl
		<< "        V" << dec2var[bcon->scope[bcon->arity-1]->getVarId()] 
		<< endl << "        [ ";
	      for(j=0; j<bcon->arity-1; ++j)
		o << "V" << dec2var[bcon->scope[j]->getVarId()] << " ";
	      o << "]\n        V" 
		<< dec2var[bcon->scope[bcon->arity]->getVarId()] 
		<< endl << "      </parameters>"
		<< endl << "    </constraint>" << endl;
	    }
	  else if( global[k] == "global:cumulative" )
	    {
	      o << "\"/>" << endl;
	    }	  
	  else if( global[k] == "global:weightedSum" )
	    {
	      o << "\"/>" << endl;
	    }
	  else if( global[k] == "global:allDifferent" )
	    {
	      o << "\"/>" << endl;
	    }
	  ++k;
	}
      }
  o << "  </constraints>" 
    << endl << endl
    << "</instance>" << endl;

}



int ConstraintStore::addConstraint(BuildObjectConstraint *c)
{
  store.push( c );
  return (store.size()-1);
}

BuildObjectConstraint* ConstraintStore::getConstraint(const int c)
{
  if( !store[c] ) {
    switch( c ) {
    case ABS         : store[c] = new BuildObjectAbs();              break;
    case ALLDIFF     : store[c] = new BuildObjectAllDiff();          break;
    case AND         : store[c] = new BuildObjectAnd();              break;
    case CLAUSE      : store[c] = new BuildObjectClause();           break;
    case DISJUNCTIVE : store[c] = new BuildObjectDisjunctive();      break;
    case DIV         : store[c] = new BuildObjectDiv();              break;
    case EDGEFINDER  : store[c] = new BuildObjectEdgeFinder();       break;
    case ELEMENT     : store[c] = new BuildObjectElement();          break;
    case EQUAL       : store[c] = new BuildObjectEqual();            break;
    case NEQ         : store[c] = new BuildObjectNotEqual();         break;
    case NEQCONSTANT : store[c] = new BuildObjectNotEqualConstant(); break;
    case EQCONSTANT  : store[c] = new BuildObjectEqualConstant();    break;
    case FACTOR      : store[c] = new BuildObjectFactor();           break;
    case GCC         : store[c] = new BuildObjectGcc();              break;
    case LEX         : store[c] = new BuildObjectLex();              break;
    case MAX         : store[c] = new BuildObjectMax();              break;
    case MIN         : store[c] = new BuildObjectMin();              break;
    case MAXOCCUR    : store[c] = new BuildObjectMaxOccurrence();    break;
    case MINOCCUR    : store[c] = new BuildObjectMinOccurrence();    break;
    case MOD         : store[c] = new BuildObjectMod();              break;
    case MODCONSTANT : store[c] = new BuildObjectModConstant();      break;
    case MUL         : store[c] = new BuildObjectMul();              break;
    case ADD         : store[c] = new BuildObjectAdd();              break;
    case SUB         : store[c] = new BuildObjectSub();              break;
    case NEG         : store[c] = new BuildObjectNegation();         break;
    case NOT         : store[c] = new BuildObjectNot();              break;
    case OFFSET      : store[c] = new BuildObjectOffset();           break;
    case OR          : store[c] = new BuildObjectOr();               break;
    case POW         : store[c] = new BuildObjectPow();              break;
    case PRECEDENCE  : store[c] = new BuildObjectPrecedence();       break;
    case SUM         : store[c] = new BuildObjectSum();              break;
    case SUMEQ       : store[c] = new BuildObjectSumEq();            break;
    case NEQW        : store[c] = new BuildObjectNotEqualWildcard(); break;
    case SLIDE       : store[c] = new BuildObjectSlide();            break;
    case LOWERBOUND  : store[c] = new BuildObjectLowerBound();       break;
    case UPPERBOUND  : store[c] = new BuildObjectUpperBound();       break;
    case IFTHENELSE  : store[c] = new BuildObjectIfThenElse();       break;
      //case REGULAR     : store[c] = new BuildObjectRegular();          break;

#ifdef _COINLP
    case MAXDIFF     : store[c] = new BuildObjectMaxDifference();    break;
    case MINDIFF     : store[c] = new BuildObjectMinDifference();    break;
#endif

    }
  }
  return store[c];
}

ConstraintStore::ConstraintStore()
{
  store.init(NUMCONS, 1024);
  for(int i=0; i<NUMCONS; ++i)
    store[i] = NULL;
}

ConstraintStore::~ConstraintStore()
{
  int i = store.size();
  while( i-- )
    if( store[i] )
      delete store[i];
}

BuildObjectConstraint *ConstraintStore::operator[]( const int c ) 
{
  return getConstraint( c );
}
