
/*
  Mistral is a constraint satisfaction and optimisation library
  Copyright (C) 2003-2005  Emmanuel Hebrard
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  The author can be contacted electronically at 
  ehebrard@cse.unsw.edu.au.
*/

#include <iostream>
#include <iomanip>
#include <assert.h>

#include <csp.h>
#include <con.h>
#include <sol.h>
#include <var.h>


BitsetIterator* static_bit_domain_it = NULL;
ListIterator* static_list_domain_it = NULL;
BoolIterator* static_bool_domain_it = NULL;
RangeIterator* static_range_domain_it = NULL;
ConstantIterator* static_constant_domain_it = NULL;


ReversibleObj::~ReversibleObj()
{
}

void ReversibleObj::init(Solver *s)
{
  solver = s;
}

inline void ReversibleBool::operator= ( const int nstat )
{
  state = nstat;
  solver->save( this );
}

void ReversibleInt::operator= ( const int x )
{
  if( solver->level != lvl_.back() ) {
    lvl_.push( solver->level );
    val_.push( value );
    solver->save( this );
  }
  value = x;
}

void ReversibleInt::operator+= ( const int x )
{
  if( solver->level != lvl_.back() ) {
    lvl_.push( solver->level );
    val_.push( value );
    solver->save( this );
  }
  value += x;
}

void ReversibleInt::operator-= ( const int x )
{
  if( solver->level != lvl_.back() ) {
    lvl_.push( solver->level );
    val_.push( value );
    solver->save( this );
  }
  value -= x;
}

void ReversibleInt::operator*= ( const int x )
{
  if( solver->level != lvl_.back() ) {
    lvl_.push( solver->level );
    val_.push( value );
    solver->save( this );
  }
  value *= x;
}

void ReversibleInt::operator/= ( const int x )
{
  if( solver->level != lvl_.back() ) {
    lvl_.push( solver->level );
    val_.push( value );
    solver->save( this );
  }
  value /= x;
}

void ReversibleInt::operator|= ( const int x )
{
  if( solver->level != lvl_.back() ) {
    lvl_.push( solver->level );
    val_.push( value );
    solver->save( this );
  }
  value |= x;
}

void ReversibleInt::operator&= ( const int x )
{
  if( solver->level != lvl_.back() ) {
    lvl_.push( solver->level );
    val_.push( value );
    solver->save( this );
  }
  value &= x;
}

void ReversibleInt::operator^= ( const int x )
{
  if( solver->level != lvl_.back() ) {
    lvl_.push( solver->level );
    val_.push( value );
    solver->save( this );
  }
  value ^= x;
}

 void ReversibleInt::operator++()
{
  if( solver->level != lvl_.back() ) {
    lvl_.push( solver->level );
    val_.push( value );
    solver->save( this );
  }
  ++value;
}

void ReversibleInt::operator--()
{
  if( solver->level != lvl_.back() ) {
    lvl_.push( solver->level );
    val_.push( value );
    solver->save( this );
  }
  --value;
}


inline void ReversibleDouble::operator= ( const double x )
{
  if( solver->level != lvl_.back() ) {
    lvl_.push( solver->level );
    val_.push( value );
    solver->save( this );
  }
  value = x;
}

 void ReversibleDouble::operator++()
{
  if( solver->level != lvl_.back() ) {
    lvl_.push( solver->level );
    val_.push(value);
    solver->save( this );
  }
  ++value;
}

void ReversibleDouble::operator--()
{
  if( solver->level != lvl_.back() ) {
    lvl_.push( solver->level );
    val_.push(value);
    solver->save( this );
  }
  --value;
}

ReversibleList::~ReversibleList()
{
  delete [] list_;
  delete [] absidx;

  value_.neg_words = 0;
  value_.table = NULL;
}

void ReversibleList::init(Solver* s, 
			  MistralSet& e, 
			  const int lb, 
			  const int ub,
			  const unsigned int l)
{
  ReversibleObj::init(s);
  
  size = l;
  size_.push(size);
  lvl_.push(0);

  value_.pos_words = e.pos_words;
  value_.neg_words = e.neg_words;
  value_.table = e.table;

  list_ = new int[size+1];
  list_[size] = INT_MAX;
  absidx = new unsigned int[ub-lb+1];
  index_ = (absidx - lb);

  int i;
  if( size == (ub-lb+1) ) 
    {
      for(i=0; i<size; ++i) 
	{
	  index_[lb+i] = i;
	  list_[i] = (lb+i);
	}
    }
  else
    {
      i=0;
      int j=value_.min();
      do {
	index_[j] = i;
	list_[i] = j;
	++i;
      } while( (j = value_.next(j)) != INT_MAX );
    }
}

void ReversibleList::checkConsistency()
{
  assert( size == value_.size() );
  for(int i=0; i<size_[0]; ++i)
    {
      //      cout << setw(2) << list_[i] <<
      assert( index_[list_[i]] == i );
    }
}

void ReversibleList::save()
{
  if( solver->level != lvl_.back() ) {
    lvl_.push( solver->level );
    size_.push( size );
    solver->save( this );
  }
}

void ReversibleList::restore()
{
  int i=size;
  size_.pop( size );
  lvl_.pop();

  while( i < size ) {
    value_.fastInvert( list_[i] );
    ++i;
  }

  //  checkConsistency();

}

int ReversibleList::next(const int v) const
{
  
//   cout << "next " << v << ": " << endl;
//   for(int i=0; i<size_[0]; ++i)
//     {
//       cout << setw(2) << list_[i] << " ";
//     }
//   cout << endl;
//   for(int i=0; i<size_[0]; ++i)
//     {
//       cout << setw(2) << index_[list_[i]] << " ";
//     }
//   cout << endl;

  int idx = index_[v]+1;
  return (idx < size ? list_[idx] : VariableInt::NOVAL);
}

void ReversibleList::setTo(int elt)
{
  save();
  size=1;
  index_[*list_] = index_[elt];
  list_[index_[elt]] = *list_;
  *list_ = elt;
  index_[elt] = 0;
  value_.clear();
  value_.fastInvert( elt );

  //  checkConsistency();
}
void ReversibleList::intersectWith(const MistralSet& s)
{
  save();
  

  //TODO
}
void ReversibleList::setminusWith(const MistralSet& s)
{
  //TODO
}
void ReversibleList::erase(const int elt)
{
  save();
  --size;
  index_[list_[size]] = index_[elt];
  list_[index_[elt]] = list_[size];
  list_[size] = elt;
  index_[elt] = size;
  value_.fastInvert(elt);

  //  checkConsistency();
}
void ReversibleList::setMin(int l)
{
  //TODO
}
void ReversibleList::setMax(int u)
{
  //TODO
}
void ReversibleList::removeInterval(int l, int u)
{
  //TODO
}

void ReversibleList::print(std::ostream& o) const 
{
  value_.print( o );
}

void ReversibleSet::init(Solver* s, const int lb, const int ub)
{
  clone = false;
  ReversibleObj::init(s);
  glvl = 0;
  value.init(lb, ub, MistralSet::full);

  int i = value.pos_words;
  int j = value.neg_words;
  lvl_ = new MistralStack<int>[i-j];
  val_ = new MistralSet[i-j];
  lvl_ -= j;
  val_ -= j;
  

  int lmax;
  while( i-- > j ) { 
    lmax = value.size(i);
    val_[i].init( lmax+1, value.table[i] );
    lvl_[i].init( 0, lmax+1 );
    lvl_[i].push(0);
  }
}

void ReversibleSet::init(Solver* s, MistralSet& e, const int l)
{
  clone = true;
  ReversibleObj::init(s);
  glvl = 0;
  value.pos_words = e.pos_words;
  value.neg_words = e.neg_words;
  value.table = e.table;
  int i = value.pos_words;
  int j = value.neg_words;
  lvl_ = new MistralStack<int>[i-j];
  val_ = new MistralSet[i-j];
  lvl_ -= e.neg_words;
  val_ -= e.neg_words;
  

  int lmax;
  while( i-- > j ) { 
    lmax = value.size(i);
    val_[i].init( lmax+1, value.table[i] );
    lvl_[i].init( 0, lmax+1 );
    lvl_[i].push(0);
  }
}


void ReversibleSet::save()
{
  int l=solver->level;
  if( glvl < l ) 
    {
      glvl = l;
      int i = value.pos_words;
      int j = value.neg_words;
      while( i-- > j ) 
	{
	  l = value.table[i];
	  if( l != val_[i].table[lvl_[i].size_-1] ) 
	    {
	      val_[i].table[lvl_[i].size_] = l;
	      lvl_[i].push(glvl);
	    }
	}
      solver->save( this );
    }
}

inline void ReversibleSet::restore()
{
  int i = value.pos_words;
  int j = value.neg_words;
  int l = solver->level;
  
  glvl = 0;
  while( i-- > j ) 
    {
      value.table[i] = val_[i].table[lvl_[i].size_-1];
      if(lvl_[i].back() == l)
	lvl_[i].pop();
      if( lvl_[i].back() > glvl )
	glvl = lvl_[i].back();
    }
}


/**********************************************
 * VariableDomain
 **********************************************/

void VariableDomain::triggerEvent( const bool bc )
{
  if( vmin == vmax ) {
    solver->triggerEvent(this, Constraint::VALUETRIGGER);
  }
  else if( bc )
    solver->triggerEvent(this, Constraint::RANGETRIGGER);
  else solver->triggerEvent(this, Constraint::DOMAINTRIGGER);
}

void VariableDomain::setVariable(MistralSet *s, const int lg, 
				 const int lo, const int up)
{
  isWord = (lo >= 0 && up < 32);
  vmax.setValue( up );
  vmin.setValue( lo );
  posval = up+1;
  negval = lo;
  values.pos_words = s->pos_words;
  values.neg_words = s->neg_words;
  values.table = s->table;
  s->table = NULL;
}

void VariableDomain::setVariable(const int lo, const int up) 
{
  isWord = (lo >= 0 && up < 32);
  vmin.setValue( lo );
  vmax.setValue( up );

  negval = lo;
  posval = up+1;

  values.init(lo, up, MistralSet::full);
}


/**********************************************
 * VariableBit
 **********************************************/

void VariableBit::check_domain(int x) const
{
  cout << "check domain " << x << endl;
  assert( values.size() == size );
  assert( values.max() == vmax );
  assert( values.min() == vmin );

  int real_size = 0;
  for(int i=negval; i<posval; ++i)
    real_size += values.member( i );

  assert( real_size == size );
}

inline ValIterator *VariableBit::getDomainIterator()
{
  static_bit_domain_it->curval = vmin;
  static_bit_domain_it->domain.pointTo( values );
  return static_bit_domain_it;
}

VariableBit::VariableBit(Solver *s, const int v) : VariableDomain(s) 
  { 
    size.init(s);
    size.setValue(v);
    setVariable(0, v-1); 
    domain.init(s, values, size);
    //if(!domain_it) domain_it = new BitsetIterator();
    //domain_it = new BitsetIterator( values );
    if( !static_bit_domain_it ) 
      static_bit_domain_it = new BitsetIterator( values );
  }
  /*! 
    Constructor, create a domain containing all values
    in the interval [l,-,u]
  */
VariableBit::VariableBit(Solver *s, const int l, const int u) : VariableDomain(s) 
  { 
    size.init(s);
    size.setValue( u-l+1 );
    setVariable(l, u); 
    domain.init(s, values, size);
    //if(!domain_it) domain_it = new BitsetIterator();
    //domain_it = new BitsetIterator( values );
    if( !static_bit_domain_it ) 
      static_bit_domain_it = new BitsetIterator( values );
  }
  /*! 
    Constructor, create a domain containing all values 
    in the set e
  */
 VariableBit::VariableBit(Solver *s, MistralSet *e, 
	     const int lg, 
	     const int lo, 
	     const int up) : VariableDomain(s)
  { 
    size.init(s);
    size.setValue( lg );
    setVariable(e, lg, lo, up); 
    domain.init(s, values, size);
    //if(!domain_it) domain_it = new BitsetIterator();
    //domain_it = new BitsetIterator( values );
    if( !static_bit_domain_it ) 
      static_bit_domain_it = new BitsetIterator( values );
  }

VariableBit::~VariableBit() 
{
  //delete domain_it;
}

bool VariableBit::remove(const int v) 
{

#ifdef _CONSISTENCYCHECK
  check_domain(1);
#endif

  if(!contain(v)) return true;
  if( 1 == size ) return false;

  domain.erase( v );
  --size;
  
  triggerEvent( boundchanged( v ) );

#ifdef _CONSISTENCYCHECK
  check_domain(2);
#endif

  return true;
}

bool VariableBit::setDomain(const int v) 
{

#ifdef _CONSISTENCYCHECK
  check_domain(3);
#endif

  if( !contain( v ) ) return false;
   
  if( 1 < size ) {

    size = 1;

    domain.setTo( v );

    vmin = v;
    vmax = v;

    solver->triggerEvent(this, Constraint::VALUETRIGGER);
  }

#ifdef _CONSISTENCYCHECK
  check_domain(4);
#endif

  return true;
}

bool VariableBit::setMax(const int up) 
{

#ifdef _CONSISTENCYCHECK
  check_domain(5);
#endif

  if(vmin > up)  return false;  
  if(vmax <= up) return true;   

  domain.setMax(up);
  size = values.size();

  if( values.fastMember(up) ) vmax = up;
  else vmax = values.max();

  triggerEvent( true );

#ifdef _CONSISTENCYCHECK
  check_domain(6);
#endif

  return true;
}

bool VariableBit::setMin(const int low) 
{

#ifdef _CONSISTENCYCHECK
  check_domain(7);
#endif

  if(vmax < low)  return false;  
  if(vmin >= low) return true;  

  domain.setMin(low);
  size = values.size();

  if( values.fastMember(low) ) vmin = low;
  else vmin = values.min();

  triggerEvent( true );

#ifdef _CONSISTENCYCHECK
  check_domain(8);
#endif

  return true;
}

bool VariableBit::setDomain(const int *d, const int n) 
{
  MistralSet aux(vmin, vmax, MistralSet::full);
  int i=n;
  while( i-- ) aux.insert( d[i] );
  return setDomain( aux );
}

bool VariableBit::setDomain(VariableInt *x) const 
{
  if( 1 == size ) 
    return x->setDomain( vmin );
  return x->setDomain( values );
}

bool VariableBit::setDomain(const MistralSet& s) 
{

#ifdef _CONSISTENCYCHECK
  check_domain(9);
#endif

  if( !values.intersect(s) ) return false;
  if( values.included(s) ) return true;

  domain.intersectWith(s);
  size = values.size();

  triggerEvent( boundchanged( s, true ) );

#ifdef _CONSISTENCYCHECK
  check_domain(10);
#endif

  return true;
}

bool VariableBit::removeRange(const int lo, const int up) 
{

#ifdef _CONSISTENCYCHECK
  check_domain(11);
#endif

  if( !(values.intersect( lo, up )) )
    return true;

  if( lo <= vmin ) 
    return setMin(up+1);
  if( up >= vmax ) 
    return setMax(lo-1);  

  domain.removeInterval(lo, up);
  size = values.size();

  solver->triggerEvent(this, Constraint::DOMAINTRIGGER);  

#ifdef _CONSISTENCYCHECK
  check_domain(12);
#endif

  return true;
}

bool VariableBit::removeSet(const MistralSet& s) 
{  

#ifdef _CONSISTENCYCHECK
  check_domain(13);
#endif

  if( values.included(s) ) return false;
  if( !values.intersect(s) ) return true;

  domain.setminusWith(s);    
  size = values.size();

  triggerEvent( boundchanged( s, false ) );

#ifdef _CONSISTENCYCHECK
  check_domain(14);
#endif

  return true;
}

inline bool VariableBit::revise( MistralSet* supports, 
				 int* residues, 
				 VariableInt* X)
{
  int vali = vmin;
  bool consistent;
  if( X->isWord )
    do consistent = (X->wordIntersect( supports[vali] ) || remove( vali ));
    while( consistent && setNext(vali) );
  else 
    do consistent = (X->intersect( supports[vali], residues[vali] ) || remove( vali ));
    while( consistent && setNext(vali) );
  return consistent;
}

inline bool VariableBit::revise( Constraint* c, const int i )
{
  int vali = vmin;
  do {
    // This procedure does two things:
    // * first it checks if we can say that the value is supported
    //   without constraint check
    // * second it ready the solution in order to be check
    //   by findSupport	
    
    if( !( c->firstSupport(i, vali) || 
	   c->findSupport(i, vali) ||
	   remove(vali) ) )
      return false;	
	
  } while( setNext(vali) );
  
  return true;
}

void VariableBit::print(std::ostream& o) const 
{
  o << "x" << id ;
  if( isGround() )
    o << " = " << min();
  else {
    o << " in ";
    values.print( o );
  }
}

void VariableBit::printshort(std::ostream& o) const 
{
  o << "x" << id ;
}


/**********************************************
 * VariableList
 **********************************************/

void VariableList::check_domain(int x) const
{
  cout << "check domain " << x << endl;
  assert( values.size() == domain.size );
  assert( values.max() == vmax );
  assert( values.min() == vmin );

  int real_size = 0;
  for(int i=negval; i<posval; ++i)
    real_size += values.member( i );

  assert( real_size == domain.size );
}

VariableList::VariableList(Solver *s, const int v) : VariableDomain(s) 
{ 
  setVariable(0, v-1); 
  domain.init(s, values, 0, v-1, v);
  if( !static_list_domain_it ) static_list_domain_it = new ListIterator();
}

VariableList::VariableList(Solver *s, const int l, const int u) : VariableDomain(s) 
{ 
  setVariable(l, u); 
  domain.init(s, values, l, u, u-l+1);
  if( !static_list_domain_it ) static_list_domain_it = new ListIterator();
}

VariableList::VariableList(Solver *s, 
			   MistralSet *e, 
			   const int lg, 
			   const int lo, 
			   const int up) : VariableDomain(s)
{ 
  setVariable(e, lg, lo, up); 
  domain.init(s, values, lo, up, lg);
  if( !static_list_domain_it ) static_list_domain_it = new ListIterator();
}

inline ValIterator *VariableList::getDomainIterator()
{
  static_list_domain_it->size = domain.size-1;
  static_list_domain_it->vals = domain.list_;
  return static_list_domain_it;
}

VariableList::~VariableList() 
{
  //delete domain_it;
}

bool VariableList::remove(const int v) 
{

#ifdef _CONSISTENCYCHECK
  check_domain(1);
#endif

  if(!contain(v)) return true;
  if( 1 == domain.size ) return false;

  domain.erase( v );
  
  triggerEvent( boundchanged( v ) );

#ifdef _CONSISTENCYCHECK
  check_domain(2);
#endif

  return true;
}

bool VariableList::setDomain(const int v) 
{

#ifdef _CONSISTENCYCHECK
  check_domain(3);
#endif

  if( !contain( v ) ) return false;
   
  if( 1 < domain.size ) {

    domain.setTo( v );

    vmin = v;
    vmax = v;

    solver->triggerEvent(this, Constraint::VALUETRIGGER);
  }

#ifdef _CONSISTENCYCHECK
  check_domain(4);
#endif

  return true;
}

bool VariableList::setMax(const int up) 
{

// #ifdef _CONSISTENCYCHECK
//   check_domain(5);
// #endif

//   if(vmin > up)  return false;  
//   if(vmax <= up) return true;   

//   domain.setMax(up);

//   if( values.member(up) ) vmax = up;
//   else vmax = values.max();

//   triggerEvent( true );

// #ifdef _CONSISTENCYCHECK
//   check_domain(6);
// #endif

//TODO

  return true;
}

bool VariableList::setMin(const int low) 
{

// #ifdef _CONSISTENCYCHECK
//   check_domain(7);
// #endif

//   if(vmax < low)  return false;  
//   if(vmin >= low) return true;  

//   domain.setMin(low);
//   size = values.size();

//   if( values.member(low) ) vmin = low;
//   else vmin = values.min();

//   triggerEvent( true );

// #ifdef _CONSISTENCYCHECK
//   check_domain(8);
// #endif

//TODO

  return true;
}

bool VariableList::setDomain(const int *d, const int n) 
{
//   MistralSet aux(vmin, vmax);
//   int i=n;
//   while( i-- ) aux.insert( d[i] );
//   return setDomain( aux );

//TODO

}

bool VariableList::setDomain(VariableInt *x) const 
{
  if( 1 == domain.size ) 
    return x->setDomain( vmin );
  return x->setDomain( values );

//TODO

}

bool VariableList::setDomain(const MistralSet& s) 
{

// #ifdef _CONSISTENCYCHECK
//   check_domain(9);
// #endif

//   if( !values.intersect(s) ) return false;
//   if( values.included(s) ) return true;

//   domain.intersectWith(s);
//   size = values.size();

//   triggerEvent( boundchanged( s, true ) );

// #ifdef _CONSISTENCYCHECK
//   check_domain(10);
// #endif

//TODO

  return true;
}

bool VariableList::removeRange(const int lo, const int up) 
{

// #ifdef _CONSISTENCYCHECK
//   check_domain(11);
// #endif

//   if( !(values.intersect( lo, up )) ) return true;
//   if( lo <= vmin ) return setMin(up+1);
//   if( up >= vmax ) return setMax(lo-1);  

//   domain.removeInterval(lo, up);
//   size = values.size();

//   solver->triggerEvent(this, Constraint::DOMAINTRIGGER);  

// #ifdef _CONSISTENCYCHECK
//   check_domain(12);
// #endif

//TODO

  return true;
}

bool VariableList::removeSet(const MistralSet& s) 
{  

// #ifdef _CONSISTENCYCHECK
//   check_domain(13);
// #endif

//   if( values.included(s) ) return false;
//   if( !values.intersect(s) ) return true;

//   domain.setminusWith(s);    
//   size = values.size();

//   triggerEvent( boundchanged( s, false ) );

// #ifdef _CONSISTENCYCHECK
//   check_domain(14);
// #endif

//TODO

  return true;
}

inline bool VariableList::revise( MistralSet* supports, 
				  int* residues, 
				  VariableInt* X)
{
  bool consistent = true;
  int *val = domain.list_, i=domain.size;
  if( X->isWord )
    while( consistent && i-- ) 
      consistent = (X->wordIntersect( supports[val[i]] ) || remove( val[i] ));
  else 
    while( consistent && i-- )
      consistent = (X->intersect( supports[val[i]], residues[val[i]] ) || remove( val[i] ));
  return consistent;

//   int *val = domain.list_, i=domain.size;
//   if( X->isWord )
//     while( i-- && (X->wordIntersect( supports[val[i]] ) || remove( val[i] )) ) ;
//   else 
//     while( i-- && (X->intersect( supports[val[i]], residues[val[i]] ) || remove( val[i] )) ) ;
//   return (i < 0);


//   //bool consistent = true;
//   int *val = domain.list_, i=domain.size;
//   if( isWord )
//     while( i-- ) {
//       if( !(X->wordIntersect( supports[val[i]] ) || remove( val[i] )) ) break;
//     } 
//   else 
//     while( i-- ) {
//       if( !(X->intersect( supports[val[i]], 
// 			  residues[val[i]] ) || remove( val[i] )) ) break;  
//     }
//   return ( i < 0 );


//   int *val = domain.list_, i=domain.size;
//   if( isWord )
//     while( i-- )
//       if(!X->wordIntersect( supports[val[i]] ) && !remove( val[i] )) break;
//   else 
//     while( i-- )
//       if(!X->intersect( supports[val[i]], 
// 			residues[val[i]] ) && !remove( val[i] )) break;
//   return (i < 0);

}

inline bool VariableList::revise( Constraint* c, const int i )
{
  bool consistent = true;
  int *val = domain.list_, j=domain.size;
  while( consistent && j-- ) {
    consistent = ( c->firstSupport(i, val[j]) || 
		   c->findSupport (i, val[j]) ||
		   remove(val[j]) );
  }
  return consistent;
}


void VariableList::print(std::ostream& o) const 
{
  o << "y" << id ;
  if( isGround() )
    o << " = " << min();
  else {
    o << " in ";
    values.print( o );
  }
}

void VariableList::printshort(std::ostream& o) const 
{
  o << "y" << id ;
}


/**********************************************
 * VariableBool
 **********************************************/

VariableBool::VariableBool(Solver *s) : VariableInt(s) 
{
  domain.init(s);
  if( !static_bool_domain_it ) static_bool_domain_it = new BoolIterator();
}

inline ValIterator *VariableBool::getDomainIterator()
{
  static_bool_domain_it->curval = !(domain.state & 1);
  return static_bool_domain_it;
}

VariableBool::~VariableBool() 
{
  //delete domain_it;
}

inline bool VariableBool::setState( const int nstat ) 
{
  if( !nstat ) return false;
  if( nstat == domain.state ) return true;
  domain = nstat;
  solver->triggerEvent(this, Constraint::VALUETRIGGER);
  return true;
}

bool VariableBool::remove(const int v) 
{
  if(!((1 << v) & 3)) return true;
  return setState( domain.state & (2-v) );
}

bool VariableBool::setDomain(const int v) 
{
  if((1 << v) & 3) 
    return setState( domain.state & (v+1) );
  else return false;
}

bool VariableBool::setMax(const int up) 
{
  if( up > 0 ) return true;
  if( up < 0 ) return false;
  return setState( domain.state & 1 );
}

bool VariableBool::setMin(const int low) 
{
  if( low < 1 ) return true;
  if( low > 1 ) return false;
  return setState( domain.state & 2 );
}

bool VariableBool::setDomain(const int *d, const int n) 
{
  int nstat = 0, i=n;
  while( i-- ) 
    if( (nstat |= ((1 << d[i]) & 3) ) == 3 )  break;
  return setState( nstat );
}

bool VariableBool::setDomain(VariableInt* x) const
{ 
  if( domain.state == 3 )
    return ( x->setMin(0) && x->setMax(1) );
  else return ( x->setDomain(domain.state-1) );
}

bool VariableBool::setDomain(const MistralSet& s) 
{
  if( s.pos_words < 1 || s.neg_words > 0 ) return false;
  return setState( domain.state & s.table[0] );
}

bool VariableBool::removeRange(const int lo, const int up) 
{
  if( lo > up || lo > 1 || up < 0 ) return true;
  int nstat = 3;
  if(lo < 1) nstat ^= 1;
  if(up > 0) nstat ^= 2;
  return setState( nstat );
}

bool VariableBool::removeSet(const MistralSet& s) 
{
  if( s.pos_words < 1 || s.neg_words > 0 ) return true;
  return setState( (domain.state ^ s.table[0]) & 3 );
}

inline bool VariableBool::revise( MistralSet* supports, 
				  int* residues, 
				  VariableInt* X)
{
  int i=2;
  while( i-- )
    if( !(X->intersect( supports[i], 
			residues[i] ) ||
	  remove( i )) ) return false;
  return true;
}

inline bool VariableBool::revise( Constraint* c, const int i )
{
  int vali=2;
  while( vali-- )    
    if( !( c->firstSupport(i, vali) || 
	   c->findSupport(i, vali) ||
	   remove(vali) ) )
      return false;
  return true;
}

void VariableBool::print(std::ostream& o) const 
{
  o << "b" << id ;
  if( domain.state == 3 )
    o << " in {0,1}";
  else o << " = " << (domain.state-1) ;
}

void VariableBool::printshort(std::ostream& o) const 
{
  o << "b" << id ;
}


/**********************************************
 * VariableRange
 **********************************************/

VariableRange::VariableRange(Solver *s, const int v) : VariableInt(s) 
{ 
  setVariable(0, v-1); 
  vmin.init(s);
  vmax.init(s);
  static_range_domain_it = new RangeIterator();
}

VariableRange::VariableRange(Solver *s, const int l, const int u) : VariableInt(s) 
{ 
  setVariable(l, u); 
  vmin.init(s);
  vmax.init(s);
  static_range_domain_it = new RangeIterator();
}

VariableRange::VariableRange(Solver *s) : VariableInt(s) 
{
  if( !static_range_domain_it ) static_range_domain_it = new RangeIterator();
}

inline ValIterator *VariableRange::getDomainIterator()
{
  static_range_domain_it->curval = vmin;
  static_range_domain_it->ubound = vmax;
  return static_range_domain_it;
}

VariableRange::~VariableRange() 
{
  //delete domain_it;
}

void VariableRange::setVariable(const int low, const int up) 
{
  vmax.setValue(up);
  vmin.setValue(low);
}


bool VariableRange::remove(const int v) 
{
  if( v > vmax || v < vmin ) return true;
  if( (int)vmin == (int)vmax ) return false;

  if( v == vmin ) 
    ++vmin;
  else if( v == vmax )
    --vmax;
  else return true;

  if((int)vmin < (int)vmax)
    solver->triggerEvent(this, Constraint::RANGETRIGGER);
  else solver->triggerEvent(this, Constraint::VALUETRIGGER);

  return true;
}

bool VariableRange::setDomain(VariableInt* x) const
{ 
  if( (int)vmin == (int)vmax )
    return x->setDomain( vmin );
  return ( x->setMin(vmin) && x->setMax(vmax) );
}

bool VariableRange::setDomain(const int v) 
{
  if( v < vmin || v > vmax )
    return false;
  if( (int)vmin < (int)vmax ) {
    solver->triggerEvent(this, Constraint::VALUETRIGGER);
    vmin = v;
    vmax = v;
  }
  return true;
}

bool VariableRange::setMax(const int up) 
{
  if(vmin > up)
    return false;
  if(vmax <= up)
    return true;  

  vmax = up;
  if( (int)vmin == (int)vmax )
    solver->triggerEvent(this, Constraint::VALUETRIGGER);
  else solver->triggerEvent(this, Constraint::RANGETRIGGER);

  return true;
}

bool VariableRange::setMin(const int low) 
{
  if(vmax < low)
    return false;
  if(vmin >= low)
    return true;  

  vmin = low;
  if( (int)vmin == (int)vmax )
    solver->triggerEvent(this, Constraint::VALUETRIGGER);
  else solver->triggerEvent(this, Constraint::RANGETRIGGER);

  return true;
}

bool VariableRange::setDomain (const int *d, const int n) 
{
  int lb = INT_MAX;
  int ub = -1*INT_MAX;
  int i=n;
  while( i-- ) {
    if(d[i] > ub) ub=d[i];
    if(d[i] < lb) lb=d[i];
  }    
  return(setMin(lb) && setMax(ub));
}

bool VariableRange::setDomain (const MistralSet& s) 
{
  int lb = s.next( vmin-1 );
  int ub = s.prev( vmax+1 );

  return(setMin(lb) && setMax(ub));
}

bool VariableRange::removeRange(const int lo, const int up) 
{
  if( lo > vmax || up < vmin ) return true;
  if( lo <= vmin )
    return setMin(up+1);
  if( up >= vmax ) 
    return setMax(lo-1);  
  return true;
}

bool VariableRange::removeSet(const MistralSet& s) 
{
  int lb = vmin;
  int ub = vmax;
  while( s.member(lb) ) ++lb; 
  while( s.member(ub) ) --ub; 
  return(setMin(lb) && setMax(ub));
}

inline bool VariableRange::revise( MistralSet* supports, 
				   int* residues, 
				   VariableInt* X)
{
  // compute ub
  int i=vmax;
  while( !(X->intersect( supports[i], 
			 residues[i] )) ) --i;
  if( !setMax(i) ) return false;

  // compute lb
  i=vmin;
  while( !(X->intersect( supports[i], 
			 residues[i] )) ) ++i;
  if( !setMin(i) ) return false;

  return true;
}


inline bool VariableRange::revise( Constraint *c, const int i )
{
  // compute ub
  int vali=vmax;
  while ( !( c->firstSupport(i, vali) || 
	     c->findSupport(i, vali) ) ) --vali;
  if( !setMax(vali) ) return false;

  // compute lb
  vali=vmin;
  while ( !( c->firstSupport(i, vali) || 
	     c->findSupport(i, vali) ) ) ++vali;
  if( !setMin(vali) ) return false;

  return true;
}

void VariableRange::print(std::ostream& o) const
{
  o << "r" << id;
  if( isGround() )
    o << " = " << vmin;
  else 
    o << "[" << vmin << ".." << vmax << "]";
}

void VariableRange::printshort(std::ostream& o) const
{
  o << "r" << id;
}


Constant::Constant(Solver *s, const int v ) : VariableInt(s) { 
  val = v; 
  if( !static_constant_domain_it ) static_constant_domain_it = new ConstantIterator( );
}

ValIterator *Constant::getDomainIterator()
{ 
  static_constant_domain_it->curval = val;
  return static_constant_domain_it;
}


