#ifndef _ExpressionParser_h_
#define _ExpressionParser_h_

#include <iostream>
#include <iomanip>
#include <sstream>
#include <exception>
#include <stdexcept>
#include <string>
#include <map>
#include <vector>
#include <cctype>

#include <sol.h>


/**
 * @file ExpressionParser.h
 * @brief Defines an Abstract Syntax Tree and a parser for
 * prefix/infix/postfix expressions.
 */

namespace CSPXMLParser
{

using namespace std;

/**
 * TODO
 *
 * - handle operator priorities (infix only)
 * - distinguish booleans from integers in the valuation of an AST
 */


/**
 * return value of a function represented in an abstract syntax tree
 * (AST).
 *
 * one day we'll switch to a union with an int/bool/...
 */
//typedef
//int FunctionValue;

/// possible types for a variable or an expression
enum VarType {TYPE_INT,TYPE_BOOL};

/// user provided information about a variable
struct VarInfo
{
  int id; // identifier corresponding to the variable name
  VarType type; // type of the variable
};

typedef
map<string,VarInfo> VariableInfo; // list of variables (when used)

// /**
//  * an interface to get the value of a variable
//  */
// class VariableValuation
// {
// public:
//   virtual ~VariableValuation() {}

//   /**
//    * return the value of a variable in the current interpretation
//    *
//    * @param id: the identifier of the variable. Whether id is a global
//    * or local identifier is dependent on what has been recorded in the
//    * AST.
//    */
//   virtual FunctionValue getVarValue(int id) const=0;

//   // ??? could we get rid of the virtual
// };

/**
 * @brief An abstract node of an abstract syntax tree (AST).
 */
class AST
{
public:
  enum Syntax {TREE,PREFIX,INFIX_C,POSTFIX,MATHML};

  virtual ~AST() 
  {
  }

  /**
   * compute the value of this expression
   *
   * @param sol: provides the value of each variable. This
   * parameter is needed so that the code be usable from several
   * threads or contexts.
   */
  virtual int value(const int* sol)=0;
  virtual int eval(int& nodes, int& score)=0;


  void expression(ostream &s, Syntax syntax=INFIX_C)
  {
    switch(syntax)
    {
    case TREE:
      throw runtime_error("can't use this method to get an AST");
      break;
    case PREFIX:
      prefixExpression(s);
      break;
    case INFIX_C:
      infixExpression(s);
      break;
    case POSTFIX:
      postfixExpression(s);
      break;
    case MATHML:
      mathmlExpression(s);
      break;
    default:
      throw runtime_error("undefined syntax for an expressioin");
    }
  }

  virtual BuildObject* getPredicate( VarArray& scope )=0;

  virtual void prefixExpression(ostream &s)=0;
  virtual void infixExpression(ostream &s)=0;
  virtual void postfixExpression(ostream &s)=0;

  void mathmlExpression(ostream &s)
  {
    s << "<math>" << endl;
    mathmlExpressionRec(s);
    s << "</math>" << endl;
  }

  virtual void mathmlExpressionRec(ostream &s)=0;

  virtual void setArg(int pos, AST *subExpr)=0;
};

/**
 * @brief A variable represented in the AST.
 */
class ASTVar : public AST
{
private:
  string name;
  int id;
public:
  ASTVar(const string &name) {this->name=name; id=-1;}

  ASTVar(const string &name, int id) {this->name=name; this->id=id;}

  const string &getName() {return name;}

  inline int value(const int* sol) 
  {
    return sol[id];
  }

  inline int eval(int& nodes, int& score )
  {
    return 0;
  }

  virtual void setArg(int pos, AST *subExpr) 
  {
    throw runtime_error("a variable cannot have any argument");
  }

  virtual BuildObject* getPredicate( VarArray& scope )
  {
    //return NULL;
    return scope[id].var_ptr_;
  }

  virtual void prefixExpression(ostream &s)
  {
    s << name 
      //<< "/$" << id 
      ;
  }

  virtual void infixExpression(ostream &s)
  {
    s // << name << "/"
      //<< "$" << id
      << "parm[" << id << "]"
      ;
  }

  virtual void postfixExpression(ostream &s)
  {
    s << name
      //<< "/$" << id 
      ;
  }

  virtual void mathmlExpressionRec(ostream &s)
  {
    s << "  <ci>" << name << "</ci>" << endl;
  }
};

/**
 * @brief An integer constant represented in the AST.
 */
class ASTInteger : public AST
{
private:
  int val;
public:
  ASTInteger(const string &expr) 
  {
    istringstream f(expr);
    f >> val;
  }

  inline int value(const int* sol) 
  {
    return val;
  }

  inline int eval(int& nodes, int& score )
  {
    return -1;
  }

  virtual void setArg(int pos, AST *subExpr) 
  {
    throw runtime_error("a constant cannot have any argument");
  }

  virtual BuildObject* getPredicate( VarArray& scope )
  {
    return new BuildObject( val, val );
  }

  virtual void prefixExpression(ostream &s)
  {
    s << val;
  }

  virtual void infixExpression(ostream &s)
  {
    s << val;
  }

  virtual void postfixExpression(ostream &s)
  {
    s << val;
  }

  virtual void mathmlExpressionRec(ostream &s)
  {
    s << "  <cn>" << val << "</cn>" << endl;
  }
};

/**
 * @brief A boolean constant represented in the AST.
 */
class ASTBoolean : public AST
{
private:
  bool val;
public:
  ASTBoolean(bool v) 
  {
    val=v;
  }

  inline int value(const int* sol) 
  {
    if (val) 
      return 1; 
    else 
      return 0;
  }

  inline int eval(int& nodes, int& score )
  {
    return 0;
  }

  virtual void setArg(int pos, AST *subExpr) 
  {
    throw runtime_error("a constant cannot have any argument");
  }

  virtual BuildObject* getPredicate( VarArray& scope )
  {
    return new BuildObject( val, val );
  }

  virtual void prefixExpression(ostream &s)
  {
    s << boolalpha << val << noboolalpha;
  }

  virtual void infixExpression(ostream &s)
  {
    s << boolalpha << val << noboolalpha;
  }

  virtual void postfixExpression(ostream &s)
  {
    s << boolalpha << val << noboolalpha;
  }

  virtual void mathmlExpressionRec(ostream &s)
  {
    if (val)
      s << "  <true/>" << endl;
    else
      s << "  <false/>" << endl;
  }

};

/**
 * @brief A generic representation of a function inside the AST
 */
class ASTAbstractFunction : public AST
{
protected:
  int nbarg;
  AST **args;

  const char *prefixSymbol;
  const char *infixSymbol;
  const char *mathMLSymbol;

public:
  ASTAbstractFunction(int nbarg, 
		      const char *prefixSymbol, 
		      const char *infixSymbol,
		      const char *mathMLSymbol)
  {
    if (nbarg<0)
      throw runtime_error("number of arguments cannot be negative");

    this->nbarg=nbarg;
    this->prefixSymbol=prefixSymbol;
    this->infixSymbol=infixSymbol;
    this->mathMLSymbol=mathMLSymbol;

    if (nbarg==0)
      args=NULL;
    else {
      args=new AST *[nbarg];
      //std::fill(args, args+nbarg, NULL);
    }
  }

  virtual ~ASTAbstractFunction()
  {
    if (args!=NULL)
    {
      for(int i=0;i<nbarg;++i) {
	delete args[i];
      }
      delete[] args;
    }
  }

  virtual AST *makeNode()=0;

  virtual void setArg(int pos, AST *subExpr) 
  {
    if (pos<0 || pos>=nbarg)
      throw runtime_error("incorrect argument number");

    args[pos]=subExpr;
  }

  const char *getPrefixSymbol() {return prefixSymbol;}
  const char *getInfixSymbol() {return infixSymbol;}

  void prefixExpression(ostream &s)
  {
    s << prefixSymbol << "(";

    if (nbarg!=0 && args[0]!=NULL)
      args[0]->prefixExpression(s);

    for(int i=1;i<nbarg;++i)
    {
      s << ",";
      if (args[i]!=NULL)
	args[i]->prefixExpression(s);
    }

    s << ")";
  }

  void infixExpression(ostream &s)
  {
    if (infixSymbol!=NULL)
    {
      s << "("; // be careful with priorities
      
      if (args[0]!=NULL)
	args[0]->infixExpression(s);
      
      s << infixSymbol;

      if (args[1]!=NULL)
	args[1]->infixExpression(s);

      s << ")";
    }
    else
    {
      s << prefixSymbol << "(";

      if (nbarg!=0 && args[0]!=NULL)
	args[0]->infixExpression(s);

      for(int i=1;i<nbarg;++i)
      {
	s << ",";
	if (args[i]!=NULL)
	  args[i]->infixExpression(s);
      }

      s << ")";
    }
  }

  void postfixExpression(ostream &s)
  {
    if (nbarg!=0 && args[0]!=NULL)
      args[0]->postfixExpression(s);

    for(int i=1;i<nbarg;++i)
    {
      s << " ";
      if (args[i]!=NULL)
	args[i]->postfixExpression(s);
    }

    s << " " << prefixSymbol;
  }

  virtual void mathmlExpressionRec(ostream &s)
  {
    s << "  <apply>" << endl;
    s << "  <" << mathMLSymbol << "/>" << endl;
    for(int i=0;i<nbarg;++i)
    {
      if (args[i]!=NULL)
	args[i]->mathmlExpressionRec(s);
    }
    s << "  </apply>" << endl;
  }

protected:
  /**
   * check that all args are defined (not NULL)
   */
  void checkargs()
  {
    for(int i=0;i<nbarg;++i)
      if (args[i]==NULL)
	throw runtime_error("missing argument for a function");
  }

};


class FunctionNeg : public ASTAbstractFunction
{
public:
  FunctionNeg(const int n) : ASTAbstractFunction(n,"neg",NULL,"minus") {}

  virtual AST *makeNode() { return new FunctionNeg(1);}

  virtual BuildObject* getPredicate( VarArray& scope )
  {
    BuildObject **x = new BuildObject*[2];
    x[0] = args[0]->getPredicate( scope );
    return new BuildObjectPredicate( x, 1, 
				     -x[0]->max(), -x[0]->min(), 
				     ENV[ConstraintStore::NEG],
				     NULL, (BuildObject::AWV | BuildObject::CRANGE) );
  }

  inline int value(const int* sol)
  {
    return -args[0]->value(sol);
  }

  inline int eval(int& nodes, int& score)
  {
    args[0]->eval( nodes, score );
    ++nodes;
    return 1;
  }

  void infixExpression(ostream &s)
  {
    s << "-";
    s << "("; // ??? avoid these parentheses unless we need them

    if (args[0]!=NULL)
      args[0]->infixExpression(s);
      
    s << ")";
  }
};

class FunctionAbs : public ASTAbstractFunction
{
public:
  FunctionAbs(const int n) : ASTAbstractFunction(n,"abs",NULL,"abs") {}

  virtual AST *makeNode() { return new FunctionAbs(1);}

  virtual BuildObject* getPredicate( VarArray& scope )
  {
    return ENV[ConstraintStore::ABS]->declare( args[0]->getPredicate( scope ) );
  }

  inline int value(const int* sol)
  {
    return abs(args[0]->value(sol));
  }

  inline int eval(int& nodes, int& score )
  {
    args[0]->eval( nodes, score );
    ++nodes;
    ++score;
    return 1;
  }
};

class FunctionAdd : public ASTAbstractFunction
{
public:
  FunctionAdd(const int n) : ASTAbstractFunction(n,"add","+","plus") {}

  AST *makeNode() { return new FunctionAdd(2); }

  virtual BuildObject* getPredicate( VarArray& scope )
  {
    return ENV[ConstraintStore::ADD]->declare( args[0]->getPredicate( scope ),
					       args[1]->getPredicate( scope ) );
  }

  inline int value(const int* sol)
  {
    return args[0]->value(sol)+args[1]->value(sol);
  }

  inline int eval(int& nodes, int& score)
  {
    args[0]->eval( nodes, score );
    args[1]->eval( nodes, score );
    ++nodes;
    ++score;
    return 2;
  }

};

class FunctionSub : public ASTAbstractFunction
{
public:
  FunctionSub(const int n) : ASTAbstractFunction(n,"sub","-","minus") {}

  AST *makeNode() { return new FunctionSub(2);}

  virtual BuildObject* getPredicate( VarArray& scope )
  {
    return ENV[ConstraintStore::SUB]->declare( args[0]->getPredicate( scope ),
					       args[1]->getPredicate( scope ) );
  }

  inline int value(const int* sol)
  {
    return args[0]->value(sol)-args[1]->value(sol);
  }

  inline int eval(int& nodes, int& score )
  {
    args[0]->eval( nodes, score );
    args[1]->eval( nodes, score );
    ++nodes;
    ++score;
    return 2;
  }
};

class FunctionMul : public ASTAbstractFunction
{
public:
  FunctionMul(const int n) : ASTAbstractFunction(n,"mul","*","times") {}

  AST *makeNode() { return new FunctionMul(2);}

  virtual BuildObject* getPredicate( VarArray& scope )
  {
    return ENV[ConstraintStore::MUL]->declare( args[0]->getPredicate( scope ),
					       args[1]->getPredicate( scope ) );
  }

  inline int value(const int* sol)
  {
    return args[0]->value(sol)*args[1]->value(sol);
  }

  inline int eval(int& nodes, int& score )
  {
    args[0]->eval( nodes, score );
    args[1]->eval( nodes, score );
    ++nodes;
    score+=5;
    return 2;
  }
};

class FunctionDiv : public ASTAbstractFunction
{
public:
  FunctionDiv(const int n) : ASTAbstractFunction(n,"div","/","quotient") {}

  AST *makeNode() { return new FunctionDiv(2);}

  virtual BuildObject* getPredicate( VarArray& scope )
  {
    return ENV[ConstraintStore::DIV]->declare( args[0]->getPredicate( scope ),
					       args[1]->getPredicate( scope ) );
  }

  inline int value(const int* sol)
  {
    int div=args[1]->value(sol);

    if (div==0)
      throw runtime_error("divide by 0");

    return args[0]->value(sol)/div;
  }

  inline int eval(int& nodes, int& score )
  {
    args[0]->eval( nodes, score );
    args[1]->eval( nodes, score );
    ++nodes;
    score+=10;
    return 2;
  }
};

class FunctionMod : public ASTAbstractFunction
{
public:
  FunctionMod(const int n) : ASTAbstractFunction(n,"mod","%","rem") {}

  AST *makeNode() { return new FunctionMod(2);}

  virtual BuildObject* getPredicate( VarArray& scope )
  {
    BuildObject **x = new BuildObject*[3];
    x[0] = args[0]->getPredicate( scope );
    x[1] = args[1]->getPredicate( scope );
    return new BuildObjectPredicate(x, 2, 0, 
				    x[1]->max(), 
				    ENV[ConstraintStore::MOD],
				    NULL, BuildObject::AWV);
  }

  inline int value(const int* sol)
  {
    int div=args[1]->value(sol);
    return args[0]->value(sol)%div;
  }

  inline int eval(int& nodes, int& score )
  {
    args[0]->eval( nodes, score );
    args[1]->eval( nodes, score );
    ++nodes;
    score+=5;
    return 2;
  }
};

class FunctionPow : public ASTAbstractFunction
{
public:
  FunctionPow(const int n) : ASTAbstractFunction(n,"pow",NULL,"power") {}

  AST *makeNode() { return new FunctionPow(2);}

  virtual BuildObject* getPredicate( VarArray& scope )
  {
    return NULL;
  }

  inline int value(const int* sol)
  {
    //checkargs();

    long long prod=1;
    long long x=args[0]->value(sol);
    int n=args[1]->value(sol);
    int test;

    if (x==0 && n==0)
      throw runtime_error("pow(0,0) is undefined");
    
    while(n!=0)
    {
      if ((test=prod)!=prod || (test=x)!=x)
	throw runtime_error("overflow in pow(x,y) computation");

      if (n&0x01)
	prod*=x;

      x*=x;
    }

    if ((test=prod)!=prod) // ??? reliable ?
      throw runtime_error("overflow in pow(x,y) computation");

    return prod;
  }

  inline int eval(int& nodes, int& score )
  {
    args[0]->eval( nodes, score );
    args[1]->eval( nodes, score );
    ++nodes;
    score+=100;
    return 2;
  }
};

class FunctionMin : public ASTAbstractFunction
{
public:
  FunctionMin(const int n) : ASTAbstractFunction(n,"min",NULL,"min") {}

  AST *makeNode() { return new FunctionMin(2);}

  virtual BuildObject* getPredicate( VarArray& scope )
  {
    BuildObject** scp = new BuildObject*[3];
    scp[0] = args[0]->getPredicate( scope );
    scp[1] = args[1]->getPredicate( scope ); 
    return ENV[ConstraintStore::MIN]->declare( scp, 2 );
  }

  inline int value(const int* sol)
  {
    return std::min(args[0]->value(sol),args[1]->value(sol));
  }

  inline int eval(int& nodes, int& score )
  {
    args[0]->eval( nodes, score );
    args[1]->eval( nodes, score );
    ++nodes;
    return 2;
  }
};

class FunctionMax : public ASTAbstractFunction
{
public:
  FunctionMax(const int n) : ASTAbstractFunction(n,"max",NULL,"max") {}

  AST *makeNode() { return new FunctionMax(2);}

  virtual BuildObject* getPredicate( VarArray& scope )
  {
    BuildObject** scp = new BuildObject*[3];
    scp[0] = args[0]->getPredicate( scope );
    scp[1] = args[1]->getPredicate( scope ); 
    return ENV[ConstraintStore::MAX]->declare( scp, 2 );
  }

  inline int value(const int* sol)
  {
    return std::max(args[0]->value(sol),args[1]->value(sol));
  }

  inline int eval(int& nodes, int& score )
  {
    args[0]->eval( nodes, score );
    args[1]->eval( nodes, score );
    ++nodes;
    return 2;
  }
};

class FunctionEQ : public ASTAbstractFunction
{
public:
  FunctionEQ(const int n) : ASTAbstractFunction(n,"eq","==","eq") {}

  AST *makeNode() { return new FunctionEQ(2);}

  virtual BuildObject* getPredicate( VarArray& scope )
  {
    BuildObject **x = new BuildObject*[3];
    x[0] = args[0]->getPredicate( scope );
    x[1] = args[1]->getPredicate( scope );
    return new BuildObjectPredicate(x, 2, 0, 1, 
				    ENV[ConstraintStore::EQUAL],
				    NULL, BuildObject::CRANGE);
  }

  inline int value(const int* sol)
  {
    return args[0]->value(sol)==args[1]->value(sol);
  }

  inline int eval(int& nodes, int& score )
  {
    args[0]->eval( nodes, score );
    args[1]->eval( nodes, score );
    ++nodes;
    return 2;
  }
};

class FunctionNE : public ASTAbstractFunction
{
public:
  FunctionNE(const int n) : ASTAbstractFunction(n,"ne","!=","neq") {}

  AST *makeNode() { return new FunctionNE(2);}

  virtual BuildObject* getPredicate( VarArray& scope )
  {
    BuildObject **x = new BuildObject*[3];
    x[0] = args[0]->getPredicate( scope );
    x[1] = args[1]->getPredicate( scope );
    BuildObjectPredicate *p = new BuildObjectPredicate(x, 2, 0, 1, 
						       ENV[ConstraintStore::NEQ],
						       NULL, BuildObject::CLIST);
    return p;
  }

  inline int value(const int* sol)
  {
    return args[0]->value(sol)!=args[1]->value(sol);
  }

  inline int eval(int& nodes, int& score )
  {
    args[0]->eval( nodes, score );
    args[1]->eval( nodes, score );
    ++nodes;
    return 2;
  }
};

class FunctionGE : public ASTAbstractFunction
{
public:
  FunctionGE(const int n) : ASTAbstractFunction(n,"ge",">=","geq") {}

  AST *makeNode() { return new FunctionGE(2);}

  virtual BuildObject* getPredicate( VarArray& scope )
  {
    return ENV[ConstraintStore::PRECEDENCE]->declare( args[1]->getPredicate( scope ), 0, 
						      args[0]->getPredicate( scope ) );
  }

  inline int value(const int* sol)
  {
    return args[0]->value(sol)>=args[1]->value(sol);
  }

  inline int eval(int& nodes, int& score )
  {
    args[0]->eval( nodes, score );
    args[1]->eval( nodes, score );
    ++nodes;
    return 2;
  }
};

class FunctionGT : public ASTAbstractFunction
{
public:
  FunctionGT(const int n) : ASTAbstractFunction(n,"gt",">","gt") {}

  AST *makeNode() { return new FunctionGT(2);}

  virtual BuildObject* getPredicate( VarArray& scope )
  {
    return ENV[ConstraintStore::PRECEDENCE]->declare( args[1]->getPredicate( scope ), 1, 
						      args[0]->getPredicate( scope ) );
  }

  inline int value(const int* sol)
  {
    if (args[0]->value(sol)>args[1]->value(sol))
      return int(1);
    else
      return int(0);
  }

  inline int eval(int& nodes, int& score )
  {
    args[0]->eval( nodes, score );
    args[1]->eval( nodes, score );
    ++nodes;
    return 2;
  }
};

class FunctionLE : public ASTAbstractFunction
{
public:
  FunctionLE(const int n) : ASTAbstractFunction(n,"le","<=","leq") {}

  AST *makeNode() { return new FunctionLE(2);}

  virtual BuildObject* getPredicate( VarArray& scope )
  {
    return ENV[ConstraintStore::PRECEDENCE]->declare( args[0]->getPredicate( scope ), 0, 
						      args[1]->getPredicate( scope ) );
  }

  inline int value(const int* sol)
  {
    return args[0]->value(sol)<=args[1]->value(sol);
  }

  inline int eval(int& nodes, int& score )
  {
    args[0]->eval( nodes, score );
    args[1]->eval( nodes, score );
    ++nodes;
    return 2;
  }
};

class FunctionLT : public ASTAbstractFunction
{
public:
  FunctionLT(const int n) : ASTAbstractFunction(n,"lt","<","lt") {}

  AST *makeNode() { return new FunctionLT(2);}

  virtual BuildObject* getPredicate( VarArray& scope )
  {
    return ENV[ConstraintStore::PRECEDENCE]->declare( args[0]->getPredicate( scope ), 1, 
						      args[1]->getPredicate( scope ) );
  }

  inline int value(const int* sol)
  {
    if (args[0]->value(sol)<args[1]->value(sol))
      return int(1);
    else
      return int(0);
  }

  inline int eval(int& nodes, int& score )
  {
    args[0]->eval( nodes, score );
    args[1]->eval( nodes, score );
    ++nodes;
    return 2;
  }
};

class FunctionNot : public ASTAbstractFunction
{
public:
  FunctionNot(const int n) : ASTAbstractFunction(n,"not","!","not") {}

  AST *makeNode() { return new FunctionNot(1);}

  virtual BuildObject* getPredicate( VarArray& scope )
  {
    BuildObject **x = new BuildObject*[2];
    x[0] = args[0]->getPredicate( scope );
    return new BuildObjectPredicate( x, 1, 0, 1, 
				     ENV[ConstraintStore::NOT],
				     NULL, 0 );
  }

  inline int value(const int* sol)
  {
    return !args[0]->value(sol);
  }

  inline int eval(int& nodes, int& score )
  {
    args[0]->eval( nodes, score );
    ++nodes;
    ++score;
    return 1;
  }
};

class FunctionAnd : public ASTAbstractFunction
{
public:
  FunctionAnd(const int n) : ASTAbstractFunction(n,"and","&&","and") {}

  AST *makeNode() { return new FunctionAnd(2);}

  virtual BuildObject* getPredicate( VarArray& scope )
  {
    BuildObject **x = new BuildObject*[3];
    x[0] = args[0]->getPredicate( scope );
    x[1] = args[1]->getPredicate( scope );
    return new BuildObjectPredicate(x, 2, 0, 1, 
				    ENV[ConstraintStore::AND],
				    NULL, 0);
  }

  inline int value(const int* sol)
  {
    return args[0]->value(sol)&&args[1]->value(sol);
  }

  inline int eval(int& nodes, int& score )
  {
    args[0]->eval( nodes, score );
    args[1]->eval( nodes, score );
    ++nodes;
    //score+=3;
    return 2;
  }
};

class FunctionOr : public ASTAbstractFunction
{
public:
  FunctionOr(const int n) : ASTAbstractFunction(n,"or","||","or") {}

  AST *makeNode() { return new FunctionOr(2);}

  virtual BuildObject* getPredicate( VarArray& scope )
  {
    BuildObject **x = new BuildObject*[3];
    x[0] = args[0]->getPredicate( scope );
    x[1] = args[1]->getPredicate( scope );
    return new BuildObjectPredicate(x, 2, 0, 1, 
				    ENV[ConstraintStore::OR],
				    NULL, 0);
  }

  inline int value(const int* sol)
  {
    return args[0]->value(sol)||args[1]->value(sol);
  }

  inline int eval(int& nodes, int& score )
  {
    args[0]->eval( nodes, score );
    args[1]->eval( nodes, score );
    ++nodes;
    //score+=7;
    score+=5;
    ////score+=2;
    return 2;
  }
};

class FunctionXor : public ASTAbstractFunction
{
public:
  FunctionXor(const int n) : ASTAbstractFunction(n,"xor","^","xor") {}

  AST *makeNode() { return new FunctionXor(2);}

  virtual BuildObject* getPredicate( VarArray& scope )
  {
    BuildObject **x = new BuildObject*[3];
    x[0] = args[0]->getPredicate( scope );
    x[1] = args[1]->getPredicate( scope );
    BuildObjectPredicate *p = new BuildObjectPredicate(x, 2, 0, 1, 
						       ENV[ConstraintStore::NEQ],
						       NULL, 0);
    return p;
  }

  inline int value(const int* sol)
  {
    return args[0]->value(sol)^args[1]->value(sol);
  }

  inline int eval(int& nodes, int& score )
  {
    args[0]->eval( nodes, score );
    args[1]->eval( nodes, score );
    ++nodes;
    score+=3;
    return 2;
  }
};




class ExpressionParser
{
private:
  static const int NBFUNCTION=20;
  //vector<ASTAbstractFunction *> functionList;
  ASTAbstractFunction **functionList;
  
  const VariableInfo *varInfo; // an optional map which provides some
			       // information on the variables that
			       // may occur in the expression

public:
  ExpressionParser()
  {
    varInfo=NULL;


    functionList = new ASTAbstractFunction*[NBFUNCTION];

    functionList[0] = (new FunctionNeg(0));
    functionList[1] = (new FunctionAbs(0));
    functionList[2] = (new FunctionAdd(0));
    functionList[3] = (new FunctionSub(0));
    functionList[4] = (new FunctionMul(0));
    functionList[5] = (new FunctionEQ(0));
    functionList[6] = (new FunctionNE(0));
    functionList[7] = (new FunctionGE(0));
    functionList[8] = (new FunctionGT(0));
    functionList[9] = (new FunctionLE(0));
    functionList[10] = (new FunctionLT(0));
    functionList[11] = (new FunctionNot(0));
    functionList[12] = (new FunctionAnd(0));
    functionList[13] = (new FunctionDiv(0));
    functionList[14] = (new FunctionMod(0));
    functionList[15] = (new FunctionPow(0));
    functionList[16] = (new FunctionMin(0));
    functionList[17] = (new FunctionMax(0));
    functionList[18] = (new FunctionOr(0));
    functionList[19] = (new FunctionXor(0));
  }

  ~ExpressionParser() 
  {
    for(unsigned int i=0;i<NBFUNCTION;++i) {
      delete functionList[i];
    }
    delete [] functionList;
  }


  void setVarInfo(const VariableInfo *varInfo=NULL)
  {
    this->varInfo=varInfo;
  }

  void unsetVarInfo()
  {
    this->varInfo=NULL;
  }

  AST *prefixParser(const string &expr)
  {
    string s;

    // remove spaces
    for(unsigned int i=0;i<expr.length();++i)
      if (!isspace(expr[i]))
	s+=expr[i];

    return recursivePrefixParser(s);
  }

  AST *infixParser(const string &expr)
  {
    return new ASTVar("infix parser unimplemented");
  }

  AST *postfixParser(const string &expr)
  {
    return new ASTVar("postfix parser unimplemented");
  }

private:
  AST *recursivePrefixParser(const string &f)
  {
    int level=0;
    int argNum=0;
    int subExprStart=0;

    AST *node=NULL;

    for(unsigned int i=0;i<f.length();++i)
    {
      if (f[i]=='(')
      {
	if (level==0)
	{
	  node=findPrefixFunction(f.substr(0,i));
	  subExprStart=i+1;
	}
	++level;
      }
      else
      {
	if (level==1 && (f[i]==',' || f[i]==')'))
	{
	  node->setArg(argNum,prefixParser(f.substr(subExprStart,
						    i-subExprStart)));
	  ++argNum;

	  subExprStart=i+1;
	}

        if (f[i]==')')
	  --level;
      }
    }

    if (level!=0)
      throw runtime_error("unbalanced parentheses");

    if (node==NULL)
    {
      // no opening parenthese found, this is a constant or a variable
      if (isalpha(f[0]))
      {
	if (f=="true")
	  node=new ASTBoolean(true);
	else
	  if (f=="false")
	    node=new ASTBoolean(false);
	  else
	  {
	    // a variable
	    if (varInfo==NULL)
	      node=new ASTVar(f);
	    else
	    {
	      VariableInfo::const_iterator it=varInfo->find(f);

	      if(it==varInfo->end())
		throw runtime_error("undefined variable found in expression");

	      node=new ASTVar(f,(*it).second.id);
	    }
	  }
      }
      else
      {
	// an int
	node=new ASTInteger(f);
      }
    }

    return node;
  }

  AST *findPrefixFunction(const string &name)
  {
    for(unsigned int i=0;i<NBFUNCTION;++i)
      if (functionList[i]->getPrefixSymbol()==name)
	return functionList[i]->makeNode();

    throw runtime_error("unknown function symbol");
  }
};

} // namespace


// Local Variables:
// mode: C++
// End:


#endif
