
/*
  Mistral is a constraint satisfaction and optimisation library
  Copyright (C) 2003-2005  Emmanuel Hebrard
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  The author can be contacted electronically at 
  ehebrard@cse.unsw.edu.au.
*/

#include <sol.h>
#include <con.h>
#include <csp.h>


/**********************************************
 * Relations described in extension
 **********************************************/

/**********************************************
 * Unary Constraint
 **********************************************/

UnaryConstraint::UnaryConstraint(Solver *s, VariableInt* x)
  : X(x)
{
  solver = s;
}

UnaryConstraint::~UnaryConstraint()
{
}

void UnaryConstraint::activate()
{
  solver->unaryCons.push( this );
}


/**********************************************
 * Unary More Constraint
 **********************************************/

UnaryConstraintMore::UnaryConstraintMore(Solver *s, VariableInt* x, int lb)
  : UnaryConstraint( s, x )
{
  bound = lb;
}

bool UnaryConstraintMore::propagate()
{
  return X->setMin( bound );
}


/**********************************************
 * Unary Less Constraint
 **********************************************/

UnaryConstraintLess::UnaryConstraintLess(Solver *s, VariableInt* x, int ub)
  : UnaryConstraint( s, x )
{
  bound = ub;
}

bool UnaryConstraintLess::propagate()
{
  return X->setMax( bound );
}


/**********************************************
 * Large Extensionalbak Constraint
 **********************************************/

ConstraintLargeExtensionalbak::ConstraintLargeExtensionalbak(Solver *s, VariableInt **v,
						       const int n, MistralStack<int*>** m)
  : Constraint(s, v, n, DOMAINTRIGGER) 
{
 
  supportList = m;
  isInDomain = new MistralSet*[arity];
  vars = new VariableInt**[arity];
  residues = new int*[arity];
  themins = new int[arity];
  int i, j, k, nval;
  for(i=0; i<arity; ++i) {

    //cout << i << endl;
    
    themins[i] = scope[i]->min();
    nval=(scope[i]->max() - themins[i] + 1);
    residues[i] = new int[nval];
    std::fill(residues[i], residues[i]+nval, INT_MAX);    
    residues[i] -= themins[i];

    //cout << 11 << endl;
    isInDomain[i] = new MistralSet[arity-1];

    //cout << 22 << endl;

    vars[i] = new VariableInt*[arity-1];

    //cout << 33 << endl;
    k=0;
    for(j=0; j<arity; ++j) {
      if( i != j) {
	
	//cout << "\t" << i << " " << j <<  endl;

	switch( scope[j]->getType() ) {
	case VariableInt::CONST : { isInDomain[i][k].init( scope[j]->min(), scope[j]->max(), MistralSet::full); } break;
	case VariableInt::RANGE : { cerr << "not handled" << endl; exit(0); } break;
	case VariableInt::BOOL : { isInDomain[i][k].pointTo( (unsigned int*)(&(((VariableBool*)scope[j])->domain.state)) ); } break;
	default : { isInDomain[i][k].pointTo( ((VariableDomain*)scope[j])->values ); }
	}
	//	isInDomain[i][k].pointTo( ((VariableDomain*)scope[j])->values );
	
	//isInDomain[i][k].print( cout );
	//cout << endl;
	
	vars[i][k] = scope[j];
	++k;
      }
    }

    //cout << endl;

    --isInDomain[i];
    --vars[i];
  }

//   isInDomain = new MistralSet[arity];
//   for(i=0; i<arity; ++i) {
//     switch( scope[i]->getType() ) {
//     case VariableInt::CONST : { cerr << "not handled" << endl; exit(0); } break;
//     case VariableInt::RANGE : { cerr << "not handled" << endl; exit(0); } break;
//     case VariableInt::BOOL : { cerr << "not handled" << endl; exit(0); } break;
//     default : { isInDomain[i].pointTo( ((VariableDomain*)scope[i])->values );
// 	cout << "domain" << endl;
//     }
//     }
//   }

}

ConstraintLargeExtensionalbak::~ConstraintLargeExtensionalbak() 
{ 
  int i, k;
  for(i=0; i<arity; ++i) {
    residues[i] += themins[i];
    delete [] residues[i];

    k = arity;
    while( --k ) {
      isInDomain[i][k].table = NULL;
      isInDomain[i][k].neg_words = 0;
    }
    ++isInDomain[i];
    delete [] isInDomain[i];
    ++vars[i];
    delete [] vars[i];
  }
  delete [] isInDomain;
  delete [] vars;
  delete [] residues;
  delete [] themins;
}

bool ConstraintLargeExtensionalbak::propagate(VariableInt* pvi, const int e) 
{
  int consistent = true, i, j, k, valid;

//   for( i=0; i<arity; ++i )
//     {
//       scope[i]->print( cout );
//       cout << " / ";
//       isInDomain[i].print( cout );
//       cout << endl;
//     }
//   cout << endl;


  ValIterator *valit;
  for( i=0; consistent && i<arity; ++i )
    if( scope[i] != pvi && scope[i]->isLinked() ) 
      {
	valit = scope[i]->getDomainIterator();
	do {
	  j = *valit;
	  initSupport( i, j );
 	  valid = (residues[i][j] < INT_MAX);
 	  k = arity;
	  while( valid && --k ) 
	    //valid = X[k]->contain(supports[residues[i][j]][k]);
	    //valid = X[k]->fastContain(supports[residues[i][j]][k]);
	    //valid = ((VariableDomain*)X[k])->values.fastMember( supports[residues[i][j]][k] );
	    valid = D_X[k].fastMember( supports[residues[i][j]][k] );

	  while( !valid && index-- )
	    {
	      valid = true;
	      k = arity;
	      while( valid && --k )
		//valid = (X[k]->contain(supports[index][k]));
		//valid = (X[k]->fastContain(supports[index][k]));
		//valid = ((VariableDomain*)X[k])->values.fastMember( supports[index][k] );
		valid = D_X[k].fastMember( supports[index][k] );

	      if( valid ) {
		residues[i][j] = index;
	      }
	    }
	  if( !valid ) consistent = scope[i]->remove( j ); 
	} while( consistent && valit->next() );
      }
  return consistent;
}


int ConstraintLargeExtensionalbak::check( const int* s ) const 
{
  int i, equal=0;
//   initSupport( 0, s[0] );
//   while( !equal && index-- )
//     {
//       i=arity;
//       equal=1;
//       while( equal && --i )
// 	equal = (s[i] == support[index][i]);
//     }
//  return !equal;
  return 0;
}

void ConstraintLargeExtensionalbak::print(std::ostream& o) const
{
  o << id << "LargeExt";
  Constraint::print( o );
}

void ConstraintLargeExtensionalbak::printlong(std::ostream& o) 
{
  o << id << "LargeExt";
  Constraint::print( o );
//   o << endl;

//   int i, j, k;
//   ValIterator *valit;
//   for( i=0; i<arity; ++i )
//     {
//       scope[i]->print( o );
//       o << endl;

//       valit = scope[i]->getDomainIterator();
//       do {
// 	j = *valit;
// 	initSupport( i, j );

// 	o << "\t : ";
// 	k = 1;
// 	while( k < arity ) {
// 	  o << k ;
// 	  o.flush();

// 	  X[k]->printshort( o );
// 	  o << " ";
// 	  o.flush();
// 	  ++k;
// 	}
// 	o << endl;	

// 	while( index-- )
// 	  {
// 	    o << "\t";
// 	    k = arity;
// 	    while( --k )
// 	      o << supports[index][k] << " ";
// 	    o << endl;
// 	  }
//       } while( valit->next() );
//     }
}

/**********************************************
 * Large Extensional Constraint
 **********************************************/

ConstraintLargeExtensional::ConstraintLargeExtensional(Solver *s, VariableInt **v,
						       const int n, MistralStack<int*>** m)
  : Constraint(s, v, n, DOMAINTRIGGER) 
{ 
  supportList = m;
  D_X = new MistralSet[arity];
  residues = new int**[arity];
  themins = new int[arity];
  order = new int[arity];
  int i, j, k, nval;
  for(i=0; i<arity; ++i) {

    order[i] = i;
    //order[i] = arity-i-1;
    //cout << order[i] << " ";
    //cout.flush();

    themins[i] = scope[i]->min();
    nval=(scope[i]->max() - themins[i] + 1);
    residues[i] = new int*[nval];
    //std::fill(residues[i], residues[i]+nval, NULL);
    for(j=0; j<nval; ++j)
      residues[i][j] = NULL;
    residues[i] -= themins[i];
    switch( scope[i]->getType() ) {
    case VariableInt::CONST : { 
      D_X[i].init( scope[i]->min(), scope[i]->max(), MistralSet::full); 
    } break;
    case VariableInt::RANGE : { 
      cerr << "not handled" << endl; exit(0); 
    } break;
    case VariableInt::BOOL : { 
      D_X[i].pointTo( (unsigned int*)(&(((VariableBool*)scope[i])->domain.state)) ); 
    } break;
    default : { 
      D_X[i].pointTo( ((VariableDomain*)scope[i])->values ); 
    }
    }
  }
  //cout << endl;
}

ConstraintLargeExtensional::~ConstraintLargeExtensional() 
{ 
  int i, k;
  for(i=0; i<arity; ++i) {
    residues[i] += themins[i];
    delete [] residues[i];
    D_X[i].table = NULL;
    D_X[i].neg_words = 0;
  }
  delete [] D_X;
  delete [] residues;
  delete [] themins;
  delete [] order;
}

bool ConstraintLargeExtensional::propagate(VariableInt* pvi, const int e) 
{
  int consistent = true, i, oi, j, k, ok, valid, index, n;

  // order variables by increasing domain size
  for(i=1; i<arity; ++i)
    {
      j=i;
      while( j && scope[order[j]]->domsize() < scope[order[j-1]]->domsize() )
	{
	  ok = order[j];
	  order[j] = order[j-1];
	  order[--j] = ok;
	}
    }

  ValIterator *valit;
  for( i=0; consistent && i<arity; ++i ) {
    oi = order[i];
    if( scope[oi] != pvi && scope[oi]->isLinked() ) 
      {

// 	cout << "Iterate on ";
// 	scope[oi]->print( cout );
// 	cout << endl;

	valit = scope[oi]->getDomainIterator();
	do {
	  j = *valit;

	  //cout << "\t" << j << endl;

	  supports_X = supportList[oi][j].stack_;

	  
	  //cout << oi << " " << j << " " << supportList[oi][j].size() << endl;

	  valid = (residues[oi][j] != NULL);
	  for(k=0; valid && k<arity; ++k) {
	    ok = order[k];
	    valid = ( oi == ok || D_X[ok].fastMember( residues[oi][j][ok] ) );
	  }
	  n = supportList[oi][j].size_;
	  for(index=0; !valid && index<n; ++index)
	    {
	      valid = true;
	      for(k=0; valid && k<arity; ++k) {
		ok = order[k];
		valid = ( oi == ok || D_X[ok].fastMember( supports_X[index][ok] ) );
	      }
	      if( valid ) {
 		k = arity;
 		while( k-- ) 
		  residues[k][supports_X[index][k]] = supports_X[index];
	      }
	    }
	  if( !valid ) consistent = scope[oi]->remove( j ); 
	} while( consistent && valit->next() );
      }
  }
  return consistent;
}


int ConstraintLargeExtensional::check( const int* s ) const 
{
  int i, equal=0;
//   initSupport( 0, s[0] );
//   while( !equal && index-- )
//     {
//       i=arity;
//       equal=1;
//       while( equal && --i )
// 	equal = (s[i] == support[index][i]);
//     }
//  return !equal;
  return 0;
}

void ConstraintLargeExtensional::print(std::ostream& o) const
{
  o << id << "LargeExt";
  Constraint::print( o );
}

void ConstraintLargeExtensional::printlong(std::ostream& o) 
{
  o << id << "LargeExt";
  Constraint::print( o );
//   o << endl;

//   int i, j, k;
//   ValIterator *valit;
//   for( i=0; i<arity; ++i )
//     {
//       scope[i]->print( o );
//       o << endl;

//       valit = scope[i]->getDomainIterator();
//       do {
// 	j = *valit;
// 	initSupport( i, j );

// 	o << "\t : ";
// 	k = 1;
// 	while( k < arity ) {
// 	  o << k ;
// 	  o.flush();

// 	  X[k]->printshort( o );
// 	  o << " ";
// 	  o.flush();
// 	  ++k;
// 	}
// 	o << endl;	

// 	while( index-- )
// 	  {
// 	    o << "\t";
// 	    k = arity;
// 	    while( --k )
// 	      o << supports[index][k] << " ";
// 	    o << endl;
// 	  }
//       } while( valit->next() );
//     }
}


/**********************************************
 * Extensional Constraint
 **********************************************/

int ConstraintExtensional::getpos(const int *sol) const 
{
  int pos = sol[arity-1], i = arity-1;
  while( i-- ) pos += (sol[i])*var_sizes[i]; 
  return pos;
}

ConstraintExtensional::ConstraintExtensional(Solver *s, VariableInt **v,
					     const int n, MistralSet& m,
					     int *vs)
  : Constraint(s, v, n, DOMAINTRIGGER) 
{
  matrix.pos_words = m.pos_words;
  matrix.neg_words = m.neg_words;
  matrix.table = m.table;
  var_sizes = vs;
}

ConstraintExtensional::~ConstraintExtensional() 
{ 
  matrix.table = NULL;
  matrix.neg_words = 0;
}

int ConstraintExtensional::check( const int* s ) const 
{
  return matrix.member(getpos(s));
}

void ConstraintExtensional::fillNogood() 
{
  matrix.fill();
}

void ConstraintExtensional::fillSupport() 
{
  matrix.clear();
}

void ConstraintExtensional::addNogood(const int* sol) 
{
  matrix.insert(getpos(sol));
}

void ConstraintExtensional::addSupport(const int* sol) 
{
  matrix.erase(getpos(sol));
}

void ConstraintExtensional::print(std::ostream& o) const
{
  o << "Ext";
  Constraint::print( o );
//   o << endl;

//   int s[3];

//   for(int i=0; i<4; ++i)
//     for(int j=0; j<4; ++j)
//       for(int k=0; k<4; ++k)
// 	{
// 	  s[0] = i;
// 	  s[1] = j;
// 	  s[2] = k;
// 	  if( !check(s) )
// 	    cout << i << " " << j << " " << k << endl; 
// 	}
  
  //  exit( 0 );

//   for(int vri=0; vri<arity; ++vri) {
//     int i=arity, vali;
//     while( i-- )
//       supp[i] = 0;
//     i=arity;
//     // sol is initialized: either to the value 
//     // a variable is already assigned to
//     // or to the first value in its domain
//     while(i >= 0) {
//       // check this assignment
//       if( !check( supp ) ) {
// 	for(int j=0; j<arity; ++j)
// 	  cout << supp[j] << " ";
// 	cout << endl;
//       } 
//       // try to assign more things
//       // find the last var whose domain we have not exhausted
//       --i;
//       while( i >= 0 ) {
// 	if( i == vri || scope[i]->isGround() ) {
// 	--i;
// 	continue;
// 	}
// 	if( scope[i]->setNext( supp[i] ) )
// 	  break;
// 	else
// 	  supp[i] = scope[i]->first();
// 	--i;
//       }
//       if( i >= 0 )
// 	i = arity;
//     }
//   }
  
}


/**********************************************
 * Extensional Binary Constraint
 **********************************************/

void ConstraintBinaryExtensional::init( )
{
  int val, i=2, j;
  MistralSet aux[2];
  while( i-- ) 
    aux[i].init( scope[1-i]->minCapacity(), 
		 scope[1-i]->maxCapacity(),
		 MistralSet::empt );

  i=2;
  while( i-- ) {
    size[i] = scope[i]->domsize();

    residualSupport[i] = new int[scope[i]->maxCapacity() - 
				 scope[i]->minCapacity()];
    residualSupport[i] -= scope[i]->minCapacity();

    minSup[i] = INT_MAX;
    val = scope[i]->first();
    do {
      residualSupport[i][val] = supportList[i][val].pos_words-1;	

      aux[i].clear();
      scope[1-i]->unionTo( aux[i] );
      aux[i].intersectWith( supportList[i][val] );

      j = aux[i].size();

      if( minSup[i] > j ) minSup[i] = j;
    } while( scope[i]->setNext( val ) );
  }
}

ConstraintBinaryExtensional::ConstraintBinaryExtensional(Solver *s, VariableInt **v, MistralSet **sl)
    : Constraint(s,v,2,DOMAINTRIGGER)
{
  supportList[0] = sl[0];
  supportList[1] = sl[1];
}

ConstraintBinaryExtensional::~ConstraintBinaryExtensional() 
{ 
  int i=2;
  while( i-- ) {
    residualSupport[i]+= scope[i]->minCapacity();
    delete [] residualSupport[i];
  }
}

int ConstraintBinaryExtensional::check( const int* s ) const 
{
  return !(supportList[0][s[0]].member( s[1] ));
}

void ConstraintBinaryExtensional::fillNogood() 
{
  int val, i=2;
  while( i-- ) {
    val = scope[i]->first();
    do supportList[i][val].clear();
    while( scope[i]->setNext( val ) );
  }  
}

void ConstraintBinaryExtensional::fillSupport() 
{
  int val, i=2;
  while( i-- ) {
    val = scope[i]->first();
    do {
      supportList[i][val].fill();
    } while( scope[i]->setNext( val ) );
  }
}

void ConstraintBinaryExtensional::addNogood(const int* sol) 
{
  int i=2;
  while( i-- ) {
    supportList[i][sol[i]].erase( sol[1-i] );
  }
}

void ConstraintBinaryExtensional::addSupport(const int* sol) 
{
  int i=2;
  while( i-- ) {
    supportList[i][sol[i]].insert( sol[1-i] );
  }
}

bool ConstraintBinaryExtensional::propagate(VariableInt* pvi, const int e) 
{
  int i = (pvi == scope[0]);
  bool consistent = true;
  if( (size[1-i] - pvi->domsize()) >= minSup[i] )
    consistent = scope[i]->revise( supportList[i], 
				   residualSupport[i], 
				   pvi );
  return consistent;
}


// bool ConstraintBinaryExtensional::propagate(VariableInt* pvi, const int e) 
// {
//   int i = (pvi == scope[0]), j;
//   bool consistent = true;
//   if( (size[1-i] - pvi->domsize()) >= minSup[i] ) {
//     ValIterator *valit = scope[i]->getDomainIterator();
//     do {
//       j = *valit;
//       consistent = (pvi->intersect( supportList[i][j], 
// 				    residualSupport[i][j] ) || scope[i]->remove( j ));
//     } while( consistent && valit->next() );
//   }
//   return consistent;
// }


void ConstraintBinaryExtensional::print(std::ostream& o) const
{
  o << "BinExt";
  Constraint::print( o );
  int i=2, val;
  while( i-- ) {
    o << " |";
    val = scope[i]->first();
    do {
      o << " " << supportList[i][val].size();
    } while( scope[i]->setNext( val ) );
  }
}


/**********************************************
 * Clause Constraint
 **********************************************/

ConstraintClause::ConstraintClause(Solver *s, VariableInt** v, const int n, const int* c) :
  Constraint(s, v, n, VALUETRIGGER) 
{
  conflict = new int[n];
  int i = n;

  while( i-- ) {
    conflict[i] = c[i];
  }
}

ConstraintClause::~ConstraintClause() 
{ 
  delete [] conflict;
}

int ConstraintClause::check( const int* s ) const 
{
  int i = arity;
  while( i-- )
    if( s[i] != conflict[i] ) return 0;
  return 1;
}

bool ConstraintClause::propagate()
{
  bool consistent = true;
//   for(int i=0; i<arity; ++i)
//     if( scope[i]->isGround() )
//       consistent = propagate( scope[i], VALUETRIGGER );
  return consistent;
} 

bool ConstraintClause::propagate(VariableInt* pvi, const int e) 
{
  int i=arity, j=-1;
  while( i-- ) {
    if( scope[i]->isLinked() ) {
      if( j < 0 ) j = i;
      else {
	return true;
      }
    } else if( scope[i]->value() != conflict[i] ) {
      return true;
    }
  }

  return scope[j]->remove( conflict[j] );
}

void ConstraintClause::print(std::ostream& o) const
{

  bool sat = false;
  int i=arity;

  while( !sat && i-- ) {
    if( scope[i]->isGround() && scope[i]->value() != conflict[i] )
      sat = true;
  }
  if( sat ) o << "T";
  else {
    o << "{ ";
    for(i=0; i<arity; ++i) {
      if( scope[i]->isGround() ) {
	if( scope[i]->value() !=  conflict[i] )
	  o << "T ";
      } else {
	scope[i]->printshort( o );
	o << "~" << conflict[i] << " ";
      }
    }
    o << "}";
  }
}


/**********************************************
 * =/= Constraint
 **********************************************/

ConstraintNotEqual::ConstraintNotEqual(Solver *s, VariableInt** v)
  : Constraint(s, v, 2, VALUETRIGGER)
{
}

ConstraintNotEqual::~ConstraintNotEqual() {}

int ConstraintNotEqual::check( const int* s ) const 
{
  return ( s[0] == s[1] );
}

bool ConstraintNotEqual::propagate()
{
  return ((!scope[1]->isGround() || scope[0]->remove(scope[1]->min())) &&
	  (!scope[0]->isGround() || scope[1]->remove(scope[0]->min())));
}

bool ConstraintNotEqual::propagate(VariableInt *v, const int e)
{
  return scope[(scope[0] == v)]->remove(v->min());  
}

void ConstraintNotEqual::print(std::ostream& o) const
{
  scope[0]->printshort( o ) ;
  o << " =/= " ;
  scope[1]->printshort( o ) ;
}


/**********************************************
 * =/= Constraint with wildcard
 **********************************************/

ConstraintNotEqualWildcard::ConstraintNotEqualWildcard(Solver *s, VariableInt** v, const int w)
  : Constraint(s, v, 2, VALUETRIGGER)
{
  wildcard = w;
}

ConstraintNotEqualWildcard::~ConstraintNotEqualWildcard() {}

int ConstraintNotEqualWildcard::check( const int* s ) const 
{
  return ( s[0] == s[1] && s[0] != wildcard );
}

bool ConstraintNotEqualWildcard::propagate()
{
  return ((!scope[0]->isGround() || propagate(scope[0], VALUETRIGGER)) &&
	  (!scope[1]->isGround() || propagate(scope[1], VALUETRIGGER)));
}

bool ConstraintNotEqualWildcard::propagate(VariableInt *v, const int e)
{
  bool consistent = true;
  int idx = (scope[0] == v), w = v->min();
  if( w != wildcard )
    consistent = scope[idx]->remove(w);  
  else if( idx )
    scope[idx]->setDomain(wildcard);

  return consistent;
}

void ConstraintNotEqualWildcard::print(std::ostream& o) const
{
  scope[0]->print( o ) ;
  o << " =/= " ;
  scope[1]->print( o ) ;
  o << " or " ;
  scope[0]->print( o ) ;
  o << " = " ;
  o << wildcard ;
}


/**********************************************
 * Or Constraint
 **********************************************/

ConstraintOr::ConstraintOr(Solver *s, VariableInt** v)
  : Constraint(s, v, 2, DOMAINTRIGGER)
{
}

ConstraintOr::~ConstraintOr() {}

int ConstraintOr::check( const int* s ) const 
{
  return ( !s[0] && !s[1] );
}

bool ConstraintOr::propagate(VariableInt *v, const int e)
{
  return( v->max() || scope[(scope[0] == v)]->setDomain(1) );  
}

void ConstraintOr::print(std::ostream& o) const
{
  scope[0]->printshort( o ) ;
  o << " || " ;
  scope[1]->printshort( o ) ;
}


// /**********************************************
//  * IfThenElse Constraint
//  **********************************************/

// ConstraintIfThenElse::ConstraintIfThenElse(Solver *s, VariableInt** v)
//   : Constraint(s, v, 3, DOMAINTRIGGER)
// {
// }

// ConstraintIfThenElse::~ConstraintIfThenElse() {}

// int ConstraintIfThenElse::check( const int* s ) const 
// {
//   return ( !(s[0] ? s[1] : s[2]) );
// }

// bool ConstraintIfThenElse::propagate(VariableInt *v, const int e)
// {
// //   bool consistent = true;
// //   if( v == scope[0] ) 
// //     {
// //       if( v->min() )
// // 	consistent = scope[1]->remove( 0 );
// //       else 
// // 	consistent = scope[2]->remove( 0 );
// //     }
// //   else if( v == scope[1] )
// //     consistent = (v->max() || (scope[0]->remove( 1 ) && scope[2]->remove( 0 )));
// //   else 
// //     consistent = (v->max() || (scope[0]->remove( 0 ) && scope[1]->remove( 0 )));
// //   return consistent;

//   bool consistent = true;
//   if( v == scope[0] ) 
//     {
//       if( v->min() )
// 	consistent = scope[1]->remove( 0 );
//       else 
// 	consistent = scope[2]->remove( 0 );
//     }
//   else if( v == scope[1] )
//     consistent = (v->max() || (scope[0]->remove( 1 ) && scope[2]->remove( 0 )));
//   else 
//     consistent = (v->max() || (scope[0]->remove( 0 ) && scope[1]->remove( 0 )));
//   return consistent;

// }

// void ConstraintIfThenElse::print(std::ostream& o) const
// {
//   scope[0]->printshort( o ) ;
//   o << " ? " ;
//   scope[1]->printshort( o ) ;
//   o << " : " ;
//   scope[2]->printshort( o ) ;
// }


/**********************************************
 * inverse Constraint
 **********************************************/
ConstraintInverse::ConstraintInverse(Solver *s, VariableInt** v, int i, int j)
  : Constraint(s, v, 2, DOMAINTRIGGER)
{
  i_[0] = i;
  i_[1] = j;
}

ConstraintInverse::~ConstraintInverse() {}

int ConstraintInverse::check( const int* s ) const 
{
  return ( (s[0] == i_[0]) != (s[1] == i_[1]) );
}

bool ConstraintInverse::propagate(VariableInt *v, const int e)
{
  int idx = (scope[0] == v);
  if( !scope[1-idx]->contain( i_[1-idx] ) ) {
    if( !scope[idx]->remove( i_[idx] ) ) return false;
  } else {
    if( scope[1-idx]->isGround() && 
	!scope[idx]->setDomain( i_[idx] ) ) return false;
  }
  return true;  
}

void ConstraintInverse::print(std::ostream& o) const
{
  o << "(" ;
  scope[0]->printshort( o ) ;
  o << " = " << i_[0] << ") <=> (";
  scope[1]->printshort( o ) ;
  o << " = " << i_[1] << ")" ;
}


/**********************************************
 * Lex Constraint
 **********************************************/

ConstraintLex::ConstraintLex(Solver *s, VariableInt** v)
  : Constraint(s, v, 4, DOMAINTRIGGER)
{  
}

ConstraintLex::~ConstraintLex() 
{
}
		   
int ConstraintLex::check( const int* s ) const 
{
  return !( ( (s[2] < s[3]) && (s[0] < s[1]) )
	    ||
	    ( (s[2] + s[3] == 0) && (s[0] == s[1]) )
	    ||
	    (s[2] + s[3] == 2) );
}

bool ConstraintLex::propagate(VariableInt *v, const int e)
{ 
  // prunes B[i+1]
  // B[i] <= B[i+1]

  if( !(scope[3]->setMin( scope[2]->min() )) || 
      !(scope[2]->setMax( scope[3]->max() )) 
      ) return false;

  if( scope[2]->min() && !(scope[3]->remove(0)) 
      ||
      !scope[3]->max() && !(scope[2]->remove(1))
      ) return false;

  // irrelevant part of the vectors
  if( scope[2]->min() && scope[3]->min() ) return true;
  
  // prunes B[i+1]
  // x[i] =/= y[i] -> b[i]+b[i+1] > 0 -> b[i+1] > 0
  if( !(scope[0]->intersect( scope[1] )) &&
      !(scope[3]->remove(0)) ) return false;
  // x[i] == y[i] -> b[i]+b[i+1] > 0 -> b[i+1] = 0
  if( scope[0]->isGround() && scope[1]->isGround() && 
      !(scope[2]->max()) &&
      (scope[0]->value() == scope[1]->value()) &&
      !(scope[3]->remove(1)) ) return false;

  // prunes B[i]
  // x[i] > y[i] -> b[i] > 0
  if( scope[0]->min() > scope[1]->max() &&
      ( !(scope[2]->setDomain(1)) ||
	!(scope[3]->setDomain(1)) ) ) return false;
  // x[i] >= y[i] -> b[i] >= b[i+1]
  if( scope[0]->min() >= scope[1]->max() &&
      ( !(scope[2]->setMin( scope[3]->min() )) || 
	!(scope[3]->setMax( scope[2]->max() )) )
      ) return false;
  // x[i] < y[i] -> b[i+1] = 1
  if( scope[0]->max() < scope[1]->min() )
    if( !scope[3]->setDomain(1) ) return false;

  // prunes X[i] and Y[i]
  if( !(scope[2]->max()) )
    if( scope[3]->min() ) {
      if( !(scope[0]->setMax(scope[1]->max()-1)) ) return false;
      if( !(scope[1]->setMin(scope[0]->min()+1)) ) return false;
    }  else {
      if( !(scope[0]->setMax(scope[1]->max())) ) return false;
      if( !(scope[1]->setMin(scope[0]->min())) ) return false;
    }

  return true;
}

void ConstraintLex::print(std::ostream& o) const 
{
  o << id << " ( (" ;
  scope[2]->printshort( o ) ; 
  o << " < " ;
  scope[3]->printshort( o ) ;
  o << " -> (" ;
  scope[0]->printshort( o ) ;
  o << " < " ; 
  scope[1]->printshort( o ) ; 
  o << ") ) & ( (" ;
  scope[2]->printshort( o ) ;
  o << " + " ;
  scope[3]->printshort( o ) ;
  o << " = 0) -> " ;
  scope[0]->printshort( o ) ; 
  o << " = " ; 
  scope[1]->printshort( o ) ;
  o << ")";
}


/**********************************************
 * <= Constraint
 **********************************************/
ConstraintLess::ConstraintLess(Solver *s, VariableInt** v)
  : Constraint(s, v, 2, RANGETRIGGER, WEIGHTTWO)
{
  offset = 0;
}

ConstraintLess::ConstraintLess(Solver *s, VariableInt** v, int o)
  : Constraint(s, v, 2, RANGETRIGGER, WEIGHTTWO)
{
  offset = o;
}

ConstraintLess::~ConstraintLess() {}

int ConstraintLess::check( const int* s ) const 
{
  return ( s[0] + offset > s[1] );
}

bool ConstraintLess::propagate(VariableInt *v, const int e)
{
  return scope[1]->setMin(scope[0]->min() + offset) &&
    scope[0]->setMax(scope[1]->max() - offset);
}

void ConstraintLess::print(std::ostream& o) const
{
  scope[0]->printshort( o ) ;
  if( !offset )
    o << " <= ";
  else if( offset == 1 )
    o << " < ";
  else 
    o << " + " << offset-1 << " < " ;
  scope[1]->printshort( o ) ;
}




/**********************************************
 * AllDiff Constraint 
 **********************************************/

const int INCONSISTENT = 0;
const int CHANGES      = 1;
const int NO_CHANGES   = 2;

ConstraintAllDiff::ConstraintAllDiff(Solver *s, VariableInt** v, const int n) 
  : Constraint(s, v, n, RANGETRIGGER) 
{
  delayed = true;
  init();
}

void ConstraintAllDiff::init() 
{
  int i;
  lastLevel = -1;
  nb = 0;

  iv        = new Interval[arity];
  minsorted = new Interval*[arity];
  maxsorted = new Interval*[arity];
  bounds    = new int[2*arity+2];
  std::fill(bounds, bounds+2*arity+2, 0);

  for( i=0; i<arity; ++i ) {
    minsorted[i] = maxsorted[i] = &iv[i];  
    iv[i].min = iv[i].max = INT_MAX;
  }

  t = new int[2*arity+2];
  d = new int[2*arity+2];
  h = new int[2*arity+2];
}


ConstraintAllDiff::~ConstraintAllDiff() 
{
  delete [] bounds;
  delete [] maxsorted;
  delete [] minsorted;
  delete [] iv;
  delete [] h;
  delete [] d;
  delete [] t;
}

void sortmin( Interval *v[], int n ) 
{
  int i, current;
  bool sorted;
  Interval *t;

  current = n-1;
  sorted = false;
  while( !sorted ) {
    sorted = true;
    for( i = 0; i < current; i++ ) {
      if( v[i]->min > v[i+1]->min ) {
        t = v[i];
        v[i] = v[i+1];
        v[i+1] = t;
        sorted = false;
      }
    }
    current--;
  }
}

void sortmax( Interval *v[], int n ) 
{
  int i, current;
  bool sorted;
  Interval *t;

  current = 0;
  sorted = false;
  while( !sorted ) {
    sorted = true;
    for( i = n-1; i > current; i-- ) {
      if( v[i]->max < v[i-1]->max ) {
        t = v[i];
        v[i] = v[i-1];
        v[i-1] = t;
        sorted = false;
      }
    }
    current++;
  }
}

void ConstraintAllDiff::sortit() 
{  
  int i,j,nb,min,max,last;

  sortmin(minsorted, arity);
  sortmax(maxsorted, arity);

  min = minsorted[0]->min;
  max = maxsorted[0]->max + 1;
  bounds[0] = last = min-2;

  for (i=j=nb=0;;) { // merge minsorted[] and maxsorted[] into bounds[]
    if (i<arity && min<=max) {	// make sure minsorted exhausted first
      if (min != last)
        bounds[++nb] = last = min;
      minsorted[i]->minrank = nb;
      if (++i < arity)
        min = minsorted[i]->min;
    } else {
      if (max != last)
	bounds[++nb] = last = max;
      maxsorted[j]->maxrank = nb;
      if (++j == arity) break;
      max = maxsorted[j]->max + 1;
    }
  }
  ConstraintAllDiff::nb = nb;
  bounds[nb+1] = bounds[nb] + 2;
}


void pathset(int *t, int start, int end, int to) 
{
  int k, l;
  for (l=start; (k=l) != end; t[k]=to) 
    l = t[k];  
}

int pathmin(int *t, int i) 
{
  for (; t[i] < i; i=t[i]) ;
  return i;
}

int pathmax(int *t, int i) 
{
  for (; t[i] > i; i=t[i]) ;  
  return i;
}


int ConstraintAllDiff::filterlower() 
{
  int i,j,w,x,y,z;
  int changes = 0;

  for (i=1; i<=nb+1; i++)
    d[i] = bounds[i] - bounds[t[i]=h[i]=i-1];
  for (i=0; i<arity; i++) { // visit Intervals in increasing max order
    x = maxsorted[i]->minrank; y = maxsorted[i]->maxrank;
    j = t[z = pathmax(t, x+1)];
    if (--d[z] == 0)
      t[z = pathmax(t, t[z]=z+1)] = j;
    pathset(t, x+1, z, z); // path compression
    if (d[z] < bounds[z]-bounds[y]) return INCONSISTENT; // no solution
    if (h[x] > x) {
      maxsorted[i]->min = bounds[w = pathmax(h, h[x])];
      pathset(h, x, w, w); // path compression
      changes = 1;
    }
    if (d[z] == bounds[z]-bounds[y]) {
      pathset(h, h[y], j-1, y); // mark hall Interval
      h[y] = j-1; //("hall Interval [%d,%d)\n",bounds[j],bounds[y]);
    }
  }
  if( changes )
    return CHANGES;
  else
    return NO_CHANGES;
}


int ConstraintAllDiff::filterupper()
{
  int i,j,w,x,y,z;
  int changes = 0;

  for (i=0; i<=nb; i++)
    d[i] = bounds[t[i]=h[i]=i+1] - bounds[i];
  for (i=arity; --i>=0; ) { // visit Intervals in decreasing min order
    x = minsorted[i]->maxrank; y = minsorted[i]->minrank;
    j = t[z = pathmin(t, x-1)];
    if (--d[z] == 0)
      t[z = pathmin(t, t[z]=z-1)] = j;
    pathset(t, x-1, z, z);
    if (d[z] < bounds[y]-bounds[z]) return INCONSISTENT; // no solution
    if (h[x] < x) {
      minsorted[i]->max = bounds[w = pathmin(h, h[x])] - 1;
      pathset(h, x, w, w);
      changes = 1;
    }
    if (d[z] == bounds[y]-bounds[z]) {
      pathset(h, h[y], j+1, y);
      h[y] = j+1;
    }
  }
  if( changes )
    return CHANGES;
  else
    return NO_CHANGES;
}


bool ConstraintAllDiff::propagate(VariableInt *v, const int e) 
{
  return propagate();
}

bool ConstraintAllDiff::propagate() 
{
  int i, a, b;

//   if( v->isGround() ) {
//     a = v->value();
//     i = arity;
//     while( i-- )
//       if( scope[i] != v && !scope[i]->remove(a) ) { return false; }
//   }
  
// //   if( e & Constraint::DELAYTRIGGER ) {
// //     while( i-- )
// //       if( scope[i].isGround() && 
// //   } 
  
  int status_lower, status_upper;
  int l, u;

  a = 0;
  b = arity;

  //if( lastLevel != ((scope[0]->solver->level) - 1) ) {
  if( lastLevel != ((solver->level) - 1) ) {
    // not incremental
    status_lower = CHANGES;
    status_upper = CHANGES;
    i = 0;
    while (i < arity) {
      iv[i].min = scope[i]->min();
      iv[i].max = scope[i]->max();
      i++;
    }
  }
  else {
    // incremental
    status_lower = NO_CHANGES;
    status_upper = NO_CHANGES;
    for( i = a; i < b; i++ ) {
      l = iv[i].min;
      u = iv[i].max;
      iv[i].min = scope[i]->min();
      iv[i].max = scope[i]->max();
      if( l != iv[i].min ) status_lower = CHANGES;
      if( u != iv[i].max ) status_upper = CHANGES;
    }
  }

  //lastLevel = (scope[0]->solver->level);
  lastLevel = (solver->level);

  if( status_lower == NO_CHANGES && status_upper == NO_CHANGES ) 
    return true;

  sortit();

  status_lower = filterlower();
  if( status_lower != INCONSISTENT )
    status_upper = filterupper();  

  if( (status_lower == INCONSISTENT) || (status_upper == INCONSISTENT) ) 
    { return false; }
  else
    if( (status_lower == CHANGES) || (status_upper == CHANGES) ) {
      i = 0;
      while (i < arity) {
	if( !scope[i]->setMin( iv[i].min ) )  { return false; }
	if( !scope[i]->setMax( iv[i].max ) )  { return false; }
	i++;
      }
    }  
  return true;
}

int ConstraintAllDiff::check( const int* s ) const 
{
  int i=arity, j;
  while(--i) {
    j=i;
    while(j--)
      if( s[i] == s[j] ) return 1;
  }
  return 0; 
}

void ConstraintAllDiff::print(std::ostream& o) const 
{    
  o << "alldiff(";
  for(int i=0; i<arity-1; ++i) {
    scope[i]->printshort( o );
    o << ", ";
  }
  scope[arity-1]->printshort( o );
  o << ")";
}

/**********************************************
 * EdgeFinder Constraint
 **********************************************/ 

bool ConstraintEdgeFinder::jacksonPSchedule()
{  
  int residue[arity], rest=arity;
  int min_rd = INT_MAX, max_dd = 0;
  int i=arity, j, x, y, tmp;
  while( i-- ) {
    sorted[i] = i;
    if( min_rd > scope[i]->min() ) min_rd = scope[i]->min();
    if( max_dd < (scope[i]->max() + durations[i]) ) 
      max_dd = (scope[i]->max() + durations[i]);
  }
  int jps[max_dd];
  int jps_length = 0;

  // Bubble Sort
  for( x=0; x<arity; x++ ) {
    for( y=0; y<arity-1-x; y++ ) {
      if( scope[sorted[y  ]]->max()+durations[sorted[y  ]] < 
	  scope[sorted[y+1]]->max()+durations[sorted[y+1]] ) {
	tmp         = sorted[y+1];
	sorted[y+1] = sorted  [y];
	sorted  [y] =         tmp;
      }
    }
    // initialisation of data-structures
    residue[arity-1-x] = durations[sorted[arity-1-x]];
    Pk[x][0] = -1; Pk[x][arity] = 0;
  }

  int dur, cur = min_rd;
  while( rest ) { // there is still something to schedule
    i = arity;
    // compute Pk
    while( i--) 
      if( Pk[i][0] < 0 && scope[i]->min() <= cur ) 
	for(j=0; j<arity; ++j) {
	  Pk[i][sorted[j]] = residue[j];
	  Pk[i][arity] += residue[j];	
	}
    // jps
    i = arity; dur = INT_MAX; tmp = INT_MAX;
    while( tmp == INT_MAX && i-- ) 
      if( residue[i] ) {
	if( scope[sorted[i]]->min() <= cur ) {
	  // this is the task to schedule next in the jps.
	  if( residue[i] < dur ) dur = residue[i];
	  tmp = sorted[i];
	  if( !(residue[i] -= dur)) --rest;	  
	} else 
	  // this activity has an earlier due date but its release date is later.
	  // we change dur so as not to miss it
	  if( scope[sorted[i]]->min() - cur < dur ) dur = (scope[sorted[i]]->min() - cur);
      }
    if( cur+dur <= max_dd )
      std::fill( jps+cur, jps+cur+dur, tmp );
    cur += dur;
  }
 
  return (cur <= max_dd);
}

ConstraintEdgeFinder::ConstraintEdgeFinder(Solver *s, VariableInt** v, 
					   const int n, const int *d)
  : Constraint(s, v, n, RANGETRIGGER)
{
  sorted = new int[arity];
  durations = new int[arity];
  Pk = new int*[arity];
  for(int i=0; i<arity; ++i) {
    durations[i] = d[i];
    Pk[i] = new int[arity+1];  
  }
  //int i=arity;
  //while( i-- )
  //  triggers[i] = scope[i]->triggerOnRange();//&(scope[i]->rangeTrigger);
}

ConstraintEdgeFinder::~ConstraintEdgeFinder()
{
  delete [] sorted;
  delete [] durations;
  for(int i=0; i<arity; ++i) 
    delete [] Pk[i];  
  delete [] Pk;
}

int ConstraintEdgeFinder::check( const int* s ) const 
{
  // todo
  return 0;
}

bool ConstraintEdgeFinder::propagate(VariableInt *v, const int e)
{
  return propagate();
}

bool ConstraintEdgeFinder::propagate()
{
  if( !jacksonPSchedule() ) return false;

  int i, j, k;
  for( i=0; i<arity; ++i ) {
    Pk[i][arity] += scope[i]->min();
    for( j=0; j<arity; ++j ) {
      k = sorted[j];
      if ( i != k && Pk[i][k] ) {
	if(  Pk[i][arity] > scope[k]->max()+durations[k] ) {
	  if( !scope[i]->setMin(scope[k]->min() + durations[k]) ) return false;
	  if( !scope[k]->setMax(scope[i]->max() - durations[k]) ) return false;
	}
	Pk[i][arity] -= Pk[i][k];
      }
    }
  }
  return true;
}

void ConstraintEdgeFinder::print(std::ostream& o) const
{
  Constraint::print( o );
  o << endl;
  for(int i=0; i<arity-1; ++i) {
    o << "[" << scope[i]->min() << "..|" << durations[i]
      << "|.." << scope[i]->max()+durations[i] << "]," ;
  }
  o << "[" << scope[arity-1]->min() << "..|" << durations[arity-1]
    << "|.." << scope[arity-1]->max()+durations[arity-1] << "]" ;
}


/**********************************************
 * Boolean Sum Equality Constraint  
 **********************************************/

ConstraintBoolSumEqual::ConstraintBoolSumEqual(Solver *s, 
					       VariableInt **v, 
					       const int n,
					       const int t)
  : Constraint(s, v, n, VALUETRIGGER)
{
  total = t;
  min_.setValue(0);
  max_.setValue(n);
  min_.init( s );
  max_.init( s );
}

ConstraintBoolSumEqual::~ConstraintBoolSumEqual()
{
}

int ConstraintBoolSumEqual::check( const int* s ) const 
{
  int i = arity, tot = 0;
   while( i-- )
    tot += s[i];

  return ( tot > total ? tot - total : total - tot );
}

bool ConstraintBoolSumEqual::propagate()
{
  bool consistent = true;
  for( int i=0; consistent && i<arity; ++i )
    if( scope[i]->isGround() )
      consistent = propagate( scope[i], VALUETRIGGER );
  return consistent;
} 

bool ConstraintBoolSumEqual::propagate(VariableInt *v, const int e) 
{

  int i, smin = 0, smax = 0;

  if( v->min() ) ++min_;
  else --max_;
  
  smin = min_;
  smax = max_;

  if( smin+1 == smax && !isLinked() ) {
    
    int idx = (watch[0] == v);
    if( watch[idx]->isGround() ) {
      if( watch[idx]->min() ) ++smin;
      else --smax;
    }
   
  }

  if( smax < total || smin > total ) return false;

  i = arity;
  bool consistent = true;

  if( smax == total ) {
    while( consistent && i-- )
      consistent = scope[i]->setMin( scope[i]->max() );
  } else if( smin == total ) {
    while( consistent && i-- )
      consistent = scope[i]->setMax( scope[i]->min() );
  }

  return consistent;
}

void ConstraintBoolSumEqual::print(std::ostream& o) const
{ 
  o<< "(" ;
  for(int i=0; i<arity-1; ++i) {
    scope[i]->print( o );
    o << " + ";
  }
  scope[arity-1]->print( o );
  o << ") = " << total ;
}


/**********************************************
 * Boolean Sum Less Constraint  
 **********************************************/

ConstraintBoolSumLess::ConstraintBoolSumLess(Solver *s, 
					     VariableInt **v, 
					     const int n,
					     const int t)
  : Constraint(s, v, n, VALUETRIGGER)
{
  total = t;
  min_.setValue(0);
  min_.init( s );
}

ConstraintBoolSumLess::~ConstraintBoolSumLess()
{
}

int ConstraintBoolSumLess::check( const int* s ) const 
{
  int i = arity, tot = 0;
   while( i-- )
    tot += s[i];

  return ( tot > total ? tot - total : 0 );
}

bool ConstraintBoolSumLess::propagate()
{
  bool consistent = true;
  for( int i=0; consistent && i<arity; ++i )
    if( scope[i]->isGround() )
      consistent = propagate( scope[i], VALUETRIGGER );
  return consistent;
} 

bool ConstraintBoolSumLess::propagate(VariableInt *v, const int e) 
{
  int i, smin = 0;

  if( v->min() ) ++min_;
  smin = min_;

  if( !isLinked() ) {
    int idx = (watch[0] == v);
    if( watch[idx]->min() ) ++smin;
  }
  
  if( total < smin ) return false;

  if( smin == total ) {
    i = arity;
    while( i-- )
      scope[i]->setMax( scope[i]->min() );
  }

  return true;
}

void ConstraintBoolSumLess::print(std::ostream& o) const
{ 
  o<< "(" ;
  for(int i=0; i<arity-1; ++i) {
    scope[i]->print( o );
    o << " + ";
  }
  scope[arity-1]->print( o );
  o << ") <= " << total ;
}


/**********************************************
 * Boolean Sum More Constraint  
 **********************************************/

ConstraintBoolSumMore::ConstraintBoolSumMore(Solver *s, 
					     VariableInt **v, 
					     const int n,
					     const int t)
  : Constraint(s, v, n, VALUETRIGGER)
{
  total = t;
  max_.setValue(arity);
  max_.init( s );
}

ConstraintBoolSumMore::~ConstraintBoolSumMore()
{
}

int ConstraintBoolSumMore::check( const int* s ) const 
{
  int i = arity, tot = 0;
   while( i-- )
    tot += s[i];

  return ( tot < total ? tot - total : 0 );
}

bool ConstraintBoolSumMore::propagate()
{
  bool consistent = true;
  for( int i=0; consistent && i<arity; ++i )
    if( scope[i]->isGround() )
      consistent = propagate( scope[i], VALUETRIGGER );
  return consistent;
} 

bool ConstraintBoolSumMore::propagate(VariableInt *v, const int e) 
{
  int i, smax = 0;

  if( !v->max() ) --max_;
  smax = max_;

  if( !isLinked() ) {
    int idx = (watch[0] == v);
    if( !watch[idx]->max() ) --smax;
  }
  
  if( smax < total ) return false;

  if( smax == total ) {
    i = arity;
    while( i-- )
      scope[i]->setMin( scope[i]->max() );
  }

  return true;
}

void ConstraintBoolSumMore::print(std::ostream& o) const
{ 
  o<< "(" ;
  for(int i=0; i<arity-1; ++i) {
    scope[i]->print( o );
    o << " + ";
  }
  scope[arity-1]->print( o );
  o << ") >= " << total ;
}


/**********************************************
 * SimpleSum Predicate 
 **********************************************/

PredicateSimpleSum::PredicateSimpleSum(Solver *s, VariableInt** v, const int n)
  : Constraint(s, v, n, RANGETRIGGER)
{
  //  delayed = true;
}

PredicateSimpleSum::~PredicateSimpleSum()
{
}

int PredicateSimpleSum::check( const int* s ) const 
{
  int i = arity-1, total = 0;
   while( i-- )
    total += s[i];

  return ( total > s[arity-1] ? total - s[arity-1] : s[arity-1] - total );
}

bool PredicateSimpleSum::propagate(VariableInt *v, const int e) 
{
  return propagate();
}

bool PredicateSimpleSum::propagate() 
{

  int i, n = arity-1;
  // compute the max and th min
  int smin = 0, smax = 0;
  for(i=0; i<n; ++i) {
    smax += scope[i]->max();
    smin += scope[i]->min();
  }

  if( !scope[n]->setMax( smax ) || 
      !scope[n]->setMin( smin ) ) return false;

  smax -= scope[n]->min();
  smin -= scope[n]->max();

  // prune domains
  int aux;
  for(i=0; i<n; ++i) {
    aux = scope[i]->max();
    if( (!scope[i]->setMax(scope[i]->min() - smin)) 
	||
	(!scope[i]->setMin(aux - smax)) ) 
      return false;
  }

  return true;
}

void PredicateSimpleSum::print(std::ostream& o) const
{ 
  o<< "(" ;
  for(int i=0; i<arity-2; ++i) {
    scope[i]->printshort( o );
    o << " + ";
  }
  scope[arity-2]->printshort( o );
  o << ") = " ;
  scope[arity-1]->printshort( o );
}



/**********************************************
 * SumEqual Constraint 
 **********************************************/

ConstraintSumEqual::ConstraintSumEqual(Solver *s, VariableInt** v, 
				       const int n,
				       const int t)
  : Constraint(s, v, n, RANGETRIGGER)
{
  total = t;
}

ConstraintSumEqual::~ConstraintSumEqual()
{
}

int ConstraintSumEqual::check( const int* s ) const 
{
  int i = arity, tot = 0;
   while( i-- )
    tot += s[i];

  return ( tot > total ? tot - total : total - tot );
}


bool ConstraintSumEqual::propagate(VariableInt *v, const int e) 
{
  return propagate();
} 

bool ConstraintSumEqual::propagate()
{
  int i;
  // compute the max and th min
  int smin = 0, smax = 0;
  for(i=0; i<arity; ++i) {
    smax += scope[i]->max();
    smin += scope[i]->min();
  }

  if( smax < total || smin > total ) return false;

  smax -= total;
  smin -= total;

  // prune domains
  int aux;
  for(i=0; i<arity; ++i) {
    aux = scope[i]->max();
    if( (!scope[i]->setMax(scope[i]->min() - smin)) 
	||
	(!scope[i]->setMin(aux - smax)) ) 
      return false;
  }

  return true;
}

void ConstraintSumEqual::print(std::ostream& o) const
{ 
  o<< "(" ;
  for(int i=0; i<arity-1; ++i) {
    scope[i]->printshort( o );
    o << " + ";
  }
  scope[arity-1]->printshort( o );
  o << ") = " << total;
}


/**********************************************
 * WeightedSumEqual Constraint 
 **********************************************/

ConstraintWeightedSumEqual::ConstraintWeightedSumEqual(Solver *s, VariableInt** v, 
						       const int n,
						       const int *w)
  : Constraint(s, v, n, RANGETRIGGER)
{
  weights = new int[n+1];
  int i;
  for(i=0; i<=n; ++i) 
    weights[i] = w[i];
}

ConstraintWeightedSumEqual::~ConstraintWeightedSumEqual()
{
  delete [] weights;
}

int ConstraintWeightedSumEqual::check( const int* s ) const 
{
  int i = arity, tot = 0;
   while( i-- )
     tot += (weights[i] * s[i]);

  return ( tot > weights[arity] ? tot - weights[arity] : weights[arity] - tot );
}

bool ConstraintWeightedSumEqual::propagate(VariableInt *v, const int e) 
{
  return propagate();
}

bool ConstraintWeightedSumEqual::propagate() 
{
  int i;
  // compute the max and th min
  int smin=0, smax=0;
  for(i=0; i<arity; ++i) {
    if( weights[i] > 0 ) {
      smax += weights[i] * scope[i]->max();
      smin += weights[i] * scope[i]->min();
    } else {
      smin += weights[i] * scope[i]->max();
      smax += weights[i] * scope[i]->min();
    }
  }

  if( smax < weights[arity] || smin > weights[arity] ) return false;

  smax -= weights[arity];
  smin -= weights[arity];

  // prune domains
  int aux;
  for(i=0; i<arity; ++i) {
    if( weights[i] > 0 ) {
      aux = scope[i]->max();
      if( (!scope[i]->setMax(scope[i]->min() - smin/weights[i])) 
	  ||
	  (!scope[i]->setMin(aux - smax/weights[i])) ) 
	return false;
    } else {
      aux = scope[i]->min();
      if( (!scope[i]->setMin(scope[i]->max() - smin/weights[i]))
	  ||
	  (!scope[i]->setMax(aux - smax/weights[i])) )
	return false;
    }
  }

  return true;
}

void ConstraintWeightedSumEqual::print(std::ostream& o) const
{ 
  o<< "(" ;
  if( weights[0] < 0 ) {
    o << " - ";
    o.flush() ; 
  }
  if( weights[0] != 1 && weights[0] != -1 ) {
    o << weights[0];
    o.flush();
  }
  scope[0]->printshort( o );
  o.flush();
  for(int i=1; i<arity; ++i) {
    if( weights[i] < 0 ) {
      o << " - ";
      o.flush();
    }
    else {
      o << " + ";
      o.flush();
    }
    if( weights[i] != 1 && weights[i] != -1 ) {
      o << weights[i];
      o.flush();
    }
    scope[i]->printshort( o );
    o.flush();
  }
  o << ") = " << weights[arity];
}



/**********************************************
 * SumLess Constraint 
 **********************************************/

ConstraintSumLess::ConstraintSumLess(Solver *s, VariableInt** v, 
				     const int n,
				     const int t)
  : Constraint(s, v, n, RANGETRIGGER)
{
  total = t;
}

ConstraintSumLess::~ConstraintSumLess()
{
}

int ConstraintSumLess::check( const int* s ) const 
{
  int i = arity, tot = 0;
  while( i-- )
    tot += s[i];
  
  return ( tot > total ? tot - total : 0 );
}

bool ConstraintSumLess::propagate(VariableInt *v, const int e) 
{
  return propagate();
}

bool ConstraintSumLess::propagate() 
{
  int i;
  // compute the max and th min
  int smin = 0;
  for(i=0; i<arity; ++i) 
    smin += scope[i]->min();
  
  if( total < smin ) return false;
  smin -= total;

  // prune domains
  for(i=0; i<arity; ++i) {
    if( !scope[i]->setMax(scope[i]->min() - smin) ) 
      return false;
  }

  return true;
}

void ConstraintSumLess::print(std::ostream& o) const
{ 
  o<< "(" ;
  for(int i=0; i<arity-2; ++i) {
    scope[i]->printshort( o );
    o << " + ";
  }
  scope[arity-2]->printshort( o );
  o << ") <= " << total;
}


// // /**********************************************
// //  * Tuple Predicate 
// //  **********************************************/

// // int exponent(int x, int e) {
// //   int res = 1;
// //   while( e-- )
// //     res = (res * x);
// //   return res;
// // }

// // void int2tuple(int x, int* tuple, int n, int* exposant) {
// //   int res = x;
// //   while( n-- ) {
// //     tuple[n] = (res/exposant[n]);
// //     res -= (tuple[n] * exposant[n]);
// //   }
// // }

// // void tuple2int(int& x, int* tuple, int n, int* exposant) {
// //   x = 0;
// //   while( n-- )
// //     x += (tuple[n] * exposant[n]);  
// // }

// // TuplePredicate::TuplePredicate(Solver *s, VariableInt** nvars, 
// // 			       const int n)
// //   : Constraint(s,), BitVar()
// // {
// //   name = "tuple";

// //   int lb=0, ub=0, i, size;
// //   VariableInt *scp[n+1];
// //   vals = new MistralSet[n];

// //   exps = new int[n];
// //   for(i=0; i<n; ++i) {
// //     scp[i] = nvars[i];
// //     if(i) {
// //       exps[i] = (scp[i]->max() - scp[i]->min() + 1) * exps[i-1];
// //     } else exps[i] = 1;

// //     lb += (scp[i]->min() * exps[i]);
// //     ub += (scp[i]->max() * exps[i]);

// //     vals[i].init( scp[i]->min(), scp[i]->max(), MistralSet::empt );
// //   }

// //   setVariable( lb, ub );
// //   scp[n] = this;
// //   initScope( scp, n+1 );
// // }

// // TuplePredicate::~TuplePredicate()
// // {
// //   delete [] exps;
// //   delete [] vals;
// // }

// // bool TuplePredicate::add( CSP* model, int toplevel )
// // {
// //   VariableInt::add( model, 0 );
// //   Constraint::add( model, 0 );
// //   return true;
// // }

// // int TuplePredicate::check( const int* s ) const 
// // {
// //   int x;
// //   tuple2int( x, s, arity-1, exps ); 
// //   return ( x != s[arity-1] );
// // }

// // bool TuplePredicate::propagate(VariableInt *v, const int e) 
// // {
// //   int vali, vnxt, i, tuple[arity-1];

// //   i = arity-1;
// //   while( i-- ) vals[i].clear();

// //   vali = first();
// //   do {
// //     vnxt = getNext( vali );
// //     int2tuple( vali, tuple, arity-1, exps );
// //     i=arity-2;
// //     while( i >= 0 && scope[i]->contain(tuple[i]) ) --i;
// //     if( i >= 0 && !remove( vali ) ) return false;
// //     vali = vnxt;

// //     if( contain(vali) ) {
// //       i = arity-1;
// //       while( i-- ) vals[i].insert( tuple[i] );
// //     }
// //   }
// //   while( good( vali ) );	

// //   i = arity-1;
// //   while( i-- ) scope[i]->setDomain( vals[i] );
  
// //   return true;
// // }

// // void TuplePredicate::print(std::ostream& o) const
// // { 
// //   BitVar::print( o );
// //   o << " := <";
// //   for(int i=0; i<arity-2; ++i) {
// //     scope[i]->printshort( o );
// //     o << ", ";
// //   }
// //   scope[arity-2]->printshort( o );
// //   o << ">";
// // }


/**********************************************
 * Max Predicate 
 **********************************************/

PredicateMax::PredicateMax(Solver *s, VariableInt** v, const int n)
  : Constraint(s, v, n, RANGETRIGGER)
{
}

PredicateMax::~PredicateMax()
{
}

int PredicateMax::check( const int* s ) const 
{
  int i = arity-2, themax = s[arity-2];
  while( i-- ) if( s[i] > themax ) themax = s[i];
  return (themax != s[arity-1] );
}

bool PredicateMax::propagate(VariableInt *v, const int e) 
{
  // NOT GAC !!
  bool consistent = true;
  int i, j=-1, m;
  if( v == scope[arity-1] ) {
    i = arity-1;
    while( consistent && i-- )
      consistent = scope[i]->setMax( v->max() );
  } else {
    consistent = scope[arity-1]->setMin( v->min() );
    if( consistent ) {
      int maxmax = scope[arity-2]->max(), maxmin = scope[arity-2]->min();
      i = arity-2;
      while( i-- ) {
	if( maxmax < scope[i]->max() )
	  maxmax = scope[i]->max();
	if( maxmin < scope[i]->min() )
	  maxmin = scope[i]->min();
      }
      consistent = (scope[arity-1]->setMax( maxmax ) && 
		    scope[arity-1]->setMin( maxmin ));
    }
  }
  i = arity-1, m = scope[i]->min();
  while( i-- ) 
    if( scope[i]->max() >= m )
      {
	if( j < 0 ) 
	  j = i;
	else break;
      }
  if( i < 0 )
    consistent = (j >= 0 && scope[j]->setMin( m ));

  return consistent;
}

void PredicateMax::print(std::ostream& o) const
{ 
  o<< "max(" ;
  for(int i=0; i<arity-2; ++i) {
    scope[i]->printshort( o );
    o << ", ";
  }
  scope[arity-2]->printshort( o );
  o << ") = " ;
  scope[arity-1]->printshort( o );
}


/**********************************************
 * Min Predicate 
 **********************************************/

PredicateMin::PredicateMin(Solver *s, VariableInt** v, const int n)
  : Constraint(s, v, n, RANGETRIGGER)
{
}

PredicateMin::~PredicateMin()
{
}

int PredicateMin::check( const int* s ) const 
{
  int i = arity-2, themin = s[arity-2];
  while( i-- ) if( s[i] < themin ) themin = s[i];
  return (themin != s[arity-1] );
}

bool PredicateMin::propagate(VariableInt *v, const int e) 
{
  bool consistent = true;

  // NOT GAC !!
  int i, j=-1, m;
  if( v == scope[arity-1] ) {
    i = arity-1;
    while( i-- )
      consistent = scope[i]->setMin( v->min() );
  } else {
    consistent = scope[arity-1]->setMax( v->max() );
    if( consistent ) {
      int minmax = scope[arity-2]->max(), minmin = scope[arity-2]->min();
      i = arity-2;
      while( i-- ) {
	if( minmax > scope[i]->max() )
	  minmax = scope[i]->max();
	if( minmin > scope[i]->min() )
	  minmin = scope[i]->min();
      }
      consistent = (scope[arity-1]->setMax( minmax ) && 
		    scope[arity-1]->setMin( minmin ));
    }
  }
  i = arity-1, m = scope[i]->max();
  while( i-- ) 
    if( scope[i]->min() <= m )
      {
	if( j < 0 ) 
	  j = i;
	else break;
      }
  if( i < 0 )
    consistent = (j >= 0 && scope[j]->setMax( m ));
  
  return consistent;
}

void PredicateMin::print(std::ostream& o) const
{ 
  o<< "min(" ;
  for(int i=0; i<arity-2; ++i) {
    scope[i]->printshort( o );
    o << ", ";
  }
  scope[arity-2]->printshort( o );
  o << ") = " ;
  scope[arity-1]->printshort( o );
}


// /**********************************************
//  * Arithmetic and Logic Constraints
//  **********************************************/

/**********************************************
 * WeightedSum Predicate 
 **********************************************/

PredicateWeightedSum::PredicateWeightedSum( Solver *s, VariableInt** v, 
					    const int n, 
					    const int* os )
  : Constraint(s, v, n, RANGETRIGGER)
{
  weights = new int[n];
  int i;
  for(i=0; i<n; ++i) 
    weights[i] = os[i];
}

PredicateWeightedSum::~PredicateWeightedSum()
{
  delete [] weights;
}

int PredicateWeightedSum::check( const int* s ) const 
{
  int i = arity-1, total = weights[i];
   while( i-- )
    total += weights[i] * s[i];

  return ( total > s[arity-1] ? total - s[arity-1] : s[arity-1] - total );
}

bool PredicateWeightedSum::propagate(VariableInt *v, const int e) 
{
  return propagate();
}

bool PredicateWeightedSum::propagate() 
{
  int i, n = arity-1;
  // compute the max and th min
  int smin = weights[n], smax=weights[n];
  for(i=0; i<n; ++i) {
    if( weights[i] > 0 ) {
      smax += weights[i] * scope[i]->max();
      smin += weights[i] * scope[i]->min();
    } else {
      smin += weights[i] * scope[i]->max();
      smax += weights[i] * scope[i]->min();
    }
  }

  if( !scope[n]->setMax( smax ) || !scope[n]->setMin( smin ) ) return false;

  smax -= scope[n]->min();
  smin -= scope[n]->max();

  // prune domains
  int aux;
  for(i=0; i<n; ++i) {
    if( weights[i] > 0 ) {
      aux = scope[i]->max();
      if( (!scope[i]->setMax(scope[i]->min() - smin/weights[i])) 
	  ||
	  (!scope[i]->setMin(aux - smax/weights[i])) ) 
	return false;
    } else {
      aux = scope[i]->min();
      if( (!scope[i]->setMin(scope[i]->max() - smin/weights[i]))
	  ||
	  (!scope[i]->setMax(aux - smax/weights[i])) )
	return false;
    }
  }

  return true;
}

void PredicateWeightedSum::print(std::ostream& o) const
{ 
  o<< "(" ;
  if( weights[0] < 0 ) {
    o << " - ";
    o.flush() ; 
  }
  if( weights[0] != 1 && weights[0] != -1 ) {
    o << weights[0];
    o.flush();
  }
  scope[0]->printshort( o );
  o.flush();
  for(int i=1; i<arity-1; ++i) {
    if( weights[i] < 0 ) {
      o << " - ";
      o.flush();
    }
    else {
      o << " + ";
      o.flush();
    }
    if( weights[i] != 1 && weights[i] != -1 ) {
      o << weights[i];
      o.flush();
    }
    scope[i]->printshort( o );
    o.flush();
  }
  if( weights[arity-1] > 0 ) o << " + " << weights[arity-1];
  if( weights[arity-1] < 0 ) o << " - " << -weights[arity-1];
  o << ") = " ;
  scope[arity-1]->printshort( o );
}



/**********************************************
 * Nested Offset Predicate
 **********************************************/
/// x0 + k
PredicateOffset::PredicateOffset(Solver *s, VariableInt** v, const int k) 
  : Constraint(s, v, 2, DOMAINTRIGGER, WEIGHTTWO)
{
  offset = k;
  if( offset < 0 )
    xdom.init(scope[0]->min()+offset, scope[0]->max(), MistralSet::empt);
  else xdom.init(scope[0]->min(), scope[0]->max()+offset, MistralSet::empt);
  //weight = 0;
  //--scope[0]->weight;
}

PredicateOffset::~PredicateOffset()
{
}

int PredicateOffset::check( const int* s ) const 
{
  return( s[0]+offset != s[1] );
}

bool PredicateOffset::awakeOnDomain( VariableInt *v )
{
  xdom.clear();
  v->unionTo( xdom );
  
  if( offset < 0 ) {
    if( v == scope[0] )
      xdom.decrement( -offset );
    else xdom.increment( -offset );
  } else {
    if( v == scope[0] )
      xdom.increment( offset );
    else xdom.decrement( offset );
  }    
  return scope[(scope[0] == v)]->setDomain( xdom );     
}

bool PredicateOffset::awakeOnRange( VariableInt *v )
{
  return( v == scope[0] ?
	  (scope[1]->setMin( v->min()+offset ) &&
	   scope[1]->setMax( v->max()+offset )) :
	  (scope[0]->setMin( v->min()-offset ) &&
	   scope[0]->setMax( v->max()-offset )) );
}

bool PredicateOffset::propagate(VariableInt *v, const int e)
{
  return ( (e & Constraint::DOMAINTRIGGER) ?
	   awakeOnDomain( v ) :
	   awakeOnRange( v ) );
}

void PredicateOffset::print(std::ostream& o) const 
{
  scope[0]->printshort( o );
  if( offset < 0 )
    o << " - " << (-offset) ;
  else
    o << " + " << offset ;
  o << " = ";
  scope[1]->printshort( o );
}


/**********************************************
 * Nested Factor Predicate
 **********************************************/
/// x0 + k
PredicateFactor::PredicateFactor(Solver *s, VariableInt** v, const int k) 
  : Constraint(s, v, 2, DOMAINTRIGGER, WEIGHTTWO)
{
  factor = k;
}

PredicateFactor::~PredicateFactor()
{
}

int PredicateFactor::check( const int* s ) const 
{
  return( s[0] * factor != s[1] );
}

bool PredicateFactor::propagate(VariableInt *v, const int e)
{
  bool consistent = true;
  int lb, ub, modulo, i = (v == scope[0]);

  if( i ) // v == scope[0] 
    {
      if( factor > 0 )
	consistent = ( scope[1]->setMin( v->min() * factor ) &&
		       scope[1]->setMax( v->max() * factor ) );
      else
	consistent = ( scope[1]->setMin( v->max() * factor ) &&
		       scope[1]->setMax( v->min() * factor ) );
    }
  else // v == scope[1]
    {
      lb = v->min();
      modulo = (lb % factor);
      if( modulo ) {
	lb -= modulo; 
	v->setMin( lb );
	lb = v->min();
      }
      
      ub = v->max();
      modulo = (ub % factor);
      if( modulo ) {
	ub -= modulo; 
	v->setMax( ub );
	ub = v->max();
      }
      
      if( factor > 0 )
	consistent = ( scope[0]->setMin( lb / factor ) &&
		       scope[0]->setMax( ub / factor ) );
      else
	consistent = ( scope[0]->setMin( ub / factor ) &&
		       scope[0]->setMax( lb / factor ) );
    }
  
  return consistent;
}

void PredicateFactor::print(std::ostream& o) const 
{
  o << factor << "." ;
  scope[0]->printshort( o );
  o << " = ";
  scope[1]->printshort( o );
}


/**********************************************
 * Nested Abs Predicate
 **********************************************/
/// -x0
PredicateAbs::PredicateAbs(Solver *s, VariableInt** v) 
  : Constraint(s, v, 2)
{
}

PredicateAbs::~PredicateAbs()
{
}

int PredicateAbs::check( const int* s ) const 
{
  return( abs(s[0]) != s[1] );
}

bool PredicateAbs::propagate(VariableInt *v, const int e)
{  
  int lb, ub, aux;
  bool consistent = true;

  if( e == VALUETRIGGER ) {
    if( v == scope[0] ) {
      consistent = scope[1]->setDomain( abs(v->value()) );
    } else {
      ub = v->value();
      lb = -ub;
      consistent = (scope[0]->setMin( lb )
		    &&
		    scope[0]->setMax( ub )
		    && 
		    scope[0]->removeRange( lb+1, ub-1 ));
    }
  } else {
    if( v->isRange() ) {
      if( v == scope[0] ) {
	lb = abs(v->min());
	ub = abs(v->max());
	if(lb > ub) {
	  aux = ub;
	  ub = lb;
	} else aux = lb;
	consistent = scope[1]->setMax( ub );
      
	if( consistent ) {
	  lb = scope[1]->min();
	  if( !(v->contain(lb) || v->contain(-1*lb)) ) {
	    --lb;
	    ub = aux;
	    aux = ((ub + lb) >> 1);
	    while( ub > lb+1 ) {
	      if( v->intersect(-1*aux, aux) ) {
		ub = aux;
		aux = ((ub + lb) >> 1);
	      } else {
		lb = aux;
		aux = ((ub + lb) >> 1);
	      }
	    }  
	    consistent = scope[1]->setMin( aux+1 );
	  }
	}
      } else {
	consistent = ((scope[0]->setMax(scope[1]->max())) 
		      &&
		      (scope[0]->setMin(-1*scope[1]->max()))
		      &&
		      (!scope[1]->min() || 
		       scope[0]->removeRange( -1*scope[1]->min()+1, 
					      scope[1]->min()-1)));
      }
    } else {
      if( v == scope[0] ) {
	lb = scope[1]->first();
	do {
	  aux = scope[1]->getNext( lb );
	  if( !(scope[0]->contain(lb) || scope[0]->contain(-lb)) )
	    consistent = scope[1]->remove( lb );
	  lb = aux;
	} while( consistent && scope[1]->good( lb ) );
      } else {
	lb = scope[0]->first();
	do {
	  aux = scope[0]->getNext( lb );
	  if( !scope[1]->contain(abs(lb)) )
	    consistent = scope[0]->remove( lb );
	  lb = aux;
	} while( consistent && scope[0]->good( lb ) );
      }
    }
  }

  return consistent;
}

void PredicateAbs::print(std::ostream& o) const 
{
  o << "abs(" ;
  scope[0]->printshort( o );
  o << ") = ";
  scope[1]->printshort( o ); 
}


/**********************************************
 * Nested Negation Predicate
 **********************************************/
/// -x0
PredicateNegation::PredicateNegation(Solver *s, VariableInt** v) 
  : Constraint(s, v, 2, DOMAINTRIGGER, WEIGHTTWO)
{
  dom[0].init(v[0]->min(), v[0]->max(), MistralSet::empt);
  dom[1].init(v[1]->min(), v[1]->max(), MistralSet::empt);
}

PredicateNegation::~PredicateNegation()
{
}

int PredicateNegation::check( const int* s ) const 
{
  return( s[0] != -s[1] );
}

bool PredicateNegation::propagate(VariableInt *v, const int e)
{
  dom[0].clear();
  dom[1].clear();

  int i = (scope[0] == v);
  scope[1-i]->unionTo( dom[1-i] );
  dom[1-i].negate( dom[i] );
  return scope[i]->setDomain( dom[i] );
}

void PredicateNegation::print(std::ostream& o) const 
{
  o << "-" ;
  scope[0]->printshort( o );
  o << " = ";
  scope[1]->printshort( o );
}


/**********************************************
 * Nested Not Predicate
 **********************************************/
/// not(x0)
PredicateNot::PredicateNot(Solver *s, VariableInt** v) 
  : Constraint(s, v, 2, DOMAINTRIGGER, WEIGHTTWO)
{
}

PredicateNot::~PredicateNot()
{
}

int PredicateNot::check( const int* s ) const 
{
  return( !(s[0]) != s[1] );
}

bool PredicateNot::propagate()
{
  bool consistent = true;
  for( int i=0; consistent && i<2; ++ i )
    consistent = ( (!scope[i]->min() || scope[1-i]->setDomain( 0 )) &&
		   (scope[i]->max() || scope[1-i]->remove( 0 ) ) );
  return consistent;
}

bool PredicateNot::propagate(VariableInt *v, const int e)
{
  bool consistent = true;
  int i = (v == scope[0]);
  consistent = ( (!v->min() || scope[i]->setDomain( 0 )) &&
		 (v->max() || scope[i]->remove( 0 ) ) );
//   if( v == scope[0] ) {
//     if( !v->contain( 0 ) ) consistent = scope[1]->setDomain( 0 );
//     else if( v->isGround() ) consistent = scope[1]->setDomain( 1 );
//   } else if( scope[1]->min() ) {
//     consistent = scope[0]->setDomain( 0 );
//   } else {
//     consistent = scope[0]->setDomain( 1 );
//   }
  return consistent;
}

void PredicateNot::print(std::ostream& o) const 
{
  o << "!" ;
  scope[0]->printshort( o );
  o << " = ";
  scope[1]->printshort( o );
}


/**********************************************
 * Nested Equal Predicate
 **********************************************/
/// x0 =?= x1

PredicateEqual::PredicateEqual(Solver *s, VariableInt** v, int sp) 
  : Constraint(s, v, 3, DOMAINTRIGGER, WEIGHTTWO)
{
  spin = sp;
}

PredicateEqual::~PredicateEqual()
{
}

int PredicateEqual::check( const int* s ) const 
{
  return( (s[0] == s[1]) == (s[2] ^ spin) );
}

bool PredicateEqual::propagate(VariableInt *v, const int e)
{

  bool consistent = true;

  if( scope[2]->isGround() ) {
    if( (spin + scope[2]->min()) != 1 ) {
      consistent = ( scope[0]->setDomain(scope[1]) &&
	      scope[1]->setDomain(scope[0]) );
    } else {
      consistent = ( (!scope[0]->isGround() || scope[1]->remove(scope[0]->value()))
	      &&
	      (!scope[1]->isGround() || scope[0]->remove(scope[1]->value())) );
    }
  } else {
    if( !(scope[0]->intersect(scope[1])) ) {
      consistent = ( scope[2]->remove( spin ) );
    } else { 
      if( scope[0]->isGround() && scope[1]->isGround() ) {
	consistent = ( scope[2]->setDomain( spin ) );
      }
    }
  }

  return consistent;
}

void PredicateEqual::print(std::ostream& o) const 
{
  o << "(";
  scope[0]->printshort( o );
  if( spin )
    o << " == ";
  else
    o << " =/= ";
  scope[1]->printshort( o );
  o << ") = ";
  scope[2]->printshort( o );
}


/**********************************************
 * Nested EqualConstant Predicate
 **********************************************/
/// x0 =?= k

PredicateEqualConstant::PredicateEqualConstant(Solver *s, 
					       VariableInt** v, 
					       const int k,
					       const int sp) 
  : Constraint(s, v, 2, DOMAINTRIGGER, WEIGHTTHREE)
{
  K = k;
  spin = sp;
}

PredicateEqualConstant::~PredicateEqualConstant()
{
}

int PredicateEqualConstant::check( const int* s ) const 
{
  return( (s[0] == K) == (s[1] ^ spin) );
}

bool PredicateEqualConstant::propagate(VariableInt *v, const int e)
{
  bool consistent = true;

  if( scope[1]->isGround() ) {
    if( (spin + scope[1]->min()) != 1 ) {
      consistent = scope[0]->setDomain(K);
    } else {
      consistent = scope[0]->remove(K);
    }
  } else {
    if( !(scope[0]->contain( K )) ) {
      consistent = scope[1]->remove( spin );
    } else { 
      if( scope[0]->isGround() ) {
	consistent = scope[1]->setDomain( spin );
      }
    }
  }

  return consistent;
}

void PredicateEqualConstant::print(std::ostream& o) const 
{
  o << "(";
  scope[0]->printshort( o );
  if( spin )
    o << " == ";
  else
    o << " =/= ";
  o << K << ") = ";
  scope[1]->printshort( o );
}


/**********************************************
 * Nested And Predicate
 **********************************************/
/// x0 &?& x1
PredicateAnd::PredicateAnd(Solver *s, VariableInt** v) 
  : Constraint(s, v, 3, DOMAINTRIGGER, WEIGHTTHREE)
{
}

PredicateAnd::~PredicateAnd()
{
}

int PredicateAnd::check( const int* s ) const 
{
  return( (s[0] && s[1]) != s[2] );
}

bool PredicateAnd::propagate(VariableInt *v, const int e)
{
  if( scope[2]->min() ) {
    if( !scope[0]->setDomain( 1 ) ||
	!scope[1]->setDomain( 1 ) )
      return false;
  } else if( !(scope[2]->max()) ) {
    if( (scope[0]->min() && !scope[1]->setDomain( 0 ))
	||
	(scope[1]->min() && !scope[0]->setDomain( 0 )) )
      return false;
  }
  if( (!scope[0]->max() || !scope[1]->max()) &&
      !scope[2]->setDomain( 0 ) ) return false;
  else if( scope[0]->min() && scope[1]->min() &&
	   !scope[2]->setDomain( 1 ) ) return false;
  return true;
}

void PredicateAnd::print(std::ostream& o) const 
{
  o << "(";
  scope[0]->printshort( o );
  o << " && " ;
  scope[1]->printshort( o );
  o << ") = ";
  scope[2]->printshort( o );
}


/**********************************************
 * Nested Or Predicate
 **********************************************/
/// x0 &?& x1
PredicateOr::PredicateOr(Solver *s, VariableInt** v) 
  : Constraint(s, v, 3)
{
}

PredicateOr::~PredicateOr()
{
}

int PredicateOr::check( const int* s ) const 
{
  return( (s[0] || s[1]) != s[2] );
}

bool PredicateOr::propagate(VariableInt *v, const int e)
{
  bool consistent = true;
  if( !scope[2]->max() ) {
    consistent = ( scope[0]->setDomain( 0 ) &&
		   scope[1]->setDomain( 0 ) );
  } else if( scope[2]->min() ) {
    consistent = ( ( scope[0]->max() || scope[1]->setDomain( 1 ) )
		   &&
		   ( scope[1]->max() || scope[0]->setDomain( 1 ) ) );
  }
  if( consistent ) {
    if( scope[0]->min() || scope[1]->min() )
      consistent = scope[2]->setDomain( 1 );
    else if( !scope[0]->max() && !scope[1]->max() )
      consistent = scope[2]->setDomain( 0 );
  }
  return consistent;
}

void PredicateOr::print(std::ostream& o) const 
{
  o << "(";
  scope[0]->printshort( o );
  o << " || " ;
  scope[1]->printshort( o );
  o << ") = ";
  scope[2]->printshort( o );
}


/**********************************************
 * IfThenElse Predicate
 **********************************************/

PredicateIfThenElse::PredicateIfThenElse(Solver *s, VariableInt** v)
  : Constraint(s, v, 4, DOMAINTRIGGER)
{
}

PredicateIfThenElse::~PredicateIfThenElse() {}

int PredicateIfThenElse::check( const int* s ) const 
{
  return ( (s[0] ? s[1] : s[2]) != s[3] );
}

bool PredicateIfThenElse::propagate(VariableInt *v, const int e)
{
  bool consistent = true;

//   print( cout );
//   cout << endl;

  if( !scope[3]->intersect( scope[1] ) )
    consistent = scope[0]->remove(1);
  else if( !scope[3]->intersect( scope[2] ) )
    consistent = scope[0]->remove(0);
      
  if( scope[0]->min() )
    consistent = (scope[3]->setDomain( scope[1] ) && scope[1]->setDomain( scope[3] ));
  else if( !scope[0]->max() )
    consistent = (scope[3]->setDomain( scope[2] ) && scope[2]->setDomain( scope[3] ));

//   print( cout );
//   cout << " " << consistent << endl << endl;

  return consistent;
}

void PredicateIfThenElse::print(std::ostream& o) const
{
  scope[3]->print( o ) ;
  o << " =";
  o << "(" ;
  scope[0]->print( o ) ;
  o << " ? " ;
  scope[1]->print( o ) ;
  o << " : " ;
  scope[2]->print( o ) ;
  o << ")";
}

/**********************************************
 * <= Predicate 
 **********************************************/
/// (x0 + k < x1)
PredicateLess::PredicateLess(Solver *s, VariableInt** v, const int offs) 
  : Constraint(s, v, 3, RANGETRIGGER, WEIGHTTWO)
{
  offset = offs;
}

PredicateLess::~PredicateLess()
{
}

int PredicateLess::check( const int* s ) const 
{
  return ( (s[0] + offset <= s[1]) != s[2] );
}

bool PredicateLess::propagate(VariableInt *v, const int e) 
{
//   if( scope[2]->domsize() < 2 ) {
//     if( scope[2]->min() ) {
//       return scope[1]->setMin(scope[0]->min() + offset) &&
// 	scope[0]->setMax(scope[1]->max() - offset);
//     } else {
//       return scope[0]->setMin(scope[1]->min() - offset) &&
// 	scope[1]->setMax(scope[0]->max() + offset);
//     }
//   } else {
//     if( scope[0]->max() + offset <= scope[1]->min() ) 
//       return ( scope[2]->remove( 0 ) );
//     else if( scope[0]->min() + offset > scope[1]->max() )
//       return ( scope[2]->remove( 1 ) );
//   }
//   return true;

  bool consistent = true;

//   print( cout );
//   cout << endl;

  if( scope[2]->domsize() < 2 ) {
    if( scope[2]->min() ) {
      consistent = (scope[1]->setMin(scope[0]->min() + offset) &&
		    scope[0]->setMax(scope[1]->max() - offset));
    } else {
//       scope[0]->print( cout );
//       cout << " >= " << (scope[1]->min() - offset) << endl;

//       scope[1]->print( cout );
//       cout << " <= " << (scope[0]->max() + offset) << endl;

      consistent = (scope[0]->setMin(scope[1]->min() - offset + 1) &&
		    scope[1]->setMax(scope[0]->max() + offset - 1));
    }
  } else {

//     cout << (scope[0]->max() + offset) << " | " << scope[1]->min() << endl;
//     cout << (scope[0]->min() + offset) << " | " << scope[1]->max() << endl;

    if( scope[0]->max() + offset <= scope[1]->min() ) 
      consistent = ( scope[2]->remove( 0 ) );
    else if( scope[0]->min() + offset > scope[1]->max() )
      consistent = ( scope[2]->remove( 1 ) );
  }

//   print( cout );
//   cout << " " << consistent << endl << endl;

  return consistent;

}

void PredicateLess::print(std::ostream& o) const
{  
//   o << "(";
//   scope[0]->printshort( o );
//   if( !offset )
//     o << " <= ";
//   else if( offset == 1 )
//     o << " < ";
//   else 
//     o << " + " << offset-1 << " < " ;
//   scope[1]->printshort( o );
//   o << ") = ";
//   scope[2]->printshort( o );

  o << "(";
  scope[0]->print( o );
  if( !offset )
    o << " <= ";
  else if( offset == 1 )
    o << " < ";
  else 
    o << " + " << offset-1 << " < " ;
  scope[1]->print( o );
  o << ") = ";
  scope[2]->print( o );
}


/**********************************************
 * >= k Predicate 
 **********************************************/
/// (x0 + k < x1)
PredicateLowerBound::PredicateLowerBound(Solver *s, VariableInt** v, const int b) 
  : Constraint(s, v, 2, RANGETRIGGER, WEIGHTTHREE)
{
  bound = b;
}

PredicateLowerBound::~PredicateLowerBound()
{
}

int PredicateLowerBound::check( const int* s ) const 
{
  return ( (s[0] >= bound) != s[1] );
}

bool PredicateLowerBound::propagate(VariableInt *v, const int e) 
{
  bool consistent = true;

//   print( cout );
//   cout << endl;

  if( scope[1]->domsize() < 2 ) {
    if( scope[1]->min() ) 
      consistent = scope[0]->setMin(bound);
    else 
      consistent = scope[0]->setMax(bound-1);    
  } else {
    if( scope[0]->max() < bound  ) 
      consistent =  scope[1]->remove( 1 );
    else if( scope[0]->min() >= bound )
      consistent = scope[1]->remove( 0 );
  }

//   print( cout );
//   cout << ( consistent ? " OK" : "INCONSISTENT" ) << endl << endl;

  return consistent;
}

void PredicateLowerBound::print(std::ostream& o) const
{  
//   o << "(";
//   scope[0]->printshort( o );
//   o << " > " << (bound-1) << ") = ";
//   scope[1]->printshort( o );

  o << "(";
  scope[0]->print( o );
  o << " > " << (bound-1) << ") = ";
  scope[1]->print( o );
}


/**********************************************
 * >= k Predicate 
 **********************************************/
/// (x0 + k < x1)
PredicateUpperBound::PredicateUpperBound(Solver *s, VariableInt** v, const int b) 
  : Constraint(s, v, 2, RANGETRIGGER, WEIGHTTHREE)
{
  bound = b;
}

PredicateUpperBound::~PredicateUpperBound()
{
}

int PredicateUpperBound::check( const int* s ) const 
{
  return ( (s[0] <= bound) != s[1] );
}

bool PredicateUpperBound::propagate(VariableInt *v, const int e) 
{
  bool consistent = true;
  if( scope[1]->domsize() < 2 ) {
    if( scope[1]->min() ) 
      consistent = scope[0]->setMax(bound);
    else 
      consistent = scope[0]->setMin(bound+1);    
  } else {
    if( scope[0]->max() <= bound  ) 
      consistent =  scope[1]->remove( 0 );
    else if( scope[0]->min() > bound )
      consistent = scope[1]->remove( 1 );
  }
  return consistent;
}

void PredicateUpperBound::print(std::ostream& o) const
{  
  o << "(";
  scope[0]->printshort( o );
  o << " < " << (bound+1) << ") = ";
  scope[1]->printshort( o );
}


/**********************************************
 * Add Predicate 
 **********************************************/
/// (x0 + x1)
PredicateAdd::PredicateAdd(Solver *s, VariableInt** v) 
  : Constraint(s, v, 3, DOMAINTRIGGER, WEIGHTTHREE)
{
  int lb = std::min(v[2]->min(), std::min(v[0]->min(), v[1]->min()));
  int ub = std::max(v[2]->max(), std::max(v[0]->max(), v[1]->max()));

  //cout << lb << " " << ub << endl;
  if( ub - lb < 1024 ) {
    xdom.init(lb, ub, MistralSet::empt);
    tdom.init(lb, ub, MistralSet::empt);
  }
}

PredicateAdd::~PredicateAdd()
{
}

int PredicateAdd::check( const int* s ) const 
{
  return ( (s[0] + s[1]) != s[2] );
}

bool PredicateAdd::propagate(VariableInt *v, const int e) 
{
  int i;
  bool consistent = true, hole =false;
  
  for(i=0; i<3 && !hole; ++i)
    if( !scope[i]->isRange() )
      hole = true;

  if( hole && xdom.table ) {
    int vali, vnxt;
    
    /// revise d(scope[2])
    if( v != scope[2] && consistent ) {
      tdom.clear();
      i = ( scope[0]->domsize() > scope[1]->domsize() );
      vali = scope[i]->first();
      do {
	vnxt = scope[i]->getNext( vali );
	xdom.clear();
	scope[1-i]->unionTo( xdom );
	if( vali < 0 )
	  xdom.decrement( -vali );
	else
	  xdom.increment( vali );
	tdom.unionWith( xdom );
	vali = vnxt;
      } while( scope[i]->good( vali ) );
      consistent = scope[2]->setDomain( tdom );
    }
    
    /// revise d(scope[0])
    if( v != scope[0] && consistent ) {  
      tdom.clear();
      vali = scope[1]->first();
      do {
	vnxt = scope[1]->getNext( vali );
	xdom.clear();
	scope[2]->unionTo( xdom );
	if( vali < 0 )
	  xdom.increment( -vali );
	else
	  xdom.decrement( vali );
	tdom.unionWith( xdom );
	vali = vnxt;
      } while( scope[1]->good( vali ) );
      consistent = scope[0]->setDomain( tdom );
    }

    /// revise d(scope[1])
    if( v != scope[1] && consistent ) {  
      tdom.clear();
      vali = scope[0]->first();
      do {
	vnxt = scope[0]->getNext( vali );
	xdom.clear();
	scope[2]->unionTo( xdom );
	if( vali < 0 )
	  xdom.increment( -vali );
	else
	  xdom.decrement( vali );
	tdom.unionWith( xdom );
	vali = vnxt;
      } while( scope[0]->good( vali ) );
      consistent = scope[1]->setDomain( tdom );
    }

  } else {

    int lb = (scope[0]->min() + scope[1]->min());
    int ub = (scope[0]->max() + scope[1]->max());

    consistent = (scope[2]->setMin( lb ) &&
		  scope[2]->setMax( ub ) &&
   
		  scope[0]->setMin( scope[2]->min() - scope[1]->max() ) &&
		  scope[0]->setMax( scope[2]->max() - scope[1]->min() ) &&
		 
		  scope[1]->setMin( scope[2]->min() - scope[0]->max() ) &&
		  scope[1]->setMax( scope[2]->max() - scope[0]->min() ) );
    
  }

  return consistent;
}

void PredicateAdd::print(std::ostream& o) const
{  
  scope[0]->printshort( o );
  o << " + ";
  scope[1]->printshort( o );
  o << " = ";
  scope[2]->printshort( o );
}


/**********************************************
 * Sub Predicate 
 **********************************************/
/// (x0 - x1)
PredicateSub::PredicateSub(Solver *s, VariableInt** v) 
  : Constraint(s, v, 3, DOMAINTRIGGER, WEIGHTTHREE)
{
  int lb = std::min(scope[2]->min(), std::min(v[0]->min(), v[1]->min()));
  int ub = std::max(scope[2]->max(), std::max(v[0]->max(), v[1]->max()));

  if( ub - lb < 1024 ) {
    xdom.init(lb, ub, MistralSet::empt);
    tdom.init(lb, ub, MistralSet::empt);
  }
//   weight = 3;
//   scope[0]->weight+=2;
//   scope[1]->weight+=2;
//   scope[2]->weight+=2;

}

PredicateSub::~PredicateSub()
{
}

int PredicateSub::check( const int* s ) const 
{
  return ( (s[0] - s[1]) != s[2] );
}

bool PredicateSub::propagate(VariableInt *v, const int e) 
{
  int i;
  bool consistent = true, hole =false;

  for(i=0; i<3 && !hole; ++i)
    hole = !(scope[i]->isRange()); 

  if( hole && xdom.table ) {
    int vali, vnxt;
    
    /// revise d(scope[2])
    if( v != scope[2] && consistent ) {
      tdom.clear();
      vali = scope[1]->first();
      do {
	vnxt = scope[1]->getNext( vali );
	xdom.clear();
	scope[0]->unionTo( xdom );
	if( vali < 0 ) 
	  xdom.increment( -vali );
	else 
	  xdom.decrement( vali );
	tdom.unionWith( xdom );
	vali = vnxt;
      } while( scope[1]->good( vali ) );
      consistent = scope[2]->setDomain( tdom );
    }
    
    /// revise d(scope[0])
    if( v != scope[0] && consistent ) {  
      tdom.clear();
      vali = scope[1]->first();
      do {
	vnxt = scope[1]->getNext( vali );
	xdom.clear();
	scope[2]->unionTo( xdom );
	if( vali < 0 )
	  xdom.decrement( -vali );
	else
	  xdom.increment( vali );
	tdom.unionWith( xdom );
	vali = vnxt;
      } while( scope[1]->good( vali ) );
      consistent = scope[0]->setDomain( tdom );
    }

    /// revise d(scope[1])
    if( v != scope[1] && consistent ) {  
      tdom.clear();
      vali = scope[2]->first();
      do {
	vnxt = scope[2]->getNext( vali );
	xdom.clear();
	scope[0]->unionTo( xdom );
	if( vali < 0 )
	  xdom.increment( -vali );
	else
	  xdom.decrement( vali );
	tdom.unionWith( xdom );
	vali = vnxt;
      } while( scope[2]->good( vali ) );
      consistent = scope[1]->setDomain( tdom );
    }
  } else {

    int lb = (scope[0]->min() - scope[1]->max());
    int ub = (scope[0]->max() - scope[1]->min());

    consistent = (scope[2]->setMin( lb ) &&
		  scope[2]->setMax( ub ) &&
   
		  scope[0]->setMin( scope[2]->min() + scope[1]->min() ) &&
		  scope[0]->setMax( scope[2]->max() + scope[1]->max() ) &&
		 
		  scope[1]->setMin( scope[0]->min() - scope[2]->max() ) &&
		  scope[1]->setMax( scope[0]->max() - scope[2]->min() ) );

  }
  
  return consistent;
}

void PredicateSub::print(std::ostream& o) const
{  
  scope[0]->printshort( o );
  o << " - ";
  scope[1]->printshort( o );
  o << " = ";
  scope[2]->printshort( o );
}


/**********************************************
 * Mul Predicate 
 **********************************************/
/// (x0 * x1)
PredicateMul::PredicateMul(Solver *s, VariableInt** v) 
  : Constraint( s, v, 3 )
{
  int lb = std::min(v[2]->min(), std::min(v[0]->min(), v[1]->min()));
  int ub = std::max(v[2]->max(), std::max(v[0]->max(), v[1]->max()));

  if( ub - lb < 1024 ) {
    xdom.init(lb, ub, MistralSet::empt);
  }
  //tdom.init(lb, ub);
}

PredicateMul::~PredicateMul()
{
}

int PredicateMul::check( const int* s ) const 
{
  return ( (s[0] * s[1]) != s[2] );
}

bool PredicateMul::propagate(VariableInt *v, const int e) 
{
  // NOT GAC !!
  int x[4], lb=0, ub=0, i, j, aux, vali;
  x[0] = x[1] = x[2] = x[3] = INT_MAX;



//   cout << endl;
//   print( cout );
//   cout << endl;
//   v->print( cout );
//   cout << endl;


  /// first, the particular cases:
  /*************************************************
   * Case 1: 2 variables are ground 
   * (so that we are sur to achieve FC)
   *************************************************/
  // x0 and x1 are ground
  if( scope[0]->isGround() && scope[1]->isGround() )
    return scope[2]->setDomain( scope[0]->value() * scope[1]->value() );

  // x2 and either x0 or x1 are ground
  if( scope[2]->isGround() ) {
    j = 2;
    while( j-- ) {
      if( scope[1-j]->isGround() )
	if( !scope[1-j]->value() ) { return scope[2]->contain( 0 ); }
	else {
	  aux = scope[2]->value() / scope[1-j]->value();
	  return( //!(value() % aux) && 
		 (scope[1-j]->value() * aux == scope[2]->value()) &&
		 scope[j]->setDomain( aux ) );
	}
    }
  }

  
  // get rid of the 0's if possible
  if( !scope[2]->contain( 0 ) ) 
    if( !scope[0]->remove( 0 ) || !scope[1]->remove( 0 ) )
      { return false; }

 
  /*************************************************
   * Case 2: 1 variable is ground,  
   * we try to achieve GAC
   *************************************************/
  if( v->isGround() ) {

    // the product is not ground
    if( v != scope[2] ) {

      if( !v->min() )
	return scope[2]->setDomain( 0 );            

      // update the domain of the product
      i = (scope[0] == v);
      aux = scope[1-i]->min();
      
      lb = aux * (aux < 0 ? scope[i]->max() : scope[i]->min());
      ub = aux * (aux < 0 ? scope[i]->min() : scope[i]->max());
      
      if( !(scope[2]->setMin(lb) && scope[2]->setMax(ub)) ) return false;

      if( !scope[i]->isRange() && xdom.table )
	{
	  xdom.clear();
	  vali = scope[i]->first();
	  do {
	    xdom.insert( vali*aux );
	    scope[i]->setNext( vali );
	  } while( vali != INT_MAX ); 

	  if( !scope[2]->setDomain( xdom ) ) return false;
	}

      // update the domain of non ground factor
      lb = (aux < 0 ? (scope[2]->max() / aux) : (scope[2]->min() / aux));
      ub = (aux < 0 ? (scope[2]->min() / aux) : (scope[2]->max() / aux));
      
      if( !(scope[i]->setMin(lb) && scope[i]->setMax(ub)) ) return false;

      if( !scope[2]->isRange() && xdom.table )
	{
	  xdom.clear();
	  vali = scope[2]->first();
	  do {
	    xdom.insert( vali/aux );
	    scope[2]->setNext( vali );
	  } while( vali != INT_MAX ); 

	  if( !scope[i]->setDomain( xdom ) ) return false;
	}

    } else {
      // the product is ground
      // update the domain of the factors
      aux = scope[2]->min();      
      
      if( !aux ) 
	return( (scope[0]->contain(0) || scope[1]->setDomain(0)) &&
		(scope[1]->contain(0) || scope[0]->setDomain(0))
		) ;

      j=2;
      while( j-- ) {
	i = 1-j;
	lb = (aux < 0 ? (aux / scope[j]->max()) : (aux / scope[j]->min()));
	ub = (aux < 0 ? (aux / scope[j]->min()) : (aux / scope[j]->max()));
	  
	if( !(scope[i]->setMin(lb) && scope[i]->setMax(ub)) ) return false;	
	
	if( !scope[j]->isRange() && xdom.table )
	{
	  xdom.clear();
	  vali = scope[j]->first();
	  do {
	    xdom.insert( aux/vali );
	    scope[j]->setNext( vali );
	  } while( vali != INT_MAX ); 

	  if( !scope[i]->setDomain( xdom ) ) return false;
	}

      }
    }
    

    /*************************************************
     * Case 2: no variable is ground,  
     * we achieve BC
     *************************************************/
  } else {
    
    // update the domain of the product
    if( v != scope[2] ) {
      
      i=4;
      lb=INT_MAX;
      ub=-1*INT_MAX;
      
      x[0] = scope[0]->min() * scope[1]->min();
      x[1] = scope[0]->min() * scope[1]->max();
      x[2] = scope[0]->max() * scope[1]->min();
      x[3] = scope[0]->max() * scope[1]->max();
      
      while( i-- ) {
	if( x[i] > ub ) ub = x[i];
	if( x[i] < lb ) lb = x[i];
      }
      
      if( scope[2]->min() > lb ) {
	j = 2;
	while( j-- )
	  if( scope[1-j]->isGround() ) {
	    aux = scope[1-j]->value();
	    i = (scope[2]->min() % aux);
	    lb += (aux - i);
	  }
      }
      
      if( scope[2]->max() < ub ) {
	j = 2;
	while( j-- )
	  if( scope[1-j]->isGround() ) {
	    aux = scope[1-j]->value();
	    i = (scope[2]->max() % aux);
	    ub -= i;
	  }
      }
      
      if( !(scope[2]->setMin( lb )) ) { return false; }
      if( !(scope[2]->setMax( ub )) ) { return false; }
    }
    
    // update the domain of the factors    
    j=2;
    while( j-- ) {
      if( v != scope[1-j] ) {
	
	if( scope[2]->contain(0) && scope[j]->contain(0) )
	  continue;
	
	i=4;
	lb=INT_MAX;
	ub=-1*INT_MAX;
	
	aux = scope[j]->min();
	x[0] = scope[2]->min() / aux;
	x[1] = scope[2]->max() / aux;
	
	aux = scope[j]->max();
	x[2] = scope[2]->min() / aux;
	x[3] = scope[2]->max() / aux;
       

	while( i-- )
	  if( x[i] > ub ) ub = x[i];
	  else if( x[i] < lb ) lb = x[i];
	
	if( !lb && !scope[2]->contain(0) ) ++lb;
	if( !ub && !scope[2]->contain(0) ) --ub;


	if( !(scope[1-j]->setMin( lb )) ) { return false; }
	if( !(scope[1-j]->setMax( ub )) ) { return false; }
      }
    }
  }
    
  return true;
}

void PredicateMul::print(std::ostream& o) const
{  
  o << "(";
  scope[0]->print( o );
  o << " * ";
  scope[1]->print( o );
  o << ") = ";
  scope[2]->print( o );
}


/**********************************************
 * Div Predicate 
 **********************************************/
/// (x0 / x1)
PredicateDiv::PredicateDiv(Solver *s, VariableInt** v) 
  : Constraint(s, v, 3)
{
}

PredicateDiv::~PredicateDiv()
{
}

int PredicateDiv::check( const int* s ) const 
{
  return ( (s[0] / s[1]) != s[2] );
}

bool PredicateDiv::propagate(VariableInt *v, const int e) 
{
  // NOT GAC !!
  if( scope[0]->isGround() && 
      !(scope[0]->max()) )
    if( !(scope[2]->setDomain( 0 )) ) return false;
    else return true;

  if( !(scope[1]->remove( 0 ) ) ) return false;

  if( scope[0]->isGround() && scope[1]->isGround() )
    {      
      if( !(scope[2]->setDomain( (scope[0]->min() / scope[1]->min()) )) )
	return false;
      else return true;
    }

  if( scope[1]->isGround() && scope[2]->isGround() )
    {
      if( !scope[2]->max() ) {
	if( scope[1]->min() < 0 ) {
	  if( !(scope[0]->setMin( scope[1]->min()+1 )) ||
	      !(scope[0]->setMax( -1*scope[1]->min()-1 )) )
	    return false;
	} else if( !(scope[0]->setMax( scope[1]->min()-1 )) ||
		   !(scope[0]->setMin( -1*scope[1]->min()+1 )) )
	  return false;
      } else {
	int aux, ub, lb = (scope[1]->min() * scope[2]->min());
	
	aux = ( scope[1]->min() > 0 ? scope[1]->min() : -1*scope[1]->min() );
	--aux;
	if( lb < 0 ) {
	  ub = lb;
	  lb = lb - aux;
	} else ub = lb + aux;
	if( !(scope[0]->setMin( lb )) ||
	    !(scope[0]->setMax( ub )) )
	  return false;
	else return true;
      }
    }

  if( scope[0]->isGround() && scope[2]->isGround() )
    {
      if( !scope[2]->max() ) {
	if( scope[0]->min() < 0 ) {
	  if( !(scope[1]->removeRange( scope[0]->min(),
				       -1*scope[0]->min() )) )
	    return false;
	} else if( !(scope[1]->removeRange( -1*scope[0]->min(),
					    scope[0]->min() )) )
	  return false;
      } else {
	int aux, lb = 0, ub = (scope[0]->min() / scope[2]->min());
	if( scope[2]->min() > 0 )
	  lb = (scope[0]->min() / (scope[2]->min()+1));
	else 
	  lb = (scope[0]->min() / (scope[2]->min()-1));
	if( ub > 0 ) ++lb;
	else --lb;
	
	if( lb > ub ) {
	  aux = lb;
	  lb = ub;
	  ub = aux;
	}
	
	if( !(scope[1]->setMin( lb )) ||
	    !(scope[1]->setMax( ub )) )
	  return false;
	else return true;
      }
    }
  return true;
}

void PredicateDiv::print(std::ostream& o) const
{  
  o << "(";
  scope[0]->printshort( o );
  o << " / ";
  scope[1]->printshort( o );
  o << ") = ";
  scope[2]->printshort( o );
}


/**********************************************
 * Mod Predicate 
 **********************************************/
/// (x0 % x1)
PredicateMod::PredicateMod(Solver *s, VariableInt** v) 
  : Constraint(s, v, 3)
{
}

PredicateMod::~PredicateMod()
{
}

int PredicateMod::check( const int* s ) const 
{
  return ( (s[0] % s[1]) != s[2] );
}

bool PredicateMod::propagate(VariableInt *v, const int e) 
{
//   if( scope[0]->id == 0 && scope[2]->id == 99 ) {
//   cout << endl << "(";
//   if( v == scope[0] ) cout << "*";
//   scope[0]->print( cout );
//   cout << " % ";
//   if( v == scope[1] ) cout << "*";
//   scope[1]->print( cout );
//   cout << ") = ";
//   if( v == scope[2] ) cout << "*";
//   scope[2]->print( cout );
//   cout << endl;
//   }

  
  bool consistent = true;
  int ub, lb, modulo, target, k;  
  if( v != scope[0] ) { // prune x0
    if( scope[1]->isGround() ) { // modulo is known
      modulo = scope[1]->min();
      if( modulo == 1 ) // special case
	consistent = scope[0]->setDomain(0);
      else if( scope[2]->isGround() ) { // target is known
	target = scope[2]->min();
	consistent = (target < modulo);
      
	// positive/negative target
	if( target < 0 )
	  consistent = scope[0]->setMax(0);
	else if( target > 0 )
	  consistent = scope[0]->setMin(0);
	else 
	  consistent = scope[0]->setDomain(0);

	if( consistent ) {
	  // remove intervals [target+1+k*modulo..target+k*(modulo+1)-1]
	  k = (scope[0]->max()-target-1)/modulo;
	  while( consistent ) {
	    lb = (target+1+k*modulo);
	    ub = (target+(k+1)*modulo-1);
	    if( ub < scope[0]->min() ) break;
	    consistent = scope[0]->removeRange( lb, ub );
	    --k;
	  }
	}      
      } else {
	// prune x0 with respect to x2
	
	ValIterator *valit = scope[0]->getDomainIterator();
	do {
	k = (*valit % modulo);
	consistent = ( scope[2]->contain( k ) || scope[0]->remove( *valit ) );
	} while( consistent && valit->next() );
	
      }
    } else {
      // modulo is not known we want to prune x0

      //cout << 11 << endl;      

    }
  } else if( v != scope[2] ) {
    // prune x2

    if( scope[1]->isGround() ) { // modulo is known

      if( scope[0]->isGround() )
	
	consistent = scope[2]->setDomain( (scope[0]->min() % scope[1]->min()) );

      else {
      
	modulo = scope[1]->min();
	ub = scope[0]->max();
	if( ub > 0 && modulo <= ub )
	  ub = modulo-1;
	else ub = 0;
	lb = scope[0]->min();
	if( lb < 0 && 1-modulo > lb )
	  lb = 1-modulo;
	else lb = 0;
	
	//cout << lb << " .. " << ub << endl;
	
	consistent = scope[2]->setMax( ub ) && scope[2]->setMin( lb );
	
	ValIterator *valit = scope[2]->getDomainIterator();
	do {
	  k = *valit;
	  //cout << k << ": ";
	  if( k > 0 ) {
	    lb = (scope[0]->min()/modulo)*modulo;
	    k = std::max( k, lb+k );
	    //cout << k << "? " ;
	    while( !scope[0]->contain(k) && k <= scope[0]->max() )  {
	      k+=modulo;
	      //cout << k << "? " ;
	    }
	    if(k && k > scope[0]->max()) {
	      //cout << " remove ";
	      consistent = scope[2]->remove( *valit );
	    } //else cout << " ok ";
	  }
	  else {
	    ub = (scope[0]->max()/modulo)*modulo;
	    if(k) k = std::min( k, ub+k );
	    else k = ub;
	    //cout << k << "? " ;
	    while( !scope[0]->contain(k) && k >= scope[0]->min() ) {
	      k-=modulo;
	      //cout << k << "? " ;
	    }
	    if(k < scope[0]->min()) {
	      //cout << " remove ";
	      consistent = scope[2]->remove( *valit );
	    } //else cout << " ok ";
	  }
	  
	  // 	cout << endl;
	  // 	scope[2]->print( cout );
	  // 	cout << endl;
	  
	} while( consistent && valit->next() );
      } 
    } else {
      // modulo is not known we want to prune x2
      
      //cout << 22 << endl;
      
    }
  } //else if( v != scope[1] ) {
    
  //if( scope[0]->isGround() && scope[2]->isGround() )
      
  //}

  //  if( scope[0]->id == 0 && scope[2]->id == 99 ) {
//   if( consistent ) {
//     cout << "(";
//     scope[0]->print( cout );
//     cout << " % ";
//     scope[1]->print( cout );
//     cout << ") = ";
//     scope[2]->print( cout );
//     cout << endl;
//   } else {
//     cout << "INCONSISTENT" << endl;
//   }
//  }

  return consistent;

}

void PredicateMod::print(std::ostream& o) const
{  
  o << "(";
  scope[0]->printshort( o );
  o << " % ";
  scope[1]->printshort( o );
  o << ") = ";
  scope[2]->printshort( o );
}


/**********************************************
 * ModConstant Predicate 
 **********************************************/
/// (x0 % x1)
PredicateModConstant::PredicateModConstant(Solver *s, VariableInt** v, const int mod) 
  : Constraint(s, v, 2)
{
  modulo = mod;
}

PredicateModConstant::~PredicateModConstant()
{
}

int PredicateModConstant::check( const int* s ) const 
{
  return ( (s[0] % modulo) != s[1] );
}

bool PredicateModConstant::propagate(VariableInt *v, const int e) 
{

//   cout << endl << "(";
//   if( v == scope[0] ) cout << "*";
//   scope[0]->print( cout );
//   cout << " % " << modulo << ") = ";
//   if( v == scope[1] ) cout << "*";
//   scope[1]->print( cout );
//   cout << endl;
  
  bool consistent = true;
  int lb, ub, target, k;  
  if( v != scope[0] ) { // prune x0

    if( modulo == 1 ) // special case
      consistent = scope[0]->setDomain(0);

    else if( scope[1]->isGround() ) { // target is known
      target = scope[1]->min();
      consistent = (target < modulo);
      
      // positive/negative target
      if( target < 0 )
	consistent = scope[0]->setMax(0);
      else if( target > 0 )
	consistent = scope[0]->setMin(0);
      else 
	consistent = scope[0]->setDomain(0);

      if( consistent ) {
	// remove intervals [target+1+k*modulo..target+k*(modulo+1)-1]
	k = (scope[0]->max()-target-1)/modulo;
	while( consistent ) {
	  lb = (target+1+k*modulo);
	  ub = (target+(k+1)*modulo-1);
	  if( ub < scope[0]->min() ) break;
	  consistent = scope[0]->removeRange( lb, ub );
	  --k;
	}
      }
    } else {
      // prune x0 with respect to x1

      ValIterator *valit = scope[0]->getDomainIterator();
      do {
	k = (*valit % modulo);
	consistent = ( scope[1]->contain( k ) || scope[0]->remove( *valit ) );
      } while( consistent && valit->next() );
      
    }
  } else {

    if( scope[0]->isGround() )
	consistent = scope[2]->setDomain( (scope[0]->min() % scope[1]->min()) );
    else {
      ub = scope[0]->max();
      if( ub > 0 && modulo <= ub )
	ub = modulo-1;
      else ub = 0;
      
      lb = scope[0]->min();
      if( lb < 0 && 1-modulo > lb )
	lb = 1-modulo;
      else lb = 0;
      
      consistent = scope[1]->setMax( ub ) && scope[1]->setMin( lb );
      
      
      ValIterator *valit = scope[1]->getDomainIterator();
      do {
	k = *valit;
	
	//cout << k << ": ";
	
	if( k > 0 ) {
	  
	  lb = (scope[0]->min()/modulo)*modulo;
	  k = std::max( k, lb+k );
	  
	  //cout << k << "? ";
	  while( !scope[0]->contain(k) && k <= scope[0]->max() )  {
	    k+=modulo;
	  //cout << k << "? ";
	  }
	  if(k && k > scope[0]->max()) {
	    //cout << " => remove ";
	    consistent = scope[1]->remove( *valit );
	  } //else cout << " ok ";
	}
	
	else {
	  ub = (scope[0]->max()/modulo)*modulo;
	  if(k) k = std::min( k, ub+k );
	  else k = ub;
	  
	  //cout << k << "? ";
	  while( !scope[0]->contain(k) && k >= scope[0]->min() ) {
	    k-=modulo;
	    //cout << k << "? ";
	  }
	  if(k < scope[0]->min()) {
	    //cout << " => remove ";
	    consistent = scope[1]->remove( *valit );
	  } //else cout << " ok ";
	}
	
	//cout << endl;
      } while( consistent && valit->next() );
    }

  }

//   if( consistent ) {
//     cout << "(";
//     scope[0]->print( cout );
//     cout << " % " << modulo << ") = ";
//     scope[1]->print( cout );
//     cout << endl;
//   } else {
//     cout << "INCONSISTENT" << endl;
//   }

  return consistent;

}

void PredicateModConstant::print(std::ostream& o) const
{  
  o << "(";
  scope[0]->printshort( o );
  o << " % " << modulo << ") = ";
  scope[1]->printshort( o );
}


/**********************************************
 * Pow Predicate 
 **********************************************/
/// (x0 ^ x1)
PredicatePow::PredicatePow(Solver *s, VariableInt** v) 
  : Constraint(s, v, 3)
{
  cerr << "Not implemented" << endl;
  exit( 0 );
}

PredicatePow::~PredicatePow()
{
}

int PredicatePow::check( const int* s ) const 
{
  return ( pow((double)(s[0]), (double)(s[1])) != (double)(s[2]) );
}

bool PredicatePow::propagate(VariableInt *v, const int e) 
{
  cerr << "Not implemented" << endl;
  exit( 0 );
  return true;
}

void PredicatePow::print(std::ostream& o) const
{  
  o << "(";
  scope[0]->printshort( o );
  o << " ^ ";
  scope[1]->printshort( o );
  o << ") = ";
  scope[2]->printshort( o );
}

// Variable Variable::operator^( Variable x )
// {
//   VariableInt *scp[2] = {varptr_, x.varptr_};
//   return Variable( new PredicatePow( scp ) );
// }


/**********************************************
 * Disjunctive Predicate 
 **********************************************/
/// 0 :-> (x0 + d0 < x1), 1 :-> (x1 + d1 < x0) 
PredicateDisjunctive::PredicateDisjunctive(Solver *s, VariableInt** v, const int* dur) 
  : Constraint(s, v, 3, RANGETRIGGER)
{
  duration[0] = dur[0];
  duration[1] = dur[1];
}

PredicateDisjunctive::~PredicateDisjunctive()
{
}

int PredicateDisjunctive::check( const int* s ) const 
{
  return ( (s[2] && (s[0] + duration[0] <= s[1])) 
	   ||
	   (!s[2] && (s[1] + duration[1] <= s[0])) );
}

bool PredicateDisjunctive::propagate(VariableInt *v, const int e) 
{
  if( v != scope[2] )
    for(int i=0; i<2; ++i)
      if( scope[i]->min() + duration[i] > scope[1-i]->max() 
	  && !scope[2]->setDomain( 1-i ) ) return false;

  //if( domain.state == 1 ) {
  if( !scope[2]->max() ) {
    // x + a <= y
    return ( scope[1]->setMin(scope[0]->min() + duration[0]) &&
	     scope[0]->setMax(scope[1]->max() - duration[0]) );

  } else //if(domain.state == 2) {
    if( scope[2]->min() ) {
    // y + b <= x
    return ( scope[0]->setMin(scope[1]->min() + duration[1]) &&
	     scope[1]->setMax(scope[0]->max() - duration[1]) );
    
  } else {
    int MaxIncX = scope[1]->min() + duration[1] -1;
    int MinIncX = scope[1]->max() - duration[0] +1;

    if( MaxIncX >= MinIncX && 
	!scope[0]->removeRange(MinIncX, MaxIncX) )
      return false;
    
    MaxIncX = scope[0]->min() + duration[0] -1;
    MinIncX = scope[0]->max() - duration[1] +1;

    if( MaxIncX >= MinIncX && 
	!scope[1]->removeRange(MinIncX, MaxIncX) )
      return false;
  }
  
  return true;
}

void PredicateDisjunctive::print(std::ostream& o) const
{  
  if( scope[2]->contain(0) ) {
    o << "(";
    scope[0]->printshort( o );
    if( !duration[0] )
      o << " <= ";
    else if( duration[0] == 1 )
      o << " < ";
    else 
      o << " + " << duration[0]-1 << " < " ;
    scope[1]->printshort( o );
    o << ")";
  }
  if( scope[2]->domsize() == 2 )
    o << " || ";
  if( scope[2]->contain(1) ) {
    o << "(";
    scope[1]->printshort( o );
    if( !duration[1] )
      o << " <= ";
    else if( duration[1] == 1 )
      o << " < ";
    else 
      o << " + " << duration[1]-1 << " < " ;
    scope[0]->printshort( o );
    o << ") = ";
  }
  //  o << " [" << vmin << ".." << vmax << "]";
  scope[2]->printshort( o );
}


/**********************************************
 * Element Predicate
 **********************************************/ 
/// x[N]
PredicateElement::PredicateElement(Solver *s, VariableInt** v, 
				   const int n)
  : Constraint(s, v, n)
{
//   aux_dom.init( std::min( 0, scope[n-1]->min() ), 
// 		std::max( n-1, scope[n-1]->max() ), 
// 		MistralSet::empt );
  aux_dom.init( std::min( 0, scope[n-1]->min() ), 
		std::max( n-1, scope[n-1]->max() ), 
		MistralSet::empt );
}

PredicateElement::~PredicateElement()
{
}

int PredicateElement::check( const int* s ) const 
{ 
  //return ( s[s[arity-2]] != s[arity-1] );
  return ( s[s[arity-2]-1] != s[arity-1] );
}

bool PredicateElement::propagate(VariableInt *v, const int e)
{
  int i = arity-2;
  VariableInt *N = scope[i], *V = scope[i+1];
  aux_dom.clear();
  bool consistent = true;

  /* propagation rule for N */ 
  if( v != N ) {
    while( i-- ) 
      if( V->intersect(scope[i]) )
	//aux_dom.insert(i);
	aux_dom.insert(i+1);
    consistent = ( N->setDomain(aux_dom) );
  }

  /* propagation rule for V */ 
  if( v != V && consistent ) {
    aux_dom.clear();
    i = N->first();
    //do scope[i]->unionTo(aux_dom); 
    do scope[i-1]->unionTo(aux_dom); 
    while( N->setNext(i) );
    consistent = ( V->setDomain(aux_dom) );
  }

  /* propagation rule for x[N] */ 
  if( N->isGround() ) 
//     consistent = ( scope[N->min()]->setDomain(V) &&
// 		   V->setDomain(scope[N->min()]) );
    consistent = ( scope[N->min()-1]->setDomain(V) &&
		   V->setDomain(scope[N->min()-1]) );

  return consistent;
}

void PredicateElement::print(std::ostream& o) const
{
  o << "x"
    // << scope[0]->id << ".." << scope[arity-2]->id
    << "[" ;
  scope[arity-2]->printshort( o );
  o << "] = " ;
  scope[arity-1]->printshort( o );
}


ConstraintLDS::ConstraintLDS(Solver* s)
  : Constraint( s, s->sequence, s->length )
{
  int i;
  threshold = -1;
  mindiff.init(s);
  maxdiff.init(s);
  mindiff.setValue( 0 );
  maxdiff.setValue( arity );
  maybeequal.init( s, 0, arity-1 );
  cannotequal.init( s, 0, arity-1 );
  

  ideal = new int[arity];

  for(i=0; i<arity; ++i) {
    ideal[i] = scope[i]->branch->getBest();
    if( scope[i]->isGround() ) {
      --maxdiff;
      cannotequal.erase( i );
    }
  }
}

ConstraintLDS::~ConstraintLDS()
{
  delete [] ideal;
}

int ConstraintLDS::check( const int* s ) const 
{
  int i, nd=0;
  for(i=0; i<arity; ++i)
    nd += ( s[i] != ideal[i] );
  return nd != threshold;
}

bool ConstraintLDS::propagate(VariableInt *v, const int e)
{
  bool consistent = true;
  int i, idx=v->id;

//   for(i=0; i<arity; ++i)
//     {
//       scope[i]->print( cout );
//       if( i == idx )
// 	cout << "*";
//       cout << " ";
//     }
//   cout << endl;
//   maybeequal.print( cout );
//   cout << " / ";
//   cannotequal.print( cout );
//   cout << " [" << (mindiff) << ".." << (maxdiff) << "]" << endl;


  if( maybeequal.member( idx ) )
    {
      if( !v->contain( ideal[idx] ) ) {
	maybeequal.erase( idx );
	++mindiff;
      } else if( e & Constraint::VALUETRIGGER ) {
	--maxdiff;
	cannotequal.erase( idx );
      }
    }

  if( mindiff > threshold || maxdiff < threshold ) consistent = false;
  else {
    if( maxdiff == threshold )
      {
	for(i=0; i<arity; ++i)
	  //if( !scope[i]->isGround() )
	  if( cannotequal.member( i ) )
	    consistent = scope[i]->remove( ideal[i] );
      }
    else if( mindiff == threshold )
      {
	for(i=0; i<arity; ++i)
	  //if( scope[i]->contain( ideal[i] ) )
	  if( maybeequal.member( i ) )
	    consistent = scope[i]->setDomain( ideal[i] );
      }
  }

//   maybeequal.print( cout );
//   cout << " / ";
//   cannotequal.print( cout );
//   cout << " [" << (mindiff) << ".." << (maxdiff) << "]" << endl;
//   for(i=0; i<arity; ++i)
//     {
//       scope[i]->print( cout );
//       if( i == idx )
// 	cout << "*";
//       cout << " ";
//     }
//   cout << endl << ( consistent ? "\t\t ok" : "\t\t inconsistent") << endl;

    
  return consistent;
}

void ConstraintLDS::print(std::ostream& o) const 
{
  cout << "LDS-" << threshold ;
}


ConstraintLDDS::ConstraintLDDS(Solver* s)
  : Constraint( s, s->sequence, s->length )
{
  int i, m, size=0;

  mindiff = new ReversibleInt[arity+1];
  maxdiff = new ReversibleInt[arity+1]; 
  distance_to_val = new int*[arity];
  val_to_distance = new int*[arity];

  for(i=0; i<arity; ++i) {

//     scope[i]->printshort( cout );
//     cout << "\t";
    
    m = (scope[i]->maxCapacity() - scope[i]->minCapacity() + 1);
    distance_to_val[i] = new int[scope[i]->domsize()]; 
    val_to_distance[i] = new int[m]; 
    val_to_distance[i] -= scope[i]->minCapacity(); 
    scope[i]->branch->getOrder( distance_to_val[i], val_to_distance[i] );
    size += scope[i]->domsize()-1;
    
//     for(int j=0; j<scope[i]->domsize(); ++j)
//       {
// 	cout << j << ":" << distance_to_val[i][j] << "  ";
// 	assert( val_to_distance[i][distance_to_val[i][j]] == j );
//       }
//     cout << endl;


    mindiff[i].init(s);
    maxdiff[i].init(s);
    mindiff[i].setValue( 0 );
    maxdiff[i].setValue( scope[i]->domsize()-1 );
  }

  threshold = -1;
  mindiff[arity].init(s);
  maxdiff[arity].init(s);
  mindiff[arity].setValue( 0 );
  maxdiff[arity].setValue( size );
}

ConstraintLDDS::~ConstraintLDDS()
{
  for(int i=0; i<arity; ++i) {
    delete [] distance_to_val[i];
    val_to_distance[i] += scope[i]->minCapacity();
    delete [] val_to_distance[i];
  }
  delete [] distance_to_val;
  delete [] val_to_distance;
}

int ConstraintLDDS::check( const int* s ) const {
  int i, nd=0;
  for(i=0; i<arity; ++i)
    nd += ( val_to_distance[i][s[i]] );
  return nd != threshold;
}

bool ConstraintLDDS::propagate(VariableInt *v, const int e)
{
  bool consistent = true;
  int idx=v->id, i = mindiff[idx], j = maxdiff[idx], k;

  //  if( i < 0 || i > 22 )
  //   cout << i << endl;

  //  assert( i >= 0 && i <= 22 );
  //  assert( j >= 0 && j <= 22 );

//   for(k=0; k<arity; ++k)
//     {
//       cout << "\t[" << mindiff[k] << ".." << maxdiff[k] << "]  " ;
//       scope[k]->print( cout );
//       if( k == idx )
// 	cout << "*";
//       cout << endl;
//     }
//   cout << "\t[" << mindiff[arity] << ".." << maxdiff[arity] << "]" << endl << endl;



  if( e & Constraint::VALUETRIGGER ) {

    //    assert( scope[idx]->value() >= 0 && scope[i]->value() <= 22 );

    k = val_to_distance[idx][scope[idx]->value()];

    //    assert( k >= 0 && k <= 22 );

    mindiff[idx] = k;
    maxdiff[idx] = k;
    mindiff[arity] += (k-i);
    maxdiff[arity] += (k-j);

  } else {
    
    if( !scope[idx]->fastContain( distance_to_val[idx][i] ) ) {
      k=i;

      //     assert( k >= 0 && k <= 22 );

      do { ++k;
	//cout << distance_to_val[idx][k] << " ";
      } while( !scope[idx]->fastContain( distance_to_val[idx][k] ) );
      //cout << endl;

 //      if( k < 0 || k > 22 ) {
// 	scope[idx]->print( cout );
// 	cout << endl;
// 	cout << i << ".." << j << "  " << k << endl;
//       }


      //      assert( k >= 0 && k <= 22 );

      mindiff[arity] += (k-i);
      mindiff[idx] = k;
    }
    
    if( !scope[idx]->fastContain( distance_to_val[idx][j] ) ) {
      k=j;

      //    assert( k >= 0 && k <= 22 );

      do --k;
      while( !scope[idx]->fastContain( distance_to_val[idx][k] ) );

      //    assert( k >= 0 && k <= 22 );

      maxdiff[arity] += (k-j);
      maxdiff[idx] = k;
    }

  }





  int dmax = maxdiff[arity];
  int dmin = mindiff[arity];

  if( dmin > threshold || dmax < threshold ) consistent = false;
  else {
    for(k=0; k<arity; ++k)
      {

	

	if( ( dmin - mindiff[k] + maxdiff[k]) > threshold )
	  {
	    i = (threshold - dmin + mindiff[k]);
	    j = maxdiff[k];
	    while( ++i <= j && consistent )
	      consistent = scope[k]->remove( distance_to_val[k][i] );
	  }

	if( ( dmax - maxdiff[k] + mindiff[k]) < threshold )
	  {
	    i = ( maxdiff[k] + threshold - dmax );
	    j = mindiff[k];
	    while( --i >= j && consistent )
	      consistent = scope[k]->remove( distance_to_val[k][i] );
	  }

      }
  }

//   cout << "\t[" << mindiff[arity] << ".." << maxdiff[arity] << "]" << endl ;
//   for(k=0; k<arity; ++k)
//     {
//       cout << "\t[" << mindiff[k] << ".." << maxdiff[k] << "]  " ;
//       scope[k]->print( cout );
//       if( k == idx )
// 	cout << "*";
//       cout << endl;
//     }
//   cout << endl << consistent << " ======================" << endl << endl;

    
  return consistent;
}

void ConstraintLDDS::print(std::ostream& o) const 
{
  cout << "LDDS-" << threshold ;
}




/******************************************************************************
 File: alldiff.cpp

 Implementation of the algorithm for bounds consistency of the
 generalized cardinality constraint described in:

	C.-G. Quimper, P. van Beek, A. Lopez-Ortiz, A. Golynski, and
	S.B. Sadjad. An efficient bounds consistency algorithm for the
	global cardinality constraint. CP-2003.

 By: Claude-Guy Quimper
 ******************************************************************************/

ConstraintGlobalCardinality::ConstraintGlobalCardinality(Solver *s,
							 VariableInt **v,
							 const int n,
							 const int firstDomainValue, 
							 const int lastDomainValue,
							 const int* minOccurrences,
							 const int* maxOccurrences )
  : Constraint(s, v, n, RANGETRIGGER )
{
  delayed = true;

  int i, range;

  range = lastDomainValue - firstDomainValue + 1;

  lastLevel = -1;

  iv        = (Interval  *)calloc(arity, sizeof(Interval  ));
  minsorted = (Interval **)calloc(arity, sizeof(Interval *));
  maxsorted = (Interval **)calloc(arity, sizeof(Interval *));
  bounds    = (int *)calloc(2*arity+2, sizeof(int));

  for( i = 0; i < arity; i++ ) {
    minsorted[i] = maxsorted[i] = &iv[i];
  }

  t = (int *)calloc(2*arity+2, sizeof(int));
  d = (int *)calloc(2*arity+2, sizeof(int));
  h = (int *)calloc(2*arity+2, sizeof(int));

  stableInterval      = (int *)calloc(2*arity+2, sizeof(int));
  potentialStableSets = (int *)calloc(2*arity+2, sizeof(int));
  newMin              = (int *)calloc(  arity,   sizeof(int));

  l = initializePartialSum(firstDomainValue, range, minOccurrences);
  u = initializePartialSum(firstDomainValue, range, maxOccurrences);
}


ConstraintGlobalCardinality::~ConstraintGlobalCardinality()
{
  free(bounds);
  free(maxsorted);
  free(minsorted);
  free(iv);
  free(h);
  free(d);
  free(t);
  free(newMin);
  free(potentialStableSets);
  free(stableInterval);
  destroyPartialSum(u);
  destroyPartialSum(l);
}

void
ConstraintGlobalCardinality::sortit()
{
  int i,j,nb,min,max,last;

  sortmin(minsorted, arity);
  sortmax(maxsorted, arity);

  min = minsorted[0]->min;
  max = maxsorted[0]->max + 1;
  //MODIFIED: bounds[0] = last = min-2;
  bounds[0] = last = l->firstValue + 1;

  for (i=j=nb=0;;) { // merge minsorted[] and maxsorted[] into bounds[]
    if (i<arity && min<=max) {	// make sure minsorted exhausted first
      if (min != last)
        bounds[++nb] = last = min;
      minsorted[i]->minrank = nb;
      if (++i < arity)
        min = minsorted[i]->min;
    } else {
      if (max != last)
         bounds[++nb] = last = max;
      maxsorted[j]->maxrank = nb;
      if (++j == arity) break;
      max = maxsorted[j]->max + 1;
    }
  }
  ConstraintGlobalCardinality::nb = nb;
  //MODIFIED: bounds[nb+1] = bounds[nb] + 2;
  bounds[nb+1] = u->lastValue + 1;
}


/* 
 * Shrink the lower bounds for the max occurrences problem.
 */
int
ConstraintGlobalCardinality::filterLowerMax()
{
  int i,j,w,x,y,z;
  int changes = 0;

  for (i=1; i<=nb+1; i++) {
    t[i] = h[i] = i-1;
    d[i] = sum(u, bounds[i-1], bounds[i]-1);
  }
  for (i=0; i<arity; i++) { // visit intervals in increasing max order
    // get interval bounds
    x = maxsorted[i]->minrank; y = maxsorted[i]->maxrank;
    j = t[z = pathmax(t, x+1)];
    if (--d[z] == 0) {
      t[z = pathmax(t, t[z]=z+1)] = j;
    }
    pathset(t, x+1, z, z);
    if (d[z] < sum(u, bounds[y], bounds[z] - 1)) {
      return INCONSISTENT; // no solution
    }
    if (h[x] > x) {
      maxsorted[i]->min = bounds[w = pathmax(h, h[x])];
      pathset(h, x, w, w);
      changes = 1;
    }
    if (d[z] == sum(u, bounds[y], bounds[z] - 1)) {
      pathset(h, h[y], j-1, y); // mark hall interval
      h[y] = j-1; //("hall interval [%d,%d)\n",bounds[j],bounds[y]);
    }
  }
  if( changes )
    return CHANGES;
  else
    return NO_CHANGES;
}

/*
 * Shrink the upper bounds for the max occurrences problem.
 */
int
ConstraintGlobalCardinality::filterUpperMax()
{
  // Assertion: filterLowerMax returns true
  int i,j,w,x,y,z;
  int changes = 0;

  for (i=0; i<=nb; i++) {
    d[i] = sum(u, bounds[i], bounds[t[i]=h[i]=i+1]-1);
  }
  for (i=arity; --i>=0; ) { // visit intervals in decreasing min order
    // get interval bounds
    x = minsorted[i]->maxrank; y = minsorted[i]->minrank;
    j = t[z = pathmin(t, x-1)];
    if (--d[z] == 0) {
      t[z = pathmin(t, t[z]=z-1)] = j;
    }
    pathset(t, x-1, z, z);
    if (d[z] < sum(u, bounds[z], bounds[y]-1)) {
      return INCONSISTENT; // no solution
    }
    if (h[x] < x) {
      minsorted[i]->max = bounds[w = pathmin(h, h[x])] - 1;
      pathset(h, x, w, w);
      changes = 1;
    }
    if (d[z] == sum(u, bounds[z], bounds[y]-1)) {
      pathset(h, h[y], j+1, y);
      h[y] = j+1;
    }
  }
  if( changes )
    return CHANGES;
  else
    return NO_CHANGES;
}


/*
 * Shrink the lower bounds for the min occurrences problem.
 * called as: filterLowerMin(t, d, h, stableInterval, potentialStableSets, newMin);
 */
int
ConstraintGlobalCardinality::filterLowerMin(int *tl, int *c,
					    int* stableAndUnstableSets,
					    int* stableInterval,
					    int* potentialStableSets,
					    int* newMin)
{
  int i,j,w,x,y,z,v;
  int changes = 0;

  for (w=i=nb+1; i>0; i--) {
    //c[i] = sum(l, bounds[potentialStableSets[i]=stableInterval[i]=i-1], bounds[i]-1);
    potentialStableSets[i]=stableInterval[i]=i-1;
    c[i] = sum(l, bounds[i-1], bounds[i]-1);
    // If the capacity between both bounds is zero, we have
    // an unstable set between these two bounds.
    if (c[i] == 0) {
      stableAndUnstableSets[i-1] = w;
    }
    else {
      w = stableAndUnstableSets[w] = i - 1;
    }
  }

  for (i = w = nb + 1; i >= 0; i--) {
    if (c[i] == 0)
      tl[i] = w;
    else
      w = tl[w] = i;
  }

  for (i = 0; i < arity; i++) { // visit intervals in increasing max order
    // Get interval bounds
    x = maxsorted[i]->minrank; y = maxsorted[i]->maxrank;
    j = tl[z = pathmax(tl, x+1)];
    if (z != x+1) {
      // if bounds[z] - 1 belongs to a stable set,
      // [bounds[x], bounds[z]) is a sub set of this stable set
      v = potentialStableSets[w = pathmax(potentialStableSets, x + 1)];
      pathset(potentialStableSets, x+1, w, w); // path compression
      w = y < z ? y : z;
      pathset(potentialStableSets, potentialStableSets[w], v , w);
      potentialStableSets[w] = v;
    }

    if (c[z] <= sum(l, bounds[y], bounds[z] - 1)) {
      // (potentialStableSets[y], y] is a stable set
      w = pathmax(stableInterval, potentialStableSets[y]);
      pathset(stableInterval, potentialStableSets[y], w, w); // Path compression
      pathset(stableInterval, stableInterval[y], v=stableInterval[w], y);
      stableInterval[y] = v;
    }
    else {
      // Decrease the capacity between the two bounds
      if (--c[z] == 0) {
	tl[z = pathmax(tl, tl[z]=z+1)] = j;
      }

      // If the lower bound belongs to an unstable or a stable set,
      // remind the new value we might assigned to the lower bound
      // in case the variable doesn't belong to a stable set.
      if (stableAndUnstableSets[x] > x) {
	w = newMin[i] = pathmax(stableAndUnstableSets, x);
	pathset(stableAndUnstableSets, x, w, w); // path compression
      }
      else {
	newMin[i] = x; // Do not shrink the variable
      }

      // If an unstable set is discovered
      if (c[z] == sum(l, bounds[y], bounds[z] - 1)) {
 	if (stableAndUnstableSets[y] > y) // Consider stable and unstable sets beyong y
 	  y = stableAndUnstableSets[y]; // Equivalent to pathmax since the path is fully compressed
	pathset(stableAndUnstableSets, stableAndUnstableSets[y], j-1, y); // mark the new unstable set
	stableAndUnstableSets[y] = j-1;
      }
    }
    pathset(tl, x+1, z, z); // path compression
  }

  // If there is a failure set
  if (stableAndUnstableSets[nb] != 0) {
    return INCONSISTENT; // no solution
  }

  // Perform path compression over all elements in
  // the stable interval data structure. This data
  // structure will no longer be modified and will be
  // accessed n or 2n times. Therefore, we can afford
  // a linear time compression.
  for (i = nb+1; i > 0; i--) {
    if (stableInterval[i] > i)
      stableInterval[i] = w;
    else
      w = i;
  }

  // For all variables that are not a subset of a stable set, shrink the lower bound
  for (i=arity-1; i>=0; i--) {
    x = maxsorted[i]->minrank; y = maxsorted[i]->maxrank;
    if ((stableInterval[x] <= x) || (y > stableInterval[x])) {
      maxsorted[i]->min = skipNonNullElementsRight(l, bounds[newMin[i]]);
      changes = 1;
    }
  }

  if( changes )
    return CHANGES;
  else
    return NO_CHANGES;
}


/*
 * Shrink the upper bounds for the min occurrences problem.
 * called as: filterUpperMin(t, d, h, stableInterval, newMin);
 */
int 
ConstraintGlobalCardinality::filterUpperMin(int *tl, int *c,
					    int* stableAndUnstableSets,
					    int* stableInterval,
					    int* newMax)
{
  // ASSERTION: filterLowerMin returns true
  int i,j,w,x,y,z;
  int changes = 0;

  for (w=i=0; i<=nb; i++) {
    //    d[i] = bounds[t[i]=h[i]=i+1] - bounds[i];
    c[i] = sum(l, bounds[i], bounds[i+1]-1);
    if (c[i] == 0)
      tl[i]=w;
    else
      w=tl[w]=i;
  }
  tl[w]=i;
  for (i = 1, w = 0; i<=nb; i++) {
    if (c[i-1] == 0)
      stableAndUnstableSets[i] = w;
    else
      w = stableAndUnstableSets[w] = i;
  }
  stableAndUnstableSets[w] = i;

  for (i=arity; --i>=0; ) { // visit intervals in decreasing min order
    // Get interval bounds
    x = minsorted[i]->maxrank; y = minsorted[i]->minrank;

    // Solve the lower bound problem
    j = tl[z = pathmin(tl, x-1)];

    // If the variable is not in a discovered stable set
    // Possible optimization: Use the array stableInterval to perform this test
    if (c[z] > sum(l, bounds[z], bounds[y]-1)) {
      if (--c[z] == 0) {
	tl[z = pathmin(tl, tl[z]=z-1)] = j;
      }
      if (stableAndUnstableSets[x] < x) {
	newMax[i] = w = pathmin(stableAndUnstableSets, stableAndUnstableSets[x]);
	pathset(stableAndUnstableSets, x, w, w); // path compression
      }
      else {
	newMax[i] = x;
      }
      if (c[z] == sum(l, bounds[z], bounds[y]-1)) {
	if (stableAndUnstableSets[y] < y) {
	  y = stableAndUnstableSets[y];
        }
	pathset(stableAndUnstableSets, stableAndUnstableSets[y], j+1, y);
	stableAndUnstableSets[y] = j+1;
      }
    }
    pathset(tl, x-1, z, z);
  }

  // For all variables that are not subsets of a stable set, shrink the lower bound
  for (i=arity-1; i>=0; i--) {
    x = minsorted[i]->minrank; y = minsorted[i]->maxrank;
    if ((stableInterval[x] <= x) || (y > stableInterval[x]))
      minsorted[i]->max = skipNonNullElementsLeft(l, bounds[newMax[i]]-1);
      changes = 1;
  }

  if( changes )
    return CHANGES;
  else
    return NO_CHANGES;
}


bool ConstraintGlobalCardinality::propagate(VariableInt *v, const int e)
{
  return propagate();
}

bool ConstraintGlobalCardinality::propagate()
{
  int statusLower, statusUpper;
  int statusLowerMin, statusUpperMin;
  int statusLowerMax, statusUpperMax;
  int i, a, b;
  int dl, du;

  a = 0;
  b = arity;
  //  a = _vars.getRangeIndexMin();
  //  b = _vars.getRangeIndexMax();

/*
  if( _prop == WithValueRemoval && (a == (b-1)) && _vars[a].isBound() ) {
    return;
  }
*/


  //if( lastLevel != scope[0]->solver->level ) {
  if( lastLevel != solver->level ) {
    // not incremental
    statusLower = CHANGES;
    statusUpper = CHANGES;

    i = arity;
    while( i-- ) {
      iv[i].min = scope[i]->min();
      iv[i].max = scope[i]->max();
    }

    //IlcVariableIntArrayIterator iter(_vars);
    //IlcVariableInt v;
    //i = 0;
    //while (iter.next(v)) {
    //  iv[i].min = v.getMin();
    //  iv[i].max = v.getMax();
    //  i++;
    //}
  }
  else {
    // incremental
    statusLower = NO_CHANGES;
    statusUpper = NO_CHANGES;
    for( i = a; i < b; i++ ) {
      dl = iv[i].min;
      du = iv[i].max;
      iv[i].min = scope[i]->min();
      iv[i].max = scope[i]->max();
      //iv[i].min = _vars[i].getMin();
      //iv[i].max = _vars[i].getMax();
      if( dl != iv[i].min ) statusLower = CHANGES;
      if( du != iv[i].max ) statusUpper = CHANGES;
    }
  }

  //lastLevel = scope[0]->solver->level;
  lastLevel = solver->level;

  if( statusLower == NO_CHANGES && statusUpper == NO_CHANGES ) {
    return true;
  }

  sortit();

  // The variable domains must be inside the domain defined by
  // the lower bounds (l) and the upper bounds (u).
  //assert(minValue(l) == minValue(u));
  //assert(maxValue(l) == maxValue(u));
  //assert(minValue(l) <= minsorted[0]->min);
  //assert(maxsorted[n-1]->max <= maxValue(u));

  // Checks if there are values that must be assigned before the
  // smallest interval or after the last interval. If this is
  // the case, there is no solution to the problem
  // This is not an optimization since
  // filterLower{Min,Max} and
  // filterUpper{Min,Max} do not check for this case.
  if ((sum(l, minValue(l), minsorted[0]->min - 1) > 0) ||
      (sum(l, maxsorted[arity-1]->max + 1, maxValue(l)) > 0)) {
    //_vars.getManager().fail();
    return false;
  }

  statusLowerMax = filterLowerMax();
  if( statusLowerMax != INCONSISTENT ) {
    statusLowerMin = filterLowerMin(t, d, h,
			stableInterval, potentialStableSets, newMin);
  }

  if( (statusLowerMax == INCONSISTENT) || (statusLowerMin == INCONSISTENT) ) {
    //_vars.getManager().fail();
    return false;
  }
  else {

    statusUpperMax = filterUpperMax();
    statusUpperMin = filterUpperMin(t, d, h, stableInterval, newMin);

    if( (statusLowerMax == CHANGES) || (statusLowerMin == CHANGES) ||
        (statusUpperMax == CHANGES) || (statusUpperMin == CHANGES) ) {
      i = arity;
      while( i-- )
	if( !(scope[i]->setMin( iv[i].min ) && scope[i]->setMax( iv[i].max )) )
	  return false;
//       IlcVariableIntArrayIterator iter(_vars);
//       IlcVariableInt v;
//       i = 0;
//       while (iter.next(v)) {
//         v.setRange( iv[i].min, iv[i].max );
//         i++;
//       }
    } // if
  } // else

  return true;
}

// Create a partial sum data structure adapted to the
// filterLower{Min,Max} and filterUpper{Min,Max} functions.
// Two elements before and after the element list will be added
// with a weight of 1. 
partialSum*
ConstraintGlobalCardinality::initializePartialSum(const int firstValue, int count, 
						  const int* elements)
{
  int i,j;
  partialSum* res = new partialSum;
  res->sum = new int[count+1+2+2];
  res->firstValue = firstValue - 3; // We add three elements at the beginning
  res->lastValue = firstValue + count + 1;
  res->sum[0] = 0;
  res->sum[1] = 1;
  res->sum[2] = 2;
  for (i = 2; i < count+2; i++) {
    res->sum[i+1] = res->sum[i] + elements[i-2];
  }
  res->sum[i+1] = res->sum[i] + 1;
  res->sum[i+2] = res->sum[i+1] + 1;

  res->ds = new int[count+1+2+2];

  for (j=(i=count+3)+1; i > 0;) {
    while (res->sum[i] == res->sum[i-1])
      res->ds[i--]=j;
    j=res->ds[j]=i--;
  }
  res->ds[j]=0;
  return res;
}

void
ConstraintGlobalCardinality::destroyPartialSum(partialSum *p)
{
  delete p->ds;
  delete p->sum;
  delete p;
}

int
ConstraintGlobalCardinality::sum(partialSum *p, int from, int to)
{
  if (from <= to) {
    //assert((p->firstValue <= from) && (to <= p->lastValue));
    return p->sum[to - p->firstValue] - p->sum[from - p->firstValue - 1];
  }
  else {
    //assert((p->firstValue <= to) && (from <= p->lastValue));
    return p->sum[to - p->firstValue - 1] - p->sum[from - p->firstValue];
  }
}

int
ConstraintGlobalCardinality::minValue(partialSum *p) {
  return p->firstValue + 3;
}

int
ConstraintGlobalCardinality::maxValue(partialSum *p) {
  return p->lastValue - 2;
}

int
ConstraintGlobalCardinality::skipNonNullElementsRight(partialSum *p, int value) {
  value -= p->firstValue;
  return (p->ds[value] < value ? value : p->ds[value]) + p->firstValue;
}

int
ConstraintGlobalCardinality::skipNonNullElementsLeft(partialSum *p, int value) {
  value -= p->firstValue;
  return (p->ds[value] > value ? p->ds[p->ds[value]] : value) + p->firstValue;
}


int ConstraintGlobalCardinality::check( const int* s ) const 
{
//   int i=arity, j;
//   while(--i) {
//     j=i;
//     while(j--)
//       if( s[i] == s[j] ) return 1;
//   }
  return 0; 
}

void ConstraintGlobalCardinality::print(std::ostream& o) const 
{    
    o << "Gcc(";
    for(int i=0; i<arity-1; ++i) {
      scope[i]->printshort( o );
      o << " " ;
    }
    scope[arity-1]->printshort( o );
    o << ", [" << (l->firstValue+3) << ".." 
      << (l->lastValue-2) << "] >={";
    o << " }, <={";
    o << " })";
}

/*
 *  End of user defined propagator for enforcing bounds consistency
 *=================================================================*/


ConstraintSlideDecomp::ConstraintSlideDecomp(Solver *s, VariableInt** v, 
					     const int n, const int m,
					     const int nv, const int sw, const int K,
					     int* cft, int* t2d, int* d2t, int* exp) 
{
  nbVals = nv; 
  sequenceWidth = sw;
  conflicts = cft;
  int2ind = t2d;
  ind2int = d2t;
  exposant = exp;

  int i, j, k, nSlideVars = m-sequenceWidth+2;
  VariableInt *slideVars[n*nSlideVars];
  for(i=0; i<n; ++i)
    for(j=0; j<nSlideVars; ++j)
      slideVars[i*nSlideVars+j] = new VariableBit( solver, 0, K-1 );
  
  VariableInt *scp[sequenceWidth];
  for(i=0; i<n; ++i) {
    for(j=0; j<nSlideVars; ++j) {
      for(k=0; k<sequenceWidth-1; ++k)
	scp[k] = v[i*m+j+k];
      scp[sequenceWidth-1] = slideVars[i*nSlideVars+j];
      new PredicateTuple( solver, scp,
			  sequenceWidth,
			  nbVals,
			  int2ind,
			  ind2int,
			  exposant );
    }
    for(j=0; j<nSlideVars-1; ++j) {
      scp[0] = slideVars[i*nSlideVars+j];
      scp[1] = slideVars[i*nSlideVars+j+1];
      new ConstraintSlide( solver, scp, this );
    }
  }
}

ConstraintSlideDecomp::~ConstraintSlideDecomp()
{
  delete [] conflicts;
  delete [] int2ind;
  delete [] ind2int;
  delete [] exposant;
}

ConstraintSlide::ConstraintSlide(Solver *s, VariableInt** v, ConstraintSlideDecomp* c) 
  : Constraint(s, v, 2) 
{
  nbVals = c->nbVals; 
  conflicts = c->conflicts;
  int2ind = c->int2ind;
  ind2int = c->ind2int;
  exposant = c->exposant;
  sequenceWidth = c->sequenceWidth;
}

ConstraintSlide::~ConstraintSlide()
{
}

int ConstraintSlide::check( const int* s ) const
{
  return conflicts[appendLeft( s[0], s[1] )];
}

bool ConstraintSlide::propagate(VariableInt *v, const int e) 
{
  int vnxt, value, tuple, support;
  bool is_supported, forward = ( v == scope[0] ), consistent=true;;
  
  if( forward ) {

    value = scope[1]->first();
    do {
      vnxt = scope[1]->getNext( value );

      is_supported = false;
      support = -1;
      while( !is_supported && ++support < nbVals ) {
	tuple = appendLeft(value,support);
	is_supported = (
			!conflicts[tuple]
			&&
			scope[0]->contain( int2ind[tuple%exposant[sequenceWidth-1]] )
			);
      }
      if( !is_supported ) 
	consistent = scope[1]->remove( value );
      vnxt = value;
    }
    while( consistent && scope[1]->good(value) );

  } else {
    
    value = scope[0]->first();
    do {
      vnxt = scope[0]->getNext( value );

	is_supported = false;
	support = -1;
	while( !is_supported && ++support < nbVals ) {
	  tuple = appendRight(value,support);
	  is_supported = (
			  !conflicts[tuple]
			  &&
			  scope[1]->contain( int2ind[tuple/nbVals] )
			  );
	}
	if( !is_supported ) 
	  consistent = scope[0]->remove( value );
	vnxt = value;
    }
    while( consistent && scope[0]->good(value) );
    
  }

  return consistent;
}

void ConstraintSlide::print(std::ostream& o) const 
{    
  o << "(";
  scope[0]->printshort( o );
  o << " >> ";
  scope[1]->printshort( o );
  o << ")";
}


PredicateTuple::PredicateTuple(Solver *s, VariableInt** v, 
			       const int n, const int nv,
			       int* t2d, int* d2t, int* exp) 
  : Constraint(s, v, n) 
{
  nbVals = nv; 
  int2ind = t2d;
  ind2int = d2t;
  exposant = exp;
}

PredicateTuple::~PredicateTuple()
{
}

int PredicateTuple::check( const int* s ) const
{
  int i, tuple=0, n=arity-1;
  for(i=0; i<n; ++i)
    tuple += (s[i] * exposant[i]);
  return ( s[n] != tuple );
}

bool PredicateTuple::propagate(VariableInt *v, const int e) 
{
  int value, vnxt, i, n=arity-1;
  bool consistent = true;

  if( v == scope[n] ) {

    for(i=0; i<n; ++i)
      vals[i].clear();
    value = scope[n]->first();
    do for(i=0; i<n; ++i) 
	 vals[i].insert((ind2int[value]/exposant[i])%nbVals);
    while( scope[n]->setNext( value ) );
    for(i=0; consistent && i<n; ++i) 
      consistent = scope[i]->setDomain( vals[i] );
 
  } else {

    for(i=0; i<n; ++i)
      if( scope[i] == v ) break;

    value = scope[n]->first();
    do {	
      vnxt = scope[n]->getNext( value );
      if( !(v->contain((ind2int[value]/exposant[i])%nbVals)) )
	consistent = scope[n]->remove( value );
      value = vnxt;
    }
    while( consistent && scope[n]->good( value ) );

  }

  return consistent;
}

void PredicateTuple::print(std::ostream& o) const 
{    
  o << "(";
  for(int i=0; i<arity-2; ++i) {
    scope[i]->printshort( o );
    o << ", ";
  }
  scope[arity-1]->printshort( o );
  o << ") = " ;
  scope[arity-1]->printshort( o );
}

////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////


#ifdef _CPLEX

CPlexConstraint::CPlexConstraint(Solver *s, VariableInt** v, const int n) 
  : Constraint(s, v, n )
{
  delayed = true;

  int i, j, k=0, l;

  epsilon = 0.000001;
  lb=INT_MAX;
  ub=-1*INT_MAX;

  // the structs asgn2index, index2valasgn, index2varasgn
  // -- values2index and index2values are initialized
  //
  // asgn2index[i][j] is the num of the column corresponding to
  // scope[i] := j
  // index2valasgn[k] is the value involved in the column k;
  // index2varasgn[k] is the variable involved in the column k;
  //
  // values2index[i] is the rank of value i
  // index2values[j] is the vale at rank j

  nRows=0;
  nColumns=0;
  asgn2index = new int*[arity-1];
  lbs = new int[arity-1];

  for(i=0; i<arity-1; ++i) {
    lbs[i] = scope[i]->minCapacity();
    nColumns += scope[i]->domsize();
    if(lb > lbs[i]) lb = lbs[i];
    if(ub < scope[i]->max()) ub = scope[i]->max();

    asgn2index[i] = new int[scope[i]->maxCapacity() - lbs[i] + 1];
    asgn2index[i] -= lbs[i];
  }

  index2valasgn = new int[nColumns];
  index2varasgn = new int[nColumns];

  nvals = ub-lb+1;
  index2values = new int[nvals];
  values2index = new int[nvals];
  values2index-=lb;
  std::fill(values2index+lb, values2index+ub+1, INT_MAX);

  k-0;
  nvals=0;
  for(i=0; i<arity-1; ++i) {
    j = scope[i]->first();
    do {

      asgn2index[i][j] = k;
      index2valasgn[k] = j;
      index2varasgn[k] = i;
      ++k;

      if( values2index[j] == INT_MAX ) {
	values2index[j] = nvals;
	index2values[nvals] = j;
	++nvals;
      }
    } while( scope[i]->setNext(j) ); 
  }
  nRows = arity-1;

  /*************************
   * CPlex structs
   *************************/  

  obj     = (double *) malloc (nColumns * sizeof(double));
  rhs     = (double *) malloc (nRows    * sizeof(double));
  sense   = (char *)   malloc (nRows    * sizeof(char)); 
  matbeg  = (int *)    malloc (nColumns * sizeof(int));
  matcnt  = (int *)    malloc (nColumns * sizeof(int));   
  matind  = (int *)    malloc (nColumns * sizeof(int));   
  matval  = (double *) malloc (nColumns * sizeof(double));
  zlb     = (double *) malloc (nColumns * sizeof(double));
  zub     = (double *) malloc (nColumns * sizeof(double));
  qmatbeg = (int *)    malloc (nColumns * sizeof(int)); 
  qmatcnt = (int *)    malloc (nColumns * sizeof(int));   

  env = NULL;
  env = CPXopenCPLEX (&status);
  if ( env == NULL ) {
    char  errmsg[1024];
    fprintf (stderr, "Could not open CPLEX environment.\n");
    CPXgeterrorstring (env, status, errmsg);
    fprintf (stderr, "%s", errmsg);
  }

  status = CPXsetintparam (env, CPX_PARAM_SCRIND, CPX_ON);
  if ( status ) {
    fprintf (stderr, 
	     "Failure to turn on screen indicator, error %d.\n", status);
  }


  lp = CPXcreateprob (env, &status, "bidule");
  if ( lp == NULL ) {
    fprintf (stderr, "Failed to create problem.\n");
  }

  //////////////// CREATING THE LP /////////////////

  // maximise (not sure if it applies to the quad objective)
  objsen = -1;

  
  for(i=0; i<nColumns; ++i) { // for each columns
    // only one constraint on each var
    matcnt[i] = 1;

    // one coeff for each column
    matbeg[i] = i;

    // the index of the row is the CSP var corresponding to each CSP assignment
    j = index2varasgn[i];
    matind[i] = j;
    matval[i] = 1.0;

    // [0..1] variables
    zlb[i] = 0.0;
    zub[i] = 1.0;

    // no linear objective
    obj[i] = 0.0;
  }

  for(i=0; i<nRows; ++i) {
    // exactly one value per domain
    rhs[i] = 1.0;

    // all equalities
    sense[i] = 'E';
  }


  status = CPXcopylp (env, lp, nColumns, nRows, objsen, obj, rhs, 
		      sense, matbeg, matcnt, matind, matval,
		      zlb, zub, NULL);

  //////////////////////////////////////////////////

  x  = new double[nColumns];
  dj = new double[nColumns];
  pi    = new double[nRows];
  slack = new double[nRows];

/*************************
 *                       *
 *************************/


}

CPlexConstraint::~CPlexConstraint()
{
  values2index += lb;
  delete [] values2index;  
  delete [] index2values;  

  for(int i=0; i<arity-1; ++i) {
    asgn2index[i] += lbs[i];
    delete [] asgn2index[i];
  }
  delete [] asgn2index;
  delete [] index2valasgn;
  delete [] index2varasgn;
  delete [] lbs;
}

int CPlexConstraint::check(int* sol) const 
{
  return 0;
}

void CPlexConstraint::updateModel()
{
}

bool CPlexConstraint::propagate(VariableInt *v, const int e)
{
}

void CPlexConstraint::print(std::ostream& o) const 
{
}

/**********************************************
 * CPlexMinDiff Predicate 
 **********************************************/

PredicateCPlexMinDiff::PredicateCPlexMinDiff(Solver *s, const int f, VariableInt** v, const int n) 
  : CPlexConstraint(s, v, n)
{ 
  filtering = f;

  int i, j, k, nRows=arity-1;

  K = arity-1;
  vars = new VariableInt*[K];
  occurrence = new int[ub-lb+1];
  occurrence -= lb;
  orderedvals = new int[ub-lb+1];

  std::fill( occurrence+lb, occurrence+ub+1, 0 );



  //////////////// CREATING THE QP /////////////////

  //cout << "here" << endl;

  int nAsgn = 0;
  for(i=0; i<nRows; ++i) {
    j = scope[i]->first();
    do {
      ++occurrence[j];
    } while( scope[i]->setNext(j) ); 
  } 

  cout << lb << ".." << ub << endl;

  for(i=lb; i<=ub; ++i) 
    if(occurrence[i])
      nAsgn += (occurrence[i] * (occurrence[i]-1));

  qmatind = (int *)    malloc (nAsgn * sizeof(int)); 
  qmatval = (double *) malloc (nAsgn * sizeof(double));   
  //qmatbeg = (int *)    malloc (nColumns * sizeof(int)); 
  //qmatcnt = (int *)    malloc (nColumns * sizeof(int)); 

  cout << nAsgn << endl;

  int str_ = 0;
  for(i=0; i<nColumns; ++i) { // for each columns
    qmatbeg[i] = str_;

    for(j=0; j<nRows; ++j) {
      if(j != index2varasgn[i] && 
	 scope[j]->contain(index2valasgn[i])) {

	qmatind[str_] = asgn2index[j][index2valasgn[i]];
	qmatval[str_++] = 0.5;
      }
    }
    
    qmatcnt[i] = (str_ - qmatbeg[i]);
    
//     cout << " => " << i << ": x" << index2varasgn[i] << "=" << index2valasgn[i] << " -> "
// 	 << (str_ - qmatbeg[i]) << " " << (occurrence[index2valasgn[i]]) 
// 	 << " " << qmatbeg[i] << endl << endl;
    //assert( (str - qmatbeg[i]) == (occurrence[index2valasgn[i]] - 1) );
  }

  status = CPXcopyquad (env, lp, qmatbeg, qmatcnt, qmatind, qmatval);
  if ( status ) {
    fprintf (stderr, "Failed to copy quadratic matrix.\n");
  }

  //////////////////////////////////////////////////

  //////////////// SOLVING /////////////////////////

  status = CPXqpopt (env, lp);
  if ( status ) {
    fprintf (stderr, "Failed to optimize QP.\n");
  }
  
  status = CPXsolution (env, lp, &solstat, &objval, x, pi, slack, dj);
  if ( status ) {
    fprintf (stderr, "Failed to obtain solution.\n");
  }

  status = CPXwriteprob (env, lp, "sae.lp", NULL);

  exit( 0 );

  //////////////////////////////////////////////////
  
  weight = 0;
}

PredicateCPlexMinDiff::~PredicateCPlexMinDiff()
{
  occurrence += lb;
  delete [] occurrence;
  delete [] vars;
  delete [] orderedvals;
}

int PredicateCPlexMinDiff::approximation( int& opt ) 
{
  int nbdiff = (((K)*(K-1)) >> 1);
  int i, j, nvals, vali;

  int nvars=K, aux;
  VariableInt *auxv;
  int *firstval = orderedvals;
    
  /// count values
  std::fill( occurrence+lb, occurrence+ub+1, 0 );
  nvals=0;
  for(i=0; i<K; ++i) {
    vars[i] = scope[i];
    vali = scope[i]->first();
    do {
      if( ++occurrence[ vali ] == 1 )
	orderedvals[nvals++] = vali;
    }
    while( scope[i]->setNext(vali) );
  }
     
  int *lastval = orderedvals+nvals;

  while( nvars ) {
    /// order values
    for(i=1; firstval+i<lastval; ++i) {
      j = i;
      while( j && occurrence[firstval[j]] > occurrence[firstval[j-1]] )
	{
	  vali = firstval[j];
	  firstval[j] = firstval[j-1];
	  firstval[j-1] = vali;
	  --j;
	}
    }

    /// use the value with largest occurrences
    aux = occurrence[*firstval];
    aux = ((aux * (aux-1)) >> 1);
    nbdiff -= aux;

    // update the data-structures
    i = nvars;
    while( i-- ) 
      if( vars[i]->contain( *firstval ) )
	{
	  // update the array 'occurrence'
	  vali = vars[i]->first();
	  do --occurrence[ vali ];
	  while( vars[i]->setNext(vali) );
	  
	  // put all used variables at the end of the array 'vars'
	  --nvars;
	  auxv = vars[nvars];
	  vars[nvars] = vars[i];
	  vars[i] = auxv;
	}
    //--nvals;
    ++firstval;

  }

  return nbdiff;
}

bool PredicateCPlexMinDiff::propagate(VariableInt *v, const int e)
{
  bool consistent = true;
  return consistent;
}
  
void PredicateCPlexMinDiff::print(std::ostream& o) const 
{
  o << "#(";
  for(int i=0; i<K-1; ++i) {
    scope[i]->printshort(o);
    o << ", ";
  }
  scope[K-1]->printshort(o);
  o << ") = ";
  scope[K]->printshort( o );
}

#endif


#ifdef _COINLP

/**********************************************
 * Lp Constraint 
 **********************************************/

LpConstraint::LpConstraint( Solver *s ) 
  : Constraint(s)
{
  delayed = true;
}

LpConstraint::LpConstraint(Solver *s, VariableInt** v, const int n, const int t) 
  : Constraint(s, v, n )
{
  init( t );
  delayed = true;
}

void LpConstraint::init(const int t) 
{ 
  type = t;
  int i, j, k=0, l;
  epsilon = 0.000001;
  lb=INT_MAX;
  ub=-1*INT_MAX;

  imodel.setLogLevel( 0 );
  smodel.setLogLevel( 0 );

  // the structs asgn2index, index2valasgn, index2varasgn
  // -- values2index and index2values are initialized
  //
  // asgn2index[i][j] is the num of the column corresponding to
  // scope[i] := j
  // index2valasgn[k] is the value involved in the column k;
  // index2varasgn[k] is the variable involved in the column k;
  //
  // values2index[i] is the rank of value i
  // index2values[j] is the vale at rank j

  nRows=0;
  nColumns=0;
  asgn2index = new int*[arity-1];
  lbs = new int[arity-1];

  for(i=0; i<arity-1; ++i) {
    lbs[i] = scope[i]->minCapacity();
    nColumns += scope[i]->domsize();
    if(lb > lbs[i]) lb = lbs[i];
    if(ub < scope[i]->max()) ub = scope[i]->max();

    asgn2index[i] = new int[scope[i]->maxCapacity() - lbs[i] + 1];
    asgn2index[i] -= lbs[i];
  }

  index2valasgn = new int[nColumns];
  index2varasgn = new int[nColumns];

  nvals = ub-lb+1;
  index2values = new int[nvals];
  values2index = new int[nvals];
  values2index-=lb;
  std::fill(values2index+lb, values2index+ub+1, INT_MAX);

  k-0;
  for(i=0; i<arity-1; ++i) {
    j = scope[i]->first();
    do {

      asgn2index[i][j] = k;
      index2valasgn[k] = j;
      index2varasgn[k] = i;
      ++k;

      if( values2index[j] == INT_MAX ) {
	values2index[j] = nRows;
	index2values[nRows] = j;
	++nRows;
      }
    } while( scope[i]->setNext(j) ); 
  }
}

LpConstraint::~LpConstraint()
{
  values2index += lb;
  delete [] values2index;  
  delete [] index2values;  

  for(int i=0; i<arity-1; ++i) {
    asgn2index[i] += lbs[i];
    delete [] asgn2index[i];
  }
  delete [] asgn2index;
  delete [] index2valasgn;
  delete [] index2varasgn;
  delete [] lbs;
}

int LpConstraint::check(int* sol) const 
{
  return 0;
}

void LpConstraint::updateModel()
{
  int i, in;
  //double *colub;
  //  if( type == LpConstraint::INTERIOR ) {
  double *colub1 = imodel.columnUpper();
    // } else {
  double *colub2 = smodel.columnUpper();

 //  cout << colub1 << endl
//        << colub2 << endl << endl;


    // }
  for(i=0; i<nColumns; ++i) {

    //    assert( colub1[i] == colub2[i] );

    in = scope[index2varasgn[i]]->contain(index2valasgn[i]);
    if( colub1 && colub1[i] != 0 ) {
      if( !in ) {
	if( colub1 )
	  colub1[i] = 0.0;
	if( colub2 )
	  colub2[i] = 0.0;
      }
    } else if( in ) {
      if( colub1 )
	colub1[i] = 1.0;
      if( colub2 )
	colub2[i] = 1.0;
    }
  }
}

bool LpConstraint::propagate(VariableInt *v, const int e)
{

  switch( type ) {
  case INTERIOR : cout << "P Interior" << endl; break;
  case SIMPLEX_PRIMAL : cout << "P Simplex Primal" << endl; break;
  case SIMPLEX_DUAL : cout << "P Simplex Dual" << endl; break;
  case SIMPLEX_BARRIER : cout << "P Simplex Barrier" << endl; break;
  case SIMPLEX_PRIMAL_IT : cout << "P Simplex Primal*" << endl; break;
  case SIMPLEX_DUAL_IT : cout << "P Simplex Dual*" << endl; break;
  case SIMPLEX_BARRIER_IT : cout << "P Simplex Barrier*" << endl; break;
  }

  updateModel();

  int i, j, jnxt, k;
  bool consistent = true;

  switch( type ) {
  case INTERIOR : {
    
    imodel.primalDual();
    if( (consistent = imodel.primalFeasible()) ) {
      const double * solution = imodel.primalColumnSolution();
      for(i=0; consistent && i<arity; ++i) {
	j = scope[i]->first();
	do {
	  jnxt = scope[i]->getNext(j);
	  k = asgn2index[i][j];
	  if( solution[k] < epsilon ) consistent = scope[i]->remove(j);
	  j = jnxt;
	} while( consistent && scope[i]->good(j) ); 
      }
    }
    
    break;
  }
    
  case SIMPLEX_PRIMAL : {
   
    consistent = smodel.primal();
    //consistent = ((ClpSimplex*)model)->primalFeasible();
    
    break;
  }

  case SIMPLEX_DUAL : {
   
    consistent = smodel.dual();
    
    break;
  }

  case SIMPLEX_BARRIER : {
   
    consistent = smodel.barrier();
    
    break;
  }

  case SIMPLEX_PRIMAL_IT : {

//     for(i=0; i<arity; ++i) {
//       cout << (scope[i]->id) << " ";
//       scope[i]->printshort( cout );
//       cout << endl;
//     }   
    
    double *collb  = smodel.columnLower();
    double *colub1 = smodel.columnUpper();
    double *colub2 = imodel.columnUpper();

//     for(i=0; i<nColumns; ++i) {
//       if(!(i%5)) cout << endl;
//       cout << "[" << collb[i] << ".." << colub1[i] << "] ";
//     }
//     cout << endl;


    bool cons;



    //	smodel.primal();
    //	cons = smodel.primalFeasible();

    //	cout << ( cons ? "OK" : "KO" ) << endl;


    for(i=0; consistent && i<arity; ++i) {
      j = scope[i]->first();
      do {

	//	cout << i << " <- " << j << endl;

	jnxt = scope[i]->getNext(j);
	k = asgn2index[i][j];
	
	collb[k] = 1.0;
	smodel.primal();
	cons = smodel.primalFeasible();

	//	cout << ( cons ? "OK" : "KO" ) << endl;

	if( !cons ) {
	  consistent = scope[i]->remove( j );
	  colub1[k] = 0.0;
	  colub2[k] = 0.0;
	}
	collb[k] = 0.0;

	j = jnxt;
      } while( consistent && scope[i]->good(j) ); 
    }


    break;    
  }

  case SIMPLEX_DUAL_IT : {
   
    double *collb  = smodel.columnLower();
    double *colub1 = smodel.columnUpper();
    double *colub2 = imodel.columnUpper();    

    bool cons;
    for(i=0; consistent && i<arity; ++i) {
      j = scope[i]->first();
      do {

	//	cout << i << " <- " << j << endl;

	jnxt = scope[i]->getNext(j);
	k = asgn2index[i][j];
	
	collb[k] = 1.0;
	smodel.dual();
	cons = smodel.primalFeasible();

	//	cout << ( cons ? "OK" : "KO" ) << endl;

	if( !cons ) {
	  consistent = scope[i]->remove( j );
	  colub1[k] = 0.0;
	  colub2[k] = 0.0;
	}
	collb[k] = 0.0;

	j = jnxt;
      } while( consistent && scope[i]->good(j) ); 
    }

    break;    
    
  }

  case SIMPLEX_BARRIER_IT : {

    double *collb = smodel.columnLower();
    double *colub1 = smodel.columnUpper();
    double *colub2 = imodel.columnUpper();
    bool cons;
    for(i=0; consistent && i<arity; ++i) {
      j = scope[i]->first();
      do {

	jnxt = scope[i]->getNext(j);
	k = asgn2index[i][j];
	
	collb[k] = 1.0;
	smodel.barrier();
	cons = smodel.primalFeasible();

	if( !cons ) {
	  consistent = scope[i]->remove( j );
	  colub1[k] = 0.0;
	  colub2[k] = 0.0;
	}
	collb[k] = 0.0;

	j = jnxt;
      } while( consistent && scope[i]->good(j) ); 
    }

    break;
  }
    

  }

  return consistent;
}
  
void LpConstraint::print(std::ostream& o) const 
{
}


/**********************************************
 * MaxDiff Predicate 
 **********************************************/

PredicateMaxDiff::PredicateMaxDiff(Solver *s, VariableInt** v, const int n, const int t) 
//: Constranint(s, v, n)//,
  : LpConstraint(s, v, n)
{
  //delayed = true;
  filtering = LpConstraint::LPA; 
  // init(t);
  initStruct(t);
}

PredicateMaxDiff::PredicateMaxDiff(Solver *s, const int ft, VariableInt** v, const int n, const int mt) 
  : LpConstraint(s, v, n)
{ 
  //delayed = true;
  filtering = ft;
  //init(t);
  initStruct(mt);
}

void PredicateMaxDiff::initStruct(const int t) 
{ 
  int i, j, n=arity-1;

  if( filtering == LpConstraint::LPA ||
      filtering == LpConstraint::LPS ||
      filtering == LpConstraint::LPO ) {
        
    imodel.resize(n+nRows+1, 0);
    smodel.resize(n+nRows,   0);
    
    CoinBuild SbuildObject;
    CoinBuild IbuildObject;
    
    // the n first rows correspond to domain constraints
    // ( sum(x[i,j] | j in D(X[i])) == 1  )
    for (i=0;i<n;i++) {
      imodel.setRowLower(i, 1.0);
      imodel.setRowUpper(i, 1.0);
      smodel.setRowLower(i, 1.0);
      smodel.setRowUpper(i, 1.0);
    }

    // there are as many additional rows as values
    // for counting values
    nRows += n;
    for (i=n;i<nRows;i++) {
      imodel.setRowLower(i, 0.0);
      imodel.setRowUpper(i, 0.0);
      smodel.setRowLower(i, 0.0);
      smodel.setRowUpper(i, 0.0);
    }
    
    // ?    
    imodel.setRowLower(nRows, 0.0);
    imodel.setRowUpper(nRows, n*(n-1)/2);  
    

    for(i=0; i<n; ++i) {
      j = scope[i]->first();
      do {
	// assignment, => Xi takes j and j is used once
	int columnIndex[2] = {i, n+values2index[j]};
	double columnValue[2] = {1.0, 1.0};     
	// 0 -> Xi =/= j, 1 -> Xi == j
	SbuildObject.addColumn(2, columnIndex, columnValue, 0.0, 1.0);
	IbuildObject.addColumn(2, columnIndex, columnValue, 0.0, 1.0);
      } while( scope[i]->setNext(j) ); 
    }
    
    int nValues[nRows-n];
    std::fill(nValues, nValues+nRows-n, 0);
    for(i=0; i<n; ++i) {
      j = scope[i]->first();
      do {
	// dummy vars used to express the objective function
	// they are symmetric, so the objective is to minimise
	// 0 * y1j + 1 * y2j + 2 * y3j + .... 
	int column1Index[1] = {n+values2index[j]};
	double column1Value[1] = {-1.0};
	SbuildObject.addColumn(1, column1Index, column1Value, 0.0, 1.0, 
			       (1.0 * nValues[values2index[j]]));
	
	int column2Index[2] = {n+values2index[j], nRows};
	double column2Value[2] = {-1.0, (1.0 * nValues[values2index[j]])};
	IbuildObject.addColumn(2, column2Index, column2Value, 0.0, 1.0);
	
	++nValues[values2index[j]];
      } while( scope[i]->setNext(j) ); 
    }
    
    imodel.addColumns(IbuildObject);
    smodel.addColumns(SbuildObject);
  }

  if( filtering != LpConstraint::UPB ) {
    groundvals = new int[(ub - lb + 1)];
    groundvals -= lb;
    orderedvals =  new int[(ub - lb + 1)];
    //    toremove.init( lb, ub, MistralSet::empt );
  }

  weight = 0;
  
  freevals.init( lb, ub, MistralSet::empt );
}

PredicateMaxDiff::~PredicateMaxDiff()
{
  groundvals += lb;
  delete [] groundvals;
  delete [] orderedvals;
}

int PredicateMaxDiff::check( const int* s ) const 
{
  return 0;
}

int PredicateMaxDiff::approximation( int& opt ) 
{
  freevals.clear();

  int K = arity-1, i, j, k, aux, nbvals=0, nbvars=K, nbdiff=0, maxfree=0;
  std::fill( groundvals + lb, groundvals + ub + 1, 0 );
  for(i=lb; i<=ub; ++i)
    orderedvals[i] = i;

  // groundvals keeps the count of values already used by assigned variables
  // freevals is the union of unassigned variables' domains    
  for(i=0; i<K; ++i) 
    if(scope[i]->isGround()) {
      j = scope[i]->min();
      if( !groundvals[j] )
	++nbvals;
      nbdiff += (K - nbvars - groundvals[j]);
      --nbvars;
      ++groundvals[j];
    }
    else 
      scope[i]->unionTo( freevals );
  
  // separate freevals from non-freevals in groundvals (orderedval)
  // in [lb..k-1] are all pure ground values
  // in [k..ub] are other (useful) values
  i = ub;
  k = lb;
  while( i >= k ) {
    if( !freevals.member( orderedvals[i] ) ) {
      aux = orderedvals[k];
      orderedvals[k] = orderedvals[i];
      orderedvals[i] = aux;
      ++k;
    } else --i;
  }
  
  // we order values by increasing 'groundvals', that is,
  // the number of time there were used in assigned variables
  i = ub;
  while( i-- > k ) {
    j = i;
    while( j < ub && groundvals[orderedvals[j+1]] < groundvals[orderedvals[j]] ) {
      aux = orderedvals[j+1];
      orderedvals[j+1] = orderedvals[j];
      orderedvals[j] = aux;
      ++j;
    }
  }

  // now, for each unassigned var, we "assign" it to the least used
  // values, and then update both groundvals and orderedvals.
  while( nbvars ) {
    // we increment the current number of diffs by the number
    // of variables assigned to something else than orderedvals[k]
    nbdiff += (K - nbvars - groundvals[orderedvals[k]]);
    ++groundvals[orderedvals[k]];
    
    // maxfree is the maximum number of equalities induced by a choice so far
    if(maxfree < groundvals[orderedvals[k]])
      maxfree = groundvals[orderedvals[k]];
    
    // when the algorithm ends, freevals contains inconsistent values
    // therefore we remove orderedvals[k].
    freevals.erase( orderedvals[k] );

    // reorder the vals
    j = k;
    while( j < ub && 
	   groundvals[orderedvals[j+1]] < 
	   groundvals[orderedvals[j]] ) {
      aux = orderedvals[j+1];
      orderedvals[j+1] = orderedvals[j];
      orderedvals[j] = aux;
      ++j;
    }
    --nbvars;
  }
  
  if(opt < nbdiff) opt = nbdiff;


  // now we remove consistent values from freevals
  if( k <= ub && !(freevals.empty()) ) {

    // for each value j in freevals;
    j = freevals.min();
    do {
      // if we swap the worse choice so far with j, we obtain:
      // (opt + maxfree - groundvals[j])
      if( (opt + maxfree - groundvals[j]) > scope[K]->min() ) {
	// if this is less differences than allowed, j can be removed
	freevals.erase( j );
      }
      j = freevals.next(j);
    } while( j != INT_MAX );
  }

  return nbdiff;
}

int PredicateMaxDiff::upperbound()
{
  int K = arity-1;
  freevals.clear();
  for(int i=0; i<K; ++i) {
    scope[i]->unionTo( freevals );
  }
  int c = freevals.size();
  
  int res = (K*(K-1)/2);
  if( c < K ) 
    res -= ((c + (K%c))*((int)(K/c)) - c);

  return res;
}


bool PredicateMaxDiff::propagate(VariableInt *v, const int e)
{
  int K = arity-1;
  bool consistent = true;

  /// only filter if a strictly positive number of differences is required
  if( scope[K]->min() ) {

    int i, j, nbdiff, jnxt, k = (((K) * (K-1)) >> 1);

    switch( filtering ) {

      // simple upper bound
    case LpConstraint::UPB : {
      nbdiff = upperbound();
      if( !scope[K]->setMax( nbdiff) ) consistent = false;
      break;
    }

      // approximation only
    case LpConstraint::APX : {
      nbdiff = scope[K]->max();
      nbdiff = approximation( nbdiff );

      if( !scope[K]->setMax( nbdiff ) ) consistent = false;


      if( consistent && !freevals.empty() ) {
	
	int isngrd[K];
	for(i=0; i<K; ++i)
	  isngrd[i] = !(scope[i]->isGround()); 
	
	for(i=0; consistent && i<K; ++i) 
	  if( isngrd[i] ) {
	    j = freevals.min();
	    do {
	      consistent = scope[i]->remove( j );
	      j = freevals.next(j);
	    } while( consistent && j != INT_MAX );
	  }
      }
      
      break;
    }

    case LpConstraint::LPO : {
      
      updateModel();
      
      smodel.dual();
      nbdiff = (k - (int)(ceil(smodel.objectiveValue())));
       
      if( !scope[K]->setMax( nbdiff ) ) consistent = false;
      
      if( consistent ) {

	imodel.setRowLower(nRows, (double)( k - scope[K]->max() ));
	imodel.setRowUpper(nRows, (double)( k - scope[K]->min() ));
	
	imodel.primalDual();  
	
	if( (consistent = imodel.primalFeasible()) ) {
	  
	  const double * solution = imodel.primalColumnSolution();
	  for(i=0; consistent && i<K; ++i) {
	    j = scope[i]->first();
	    do {
	      jnxt = scope[i]->getNext(j);
	      k = asgn2index[i][j];	    
	      if( solution[k] < epsilon ) 
		consistent = scope[i]->remove(j);
	      j = jnxt;
	    } while( consistent && scope[i]->good(j) ); 
	  }
	}
      }
      
      break;
    }

    case LpConstraint::LPA : {

      nbdiff = scope[K]->max();
      nbdiff = approximation( nbdiff );

      if( !scope[K]->setMax( nbdiff ) ) consistent = false;
      
      if( consistent && !freevals.empty() ) {
	
	int isngrd[K];
	for(i=0; i<K; ++i)
	  isngrd[i] = !(scope[i]->isGround()); 
	
	for(i=0; consistent && i<K; ++i) 
	  if( isngrd[i] ) {
	    j = freevals.min();
	    do {
	      consistent = scope[i]->remove( j );
	      j = freevals.next(j);
	    } while( consistent && j != INT_MAX );
	  }
      }

      if( consistent ) {
	
	updateModel();
	
	smodel.dual();
	nbdiff = (k - (int)(ceil(smodel.objectiveValue())));
	
	if( !scope[K]->setMax( nbdiff ) ) return false;
	
	imodel.setRowLower(nRows, (double)( k - scope[K]->max() ));
	imodel.setRowUpper(nRows, (double)( k - scope[K]->min() ));
	
	imodel.primalDual();  
	
	if( (consistent = imodel.primalFeasible()) ) {
	  
	  const double * solution = imodel.primalColumnSolution();
	  for(i=0; consistent && i<K; ++i) {
	    j = scope[i]->first();
	    do {
	      jnxt = scope[i]->getNext(j);
	      k = asgn2index[i][j];	    
	      if( solution[k] < epsilon ) 
		consistent = scope[i]->remove(j);
	      j = jnxt;
	    } while( consistent && scope[i]->good(j) ); 
	  }
	}
      }
      
      break;
    }

    case LpConstraint::LPS : {

      updateModel();
      
      smodel.dual();
      nbdiff = (k - (int)(ceil(smodel.objectiveValue())));
       
      if( !scope[K]->setMax( nbdiff ) ) consistent = false;

      if( consistent ) {
	
	nbdiff = scope[K]->max();
	nbdiff = approximation( nbdiff );
	
	if( !freevals.empty() ) {

	  if( !scope[K]->setMax( nbdiff ) ) consistent = false;
	  
	  if( consistent ) {
	    int isngrd[K];
	    for(i=0; i<K; ++i)
	      isngrd[i] = !(scope[i]->isGround()); 
	    
	    for(i=0; consistent && i<K; ++i) 
	      if( isngrd[i] ) {
		j = freevals.min();
		do {
		  consistent = scope[i]->remove( j );
		  j = freevals.next(j);
		} while( consistent && j != INT_MAX );
	      }
	  }
	}
      }
	
      break;
    }

    }
  }
    
  return consistent;
}
  
void PredicateMaxDiff::print(std::ostream& o) const 
{
  int K=arity-1;
  o << "#(";
  for(int i=0; i<K-1; ++i) {
    scope[i]->printshort(o);
    o << ", ";
  }
  scope[K-1]->printshort(o);
  o << ") = ";
  scope[K]->printshort( o );
}


/**********************************************
 * MinDiff Predicate 
 **********************************************/

PredicateMinDiff::PredicateMinDiff(Solver *s, VariableInt** v, const int n, const int t) 
  : LpConstraint(s, v, n)
{
  filtering = LpConstraint::APX; 
  initStruct(t);
}

PredicateMinDiff::PredicateMinDiff(Solver *s, const int f, VariableInt** v, const int n, const int t) 
  : LpConstraint(s, v, n)
{ 
  filtering = f;
  initStruct(t);
}

void PredicateMinDiff::initStruct(const int t) 
{ 

  int i, j, k, nRows=arity-1;

  K = arity-1;
  vars = new VariableInt*[K];
  occurrence = new int[ub-lb+1];
  occurrence -= lb;
  orderedvals = new int[ub-lb+1];

  std::fill( occurrence+lb, occurrence+ub+1, 0 );
  
  weight = 0;


  if( filtering == LpConstraint::LPA ||
      filtering == LpConstraint::LPS ||
      filtering == LpConstraint::LPO ) 
    {
      
      imodel.resize(nRows, 0);
      smodel.resize(nRows, 0);
      
      CoinBuild IbuildObject;
      CoinBuild SbuildObject;
      
      for (i=0;i<nRows;i++) {
	imodel.setRowLower(i, 1.0);
	imodel.setRowUpper(i, 1.0);
	smodel.setRowLower(i, 1.0);
	smodel.setRowUpper(i, 1.0);
      }
      
      nColumns = 0;

      for(i=0; i<nRows; ++i) {
	j = scope[i]->first();
	do {
	  ++occurrence[j];
	  ++nColumns;
	  int columnIndex[1] = {i};
	  double columnValue[1] = {1.0};     
	  IbuildObject.addColumn(1, columnIndex, columnValue, 0.0, 1.0);
	  SbuildObject.addColumn(1, columnIndex, columnValue, 0.0, 1.0);
	} while( scope[i]->setNext(j) ); 
      } 
      imodel.addColumns(IbuildObject);
      smodel.addColumns(SbuildObject);

      int nProd = 0;
      for(i=lb; i<=ub; ++i)
	nProd += ((occurrence[i] - 1) * (occurrence[i]) / 2);
      

 //      //int start[nColumns-nRows];
//       int start[nColumns];
//       int column[nProd];
//       double element[nProd];

      
      int *start = new int[nColumns+1];
      int *column = new int[nProd];
      double *element = new double[nProd];
      
//       for(i=0; i<nRows; ++i) {
// 	j = scope[i]->first();
// 	do {
	  
// 	  for(k=i+1; k<nRows; ++k) {
// 	    if( scope[k]->contain(j) )
// 	      cout << "x" << i << "=" << j << " * x" << k << "=" << j << " " << endl
// 		   << "cl" << setw(2) << asgn2index[i][j] << " * cl" << setw(2) << asgn2index[k][j] << " " << endl;
// 	  }
// 	  cout << endl;
	  
// 	} while( scope[i]->setNext(j) ); 
//       }

//       for(i=0; i<nvals; ++i) {
// 	for(j=0; j<nRows; ++j) {
// 	  for(k=j+1; k<nRows; ++k)
// 	    cout << "x" << j << "=" << index2values[i] << "*x"
// 		 << k << "=" << index2values[i] << "  ";
// 	  cout << endl;
// 	}
// 	cout << endl;
//       }

      int idx = 0, str = 0;
      for(j=0; j<nRows; ++j) {
	for(i=0; i<nvals; ++i) {
	  if(scope[j]->contain(index2values[i])) {
	    start[idx] = str;
	    for(k=j+1; k<nRows; ++k)
	      if(scope[k]->contain(index2values[i])) {
		

		column[str] = asgn2index[k][index2values[i]];
		
		assert( index2varasgn[column[str]] == k );
		assert( index2varasgn[column[str]] == k );
		assert( index2valasgn[column[str]] == index2values[i] );
		
		
		element[str] = -1.0;
		++str;
	      }
	    ++idx;
	    // 	for(j=0; j<occurrence[index2values[i]]; ++j) {
	    // 	  cout << str << " ";
	    // 	  start[idx] = str;
	    // 	  str += (occurrence[index2values[i]]-j-1);
	    // 	  ++idx;
	    // 	}
	  }
	  //cout << endl;
	}
      }
      start[idx] = str;

//       for(i=0; i<=nColumns; ++i)
// 	cout << start[i] << " ";
//       cout << endl;
//       for(i=0; i<nProd; ++i)
// 	cout << column[i] << " ";
//       cout << endl;

//       for (j=0;j<nColumns-1;j++) {
// 	if (start[j]<start[j+1]) {
// 	  printf("Column %d ",j);
// 	  for (i=start[j];i<start[j+1];i++) {
// 	    printf("( %d, %g) ",column[i],element[i]);
// 	  }
// 	  printf("\n");
// 	}
//       }

//       for (j=0;j<nColumns-1;j++) {
// 	if (start[j]<start[j+1]) {
// 	  cout << "x" << index2varasgn[j] << "=" << index2valasgn[j] << ": ";
// 	  for (i=start[j];i<start[j+1];i++) {
// 	    cout << "x" << index2varasgn[column[i]] << "="
// 		 << index2valasgn[column[i]] << " ";
// 	  }
// 	  printf("\n");
// 	}
//       }

      imodel.loadQuadraticObjective(nColumns,start,column,element);
      smodel.loadQuadraticObjective(nColumns,start,column,element);

 //      ClpCholeskyBase * 
      cholesky = new ClpCholeskyBase();
      cholesky->setKKT(true);
      imodel.setCholesky(cholesky);

      delete [] start;
      delete [] column;
      delete [] element;

    }
  
//   switch( filtering ) {
//   case LpConstraint::APX : {
//     cout << "Approximation algorithm" << endl;
//     break;
//   }
//   case LpConstraint::LPO : {
//     cout << "Linear Relaxation" << endl;
//     break;
//   }
//   case LpConstraint::LPS : {
//     cout << "Linear Relaxation + approximation" << endl;
//     break;
//   }
//   case LpConstraint::LPA : {
//     cout << "approximation + Linear Relaxation" << endl;
//     break;
//   }
//   default : {
//     cout << "Decomposition" << endl;
//     break;
//   }
//   }



}

PredicateMinDiff::~PredicateMinDiff()
{
  occurrence += lb;
  delete [] occurrence;
  delete [] vars;
  delete [] orderedvals;
}

int PredicateMinDiff::approximation( int& opt ) 
{
  //  cout << "approximation [" << lb << ".." << ub << "]" << endl;

  int nbdiff = (((K)*(K-1)) >> 1);
  int i, j, nvals, vali;

//   for(i=0; i<K; ++i) {
//     scope[i]->print( cout );
//     cout << endl;
//   }
//   cout << endl;


  int nvars=K, aux;
  VariableInt *auxv;
  int *firstval = orderedvals;

    
  /// count values
  std::fill( occurrence+lb, occurrence+ub+1, 0 );
  nvals=0;
  for(i=0; i<K; ++i) {
    vars[i] = scope[i];
    vali = scope[i]->first();
    do {
      if( ++occurrence[ vali ] == 1 )
	orderedvals[nvals++] = vali;
    }
    while( scope[i]->setNext(vali) );
  }
     
  int *lastval = orderedvals+nvals;


 //  ////
//   for(i=lb; i<=ub; ++i)
//     cout << i << ": " << occurrence[i] << endl;

  while( nvars ) {

//     cout << "There are still " << nvars << " unassgigned variables " << endl;
//     for(i=0; i<nvars; ++i) {
//       vars[i]->printshort( cout );
//       cout << " ";
//     }
//     cout << endl;

    /// order values
    for(i=1; firstval+i<lastval; ++i) {
      j = i;
      while( j && occurrence[firstval[j]] > occurrence[firstval[j-1]] )
	{
	  vali = firstval[j];
	  firstval[j] = firstval[j-1];
	  firstval[j-1] = vali;
	  --j;
	}
    }

//   i = ub;
//   while( i-- > k ) {
//     j = i;
//     while( j < ub && groundvals[firstval[j+1]] < groundvals[firstval[j]] ) {
//       aux = firstval[j+1];
//       firstval[j+1] = firstval[j];
//       firstval[j] = aux;
//       ++j;
//     }
//   }


//     for(i=0; firstval+i<lastval; ++i)
//       cout << firstval[i] << " (" 
// 	   << occurrence[firstval[i]] << ")  ";
//     cout << endl;
    
    /// use the value with largest occurrences
    aux = occurrence[*firstval];
    aux = ((aux * (aux-1)) >> 1);
    nbdiff -= aux;

    //cout << "ASSIGN => " << (*firstval) << endl;

    // update the data-structures
    i = nvars;
    while( i-- ) 
      if( vars[i]->contain( *firstval ) )
	{
	  // update the array 'occurrence'
	  vali = vars[i]->first();
	  do --occurrence[ vali ];
	  while( vars[i]->setNext(vali) );
	  
	  // put all used variables at the end of the array 'vars'
	  --nvars;
	  auxv = vars[nvars];
	  vars[nvars] = vars[i];
	  vars[i] = auxv;
	}
    //--nvals;
    ++firstval;


 //    for(i=0; firstval+i<lastval; ++i)
//       cout << firstval[i] << " (" 
// 	   << occurrence[firstval[i]] << ")  ";
//     cout << endl << endl;

  }
  
  // cout << " => " << nbdiff << endl;

  return nbdiff;
}

bool PredicateMinDiff::propagate(VariableInt *v, const int e)
{
  bool consistent = true;
  int i, j, nbdiff, jnxt, k = (((K) * (K-1)) >> 1);
  
  if( scope[K]->max() < k ) {

    switch( filtering ) {

    case LpConstraint::APX : {
      nbdiff = scope[K]->min();
      nbdiff = approximation( nbdiff );

      if( !scope[K]->setMin( nbdiff ) ) consistent = false;
     
      break;
    }

    case LpConstraint::LPO : {

      updateModel();

//       ClpInterior barrier;
//       barrier.borrowModel(imodel);
//       ClpCholeskyBase * cholesky = new ClpCholeskyBase();
//       cholesky->setKKT(true);
//       barrier.setCholesky(cholesky);
//       barrier.primalDual();
//       int barrierobj = ((K * (K-1) / 2) + (int)(barrier.objectiveValue()));
//       barrier.returnModel(imodel);


      smodel.primal();
      int simplexobj = (k + (int)round(smodel.objectiveValue()));
      
      if( !(smodel.primalFeasible()) ) simplexobj = -1;
      
      
      //       imodel.writeMps( "xx.mps" );
      
      
      //ClpCholeskyBase * cholesky = new ClpCholeskyBase();
      //cholesky->setKKT(true);

      imodel.primalDual(); 

      nbdiff = scope[K]->min();
      nbdiff = approximation( nbdiff );

      int objective = (k + (int)round(imodel.objectiveValue()));

      if( !(imodel.primalFeasible()) ) objective = -1;

      if( simplexobj < 0 ) cout << " - ";
      else cout << setw(2) << simplexobj << " ";

      if( objective < 0 ) cout << " - ";
      else cout << setw(2) << objective << " ";

      cout << setw(2) << nbdiff << endl; 



      if( objective != nbdiff ) {//|| objective != simplexobj || simplexobj != nbdiff ) {

	cout << "I-model" << endl;

	const double * solution = imodel.primalColumnSolution();
	for(j=0; j<nColumns; ++j) 
	  if( solution[j] )
	    cout << setw(12) << solution[j] << ".x" << index2varasgn[j] << " = "
		 << index2valasgn[j] << endl;
	cout << "\t" << (imodel.objectiveValue()) << endl;

	cout << smodel.objectiveValue() << " " 
	     << (floor(smodel.objectiveValue())) << " " 
	     << (ceil(smodel.objectiveValue())) << " " 
	     << ((int)(smodel.objectiveValue())) << " " 
	     << endl
	     << imodel.objectiveValue() << " " 
	     << (floor(imodel.objectiveValue())) << " " 
	     << (ceil(imodel.objectiveValue())) << " " 
	     << ((int)(imodel.objectiveValue())) << " " 
	     << endl;

	for(i=0; i<K; ++i) {
	  scope[i]->print( cout );
	  cout << endl;
	}
	cout << endl;
	
	imodel.writeMps( "imodel.mps" );
	smodel.writeMps( "smodel.mps" );

	ifstream imod("imodel.mps");
	ifstream smod("smodel.mps");
	char ci, cs;
	while( imod.good() && smod.good() )
	  {
	    imod.get( ci );
	    smod.get( cs );
	    if( ci != cs ) {
	      cout << "DIFF" << endl;
	      exit(0);
	    }
	  }
	if( imod.good() || smod.good() ) {
	  cout << "DIFF" << endl;
	}
	//exit(0);
      }


     if( simplexobj != nbdiff ) {//|| objective != simplexobj || simplexobj != nbdiff ) {

	cout << "S-model" << endl;

	for(i=0; i<K; ++i) {
	  scope[i]->print( cout );
	  cout << endl;
	}
	cout << endl;
	

	const double * solution = smodel.primalColumnSolution();
	for(j=0; j<nColumns; ++j) 
	  if( solution[j] )
	    cout << setw(12) << solution[j] << ".x" << index2varasgn[j] << " = "
		 << index2valasgn[j] << endl;
	cout << "\t" << (smodel.objectiveValue()) << endl;


	smodel.writeMps( "smodel.mps" );
	
	ClpSimplex secondmodel;
	secondmodel.setLogLevel( 0 );
	
	secondmodel.readMps("smodel.mps");

	secondmodel.primal();

	const double * solution2 = secondmodel.primalColumnSolution();
	for(j=0; j<nColumns; ++j) 
	  if( solution2[j] )
	    cout << setw(12) << solution2[j] << ".x" << index2varasgn[j] << " = "
		 << index2valasgn[j] << endl;
	cout << "\t" << (secondmodel.objectiveValue()) << endl;


// 	ifstream imod("imodel.mps");
// 	ifstream smod("smodel.mps");
// 	char ci, cs;
// 	while( imod.good() && smod.good() )
// 	  {
// 	    imod.get( ci );
// 	    smod.get( cs );
// 	    if( ci != cs ) {
// 	      cout << "DIFF" << endl;
// 	      exit(0);
// 	    }
// 	  }
// 	if( imod.good() || smod.good() ) {
// 	  cout << "DIFF" << endl;
// 	}
	exit(0);
      }
      
      consistent = scope[K]->setMin( nbdiff );
     
      break;

 //      for(i=0; i<K; ++i) {
// 	scope[i]->print( cout );
// 	cout << endl;
//       }
//       cout << endl;

//       updateModel();

//       imodel.writeMps( "xx.mps" );

//       imodel.primalDual(); 

//       if( (consistent = imodel.primalFeasible()) ) {

// 	cout << imodel.objectiveValue() << endl; 

//       } else {
	
// 	cout << "Damn" << endl;

//       }

//       exit( 0 );
      
  //     //smodel.dual();
//       nbdiff = (k - (int)(ceil(smodel.objectiveValue())));
       
//       if( !scope[K]->setMax( nbdiff ) ) consistent = false;
      
//       if( consistent ) {

// 	imodel.setRowLower(nRows, (double)( k - scope[K]->max() ));
// 	imodel.setRowUpper(nRows, (double)( k - scope[K]->min() ));
	
// 	imodel.primalDual();  
	
// 	if( (consistent = imodel.primalFeasible()) ) {
	  
// 	  const double * solution = imodel.primalColumnSolution();
// 	  for(i=0; consistent && i<K; ++i) {
// 	    j = scope[i]->first();
// 	    do {
// 	      jnxt = scope[i]->getNext(j);
// 	      k = asgn2index[i][j];	    
// 	      if( solution[k] < epsilon ) 
// 		consistent = scope[i]->remove(j);
// 	      j = jnxt;
// 	    } while( consistent && scope[i]->good(j) ); 
// 	  }
// 	}
//       }
      
//      break;
    }

      
    }
  }
    
  return consistent;
}
  
void PredicateMinDiff::print(std::ostream& o) const 
{
  o << "#(";
  for(int i=0; i<K-1; ++i) {
    scope[i]->printshort(o);
    o << ", ";
  }
  scope[K-1]->printshort(o);
  o << ") = ";
  scope[K]->printshort( o );
}

#endif
